@@ -57,7 +57,6 @@ Street, Fifth Floor, Boston, MA 02110-1335 USA
 #include "backup_copy.h"
 #include "backup_mysql.h"
 #include <btr0btr.h>
-#include "xb0xb.h"
 
 #define ROCKSDB_BACKUP_DIR "#rocksdb"
 
@@ -479,7 +478,7 @@ struct datafile_cur_t {
 	{
 		memset(rel_path, 0, sizeof rel_path);
 		if (filename) {
-			strncpy(abs_path, filename, sizeof abs_path);
+			strncpy(abs_path, filename, sizeof abs_path - 1);
 			abs_path[(sizeof abs_path) - 1] = 0;
 		} else {
 			abs_path[0] = '\0';
@@ -642,12 +641,10 @@ int
 mkdirp(const char *pathname, int Flags, myf MyFlags)
 {
 	char *parent, *p;
-	int len = strlen(pathname) + 1;
 
 	/* make a parent directory path */
-	if (!(parent= (char *)malloc(len)))
+	if (!(parent= strdup(pathname)))
           return(-1);
-	memcpy(parent, pathname, len);
 
 	for (p = parent + strlen(parent);
 	    !is_path_separator(*p) && p != parent; p--) ;
@@ -1012,7 +1009,6 @@ static int fix_win_file_permissions(const char *file)
 	ACL* pOldDACL;
 	SECURITY_DESCRIPTOR* pSD = NULL;
 	EXPLICIT_ACCESS ea = { 0 };
-	BOOL isWellKnownSID = FALSE;
 	PSID pSid = NULL;
 
 	GetSecurityInfo(hFile, SE_FILE_OBJECT, DACL_SECURITY_INFORMATION, NULL, NULL,
@@ -1033,15 +1029,15 @@ static int fix_win_file_permissions(const char *file)
 	ea.Trustee.TrusteeType = TRUSTEE_IS_UNKNOWN;
 	ACL* pNewDACL = 0;
 	DWORD err = SetEntriesInAcl(1, &ea, pOldDACL, &pNewDACL);
-	if (pNewDACL)
+	if (!err)
 	{
+		DBUG_ASSERT(pNewDACL);
 		SetSecurityInfo(hFile, SE_FILE_OBJECT, DACL_SECURITY_INFORMATION, NULL, NULL,
 			pNewDACL, NULL);
+		LocalFree((HLOCAL)pNewDACL);
 	}
 	if (pSD != NULL)
 		LocalFree((HLOCAL)pSD);
-	if (pNewDACL != NULL)
-		LocalFree((HLOCAL)pNewDACL);
 	CloseHandle(hFile);
 	return 0;
 }
@@ -1610,7 +1606,8 @@ bool backup_finish()
 		return(false);
 	}
 
-	if (!write_xtrabackup_info(mysql_connection, XTRABACKUP_INFO, opt_history != 0)) {
+	if (!write_xtrabackup_info(mysql_connection, XTRABACKUP_INFO,
+				    opt_history != 0, true)) {
 		return(false);
 	}
 
