@@ -1,6 +1,6 @@
 /* hmac.c
  *
- * Copyright (C) 2006-2020 wolfSSL Inc.
+ * Copyright (C) 2006-2019 wolfSSL Inc.
  *
  * This file is part of wolfSSL.
  *
@@ -65,7 +65,8 @@
     {
         if (hmac == NULL || (key == NULL && keySz != 0) ||
            !(type == WC_MD5 || type == WC_SHA || type == WC_SHA256 ||
-                type == WC_SHA384 || type == WC_SHA512)) {
+                type == WC_SHA384 || type == WC_SHA512 ||
+                type == BLAKE2B_ID)) {
             return BAD_FUNC_ARG;
         }
 
@@ -130,7 +131,8 @@ int wc_HmacSizeByType(int type)
             type == WC_SHA224 || type == WC_SHA256 ||
             type == WC_SHA384 || type == WC_SHA512 ||
             type == WC_SHA3_224 || type == WC_SHA3_256 ||
-            type == WC_SHA3_384 || type == WC_SHA3_512)) {
+            type == WC_SHA3_384 || type == WC_SHA3_512 ||
+            type == BLAKE2B_ID)) {
         return BAD_FUNC_ARG;
     }
 
@@ -170,6 +172,12 @@ int wc_HmacSizeByType(int type)
             break;
     #endif /* WOLFSSL_SHA512 */
 
+    #ifdef HAVE_BLAKE2
+        case BLAKE2B_ID:
+            ret = BLAKE2B_OUTBYTES;
+            break;
+    #endif /* HAVE_BLAKE2 */
+
     #ifdef WOLFSSL_SHA3
         case WC_SHA3_224:
             ret = WC_SHA3_224_DIGEST_SIZE;
@@ -237,6 +245,12 @@ int _InitHmac(Hmac* hmac, int type, void* heap)
             break;
     #endif /* WOLFSSL_SHA512 */
 
+    #ifdef HAVE_BLAKE2
+        case BLAKE2B_ID:
+            ret = wc_InitBlake2b(&hmac->hash.blake2b, BLAKE2B_256);
+            break;
+    #endif /* HAVE_BLAKE2 */
+
     #ifdef WOLFSSL_SHA3
     #ifndef WOLFSSL_NOSHA3_224
         case WC_SHA3_224:
@@ -289,20 +303,15 @@ int wc_HmacSetKey(Hmac* hmac, int type, const byte* key, word32 length)
             type == WC_SHA224 || type == WC_SHA256 ||
             type == WC_SHA384 || type == WC_SHA512 ||
             type == WC_SHA3_224 || type == WC_SHA3_256 ||
-            type == WC_SHA3_384 || type == WC_SHA3_512)) {
+            type == WC_SHA3_384 || type == WC_SHA3_512 ||
+            type == BLAKE2B_ID)) {
         return BAD_FUNC_ARG;
     }
 
-#ifndef HAVE_FIPS
     /* if set key has already been run then make sure and free existing */
-    /* This is for async and PIC32MZ situations, and just normally OK,
-       provided the user calls wc_HmacInit() first. That function is not
-       available in FIPS builds. In current FIPS builds, the hashes are
-       not allocating resources. */
-    if (hmac->macType != WC_HASH_TYPE_NONE) {
+    if (hmac->macType != 0) {
         wc_HmacFree(hmac);
     }
-#endif
 
     hmac->innerHashKeyed = 0;
     hmac->macType = (byte)type;
@@ -448,6 +457,27 @@ int wc_HmacSetKey(Hmac* hmac, int type, const byte* key, word32 length)
             break;
     #endif /* WOLFSSL_SHA512 */
 
+    #ifdef HAVE_BLAKE2
+        case BLAKE2B_ID:
+            hmac_block_size = BLAKE2B_BLOCKBYTES;
+            if (length <= BLAKE2B_BLOCKBYTES) {
+                if (key != NULL) {
+                    XMEMCPY(ip, key, length);
+                }
+            }
+            else {
+                ret = wc_Blake2bUpdate(&hmac->hash.blake2b, key, length);
+                if (ret != 0)
+                    break;
+                ret = wc_Blake2bFinal(&hmac->hash.blake2b, ip, BLAKE2B_256);
+                if (ret != 0)
+                    break;
+
+                length = BLAKE2B_256;
+            }
+            break;
+    #endif /* HAVE_BLAKE2 */
+
     #ifdef WOLFSSL_SHA3
     #ifndef WOLFSSL_NOSHA3_224
         case WC_SHA3_224:
@@ -613,6 +643,13 @@ static int HmacKeyInnerHash(Hmac* hmac)
             break;
     #endif /* WOLFSSL_SHA512 */
 
+    #ifdef HAVE_BLAKE2
+        case BLAKE2B_ID:
+            ret = wc_Blake2bUpdate(&hmac->hash.blake2b, (byte*)hmac->ipad,
+                                                            BLAKE2B_BLOCKBYTES);
+            break;
+    #endif /* HAVE_BLAKE2 */
+
     #ifdef WOLFSSL_SHA3
     #ifndef WOLFSSL_NOSHA3_224
         case WC_SHA3_224:
@@ -723,6 +760,12 @@ int wc_HmacUpdate(Hmac* hmac, const byte* msg, word32 length)
             break;
     #endif /* WOLFSSL_SHA512 */
 
+    #ifdef HAVE_BLAKE2
+        case BLAKE2B_ID:
+            ret = wc_Blake2bUpdate(&hmac->hash.blake2b, msg, length);
+            break;
+    #endif /* HAVE_BLAKE2 */
+
     #ifdef WOLFSSL_SHA3
     #ifndef WOLFSSL_NOSHA3_224
         case WC_SHA3_224:
@@ -896,6 +939,24 @@ int wc_HmacFinal(Hmac* hmac, byte* hash)
             break;
     #endif /* WOLFSSL_SHA512 */
 
+    #ifdef HAVE_BLAKE2
+        case BLAKE2B_ID:
+            ret = wc_Blake2bFinal(&hmac->hash.blake2b, (byte*)hmac->innerHash,
+                                                                   BLAKE2B_256);
+            if (ret != 0)
+                break;
+            ret = wc_Blake2bUpdate(&hmac->hash.blake2b, (byte*)hmac->opad,
+                                                            BLAKE2B_BLOCKBYTES);
+            if (ret != 0)
+                break;
+            ret = wc_Blake2bUpdate(&hmac->hash.blake2b, (byte*)hmac->innerHash,
+                                                                   BLAKE2B_256);
+            if (ret != 0)
+                break;
+            ret = wc_Blake2bFinal(&hmac->hash.blake2b, hash, BLAKE2B_256);
+            break;
+    #endif /* HAVE_BLAKE2 */
+
     #ifdef WOLFSSL_SHA3
     #ifndef WOLFSSL_NOSHA3_224
         case WC_SHA3_224:
@@ -985,7 +1046,6 @@ int wc_HmacInit(Hmac* hmac, void* heap, int devId)
         return BAD_FUNC_ARG;
 
     XMEMSET(hmac, 0, sizeof(Hmac));
-    hmac->macType = WC_HASH_TYPE_NONE;
     hmac->heap = heap;
 #ifdef WOLF_CRYPTO_CB
     hmac->devId = devId;
@@ -1076,6 +1136,11 @@ void wc_HmacFree(Hmac* hmac)
             break;
     #endif /* WOLFSSL_SHA512 */
 
+    #ifdef HAVE_BLAKE2
+        case BLAKE2B_ID:
+            break;
+    #endif /* HAVE_BLAKE2 */
+
     #ifdef WOLFSSL_SHA3
     #ifndef WOLFSSL_NOSHA3_224
         case WC_SHA3_224:
