@@ -1,5 +1,5 @@
 /* Copyright (c) 2000, 2014, Oracle and/or its affiliates.
-   Copyright (c) 2009, 2017, MariaDB Corporation.
+   Copyright (c) 2009, 2019, MariaDB Corporation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -19,12 +19,12 @@
 #ifndef _my_pthread_h
 #define _my_pthread_h
 
-#include "my_global.h"                          /* myf */
-
 #ifndef ETIME
 #define ETIME ETIMEDOUT				/* For FreeBSD */
 #endif
 
+#include <my_atomic.h>
+
 #ifdef  __cplusplus
 #define EXTERNC extern "C"
 extern "C" {
@@ -82,7 +82,7 @@ int pthread_cond_signal(pthread_cond_t *cond);
 int pthread_cond_broadcast(pthread_cond_t *cond);
 int pthread_cond_destroy(pthread_cond_t *cond);
 int pthread_attr_init(pthread_attr_t *connect_att);
-int pthread_attr_setstacksize(pthread_attr_t *connect_att,DWORD stack);
+int pthread_attr_setstacksize(pthread_attr_t *connect_att,size_t stack);
 int pthread_attr_destroy(pthread_attr_t *connect_att);
 int my_pthread_once(my_pthread_once_t *once_control,void (*init_routine)(void));
 
@@ -191,14 +191,26 @@ extern int my_pthread_create_detached;
 int sigwait(sigset_t *set, int *sig);
 #endif
 
-#define my_sigwait(A,B) sigwait((A),(B))
+static inline int my_sigwait(sigset_t *set, int *sig, int *code)
+{
+#ifdef HAVE_SIGWAITINFO
+  siginfo_t siginfo;
+  *sig= sigwaitinfo(set, &siginfo);
+  *code= siginfo.si_code;
+  return *sig < 0 ?  errno : 0;
+#else
+#define SI_KERNEL 128
+  *code= 0;
+  return sigwait(set, sig);
+#endif
+}
 
 #if defined(HAVE_SIGTHREADMASK) && !defined(HAVE_PTHREAD_SIGMASK)
 #define pthread_sigmask(A,B,C) sigthreadmask((A),(B),(C))
 #endif
 
 #if !defined(HAVE_SIGWAIT) && !defined(HAVE_rts_threads) && !defined(sigwait) && !defined(alpha_linux_port) && !defined(_AIX)
-int sigwait(sigset_t *setp, int *sigp);		/* Use our implemention */
+int sigwait(sigset_t *setp, int *sigp);		/* Use our implementation */
 #endif
 
 
@@ -300,6 +312,8 @@ int my_pthread_mutex_trylock(pthread_mutex_t *mutex);
 #endif
 #endif
 
+size_t my_setstacksize(pthread_attr_t *attr, size_t stacksize);
+
 /*
   The defines set_timespec and set_timespec_nsec should be used
   for calculating an absolute time at which
@@ -309,7 +323,7 @@ int my_pthread_mutex_trylock(pthread_mutex_t *mutex);
 
 #ifndef set_timespec_nsec
 #define set_timespec_nsec(ABSTIME,NSEC)                                 \
-  set_timespec_time_nsec((ABSTIME), my_hrtime().val*1000 + (NSEC))
+  set_timespec_time_nsec((ABSTIME), my_hrtime_coarse().val*1000 + (NSEC))
 #endif /* !set_timespec_nsec */
 
 /* adapt for two different flavors of struct timespec */
@@ -395,7 +409,7 @@ typedef struct st_safe_mutex_deadlock_t
 
 #ifdef SAFE_MUTEX_DETECT_DESTROY
 /*
-  Used to track the destroying of mutexes. This needs to be a seperate
+  Used to track the destroying of mutexes. This needs to be a separate
   structure because the safe_mutex_t structure could be freed before
   the mutexes are destroyed.
 */
@@ -503,9 +517,9 @@ void safe_mutex_free_deadlock_data(safe_mutex_t *mp);
      MDL subsystem deadlock detector relies on this property for
      its correctness.
   2) They are optimized for uncontended wr-lock/unlock case.
-     This is scenario in which they are most oftenly used
+     This is a scenario in which they are most often used
      within MDL subsystem. Optimizing for it gives significant
-     performance improvements in some of tests involving many
+     performance improvements in some of the tests involving many
      connections.
 
   Another important requirement imposed on this type of rwlock
@@ -569,36 +583,13 @@ extern int rw_pr_destroy(rw_pr_lock_t *);
 /**
   Implementation of Windows rwlock.
 
-  We use native (slim) rwlocks on Win7 and later, and fallback to  portable
-  implementation on earlier Windows.
-
-  slim rwlock are also available on Vista/WS2008, but we do not use it
-  ("trylock" APIs are missing on Vista)
+  We use native (slim) rwlocks on Windows, which requires Win7
+  or later.
 */
-typedef union
+typedef struct _my_rwlock_t
 {
-  /* Native rwlock (is_srwlock == TRUE) */
-  struct 
-  {
-    SRWLOCK srwlock;             /* native reader writer lock */
-    BOOL have_exclusive_srwlock; /* used for unlock */
-  };
-
-  /*
-    Portable implementation (is_srwlock == FALSE)
-    Fields are identical with Unix my_rw_lock_t fields.
-  */
-  struct 
-  {
-    pthread_mutex_t lock;       /* lock for structure		*/
-    pthread_cond_t  readers;    /* waiting readers		*/
-    pthread_cond_t  writers;    /* waiting writers		*/
-    int state;                  /* -1:writer,0:free,>0:readers	*/
-    int waiters;                /* number of waiting writers	*/
-#ifdef SAFE_MUTEX
-    pthread_t  write_thread;
-#endif
-  };
+  SRWLOCK srwlock;             /* native reader writer lock */
+  BOOL have_exclusive_srwlock; /* used for unlock */
 } my_rw_lock_t;
 
 
@@ -693,10 +684,10 @@ extern void my_mutex_end(void);
   We need to have at least 256K stack to handle calls to myisamchk_init()
   with the current number of keys and key parts.
 */
-#if defined(__SANITIZE_ADDRESS__) || defined(WITH_UBSAN)
-#define DEFAULT_THREAD_STACK	(364*1024L)
+#ifdef __SANITIZE_ADDRESS__
+#define DEFAULT_THREAD_STACK	(383*1024L) /* 392192 */
 #else
-#define DEFAULT_THREAD_STACK	(292*1024L)
+#define DEFAULT_THREAD_STACK	(292*1024L) /* 299008 */
 #endif
 #endif
 
@@ -707,36 +698,47 @@ extern void my_mutex_end(void);
 
 #define INSTRUMENT_ME 0
 
+/*
+  Thread specific variables
+
+  Aria key cache is using the following variables for keeping track of
+  state:
+  suspend, next, prev, keycache_link, keycache_file, suspend, lock_type
+
+  MariaDB uses the following to
+  mutex, current_mutex, current_cond, abort
+*/
+
 struct st_my_thread_var
 {
   int thr_errno;
   mysql_cond_t suspend;
   mysql_mutex_t mutex;
+  struct st_my_thread_var *next,**prev;
   mysql_mutex_t * volatile current_mutex;
   mysql_cond_t * volatile current_cond;
+  void *keycache_link;
+  void *keycache_file;
+  void *stack_ends_here;
+  safe_mutex_t *mutex_in_use;
   pthread_t pthread_self;
   my_thread_id id, dbug_id;
   int volatile abort;
+  uint lock_type; /* used by conditional release the queue */
   my_bool init;
-  struct st_my_thread_var *next,**prev;
-  void *keycache_link;
-  uint  lock_type; /* used by conditional release the queue */
-  void  *stack_ends_here;
-  safe_mutex_t *mutex_in_use;
 #ifndef DBUG_OFF
   void *dbug;
   char name[THREAD_NAME_SIZE+1];
 #endif
 };
 
-struct st_my_thread_var *_my_thread_var(void);
+extern struct st_my_thread_var *_my_thread_var(void) __attribute__ ((const));
 extern void **my_thread_var_dbug(void);
 extern safe_mutex_t **my_thread_var_mutex_in_use(void);
 extern uint my_thread_end_wait_time;
 extern my_bool safe_mutex_deadlock_detector;
 #define my_thread_var (_my_thread_var())
 #define my_errno my_thread_var->thr_errno
-int set_mysys_var(struct st_my_thread_var *mysys_var);
 /*
   Keep track of shutdown,signal, and main threads so that my_end() will not
   report errors with them
@@ -806,6 +808,26 @@ extern uint thd_lib_detected;
 #define statistic_sub(V,C,L)     (V)-=(C)
 #endif /* SAFE_STATISTICS */
 
+static inline void thread_safe_increment32(int32 *value)
+{
+  (void) my_atomic_add32_explicit(value, 1, MY_MEMORY_ORDER_RELAXED);
+}
+
+static inline void thread_safe_decrement32(int32 *value)
+{
+  (void) my_atomic_add32_explicit(value, -1, MY_MEMORY_ORDER_RELAXED);
+}
+
+static inline void thread_safe_increment64(int64 *value)
+{
+  (void) my_atomic_add64_explicit(value, 1, MY_MEMORY_ORDER_RELAXED);
+}
+
+static inline void thread_safe_decrement64(int64 *value)
+{
+  (void) my_atomic_add64_explicit(value, -1, MY_MEMORY_ORDER_RELAXED);
+}
+
 /*
   No locking needed, the counter is owned by the thread
 */
