@@ -1,4 +1,5 @@
 /* Copyright (C) 2013, 2015, Alexey Botchkov and SkySQL Ab
+   Copyright (c) 2019, MariaDB Corporation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -15,15 +16,11 @@
 
 
 #define PLUGIN_VERSION 0x104
-#define PLUGIN_STR_VERSION "1.4.9"
+#define PLUGIN_STR_VERSION "1.4.8"
 
 #define _my_thread_var loc_thread_var
 
 #include <my_config.h>
-#include <stdio.h>
-#include <time.h>
-#include <string.h>
-#include <fcntl.h>
 #include <assert.h>
 
 #ifndef _WIN32
@@ -81,6 +78,7 @@ static void closelog() {}
 */
 
 #if !defined(MYSQL_DYNAMIC_PLUGIN) && !defined(MARIADB_ONLY)
+#include <typelib.h>
 #define MARIADB_ONLY
 #endif /*MYSQL_PLUGIN_DYNAMIC*/
 
@@ -88,11 +86,12 @@ static void closelog() {}
 #define MYSQL_SERVICE_LOGGER_INCLUDED
 #endif /*MARIADB_ONLY*/
 
+#include <my_global.h>
 #include <my_base.h>
-//#include <my_dir.h>
 #include <typelib.h>
 #include <mysql/plugin.h>
 #include <mysql/plugin_audit.h>
+#include <string.h>
 #include "../../mysys/mysys_priv.h"
 #ifndef RTLD_DEFAULT
 #define RTLD_DEFAULT NULL
@@ -328,10 +327,6 @@ struct connection_info
   char query_buffer[1024];
   time_t query_time;
   int log_always;
-  char proxy[64];
-  int proxy_length;
-  char proxy_host[64];
-  int proxy_host_length;
 };
 
 #define DEFAULT_FILENAME_LEN 16
@@ -465,6 +460,7 @@ static const char *syslog_facility_names[]=
   "LOG_LOCAL4", "LOG_LOCAL5", "LOG_LOCAL6", "LOG_LOCAL7",
   0
 };
+#ifndef _WIN32
 static unsigned int syslog_facility_codes[]=
 {
   LOG_USER, LOG_MAIL, LOG_DAEMON, LOG_AUTH,
@@ -479,6 +475,7 @@ static unsigned int syslog_facility_codes[]=
   LOG_LOCAL0, LOG_LOCAL1, LOG_LOCAL2, LOG_LOCAL3,
   LOG_LOCAL4, LOG_LOCAL5, LOG_LOCAL6, LOG_LOCAL7,
 };
+#endif
 static TYPELIB syslog_facility_typelib=
 {
     array_elements(syslog_facility_names) - 1, "syslog_facility_typelib",
@@ -496,11 +493,13 @@ static const char *syslog_priority_names[]=
   0
 };
 
+#ifndef _WIN32
 static unsigned int syslog_priority_codes[]=
 {
   LOG_EMERG, LOG_ALERT, LOG_CRIT, LOG_ERR,
   LOG_WARNING, LOG_NOTICE, LOG_INFO, LOG_DEBUG,
 };
+#endif
 
 static TYPELIB syslog_priority_typelib=
 {
@@ -765,9 +764,9 @@ static int user_coll_fill(struct user_coll *c, char *users,
       if (cmp_user && take_over_cmp)
       {
         ADD_ATOMIC(internal_stop_logging, 1);
-        CLIENT_ERROR(1, "User '%.*b' was removed from the"
+        CLIENT_ERROR(1, "User '%.*s' was removed from the"
             " server_audit_excl_users.",
-            MYF(ME_JUST_WARNING), (int) cmp_length, users);
+            MYF(ME_WARNING), (int) cmp_length, users);
         ADD_ATOMIC(internal_stop_logging, -1);
         blank_user(cmp_user);
         refill_cmp_coll= 1;
@@ -775,8 +774,8 @@ static int user_coll_fill(struct user_coll *c, char *users,
       else if (cmp_user)
       {
         ADD_ATOMIC(internal_stop_logging, 1);
-        CLIENT_ERROR(1, "User '%.*b' is in the server_audit_incl_users, "
-            "so wasn't added.", MYF(ME_JUST_WARNING), (int) cmp_length, users);
+        CLIENT_ERROR(1, "User '%.*s' is in the server_audit_incl_users, "
+            "so wasn't added.", MYF(ME_WARNING), (int) cmp_length, users);
         ADD_ATOMIC(internal_stop_logging, -1);
         remove_user(users);
         continue;
@@ -1084,7 +1083,7 @@ static int start_logging()
                   "Could not create file '%s'.", alt_fname);
       is_active= 0;
       CLIENT_ERROR(1, "SERVER AUDIT plugin can't create file '%s'.",
-          MYF(ME_JUST_WARNING), alt_fname);
+          MYF(ME_WARNING), alt_fname);
       return 1;
     }
     error_header();
@@ -1132,13 +1131,9 @@ static void setup_connection_simple(struct connection_info *ci)
   ci->ip_length= 0;
   ci->query_length= 0;
   ci->header= 0;
-  ci->proxy_length= 0;
 }
 
 
-#define MAX_HOSTNAME 61
-#define USERNAME_LENGTH 384
-
 static void setup_connection_connect(struct connection_info *cn,
     const struct mysql_event_connection *event)
 {
@@ -1147,7 +1142,7 @@ static void setup_connection_connect(struct connection_info *cn,
   cn->log_always= 0;
   cn->thread_id= event->thread_id;
   get_str_n(cn->db, &cn->db_length, sizeof(cn->db),
-            event->database, event->database_length);
+            event->database.str, event->database.length);
   get_str_n(cn->user, &cn->user_length, sizeof(cn->db),
             event->user, event->user_length);
   get_str_n(cn->host, &cn->host_length, sizeof(cn->host),
@@ -1155,29 +1150,6 @@ static void setup_connection_connect(struct connection_info *cn,
   get_str_n(cn->ip, &cn->ip_length, sizeof(cn->ip),
             event->ip, event->ip_length);
   cn->header= 0;
-  if (event->proxy_user && event->proxy_user[0])
-  {
-    const char *priv_host= event->proxy_user +
-            sizeof(char[MAX_HOSTNAME+USERNAME_LENGTH+5]);
-    size_t priv_host_length;
-
-    if (mysql_57_started)
-    {
-      priv_host+= sizeof(size_t);
-      priv_host_length= *(size_t *) (priv_host + MAX_HOSTNAME);
-    }
-    else
-      priv_host_length= strlen(priv_host);
-
-
-    get_str_n(cn->proxy, &cn->proxy_length, sizeof(cn->proxy),
-              event->priv_user, event->priv_user_length);
-    get_str_n(cn->proxy_host, &cn->proxy_host_length,
-              sizeof(cn->proxy_host),
-              priv_host, priv_host_length);
-  }
-  else
-    cn->proxy_length= 0;
 }
 
 
@@ -1259,7 +1231,7 @@ static void setup_connection_table(struct connection_info *cn,
   cn->log_always= 0;
   cn->query_length= 0;
   get_str_n(cn->db, &cn->db_length, sizeof(cn->db),
-            event->database, event->database_length);
+            event->database.str, event->database.length);
   get_str_n(cn->user, &cn->user_length, sizeof(cn->db),
             event->user, SAFE_STRLEN(event->user));
   get_str_n(cn->host, &cn->host_length, sizeof(cn->host),
@@ -1377,31 +1349,6 @@ static size_t log_header(char *message, size_t message_len,
 }
 
 
-static int log_proxy(const struct connection_info *cn,
-                     const struct mysql_event_connection *event)
-                   
-{
-  time_t ctime;
-  size_t csize;
-  char message[1024];
-
-  (void) time(&ctime);
-  csize= log_header(message, sizeof(message)-1, &ctime,
-                    servhost, servhost_len,
-                    cn->user, cn->user_length,
-                    cn->host, cn->host_length,
-                    cn->ip, cn->ip_length,
-                    event->thread_id, 0, "PROXY_CONNECT");
-  csize+= my_snprintf(message+csize, sizeof(message) - 1 - csize,
-    ",%.*s,`%.*s`@`%.*s`,%d", cn->db_length, cn->db,
-                     cn->proxy_length, cn->proxy,
-                     cn->proxy_host_length, cn->proxy_host,
-                     event->status);
-  message[csize]= '\n';
-  return write_log(message, csize + 1, 1);
-}
-
-
 static int log_connection(const struct connection_info *cn,
                           const struct mysql_event_connection *event,
                           const char *type)
@@ -1439,7 +1386,7 @@ static int log_connection_event(const struct mysql_event_connection *event,
                     event->ip, event->ip_length,
                     event->thread_id, 0, type);
   csize+= my_snprintf(message+csize, sizeof(message) - 1 - csize,
-    ",%.*s,,%d", event->database_length, event->database, event->status);
+    ",%.*s,,%d", event->database.length, event->database.str, event->status);
   message[csize]= '\n';
   return write_log(message, csize + 1, 1);
 }
@@ -1484,6 +1431,7 @@ static size_t escape_string_hide_passwords(const char *str, unsigned int len,
   const char *res_start= result;
   const char *res_end= result + result_len - 2;
   size_t d_len;
+  char b_char;
 
   while (len)
   {
@@ -1521,28 +1469,27 @@ static size_t escape_string_hide_passwords(const char *str, unsigned int len,
 
       if (*next_s)
       {
-        const char b_char= *next_s++;
-        memset(result + d_len, '*', 5);
+        memmove(result + d_len, "*****", 5);
         result+= d_len + 5;
-
-        while (*next_s)
-        {
-          if (*next_s == b_char)
-          {
-            ++next_s;
-            break;
-          }
-          if (*next_s == '\\')
-          {
-            if (next_s[1])
-              next_s++;
-          }
-          next_s++;
-        }
+        b_char= *(next_s++);
       }
       else
         result+= d_len;
 
+      while (*next_s)
+      {
+        if (*next_s == b_char)
+        {
+          ++next_s;
+          break;
+        }
+        if (*next_s == '\\')
+        {
+          if (next_s[1])
+            next_s++;
+        }
+        next_s++;
+      }
       len-= (uint)(next_s - str);
       str= next_s;
       continue;
@@ -1550,23 +1497,19 @@ static size_t escape_string_hide_passwords(const char *str, unsigned int len,
 no_password:
     if (result >= res_end)
       break;
-    else
+    if ((b_char= escaped_char(*str)))
     {
-      const char b_char= escaped_char(*str);
-      if (b_char)
-      {
-        if (result+1 >= res_end)
-          break;
-        *(result++)= '\\';
-        *(result++)= b_char;
-      }
-      else if (is_space(*str))
-        *(result++)= ' ';
-      else
-        *(result++)= *str;
-      str++;
-      len--;
+      if (result+1 >= res_end)
+        break;
+      *(result++)= '\\';
+      *(result++)= b_char;
     }
+    else if (is_space(*str))
+      *(result++)= ' ';
+    else
+      *(result++)= *str;
+    str++;
+    len--;
   }
   *result= 0;
   return result - res_start;
@@ -1873,8 +1816,8 @@ static int log_table(const struct connection_info *cn,
                     event->ip, SAFE_STRLEN_UI(event->ip),
                     event->thread_id, cn->query_id, type);
   csize+= my_snprintf(message+csize, sizeof(message) - 1 - csize,
-            ",%.*s,%.*s,",event->database_length, event->database,
-                          event->table_length, event->table);
+            ",%.*s,%.*s,",event->database.length, event->database.str,
+                          event->table.length, event->table.str);
   message[csize]= '\n';
   return write_log(message, csize + 1, 1);
 }
@@ -1895,10 +1838,10 @@ static int log_rename(const struct connection_info *cn,
                     event->ip, SAFE_STRLEN_UI(event->ip),
                     event->thread_id, cn->query_id, "RENAME");
   csize+= my_snprintf(message+csize, sizeof(message) - 1 - csize,
-            ",%.*s,%.*s|%.*s.%.*s,",event->database_length, event->database,
-                         event->table_length, event->table,
-                         event->new_database_length, event->new_database,
-                         event->new_table_length, event->new_table);
+            ",%.*s,%.*s|%.*s.%.*s,",event->database.length, event->database.str,
+                         event->table.length, event->table.str,
+                         event->new_database.length, event->new_database.str,
+                         event->new_table.length, event->new_table.str);
   message[csize]= '\n';
   return write_log(message, csize + 1, 1);
 }
@@ -1963,7 +1906,7 @@ static void update_connection_info(struct connection_info *cn,
             /* Change DB */
             if (mysql_57_started)
               get_str_n(cn->db, &cn->db_length, sizeof(cn->db),
-                  event->database, event->database_length);
+                  event->database.str, event->database.length);
             else
               get_str_n(cn->db, &cn->db_length, sizeof(cn->db),
                   event->general_query, event->general_query_length);
@@ -1989,9 +1932,9 @@ static void update_connection_info(struct connection_info *cn,
           if (ci_needs_setup(cn))
             setup_connection_query(cn, event);
 
-          if (mode == 0 && cn->db_length == 0 && event->database_length > 0)
+          if (mode == 0 && cn->db_length == 0 && event->database.length > 0)
             get_str_n(cn->db, &cn->db_length, sizeof(cn->db),
-                      event->database, event->database_length);
+                      event->database.str, event->database.length);
 
           if (event->general_error_code == 0)
           {
@@ -2006,7 +1949,7 @@ static void update_connection_info(struct connection_info *cn,
                     event->general_query + 4, event->general_query_length - 4);
               else
                 get_str_n(cn->db, &cn->db_length, sizeof(cn->db),
-                    event->database, event->database_length);
+                    event->database.str, event->database.length);
             }
           }
           update_general_user(cn, event);
@@ -2047,9 +1990,9 @@ static void update_connection_info(struct connection_info *cn,
                 event->ip, SAFE_STRLEN(event->ip));
     }
 
-    if (cn->db_length == 0 && event->database_length != 0)
+    if (cn->db_length == 0 && event->database.length != 0)
       get_str_n(cn->db, &cn->db_length, sizeof(cn->db),
-                event->database, event->database_length);
+                event->database.str, event->database.length);
 
     if (mode == 0)
       cn->query_id= event->query_id;
@@ -2063,13 +2006,9 @@ static void update_connection_info(struct connection_info *cn,
     {
       case MYSQL_AUDIT_CONNECTION_CONNECT:
         setup_connection_connect(cn, event);
-        if (event->status == 0 && event->proxy_user && event->proxy_user[0])
-          log_proxy(cn, event);
         break;
       case MYSQL_AUDIT_CONNECTION_CHANGE_USER:
         *after_action= AA_CHANGE_USER;
-        if (event->proxy_user && event->proxy_user[0])
-          log_proxy(cn, event);
         break;
       default:;
     }
@@ -2263,8 +2202,7 @@ static void auditing_v8(MYSQL_THD thd, struct mysql_event_general_v8 *ev_v8)
   static const int db_len_off= 68;
   #endif /*x86_64*/
 #endif /*DBUG_OFF*/
-#endif /*__linux*/
-
+#endif /* __linux__ */
   struct mysql_event_general event;
 
   if (ev_v8->event_class != MYSQL_AUDIT_GENERAL_CLASS)
@@ -2282,8 +2220,8 @@ static void auditing_v8(MYSQL_THD thd, struct mysql_event_general_v8 *ev_v8)
   event.general_charset= ev_v8->general_charset;
   event.general_time= ev_v8->general_time;
   event.general_rows= ev_v8->general_rows;
-  event.database= 0;
-  event.database_length= 0;
+  event.database.str= 0;
+  event.database.length= 0;
 
   if (event.general_query_length > 0)
   {
@@ -2291,8 +2229,8 @@ static void auditing_v8(MYSQL_THD thd, struct mysql_event_general_v8 *ev_v8)
     event.general_command= "Query";
     event.general_command_length= 5;
 #ifdef __linux__
-    event.database= *(char **) (((char *) thd) + db_off);
-    event.database_length= *(size_t *) (((char *) thd) + db_len_off);
+    event.database.str= *(char **) (((char *) thd) + db_off);
+    event.database.length= *(size_t *) (((char *) thd) + db_len_off);
 #endif /*__linux*/
   }
 #ifdef __linux__
@@ -2323,9 +2261,8 @@ static void auditing_v13(MYSQL_THD thd, unsigned int *ev_v0)
 }
 
 
-int get_db_mysql57(MYSQL_THD thd, char **name, int *len)
+int get_db_mysql57(MYSQL_THD thd, char **name, size_t *len)
 {
-#ifdef __linux__
   int db_off;
   int db_len_off;
   if (debug_server_started)
@@ -2349,8 +2286,9 @@ int get_db_mysql57(MYSQL_THD thd, char **name, int *len)
 #endif /*x86_64*/
   }
 
+#ifdef __linux__
   *name= *(char **) (((char *) thd) + db_off);
-  *len= *((int *) (((char*) thd) + db_len_off));
+  *len= *((size_t *) (((char*) thd) + db_len_off));
   if (*name && (*name)[*len] != 0)
     return 1;
   return 0;
@@ -2706,7 +2644,7 @@ static void update_file_path(MYSQL_THD thd,
       {
         error_header();
         fprintf(stderr, "Logging was disabled..\n");
-        CLIENT_ERROR(1, "Logging was disabled.", MYF(ME_JUST_WARNING));
+        CLIENT_ERROR(1, "Logging was disabled.", MYF(ME_WARNING));
       }
       goto exit_func;
     }
@@ -2921,7 +2859,7 @@ static void update_logging(MYSQL_THD thd,
     start_logging();
     if (!logging)
     {
-      CLIENT_ERROR(1, "Logging was disabled.", MYF(ME_JUST_WARNING));
+      CLIENT_ERROR(1, "Logging was disabled.", MYF(ME_WARNING));
     }
     mark_always_logged(thd);
   }
