@@ -1,5 +1,4 @@
-/* Copyright (c) 2006, 2019, Oracle and/or its affiliates.
-   Copyright (c) 2009, 2020, MariaDB Corporation
+/* Copyright (c) 2006, 2013, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -14,7 +13,7 @@
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1335  USA */
 
-#include <my_global.h>
+#include "mariadb.h"
 #include "sql_priv.h"
 #include "unireg.h"
 #include "event_scheduler.h"
@@ -50,11 +49,11 @@ Event_db_repository *Event_worker_thread::db_repository;
 
 
 static
-const LEX_STRING scheduler_states_names[] =
+const LEX_CSTRING scheduler_states_names[] =
 {
-  { C_STRING_WITH_LEN("INITIALIZED") },
-  { C_STRING_WITH_LEN("RUNNING") },
-  { C_STRING_WITH_LEN("STOPPING") }
+  { STRING_WITH_LEN("INITIALIZED") },
+  { STRING_WITH_LEN("RUNNING") },
+  { STRING_WITH_LEN("STOPPING") }
 };
 
 struct scheduler_param {
@@ -86,7 +85,7 @@ Event_worker_thread::print_warnings(THD *thd, Event_job_data *et)
   char prefix_buf[5 * STRING_BUFFER_USUAL_SIZE];
   String prefix(prefix_buf, sizeof(prefix_buf), system_charset_info);
   prefix.length(0);
-  prefix.append("Event Scheduler: [");
+  prefix.append(STRING_WITH_LEN("Event Scheduler: ["));
 
   prefix.append(et->definer.str, et->definer.length, system_charset_info);
   prefix.append("][", 2);
@@ -151,7 +150,7 @@ deinit_event_thread(THD *thd)
 {
   thd->proc_info= "Clearing";
   DBUG_PRINT("exit", ("Event thread finishing"));
-  unlink_not_visible_thd(thd);
+  server_threads.erase(thd);
   delete thd;
 }
 
@@ -186,7 +185,7 @@ pre_init_event_thread(THD* thd)
   thd->net.read_timeout= slave_net_timeout;
   thd->variables.option_bits|= OPTION_AUTO_IS_NULL;
   thd->client_capabilities|= CLIENT_MULTI_RESULTS;
-  add_to_active_threads(thd);
+  server_threads.insert(thd);
 
   /*
     Guarantees that we will see the thread in SHOW PROCESSLIST though its
@@ -298,11 +297,10 @@ Event_worker_thread::run(THD *thd, Event_queue_element_for_exec *event)
   DBUG_ENTER("Event_worker_thread::run");
   DBUG_PRINT("info", ("Time is %u, THD: %p", (uint)my_time(0), thd));
 
-  inc_thread_running();
   if (res)
     goto end;
 
-  if ((res= db_repository->load_named_event(thd, event->dbname, event->name,
+  if ((res= db_repository->load_named_event(thd, &event->dbname, &event->name,
                                             &job_data)))
   {
     DBUG_PRINT("error", ("Got error from load_named_event"));
@@ -327,7 +325,6 @@ end:
              event->name.str));
 
   delete event;
-  dec_thread_running();
   deinit_event_thread(thd);
 
   DBUG_VOID_RETURN;
@@ -498,7 +495,6 @@ Event_scheduler::run(THD *thd)
       DBUG_PRINT("info", ("job_data is NULL, the thread was killed"));
     }
     DBUG_PRINT("info", ("state=%s", scheduler_states_names[state].str));
-    free_root(thd->mem_root, MYF(0));
   }
 
   LOCK_DATA();
@@ -650,14 +646,11 @@ Event_scheduler::stop()
     state= STOPPING;
     DBUG_PRINT("info", ("Scheduler thread has id %lu",
                         (ulong) scheduler_thd->thread_id));
-    /* Lock from delete */
-    mysql_mutex_lock(&scheduler_thd->LOCK_thd_data);
     /* This will wake up the thread if it waits on Queue's conditional */
     sql_print_information("Event Scheduler: Killing the scheduler thread, "
                           "thread id %lu",
                           (ulong) scheduler_thd->thread_id);
     scheduler_thd->awake(KILL_CONNECTION);
-    mysql_mutex_unlock(&scheduler_thd->LOCK_thd_data);
 
     /* thd could be 0x0, when shutting down */
     sql_print_information("Event Scheduler: "
@@ -686,20 +679,20 @@ end:
     Event_scheduler::workers_count()
 */
 
+static my_bool workers_count_callback(THD *thd, uint32_t *count)
+{
+  if (thd->system_thread == SYSTEM_THREAD_EVENT_WORKER)
+    ++*count;
+  return 0;
+}
+
+
 uint
 Event_scheduler::workers_count()
 {
-  THD *tmp;
-  uint count= 0;
-
+  uint32_t count= 0;
   DBUG_ENTER("Event_scheduler::workers_count");
-  mysql_mutex_lock(&LOCK_thread_count);       // For unlink from list
-  I_List_iterator<THD> it(threads);
-  while ((tmp=it++))
-    if (tmp->system_thread == SYSTEM_THREAD_EVENT_WORKER)
-      ++count;
-  mysql_mutex_unlock(&LOCK_thread_count);
-  DBUG_PRINT("exit", ("%d", count));
+  server_threads.iterate(workers_count_callback, &count);
   DBUG_RETURN(count);
 }
 
