@@ -14,7 +14,7 @@
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */
 
 
-#include <my_global.h>
+#include "mariadb.h"
 #include "sql_priv.h"
 #include "sql_class.h"
 #include "item.h"
@@ -173,7 +173,7 @@ static int json_nice(json_engine_t *je, String *nice_js,
           key_end= je->s.c_str;
         } while (json_read_keyname_chr(je) == 0);
         
-        if (je->s.error)
+        if (unlikely(je->s.error))
           goto error;
 
         if (!first_value)
@@ -374,23 +374,17 @@ static int path_setup_nwc(json_path_t *p, CHARSET_INFO *i_cs,
 longlong Item_func_json_valid::val_int()
 {
   String *js= args[0]->val_json(&tmp_value);
-  json_engine_t je;
 
   if ((null_value= args[0]->null_value))
     return 0;
 
-  json_scan_start(&je, js->charset(), (const uchar *) js->ptr(),
-                  (const uchar *) js->ptr()+js->length());
-
-  while (json_scan_next(&je) == 0) {}
-
-  return je.s.error == 0;
+  return json_valid(js->ptr(), js->length(), js->charset());
 }
 
 
 bool Item_func_json_exists::fix_length_and_dec()
 {
-  if (Item_int_func::fix_length_and_dec())
+  if (Item_bool_func::fix_length_and_dec())
     return TRUE;
   maybe_null= 1;
   path.set_constant_flag(args[1]->const_item());
@@ -445,7 +439,17 @@ bool Item_func_json_value::fix_length_and_dec()
 {
   collation.set(args[0]->collation);
   max_length= args[0]->max_length;
-  path.set_constant_flag(args[1]->const_item());
+  set_constant_flag(args[1]->const_item());
+  maybe_null= 1;
+  return FALSE;
+}
+
+
+bool Item_func_json_query::fix_length_and_dec()
+{
+  collation.set(args[0]->collation);
+  max_length= args[0]->max_length;
+  set_constant_flag(args[1]->const_item());
   maybe_null= 1;
   return FALSE;
 }
@@ -455,115 +459,100 @@ bool Item_func_json_value::fix_length_and_dec()
   Returns NULL, not an error if the found value
   is not a scalar.
 */
-String *Item_func_json_value::val_str(String *str)
+bool Json_path_extractor::extract(String *str, Item *item_js, Item *item_jp,
+                                  CHARSET_INFO *cs)
 {
-  json_engine_t je;
-  String *js= args[0]->val_json(&tmp_js);
+  String *js= item_js->val_json(&tmp_js);
   int error= 0;
   uint array_counters[JSON_DEPTH_LIMIT];
 
-  if (!path.parsed)
+  if (!parsed)
   {
-    String *s_p= args[1]->val_str(&tmp_path);
+    String *s_p= item_jp->val_str(&tmp_path);
     if (s_p &&
-        json_path_setup(&path.p, s_p->charset(), (const uchar *) s_p->ptr(),
+        json_path_setup(&p, s_p->charset(), (const uchar *) s_p->ptr(),
                         (const uchar *) s_p->ptr() + s_p->length()))
-      goto err_return;
-    path.parsed= path.constant;
+      return true;
+    parsed= constant;
   }
 
-  if ((null_value= args[0]->null_value || args[1]->null_value))
-    return NULL;
-
-  json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
-                  (const uchar *) js->ptr() + js->length());
+  if (item_js->null_value || item_jp->null_value)
+    return true;
 
+  Json_engine_scan je(*js);
   str->length(0);
-  str->set_charset(collation.collation);
+  str->set_charset(cs);
 
-  path.cur_step= path.p.steps;
+  cur_step= p.steps;
 continue_search:
-  if (json_find_path(&je, &path.p, &path.cur_step, array_counters))
-  {
-    if (je.s.error)
-      goto err_return;
-
-    null_value= 1;
-    return 0;
-  }
+  if (json_find_path(&je, &p, &cur_step, array_counters))
+    return true;
 
   if (json_read_value(&je))
-    goto err_return;
+    return true;
 
-  if (check_and_get_value(&je, str, &error))
+  if (unlikely(check_and_get_value(&je, str, &error)))
   {
     if (error)
-      goto err_return;
+      return true;
     goto continue_search;
   }
 
-  return str;
-
-err_return:
-  null_value= 1;
-  return 0;
+  return false;
 }
 
 
-bool Item_func_json_value::check_and_get_value(json_engine_t *je, String *res,
-                                               int *error)
+bool Json_engine_scan::check_and_get_value_scalar(String *res, int *error)
 {
   CHARSET_INFO *json_cs;
   const uchar *js;
   uint js_len;
 
-  if (!json_value_scalar(je))
+  if (!json_value_scalar(this))
   {
     /* We only look for scalar values! */
-    if (json_skip_level(je) || json_scan_next(je))
+    if (json_skip_level(this) || json_scan_next(this))
       *error= 1;
     return true;
   }
 
-  if (je->value_type == JSON_VALUE_TRUE ||
-      je->value_type == JSON_VALUE_FALSE)
+  if (value_type == JSON_VALUE_TRUE ||
+      value_type == JSON_VALUE_FALSE)
   {
     json_cs= &my_charset_utf8mb4_bin;
-    js= (const uchar *) ((je->value_type == JSON_VALUE_TRUE) ? "1" : "0");
+    js= (const uchar *) ((value_type == JSON_VALUE_TRUE) ? "1" : "0");
     js_len= 1;
   }
   else
   {
-    json_cs= je->s.cs;
-    js= je->value;
-    js_len= je->value_len;
+    json_cs= s.cs;
+    js= value;
+    js_len= value_len;
   }
 
 
-  return st_append_json(res, json_cs, js, js_len); 
+  return st_append_json(res, json_cs, js, js_len);
 }
 
 
-bool Item_func_json_query::check_and_get_value(json_engine_t *je, String *res,
-                                               int *error)
+bool Json_engine_scan::check_and_get_value_complex(String *res, int *error)
 {
-  const uchar *value;
-  if (json_value_scalar(je))
+  if (json_value_scalar(this))
   {
     /* We skip scalar values. */
-    if (json_scan_next(je))
+    if (json_scan_next(this))
       *error= 1;
     return true;
   }
 
-  value= je->value;
-  if (json_skip_level(je))
+  const uchar *tmp_value= value;
+  if (json_skip_level(this))
   {
     *error= 1;
     return true;
   }
 
-  res->set((const char *) je->value, (uint32)(je->s.c_str - value), je->s.cs);
+  res->set((const char *) value, (uint32)(s.c_str - tmp_value), s.cs);
   return false;
 }
 
@@ -606,7 +595,7 @@ String *Item_func_json_quote::val_str(String *str)
 
 bool Item_func_json_unquote::fix_length_and_dec()
 {
-  collation.set(&my_charset_utf8_general_ci,
+  collation.set(&my_charset_utf8mb3_general_ci,
                 DERIVATION_COERCIBLE, MY_REPERTOIRE_ASCII);
   max_length= args[0]->max_length;
   maybe_null= 1;
@@ -647,16 +636,16 @@ String *Item_func_json_unquote::val_str(String *str)
   if (!(js= read_json(&je)))
     return NULL;
 
-  if (je.s.error || je.value_type != JSON_VALUE_STRING)
+  if (unlikely(je.s.error) || je.value_type != JSON_VALUE_STRING)
     return js;
 
   str->length(0);
-  str->set_charset(&my_charset_utf8_general_ci);
+  str->set_charset(&my_charset_utf8mb3_general_ci);
 
   if (str->realloc_with_extra_if_needed(je.value_len) ||
       (c_len= json_unescape(js->charset(),
         je.value, je.value + je.value_len,
-        &my_charset_utf8_general_ci,
+        &my_charset_utf8mb3_general_ci,
         (uchar *) str->ptr(), (uchar *) (str->ptr() + je.value_len))) < 0)
     goto error;
 
@@ -686,7 +675,7 @@ static int alloc_tmp_paths(THD *thd, uint n_paths,
         return 1;
 
       for (uint c_path=0; c_path < n_paths; c_path++)
-        (*tmp_paths)[c_path].set_charset(&my_charset_utf8_general_ci);
+        (*tmp_paths)[c_path].set_charset(&my_charset_utf8mb3_general_ci);
     }
 
     return 0;
@@ -860,7 +849,7 @@ String *Item_func_json_extract::read_json(String *str,
     }
   }
 
-  if (je.s.error)
+  if (unlikely(je.s.error))
     goto error;
 
   if (!not_first_value)
@@ -962,41 +951,6 @@ double Item_func_json_extract::val_real()
 }
 
 
-my_decimal *Item_func_json_extract::val_decimal(my_decimal *to)
-{
-  json_value_types type;
-  char *value;
-  int value_len;
-
-  if (read_json(NULL, &type, &value, &value_len) != NULL)
-  {
-    switch (type)
-    {
-      case JSON_VALUE_STRING:
-      case JSON_VALUE_NUMBER:
-      {
-        my_decimal *res= decimal_from_string_with_check(to, collation.collation,
-                                                        value,
-                                                        value + value_len);
-        null_value= res == NULL;
-        return res;
-      }
-      case JSON_VALUE_TRUE:
-        int2my_decimal(E_DEC_FATAL_ERROR, 1, false/*unsigned_flag*/, to);
-        return to;
-      case JSON_VALUE_OBJECT:
-      case JSON_VALUE_ARRAY:
-      case JSON_VALUE_FALSE:
-      case JSON_VALUE_NULL:
-        break;
-    };
-  }
-  int2my_decimal(E_DEC_FATAL_ERROR, 0, false/*unsigned_flag*/, to);
-  return to;
-}
-
-
-
 bool Item_func_json_contains::fix_length_and_dec()
 {
   a2_constant= args[1]->const_item();
@@ -1004,7 +958,7 @@ bool Item_func_json_contains::fix_length_and_dec()
   maybe_null= 1;
   if (arg_count > 2)
     path.set_constant_flag(args[2]->const_item());
-  return Item_int_func::fix_length_and_dec();
+  return Item_bool_func::fix_length_and_dec();
 }
 
 
@@ -1054,7 +1008,7 @@ static int check_contains(json_engine_t *js, json_engine_t *value)
         k_end= value->s.c_str;
       } while (json_read_keyname_chr(value) == 0);
 
-      if (value->s.error || json_read_value(value))
+      if (unlikely(value->s.error) || json_read_value(value))
         return FALSE;
 
       if (set_js)
@@ -1097,7 +1051,7 @@ static int check_contains(json_engine_t *js, json_engine_t *value)
             return FALSE;
           return TRUE;
         }
-        if (value->s.error || js->s.error ||
+        if (unlikely(value->s.error) || unlikely(js->s.error) ||
             (!v_scalar && json_skip_to_level(js, c_level)))
           return FALSE;
       }
@@ -1225,7 +1179,7 @@ longlong Item_func_json_contains::val_int()
     goto error;
 
   result= check_contains(&je, &ve);
-  if (je.s.error || ve.s.error)
+  if (unlikely(je.s.error || ve.s.error))
     goto error;
 
   return result;
@@ -1256,7 +1210,7 @@ bool Item_func_json_contains_path::fix_length_and_dec()
   ooa_parsed= FALSE;
   maybe_null= 1;
   mark_constant_paths(paths, args+2, arg_count-2);
-  return Item_int_func::fix_length_and_dec();
+  return Item_bool_func::fix_length_and_dec();
 }
 
 
@@ -1379,6 +1333,7 @@ longlong Item_func_json_contains_path::val_int()
   longlong result;
   json_path_t p;
   int n_found;
+  LINT_INIT(n_found);
 
   if ((null_value= args[0]->null_value))
     return 0;
@@ -1444,7 +1399,7 @@ longlong Item_func_json_contains_path::val_int()
     }
   }
 
-  if (je.s.error == 0)
+  if (likely(je.s.error == 0))
     return result;
 
   report_json_error(js, &je, 0);
@@ -1456,7 +1411,7 @@ null_return:
 
 static int append_json_value(String *str, Item *item, String *tmp_val)
 {
-  if (item->is_bool_type())
+  if (item->type_handler()->is_bool_type())
   {
     longlong v_int= item->val_int();
     const char *t_f;
@@ -1814,7 +1769,7 @@ String *Item_func_json_array_insert::val_str(String *str)
         goto js_error;
     }
 
-    if (je.s.error)
+    if (unlikely(je.s.error))
       goto js_error;
 
     str->length(0);
@@ -1946,7 +1901,7 @@ static int do_merge(String *str, json_engine_t *je1, json_engine_t *je2)
         key_end= je1->s.c_str;
       } while (json_read_keyname_chr(je1) == 0);
 
-      if (je1->s.error)
+      if (unlikely(je1->s.error))
         return 1;
 
       if (first_key)
@@ -1981,7 +1936,7 @@ static int do_merge(String *str, json_engine_t *je1, json_engine_t *je2)
           return ires;
         goto merged_j1;
       }
-      if (je2->s.error)
+      if (unlikely(je2->s.error))
         return 2;
 
       key_start= je1->s.c_str;
@@ -2011,7 +1966,7 @@ merged_j1:
         key_end= je2->s.c_str;
       } while (json_read_keyname_chr(je2) == 0);
 
-      if (je2->s.error)
+      if (unlikely(je2->s.error))
         return 1;
 
       *je1= sav_je1;
@@ -2022,7 +1977,7 @@ merged_j1:
         json_string_set_str(&key_name, key_start, key_end);
         if (!json_key_matches(je1, &key_name))
         {
-          if (je1->s.error || json_skip_key(je1))
+          if (unlikely(je1->s.error || json_skip_key(je1)))
             return 2;
           continue;
         }
@@ -2032,7 +1987,7 @@ merged_j1:
         goto continue_j2;
       }
 
-      if (je1->s.error)
+      if (unlikely(je1->s.error))
         return 2;
 
       if (first_key)
@@ -2127,6 +2082,7 @@ String *Item_func_json_merge::val_str(String *str)
   json_engine_t je1, je2;
   String *js1= args[0]->val_json(&tmp_js1), *js2=NULL;
   uint n_arg;
+  LINT_INIT(js2);
 
   if (args[0]->null_value)
     goto null_return;
@@ -2262,7 +2218,7 @@ static int do_merge_patch(String *str, json_engine_t *je1, json_engine_t *je2,
     json_string_t key_name;
     size_t sav_len;
     bool mrg_empty;
-  
+
     *empty_result= FALSE;
     json_string_set_cs(&key_name, je1->s.cs);
 
@@ -2429,8 +2385,6 @@ String *Item_func_json_merge_patch::val_str(String *str)
   uint n_arg;
   bool empty_result, merge_to_null;
 
-  /* To report errors properly if some JSON is invalid. */
-  je1.s.error= je2.s.error= 0;
   merge_to_null= args[0]->null_value;
 
   for (n_arg=1; n_arg < arg_count; n_arg++)
@@ -2517,6 +2471,7 @@ bool Item_func_json_length::fix_length_and_dec()
   if (arg_count > 1)
     path.set_constant_flag(args[1]->const_item());
   maybe_null= 1;
+  max_length= 10;
   return FALSE;
 }
 
@@ -2594,7 +2549,7 @@ longlong Item_func_json_length::val_int()
     while (json_scan_next(&je) == 0) {}
   }
 
-  if (!je.s.error)
+  if (likely(!je.s.error))
     return length;
 
 err_return:
@@ -2648,7 +2603,7 @@ longlong Item_func_json_depth::val_int()
     }
   } while (json_scan_next(&je) == 0);
 
-  if (!je.s.error)
+  if (likely(!je.s.error))
     return depth;
 
   report_json_error(js, &je, 0);
@@ -2659,7 +2614,7 @@ longlong Item_func_json_depth::val_int()
 
 bool Item_func_json_type::fix_length_and_dec()
 {
-  collation.set(&my_charset_utf8_general_ci);
+  collation.set(&my_charset_utf8mb3_general_ci);
   max_length= 12;
   maybe_null= 1;
   return FALSE;
@@ -2705,7 +2660,7 @@ String *Item_func_json_type::val_str(String *str)
     break;
   }
 
-  str->set(type, strlen(type), &my_charset_utf8_general_ci);
+  str->set(type, strlen(type), &my_charset_utf8mb3_general_ci);
   return str;
 
 error:
@@ -2866,7 +2821,7 @@ String *Item_func_json_insert::val_str(String *str)
         }
       }
 
-      if (je.s.error)
+      if (unlikely(je.s.error))
         goto js_error;
 
       if (!mode_insert)
@@ -2904,7 +2859,7 @@ String *Item_func_json_insert::val_str(String *str)
         }
       }
 
-      if (je.s.error)
+      if (unlikely(je.s.error))
         goto js_error;
 
       if (!mode_insert)
@@ -3006,7 +2961,7 @@ String *Item_func_json_remove::val_str(String *str)
   {
     uint array_counters[JSON_DEPTH_LIMIT];
     json_path_with_flags *c_path= paths + n_path;
-    const char *rem_start, *rem_end;
+    const char *rem_start= 0, *rem_end;
     const json_path_step_t *lp;
     uint n_item= 0;
 
@@ -3077,7 +3032,7 @@ String *Item_func_json_remove::val_str(String *str)
         }
       }
 
-      if (je.s.error)
+      if (unlikely(je.s.error))
         goto js_error;
 
       continue;
@@ -3109,7 +3064,7 @@ String *Item_func_json_remove::val_str(String *str)
         }
       }
 
-      if (je.s.error)
+      if (unlikely(je.s.error))
         goto js_error;
 
       continue;
@@ -3275,7 +3230,7 @@ skip_search:
       {
         key_end= je.s.c_str;
       } while (json_read_keyname_chr(&je) == 0);
-      if (je.s.error)
+      if (unlikely(je.s.error))
         goto err_return;
       key_len= (int)(key_end - key_start);
 
@@ -3299,7 +3254,7 @@ skip_search:
     }
   }
 
-  if (je.s.error || str->append("]", 1))
+  if (unlikely(je.s.error || str->append("]", 1)))
     goto err_return;
 
   null_value= 0;
@@ -3484,7 +3439,7 @@ String *Item_func_json_search::val_str(String *str)
     }
   }
 
-  if (je.s.error)
+  if (unlikely(je.s.error))
     goto js_error;
 
 end:
@@ -3607,7 +3562,7 @@ int Arg_comparator::compare_json_str_basic(Item *j, Item *s)
        if (value2.realloc_with_extra_if_needed(je.value_len) ||
          (c_len= json_unescape(js->charset(), je.value,
                                je.value + je.value_len,
-                               &my_charset_utf8_general_ci,
+                               &my_charset_utf8mb3_general_ci,
                                (uchar *) value2.ptr(),
                                (uchar *) (value2.ptr() + je.value_len))) < 0)
          goto error;
@@ -3656,7 +3611,7 @@ int Arg_comparator::compare_e_json_str_basic(Item *j, Item *s)
     if (value1.realloc_with_extra_if_needed(value_len) ||
         (c_len= json_unescape(value1.charset(), (uchar *) value,
                               (uchar *) value+value_len,
-                              &my_charset_utf8_general_ci,
+                              &my_charset_utf8mb3_general_ci,
                               (uchar *) value1.ptr(),
                               (uchar *) (value1.ptr() + value_len))) < 0)
       return 1;
@@ -3668,3 +3623,139 @@ int Arg_comparator::compare_e_json_str_basic(Item *j, Item *s)
 }
 
 
+String* Item_func_json_arrayagg::convert_to_json(Item *item, String *res)
+{
+  String tmp;
+  res->length(0);
+  append_json_value(res, item, &tmp);
+  return res;
+}
+
+
+String* Item_func_json_arrayagg::val_str(String *str)
+{
+  str= Item_func_group_concat::val_str(str);
+  String s;
+  s.append('[');
+  s.swap(*str);
+  str->append(s);
+  str->append(']');
+
+  return str;
+}
+
+
+Item_func_json_objectagg::
+Item_func_json_objectagg(THD *thd, Item_func_json_objectagg *item)
+  :Item_sum(thd, item)
+{
+  result.set_charset(collation.collation);
+  result.append("{");
+}
+
+
+bool
+Item_func_json_objectagg::fix_fields(THD *thd, Item **ref)
+{
+  uint i;                       /* for loop variable */
+  DBUG_ASSERT(fixed == 0);
+
+  if (init_sum_func_check(thd))
+    return TRUE;
+
+  maybe_null= 1;
+
+  /*
+    Fix fields for select list and ORDER clause
+  */
+
+  for (i=0 ; i < arg_count ; i++)
+  {
+    if (args[i]->fix_fields_if_needed_for_scalar(thd, &args[i]))
+      return TRUE;
+    m_with_subquery|= args[i]->with_subquery();
+    with_param|= args[i]->with_param;
+    with_window_func|= args[i]->with_window_func;
+  }
+
+  /* skip charset aggregation for order columns */
+  if (agg_arg_charsets_for_string_result(collation, args, arg_count))
+    return 1;
+
+  result.set_charset(collation.collation);
+  result_field= 0;
+  null_value= 1;
+  max_length= (uint32)(thd->variables.group_concat_max_len
+              / collation.collation->mbminlen
+              * collation.collation->mbmaxlen);
+
+  if (check_sum_func(thd, ref))
+    return TRUE;
+
+  fixed= 1;
+  return FALSE;
+}
+
+
+void Item_func_json_objectagg::cleanup()
+{
+  DBUG_ENTER("Item_func_json_objectagg::cleanup");
+  Item_sum::cleanup();
+
+  result.length(1);
+  DBUG_VOID_RETURN;
+}
+
+
+Item *Item_func_json_objectagg::copy_or_same(THD* thd)
+{
+  return new (thd->mem_root) Item_func_json_objectagg(thd, this);
+}
+
+
+void Item_func_json_objectagg::clear()
+{
+  result.length(1);
+  null_value= 1;
+}
+
+
+bool Item_func_json_objectagg::add()
+{
+  StringBuffer<MAX_FIELD_WIDTH> buf;
+  String *key;
+
+  key= args[0]->val_str(&buf);
+  if (args[0]->is_null())
+    return 0;
+
+  null_value= 0;
+  if (result.length() > 1)
+    result.append(", ");
+
+  result.append("\"");
+  result.append(*key);
+  result.append("\":");
+
+  buf.length(0);
+  append_json_value(&result, args[1], &buf);
+
+  return 0;
+}
+
+
+String* Item_func_json_objectagg::val_str(String* str)
+{
+  DBUG_ASSERT(fixed == 1);
+  if (null_value)
+    return 0;
+
+  result.append("}");
+  return &result;
+}
+
+
+void Item_func_json_objectagg::print(String *str, enum_query_type query_type)
+{
+}
+
