@@ -112,17 +112,7 @@ extern "C" my_bool wsrep_get_debug()
 
 extern "C" my_bool wsrep_thd_is_local(const THD *thd)
 {
-  /*
-    async replication IO and background threads have nothing to replicate in the cluster,
-    marking them as non-local here to prevent write set population and replication
-
-    async replication SQL thread, applies client transactions from mariadb master
-    and will be replicated into cluster
-   */
-  return (
-          thd->system_thread != SYSTEM_THREAD_SLAVE_BACKGROUND &&
-          thd->system_thread != SYSTEM_THREAD_SLAVE_IO &&
-          thd->wsrep_cs().mode() == wsrep::client_state::m_local);
+  return thd->wsrep_cs().mode() == wsrep::client_state::m_local;
 }
 
 extern "C" my_bool wsrep_thd_is_applying(const THD *thd)
@@ -168,7 +158,6 @@ extern "C" void wsrep_handle_SR_rollback(THD *bf_thd,
                                          THD *victim_thd)
 {
   DBUG_ASSERT(victim_thd);
-  DBUG_ASSERT(wsrep_thd_is_SR(victim_thd));
   if (!victim_thd || !wsrep_on(bf_thd)) return;
 
   WSREP_DEBUG("handle rollback, for deadlock: thd %llu trx_id %" PRIu64 " frags %zu conf %s",
@@ -197,47 +186,25 @@ extern "C" void wsrep_handle_SR_rollback(THD *bf_thd,
   }
 }
 
-extern "C" my_bool wsrep_thd_bf_abort(THD *bf_thd, THD *victim_thd,
+extern "C" my_bool wsrep_thd_bf_abort(const THD *bf_thd, THD *victim_thd,
                                       my_bool signal)
 {
-  DBUG_EXECUTE_IF("sync.before_wsrep_thd_abort",
-                 {
-                   const char act[]=
-                     "now "
-                     "SIGNAL sync.before_wsrep_thd_abort_reached "
-                     "WAIT_FOR signal.before_wsrep_thd_abort";
-                   DBUG_ASSERT(!debug_sync_set_action(bf_thd,
-                                                      STRING_WITH_LEN(act)));
-                 };);
-
+  /* Note: do not store/reset globals before wsrep_bf_abort() call
+     to avoid losing BF thd context. */
+  if (WSREP(victim_thd) && !victim_thd->wsrep_trx().active())
+  {
+    WSREP_DEBUG("BF abort for non active transaction");
+    wsrep_start_transaction(victim_thd, victim_thd->wsrep_next_trx_id());
+  }
   my_bool ret= wsrep_bf_abort(bf_thd, victim_thd);
+  wsrep_store_threadvars((THD*)bf_thd);
   /*
     Send awake signal if victim was BF aborted or does not
     have wsrep on. Note that this should never interrupt RSU
     as RSU has paused the provider.
    */
   if ((ret || !wsrep_on(victim_thd)) && signal)
-  {
-    mysql_mutex_assert_not_owner(&victim_thd->LOCK_thd_data);
-    mysql_mutex_assert_not_owner(&victim_thd->LOCK_thd_kill);
-    mysql_mutex_lock(&victim_thd->LOCK_thd_data);
-
-    if (victim_thd->wsrep_aborter && victim_thd->wsrep_aborter != bf_thd->thread_id)
-    {
-      WSREP_DEBUG("victim is killed already by %llu, skipping awake",
-                  victim_thd->wsrep_aborter);
-      mysql_mutex_unlock(&victim_thd->LOCK_thd_data);
-      return false;
-    }
-
-    mysql_mutex_lock(&victim_thd->LOCK_thd_kill);
-    victim_thd->wsrep_aborter= bf_thd->thread_id;
-    victim_thd->awake_no_mutex(KILL_QUERY);
-    mysql_mutex_unlock(&victim_thd->LOCK_thd_kill);
-    mysql_mutex_unlock(&victim_thd->LOCK_thd_data);
-  } else {
-    WSREP_DEBUG("wsrep_thd_bf_abort skipped awake");
-  }
+    victim_thd->awake(KILL_QUERY);
   return ret;
 }
 
@@ -321,42 +288,16 @@ extern "C" void wsrep_commit_ordered(THD *thd)
       thd->wsrep_trx().state() == wsrep::transaction::s_committing &&
       !wsrep_commit_will_write_binlog(thd))
   {
-    DEBUG_SYNC(thd, "before_wsrep_ordered_commit");
     thd->wsrep_cs().ordered_commit();
   }
 }
 
-extern "C" bool wsrep_thd_set_wsrep_aborter(THD *bf_thd, THD *victim_thd)
+extern "C" my_bool wsrep_thd_has_ignored_error(const THD *thd)
 {
-  WSREP_DEBUG("wsrep_thd_set_wsrep_aborter called");
-  mysql_mutex_assert_owner(&victim_thd->LOCK_thd_data);
-  if (victim_thd->wsrep_aborter && victim_thd->wsrep_aborter != bf_thd->thread_id)
-  {
-    return true;
-  }
-  victim_thd->wsrep_aborter = bf_thd->thread_id;
-  return false;
+  return thd->wsrep_has_ignored_error;
 }
 
-extern "C" void wsrep_report_bf_lock_wait(const THD *thd,
-                                          unsigned long long trx_id)
+extern "C" void wsrep_thd_set_ignored_error(THD *thd, my_bool val)
 {
-  if (thd)
-  {
-    WSREP_ERROR("Thread %s trx_id: %llu thread: %ld "
-                "seqno: %lld client_state: %s client_mode: %s transaction_mode: %s "
-                "applier: %d toi: %d local: %d "
-                "query: %s",
-                wsrep_thd_is_BF(thd, false) ? "BF" : "normal",
-                trx_id,
-                thd_get_thread_id(thd),
-                wsrep_thd_trx_seqno(thd),
-                wsrep_thd_client_state_str(thd),
-                wsrep_thd_client_mode_str(thd),
-                wsrep_thd_transaction_state_str(thd),
-                wsrep_thd_is_applying(thd),
-                wsrep_thd_is_toi(thd),
-                wsrep_thd_is_local(thd),
-                wsrep_thd_query(thd));
-  }
+  thd->wsrep_has_ignored_error= val;
 }
