@@ -175,7 +175,7 @@ static int prepare_for_repair(THD *thd, TABLE_LIST *table_list,
   /*
     Check if this is a table type that stores index and data separately,
     like ISAM or MyISAM. We assume fixed order of engine file name
-    extensions array. First element of engine file name extensions array
+    extentions array. First element of engine file name extentions array
     is meta/index file extention. Second element - data file extention. 
   */
   ext= table->file->bas_ext();
@@ -768,20 +768,35 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
       {
         compl_result_code= result_code= HA_ADMIN_INVALID;
       }
-
-      /*
-        The check for ALTER_PARTITION_ADMIN implements this logic:
-        do not collect EITS STATS for this syntax:
-          ALTER TABLE ... ANALYZE PARTITION p
-        EITS statistics is global (not per-partition). Collecting global stats
-        is much more expensive processing just one partition, so the most
-        appropriate action is to just not collect EITS stats for this command.
-      */
       collect_eis=
         (table->table->s->table_category == TABLE_CATEGORY_USER &&
-        !(lex->alter_info.flags & ALTER_PARTITION_ADMIN) &&
          (check_eits_collection_allowed(thd) ||
           lex->with_persistent_for_clause));
+
+
+      if (!lex->index_list)
+      {
+        tab->keys_in_use_for_query.init(tab->s->keys);
+      }
+      else
+      {
+        int pos;
+        LEX_STRING *index_name;
+        List_iterator_fast<LEX_STRING> it(*lex->index_list);
+   
+        tab->keys_in_use_for_query.clear_all();  
+        while ((index_name= it++))
+	{
+          if (tab->s->keynames.type_names == 0 ||
+              (pos= find_type(&tab->s->keynames, index_name->str,
+                              index_name->length, 1)) <= 0)
+          {
+            compl_result_code= result_code= HA_ADMIN_INVALID;
+            break;
+          }
+          tab->keys_in_use_for_query.set_bit(--pos);
+        }  
+      }
     }
 
     if (result_code == HA_ADMIN_OK)
@@ -866,27 +881,6 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
           }
           tab->file->column_bitmaps_signal();
         }
-        if (!lex->index_list)
-          tab->keys_in_use_for_query.init(tab->s->keys);
-        else
-        {
-          int pos;
-          LEX_STRING *index_name;
-          List_iterator_fast<LEX_STRING> it(*lex->index_list);
-
-          tab->keys_in_use_for_query.clear_all();
-          while ((index_name= it++))
-          {
-            if (tab->s->keynames.type_names == 0 ||
-                (pos= find_type(&tab->s->keynames, index_name->str,
-                                index_name->length, 1)) <= 0)
-            {
-              compl_result_code= result_code= HA_ADMIN_INVALID;
-              break;
-            }
-            tab->keys_in_use_for_query.set_bit(--pos);
-          }
-        }
         if (!(compl_result_code=
               alloc_statistics_for_table(thd, table->table)) &&
             !(compl_result_code=
@@ -1100,7 +1094,7 @@ send_result_message:
         }
         /* Make sure this table instance is not reused after the operation. */
         if (table->table)
-          table->table->mark_table_for_reopen();
+          table->table->m_needs_reopen= true;
       }
       result_code= result_code ? HA_ADMIN_FAILED : HA_ADMIN_OK;
       table->next_local= save_next_local;
@@ -1225,7 +1219,7 @@ err:
     trans_rollback(thd);
   if (table && table->table)
   {
-    table->table->mark_table_for_reopen();
+    table->table->m_needs_reopen= true;
     table->table= 0;
   }
   close_thread_tables(thd);			// Shouldn't be needed
