@@ -972,7 +972,7 @@ With_element::rename_columns_of_derived_unit(THD *thd,
     /* Rename the columns of the first select in the unit */
     while ((item= it++, name= nm++))
     {
-      item->set_name(thd, name->str, (uint) name->length, system_charset_info);
+      item->set_name(thd, *name);
       item->is_autogenerated_name= false;
     }
 
@@ -1123,7 +1123,6 @@ bool TABLE_LIST::set_as_with_table(THD *thd, With_element *with_elem)
   {
     derived= with_elem->spec;
     if (derived != select_lex->master_unit() &&
-        !with_elem->is_recursive &&
         !is_with_table_recursive_reference())
     {
        derived->move_as_slave(select_lex);
