@@ -1811,10 +1811,11 @@ void Explain_table_access::print_explain_json(Explain_query *query,
       double total_time= op_tracker.get_time_ms();
       if (rowid_filter)
         total_time+= rowid_filter->tracker->get_time_fill_container_ms();
-      writer->add_member("r_total_time_ms").add_double(total_time);
+      writer->add_member("r_table_time_ms").add_double(total_time);
+      writer->add_member("r_other_time_ms").add_double(extra_time_tracker.get_time_ms());
     }
   }
-  
+
   /* `filtered` */
   if (filtered_set)
     writer->add_member("filtered").add_double(filtered);
@@ -1883,11 +1884,6 @@ void Explain_table_access::print_explain_json(Explain_query *query,
     /* This is a derived table. Print its contents here */
     writer->add_member("materialized").start_object();
     Explain_node *node= query->get_node(derived_select_number);
-    if (node->get_type() == Explain_node::EXPLAIN_SELECT &&
-        ((Explain_select*)node)->is_lateral)
-    {
-      writer->add_member("lateral").add_ll(1);
-    }
     node->print_explain_json(query, writer, is_analyze);
     writer->end_object();
   }
