@@ -48,20 +48,20 @@ struct Field_definition
 
 static Field_definition sequence_structure[]=
 {
-  {"next_not_cached_value", 21, &type_handler_longlong,
+  {"next_not_cached_value", 21, &type_handler_slonglong,
    {STRING_WITH_LEN("")}, FL},
-  {"minimum_value", 21, &type_handler_longlong, {STRING_WITH_LEN("")}, FL},
-  {"maximum_value", 21, &type_handler_longlong, {STRING_WITH_LEN("")}, FL},
-  {"start_value", 21, &type_handler_longlong, {STRING_WITH_LEN("start value when sequences is created or value if RESTART is used")},  FL},
-  {"increment", 21, &type_handler_longlong,
+  {"minimum_value", 21, &type_handler_slonglong, {STRING_WITH_LEN("")}, FL},
+  {"maximum_value", 21, &type_handler_slonglong, {STRING_WITH_LEN("")}, FL},
+  {"start_value", 21, &type_handler_slonglong, {STRING_WITH_LEN("start value when sequences is created or value if RESTART is used")},  FL},
+  {"increment", 21, &type_handler_slonglong,
    {STRING_WITH_LEN("increment value")}, FL},
-  {"cache_size", 21, &type_handler_longlong, {STRING_WITH_LEN("")},
+  {"cache_size", 21, &type_handler_ulonglong, {STRING_WITH_LEN("")},
    FL | UNSIGNED_FLAG},
-  {"cycle_option", 1, &type_handler_tiny, {STRING_WITH_LEN("0 if no cycles are allowed, 1 if the sequence should begin a new cycle when maximum_value is passed")},
+  {"cycle_option", 1, &type_handler_utiny, {STRING_WITH_LEN("0 if no cycles are allowed, 1 if the sequence should begin a new cycle when maximum_value is passed")},
    FL | UNSIGNED_FLAG },
-  {"cycle_count", 21, &type_handler_longlong,
+  {"cycle_count", 21, &type_handler_slonglong,
    {STRING_WITH_LEN("How many cycles have been done")}, FL},
-  {NULL, 0, &type_handler_longlong, {STRING_WITH_LEN("")}, 0}
+  {NULL, 0, &type_handler_slonglong, {STRING_WITH_LEN("")}, 0}
 };
 
 #undef FL
@@ -203,16 +203,6 @@ bool check_sequence_fields(LEX *lex, List<Create_field> *fields)
     reason= "Sequence tables cannot have any keys";
     goto err;
   }
-  if (lex->alter_info.check_constraint_list.elements > 0)
-  {
-    reason= "Sequence tables cannot have any constraints";
-    goto err;
-  }
-  if (lex->alter_info.flags & ALTER_ORDER)
-  {
-    reason= "ORDER BY";
-    goto err;
-  }
 
   for (field_no= 0; (field= it++); field_no++)
   {
@@ -220,8 +210,7 @@ bool check_sequence_fields(LEX *lex, List<Create_field> *fields)
     if (my_strcasecmp(system_charset_info, field_def->field_name,
                       field->field_name.str) ||
         field->flags != field_def->flags ||
-        field->type_handler() != field_def->type_handler ||
-        field->check_constraint || field->vcol_info)
+        field->type_handler() != field_def->type_handler)
     {
       reason= field->field_name.str;
       goto err;
@@ -231,8 +220,8 @@ bool check_sequence_fields(LEX *lex, List<Create_field> *fields)
 
 err:
   my_error(ER_SEQUENCE_INVALID_TABLE_STRUCTURE, MYF(0),
-           lex->select_lex.table_list.first->db.str,
-           lex->select_lex.table_list.first->table_name.str, reason);
+           lex->first_select_lex()->table_list.first->db.str,
+           lex->first_select_lex()->table_list.first->table_name.str, reason);
   DBUG_RETURN(TRUE);
 }
 
@@ -366,10 +355,8 @@ bool sequence_insert(THD *thd, LEX *lex, TABLE_LIST *org_table_list)
   seq->reserved_until= seq->start;
   error= seq->write_initial_sequence(table);
 
-  if (trans_commit_stmt(thd))
-    error= 1;
-  if (trans_commit_implicit(thd))
-    error= 1;
+  trans_commit_stmt(thd);
+  trans_commit_implicit(thd);
 
   if (!temporary_table)
   {
@@ -485,10 +472,6 @@ int SEQUENCE::read_initial_values(TABLE *table)
       if (mdl_lock_used)
         thd->mdl_context.release_lock(mdl_request.ticket);
       write_unlock(table);
-
-      if (!has_active_transaction && !thd->transaction.stmt.is_empty() &&
-          !thd->in_sub_stmt)
-        trans_commit_stmt(thd);
       DBUG_RETURN(HA_ERR_LOCK_WAIT_TIMEOUT);
     }
     DBUG_ASSERT(table->reginfo.lock_type == TL_READ);
@@ -504,12 +487,9 @@ int SEQUENCE::read_initial_values(TABLE *table)
       Doing mysql_lock_tables() may have started a read only transaction.
       If that happend, it's better that we commit it now, as a lot of
       code assumes that there is no active stmt transaction directly after
-      open_tables().
-      But we also don't want to commit the stmt transaction while in a
-      substatement, see MDEV-15977.
+      open_tables()
     */
-    if (!has_active_transaction && !thd->transaction.stmt.is_empty() &&
-        !thd->in_sub_stmt)
+    if (!has_active_transaction && !thd->transaction.stmt.is_empty())
       trans_commit_stmt(thd);
   }
   write_unlock(table);
