@@ -1,5 +1,5 @@
 /* Copyright (c) 2000, 2010, Oracle and/or its affiliates.
-   Copyright (c) 2009, 2020, MariaDB Corporation.
+   Copyright (c) 2009, 2013 Monty Program Ab.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -17,7 +17,8 @@
 
 /* Functions to handle date and time */
 
-#include <my_global.h>
+#include "mariadb.h"
+#include "sql_priv.h"
 #include "sql_time.h"
 #include "tztime.h"                      // struct Time_zone
 #include "sql_class.h"                   // THD
@@ -36,34 +37,108 @@
   Order of elements in 'interval_type_to_name' should correspond to 
   the order of elements in 'interval_type' enum
   
-  See also interval_type, interval_names
+  See also interval_type, interval_names, append_interval
 */
 
-LEX_STRING interval_type_to_name[INTERVAL_LAST] = {
-  { C_STRING_WITH_LEN("YEAR")},
-  { C_STRING_WITH_LEN("QUARTER")},
-  { C_STRING_WITH_LEN("MONTH")},
-  { C_STRING_WITH_LEN("WEEK")},
-  { C_STRING_WITH_LEN("DAY")},
-  { C_STRING_WITH_LEN("HOUR")},
-  { C_STRING_WITH_LEN("MINUTE")},
-  { C_STRING_WITH_LEN("SECOND")},
-  { C_STRING_WITH_LEN("MICROSECOND")},
-  { C_STRING_WITH_LEN("YEAR_MONTH")},
-  { C_STRING_WITH_LEN("DAY_HOUR")},
-  { C_STRING_WITH_LEN("DAY_MINUTE")},
-  { C_STRING_WITH_LEN("DAY_SECOND")},
-  { C_STRING_WITH_LEN("HOUR_MINUTE")},
-  { C_STRING_WITH_LEN("HOUR_SECOND")},
-  { C_STRING_WITH_LEN("MINUTE_SECOND")},
-  { C_STRING_WITH_LEN("DAY_MICROSECOND")},
-  { C_STRING_WITH_LEN("HOUR_MICROSECOND")},
-  { C_STRING_WITH_LEN("MINUTE_MICROSECOND")},
-  { C_STRING_WITH_LEN("SECOND_MICROSECOND")}
-}; 
+LEX_CSTRING interval_type_to_name[INTERVAL_LAST] = {
+  { STRING_WITH_LEN("YEAR")},
+  { STRING_WITH_LEN("QUARTER")},
+  { STRING_WITH_LEN("MONTH")},
+  { STRING_WITH_LEN("WEEK")},
+  { STRING_WITH_LEN("DAY")},
+  { STRING_WITH_LEN("HOUR")},
+  { STRING_WITH_LEN("MINUTE")},
+  { STRING_WITH_LEN("SECOND")},
+  { STRING_WITH_LEN("MICROSECOND")},
+  { STRING_WITH_LEN("YEAR_MONTH")},
+  { STRING_WITH_LEN("DAY_HOUR")},
+  { STRING_WITH_LEN("DAY_MINUTE")},
+  { STRING_WITH_LEN("DAY_SECOND")},
+  { STRING_WITH_LEN("HOUR_MINUTE")},
+  { STRING_WITH_LEN("HOUR_SECOND")},
+  { STRING_WITH_LEN("MINUTE_SECOND")},
+  { STRING_WITH_LEN("DAY_MICROSECOND")},
+  { STRING_WITH_LEN("HOUR_MICROSECOND")},
+  { STRING_WITH_LEN("MINUTE_MICROSECOND")},
+  { STRING_WITH_LEN("SECOND_MICROSECOND")}
+};
 
-	/* Calc weekday from daynr */
-	/* Returns 0 for monday, 1 for tuesday .... */
+int append_interval(String *str, interval_type int_type, const INTERVAL &interval)
+{
+  char buf[64];
+  size_t len;
+  switch (int_type) {
+  case INTERVAL_YEAR:
+    len= my_snprintf(buf,sizeof(buf),"%u", interval.year);
+    break;
+  case INTERVAL_QUARTER:
+  case INTERVAL_MONTH:
+    len= my_snprintf(buf,sizeof(buf),"%u", interval.month);
+    int_type=INTERVAL_MONTH;
+    break;
+  case INTERVAL_WEEK:
+  case INTERVAL_DAY:
+    len= my_snprintf(buf,sizeof(buf),"%u", interval.day);
+    int_type=INTERVAL_DAY;
+    break;
+  case INTERVAL_HOUR:
+    len= my_snprintf(buf,sizeof(buf),"%u", interval.hour);
+    break;
+  case INTERVAL_MINUTE:
+    len= my_snprintf(buf,sizeof(buf),"%u", interval.minute);
+    break;
+  case INTERVAL_SECOND:
+    len= my_snprintf(buf,sizeof(buf),"%u", interval.second);
+    break;
+  case INTERVAL_MICROSECOND:
+    len= my_snprintf(buf,sizeof(buf),"%u", interval.second_part);
+    break;
+  case INTERVAL_YEAR_MONTH:
+    len= my_snprintf(buf,sizeof(buf),"%u-%02u", interval.day, interval.month);
+    break;
+  case INTERVAL_DAY_HOUR:
+    len= my_snprintf(buf,sizeof(buf),"%u %u", interval.day, interval.hour);
+    break;
+  case INTERVAL_DAY_MINUTE:
+    len= my_snprintf(buf,sizeof(buf),"%u %u:%02u", interval.day, interval.hour, interval.minute);
+    break;
+  case INTERVAL_DAY_SECOND:
+    len= my_snprintf(buf,sizeof(buf),"%u %u:%02u:%02u", interval.day, interval.hour, interval.minute, interval.second);
+    break;
+  case INTERVAL_HOUR_MINUTE:
+    len= my_snprintf(buf,sizeof(buf),"%u:%02u", interval.hour, interval.minute);
+    break;
+  case INTERVAL_HOUR_SECOND:
+    len= my_snprintf(buf,sizeof(buf),"%u:%02u:%02u", interval.hour, interval.minute, interval.second);
+    break;
+  case INTERVAL_MINUTE_SECOND:
+    len= my_snprintf(buf,sizeof(buf),"%u:%02u", interval.minute, interval.second);
+    break;
+  case INTERVAL_DAY_MICROSECOND:
+    len= my_snprintf(buf,sizeof(buf),"%u %u:%02u:%02u.%06u", interval.day, interval.hour, interval.minute, interval.second, interval.second_part);
+    break;
+  case INTERVAL_HOUR_MICROSECOND:
+    len= my_snprintf(buf,sizeof(buf),"%u:%02u:%02u.%06u", interval.hour, interval.minute, interval.second, interval.second_part);
+    break;
+  case INTERVAL_MINUTE_MICROSECOND:
+    len= my_snprintf(buf,sizeof(buf),"%u:%02u.%06u", interval.minute, interval.second, interval.second_part);
+    break;
+  case INTERVAL_SECOND_MICROSECOND:
+    len= my_snprintf(buf,sizeof(buf),"%u.%06u", interval.second, interval.second_part);
+    break;
+  default:
+    DBUG_ASSERT(0);
+    len= 0;
+  }
+  return str->append(buf, len) || str->append(' ') ||
+         str->append(interval_type_to_name + int_type);
+}
+
+
+/*
+  Calc weekday from daynr
+  Returns 0 for monday, 1 for tuesday ...
+*/
 
 int calc_weekday(long daynr,bool sunday_first_day_of_week)
 {
@@ -100,7 +175,7 @@ int calc_weekday(long daynr,bool sunday_first_day_of_week)
 	next week is week 1.
 */
 
-uint calc_week(MYSQL_TIME *l_time, uint week_behaviour, uint *year)
+uint calc_week(const MYSQL_TIME *l_time, uint week_behaviour, uint *year)
 {
   uint days;
   ulong daynr=calc_daynr(l_time->year,l_time->month,l_time->day);
@@ -214,15 +289,15 @@ ulong convert_month_to_period(ulong month)
 
 
 bool
-check_date_with_warn(const MYSQL_TIME *ltime, ulonglong fuzzy_date,
-                     timestamp_type ts_type)
+check_date_with_warn(THD *thd, const MYSQL_TIME *ltime,
+                     date_conv_mode_t fuzzydate, timestamp_type ts_type)
 {
   int unused;
-  if (check_date(ltime, fuzzy_date, &unused))
+  if (check_date(ltime, fuzzydate, &unused))
   {
     ErrConvTime str(ltime);
-    make_truncated_value_warning(current_thd, Sql_condition::WARN_LEVEL_WARN,
-                                 &str, ts_type, NULL, NULL, NULL);
+    make_truncated_value_warning(thd, Sql_condition::WARN_LEVEL_WARN,
+                                 &str, ts_type, 0, 0);
     return true;
   }
   return false;
@@ -230,7 +305,7 @@ check_date_with_warn(const MYSQL_TIME *ltime, ulonglong fuzzy_date,
 
 
 bool
-adjust_time_range_with_warn(MYSQL_TIME *ltime, uint dec)
+adjust_time_range_with_warn(THD *thd, MYSQL_TIME *ltime, uint dec)
 {
   MYSQL_TIME copy= *ltime;
   ErrConvTime str(&copy);
@@ -238,8 +313,7 @@ adjust_time_range_with_warn(MYSQL_TIME *ltime, uint dec)
   if (check_time_range(ltime, dec, &warnings))
     return true;
   if (warnings)
-    make_truncated_value_warning(current_thd, Sql_condition::WARN_LEVEL_WARN,
-                                 &str, MYSQL_TIMESTAMP_TIME, NULL, NULL, NULL);
+    thd->push_warning_truncated_wrong_value("time", str.ptr());
   return false;
 }
 
@@ -255,8 +329,8 @@ adjust_time_range_with_warn(MYSQL_TIME *ltime, uint dec)
 */
 static uint
 to_ascii(CHARSET_INFO *cs,
-         const char *src, uint src_length,
-         char *dst, uint dst_length)
+         const char *src, size_t src_length,
+         char *dst, size_t dst_length)
                      
 {
   int cnvres;
@@ -277,33 +351,70 @@ to_ascii(CHARSET_INFO *cs,
 }
 
 
-/* Character set-aware version of str_to_time() */
-bool
-str_to_time(CHARSET_INFO *cs, const char *str,uint length,
-            MYSQL_TIME *l_time, ulonglong fuzzydate, MYSQL_TIME_STATUS *status)
+class TemporalAsciiBuffer: public LEX_CSTRING
 {
   char cnv[32];
-  if ((cs->state & MY_CS_NONASCII) != 0)
+public:
+  TemporalAsciiBuffer(const char *str, size_t length, CHARSET_INFO *cs)
   {
-    length= to_ascii(cs, str, length, cnv, sizeof(cnv));
-    str= cnv;
+    if ((cs->state & MY_CS_NONASCII) != 0)
+    {
+      LEX_CSTRING::str= cnv;
+      LEX_CSTRING::length= to_ascii(cs, str, length, cnv, sizeof(cnv));
+    }
+    else
+    {
+      LEX_CSTRING::str= str;
+      LEX_CSTRING::length= length;
+    }
   }
-  return str_to_time(str, length, l_time, fuzzydate, status);
+};
+
+
+/* Character set-aware version of ascii_to_datetime_or_date_or_time() */
+bool Temporal::str_to_datetime_or_date_or_time(THD *thd, MYSQL_TIME_STATUS *st,
+                                               const char *str, size_t length,
+                                               CHARSET_INFO *cs,
+                                               date_mode_t fuzzydate)
+{
+  TemporalAsciiBuffer tmp(str, length, cs);
+  return ascii_to_datetime_or_date_or_time(st, tmp.str, tmp.length, fuzzydate)||
+         add_nanoseconds(thd, &st->warnings, fuzzydate, st->nanoseconds);
 }
 
 
-/* Character set-aware version of str_to_datetime() */
-bool str_to_datetime(CHARSET_INFO *cs, const char *str, uint length,
-                     MYSQL_TIME *l_time, ulonglong flags,
-                     MYSQL_TIME_STATUS *status)
+/* Character set-aware version of str_to_datetime_or_date() */
+bool Temporal::str_to_datetime_or_date(THD *thd, MYSQL_TIME_STATUS *status,
+                                       const char *str, size_t length,
+                                       CHARSET_INFO *cs,
+                                       date_mode_t flags)
 {
-  char cnv[32];
-  if ((cs->state & MY_CS_NONASCII) != 0)
-  {
-    length= to_ascii(cs, str, length, cnv, sizeof(cnv));
-    str= cnv;
-  }
-  return str_to_datetime(str, length, l_time, flags, status);
+  TemporalAsciiBuffer tmp(str, length, cs);
+  return ascii_to_datetime_or_date(status, tmp.str, tmp.length, flags) ||
+         add_nanoseconds(thd, &status->warnings, flags, status->nanoseconds);
+}
+
+
+/* Character set-aware version of ascii_to_temporal() */
+bool Temporal::str_to_temporal(THD *thd, MYSQL_TIME_STATUS *status,
+                               const char *str, size_t length, CHARSET_INFO *cs,
+                               date_mode_t flags)
+{
+  TemporalAsciiBuffer tmp(str, length, cs);
+  return ascii_to_temporal(status, tmp.str, tmp.length, flags) ||
+         add_nanoseconds(thd, &status->warnings, flags, status->nanoseconds);
+}
+
+
+/* Character set-aware version of str_to_DDhhmmssff() */
+bool Interval_DDhhmmssff::str_to_DDhhmmssff(MYSQL_TIME_STATUS *status,
+                                            const char *str, size_t length,
+                                            CHARSET_INFO *cs, ulong max_hour)
+{
+  TemporalAsciiBuffer tmp(str, length, cs);
+  bool rc= ::str_to_DDhhmmssff(tmp.str, tmp.length, this, UINT_MAX32, status);
+  DBUG_ASSERT(status->warnings || !rc);
+  return rc;
 }
 
 
@@ -312,128 +423,53 @@ bool str_to_datetime(CHARSET_INFO *cs, const char *str, uint length,
   if string was truncated during conversion.
 
   NOTE
-    See description of str_to_datetime() for more information.
+    See description of str_to_datetime_xxx() for more information.
 */
 
 bool
-str_to_datetime_with_warn(CHARSET_INFO *cs,
-                          const char *str, uint length, MYSQL_TIME *l_time,
-                          ulonglong flags)
+str_to_datetime_with_warn(THD *thd, CHARSET_INFO *cs,
+                          const char *str, size_t length, MYSQL_TIME *to,
+                          date_mode_t mode)
 {
-  MYSQL_TIME_STATUS status;
-  THD *thd= current_thd;
-  bool ret_val= str_to_datetime(cs, str, length, l_time, flags, &status);
-  if (ret_val || status.warnings)
-    make_truncated_value_warning(thd,
-                                 ret_val ? Sql_condition::WARN_LEVEL_WARN :
-                                 Sql_condition::time_warn_level(status.warnings),
-                                 str, length, flags & TIME_TIME_ONLY ?
-                                 MYSQL_TIMESTAMP_TIME : l_time->time_type, 
-                                 NULL, NULL, NULL);
-  DBUG_EXECUTE_IF("str_to_datetime_warn",
-                  push_warning(thd, Sql_condition::WARN_LEVEL_NOTE,
-                               ER_YES, str););
-  return ret_val;
+  Temporal::Warn_push warn(thd, NULL, NullS, to, mode);
+  Temporal_hybrid *t= new(to) Temporal_hybrid(thd, &warn, str, length, cs, mode);
+  return !t->is_valid_temporal();
 }
 
 
-/**
-  converts a pair of numbers (integer part, microseconds) to MYSQL_TIME
-
-  @param neg           sign of the time value
-  @param nr            integer part of the number to convert
-  @param sec_part      microsecond part of the number
-  @param ltime         converted value will be written here
-  @param fuzzydate     conversion flags (TIME_INVALID_DATE, etc)
-  @param str           original number, as an ErrConv. For the warning
-  @param field_name    field name or NULL if not a field. For the warning
-  
-  @returns 0 for success, 1 for a failure
-*/
-static bool number_to_time_with_warn(bool neg, ulonglong nr, ulong sec_part,
-                                     MYSQL_TIME *ltime, ulonglong fuzzydate,
-                                     const ErrConv *str,
-                                     const TABLE_SHARE *s, const char *field_name)
-{
-  int was_cut;
-  longlong res;
-  enum_field_types f_type;
-  bool have_warnings= false;
-
-  if (fuzzydate & TIME_TIME_ONLY)
-  {
-    fuzzydate= TIME_TIME_ONLY; // clear other flags
-    f_type= MYSQL_TYPE_TIME;
-    res= number_to_time(neg, nr, sec_part, ltime, &was_cut);
-    have_warnings= MYSQL_TIME_WARN_HAVE_WARNINGS(was_cut);
-  }
-  else
-  {
-    f_type= MYSQL_TYPE_DATETIME;
-    if (neg)
-    {
-      res= -1;
-    }
-    else
-    {
-      res= number_to_datetime(nr, sec_part, ltime, fuzzydate, &was_cut);
-      have_warnings= was_cut && (fuzzydate & TIME_NO_ZERO_IN_DATE);
-    }
-  }
-
-  if (res < 0 || have_warnings)
-  {
-    make_truncated_value_warning(current_thd,
-                                 Sql_condition::WARN_LEVEL_WARN, str,
-                                 res < 0 ? MYSQL_TIMESTAMP_ERROR
-                                         : mysql_type_to_time_type(f_type),
-                                 s ? s->db.str : NULL,
-                                 s ? s->table_name.str : NULL, field_name);
-  }
-  return res < 0;
-}
-
-
-bool double_to_datetime_with_warn(double value, MYSQL_TIME *ltime,
-                                  ulonglong fuzzydate,
+bool double_to_datetime_with_warn(THD *thd, double value, MYSQL_TIME *ltime,
+                                  date_mode_t fuzzydate,
                                   const TABLE_SHARE *s, const char *field_name)
 {
-  const ErrConvDouble str(value);
-  bool neg= value < 0;
-
-  if (neg)
-    value= -value;
-
-  if (value > static_cast<double>(LONGLONG_MAX))
-    value= static_cast<double>(LONGLONG_MAX);
-
-  longlong nr= static_cast<ulonglong>(floor(value));
-  uint sec_part= static_cast<ulong>((value - floor(value))*TIME_SECOND_PART_FACTOR);
-  return number_to_time_with_warn(neg, nr, sec_part, ltime, fuzzydate, &str,
-                                  s, field_name);
+  Temporal::Warn_push warn(thd, s, field_name, ltime, fuzzydate);
+  Temporal_hybrid *t= new (ltime) Temporal_hybrid(thd, &warn, value, fuzzydate);
+  return !t->is_valid_temporal();
 }
 
 
-bool decimal_to_datetime_with_warn(const my_decimal *value, MYSQL_TIME *ltime,
-                                   ulonglong fuzzydate,
+bool decimal_to_datetime_with_warn(THD *thd, const my_decimal *value,
+                                   MYSQL_TIME *ltime,
+                                   date_mode_t fuzzydate,
                                    const TABLE_SHARE *s, const char *field_name)
 {
-  const ErrConvDecimal str(value);
-  ulonglong nr;
-  ulong sec_part;
-  bool neg= my_decimal2seconds(value, &nr, &sec_part);
-  return number_to_time_with_warn(neg, nr, sec_part, ltime, fuzzydate, &str,
-                                  s, field_name);
+  Temporal::Warn_push warn(thd, s, field_name, ltime, fuzzydate);
+  Temporal_hybrid *t= new (ltime) Temporal_hybrid(thd, &warn, value, fuzzydate);
+  return !t->is_valid_temporal();
 }
 
 
-bool int_to_datetime_with_warn(bool neg, ulonglong value, MYSQL_TIME *ltime,
-                               ulonglong fuzzydate,
+bool int_to_datetime_with_warn(THD *thd, const Longlong_hybrid &nr,
+                               MYSQL_TIME *ltime,
+                               date_mode_t fuzzydate,
                                const TABLE_SHARE *s, const char *field_name)
 {
-  const ErrConvInteger str(neg ? - (longlong) value : (longlong) value, !neg);
-  return number_to_time_with_warn(neg, value, 0, ltime,
-                                  fuzzydate, &str, s, field_name);
+  /*
+    Note: conversion from an integer to TIME can overflow to '838:59:59.999999',
+    so the conversion result can have fractional digits.
+  */
+  Temporal::Warn_push warn(thd, s, field_name, ltime, fuzzydate);
+  Temporal_hybrid *t= new (ltime) Temporal_hybrid(thd, &warn, nr, fuzzydate);
+  return !t->is_valid_temporal();
 }
 
 
@@ -479,7 +515,8 @@ void localtime_to_TIME(MYSQL_TIME *to, struct tm *from)
   to->second=   (int) from->tm_sec;
 }
 
-void calc_time_from_sec(MYSQL_TIME *to, long seconds, long microseconds)
+
+void calc_time_from_sec(MYSQL_TIME *to, ulong seconds, ulong microseconds)
 {
   long t_seconds;
   // to->neg is not cleared, it may already be set to a useful value
@@ -762,7 +799,7 @@ DATE_TIME_FORMAT
       !parse_date_time_format(format_type, format_str,
 			      format_length, &tmp))
   {
-    tmp.format.str=    (char*) format_str;
+    tmp.format.str=    format_str;
     tmp.format.length= format_length;
     return date_time_format_copy((THD *)0, &tmp);
   }
@@ -789,7 +826,8 @@ DATE_TIME_FORMAT
 DATE_TIME_FORMAT *date_time_format_copy(THD *thd, DATE_TIME_FORMAT *format)
 {
   DATE_TIME_FORMAT *new_format;
-  ulong length= sizeof(*format) + format->format.length + 1;
+  size_t length= sizeof(*format) + format->format.length + 1;
+  char *format_pos;
 
   if (thd)
     new_format= (DATE_TIME_FORMAT *) thd->alloc(length);
@@ -798,14 +836,13 @@ DATE_TIME_FORMAT *date_time_format_copy(THD *thd, DATE_TIME_FORMAT *format)
   if (new_format)
   {
     /* Put format string after current pos */
-    new_format->format.str= (char*) (new_format+1);
+    new_format->format.str= format_pos= (char*) (new_format+1);
     memcpy((char*) new_format->positions, (char*) format->positions,
 	   sizeof(format->positions));
     new_format->time_separator= format->time_separator;
     /* We make the string null terminated for easy printf in SHOW VARIABLES */
-    memcpy((char*) new_format->format.str, format->format.str,
-	   format->format.length);
-    new_format->format.str[format->format.length]= 0;
+    memcpy(format_pos, format->format.str, format->format.length);
+    format_pos[format->format.length]= 0;
     new_format->format.length= format->format.length;
   }
   return new_format;
@@ -860,50 +897,12 @@ void make_truncated_value_warning(THD *thd,
                                   Sql_condition::enum_warning_level level,
                                   const ErrConv *sval,
 				  timestamp_type time_type,
-                                  const char *db_name, const char *table_name,
-                                  const char *field_name)
+                                  const TABLE_SHARE *s, const char *field_name)
 {
-  char warn_buff[MYSQL_ERRMSG_SIZE];
-  const char *type_str;
-  CHARSET_INFO *cs= &my_charset_latin1;
-
-  switch (time_type) {
-    case MYSQL_TIMESTAMP_DATE: 
-      type_str= "date";
-      break;
-    case MYSQL_TIMESTAMP_TIME:
-      type_str= "time";
-      break;
-    case MYSQL_TIMESTAMP_DATETIME:  // FALLTHROUGH
-    default:
-      type_str= "datetime";
-      break;
-  }
-  if (field_name)
-  {
-    if (!db_name)
-      db_name= "";
-    if (!table_name)
-      table_name= "";
-
-    cs->cset->snprintf(cs, warn_buff, sizeof(warn_buff),
-                       ER_THD(thd, ER_TRUNCATED_WRONG_VALUE_FOR_FIELD),
-                       type_str, sval->ptr(),
-                       db_name, table_name, field_name,
-                       (ulong) thd->get_stmt_da()->current_row_for_warning());
-  }
-  else
-  {
-    if (time_type > MYSQL_TIMESTAMP_ERROR)
-      cs->cset->snprintf(cs, warn_buff, sizeof(warn_buff),
-                         ER_THD(thd, ER_TRUNCATED_WRONG_VALUE),
-                         type_str, sval->ptr());
-    else
-      cs->cset->snprintf(cs, warn_buff, sizeof(warn_buff),
-                         ER_THD(thd, ER_WRONG_VALUE), type_str, sval->ptr());
-  }
-  push_warning(thd, level,
-               ER_TRUNCATED_WRONG_VALUE, warn_buff);
+  const char *type_str= Temporal::type_name_by_timestamp_type(time_type);
+  return thd->push_warning_wrong_or_truncated_value
+    (level, time_type <= MYSQL_TIMESTAMP_ERROR, type_str, sval->ptr(),
+     s, field_name);
 }
 
 
@@ -914,12 +913,12 @@ void make_truncated_value_warning(THD *thd,
                    (X)->second_part)
 #define GET_PART(X, N) X % N ## LL; X/= N ## LL
 
-bool date_add_interval(MYSQL_TIME *ltime, interval_type int_type,
-                       INTERVAL interval)
+bool date_add_interval(THD *thd, MYSQL_TIME *ltime, interval_type int_type,
+                       const INTERVAL &interval, bool push_warn)
 {
   long period, sign;
 
-  sign= (interval.neg == ltime->neg ? 1 : -1);
+  sign= (interval.neg == (bool)ltime->neg ? 1 : -1);
 
   switch (int_type) {
   case INTERVAL_SECOND:
@@ -1028,8 +1027,8 @@ bool date_add_interval(MYSQL_TIME *ltime, interval_type int_type,
     return 0;                                   // Ok
 
 invalid_date:
+  if (push_warn)
   {
-    THD *thd= current_thd;
     push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
                         ER_DATETIME_FUNCTION_OVERFLOW,
                         ER_THD(thd, ER_DATETIME_FUNCTION_OVERFLOW),
@@ -1069,7 +1068,7 @@ null_date:
 
 bool
 calc_time_diff(const MYSQL_TIME *l_time1, const MYSQL_TIME *l_time2,
-               int l_sign, longlong *seconds_out, long *microseconds_out)
+               int l_sign, ulonglong *seconds_out, ulong *microseconds_out)
 {
   long days;
   bool neg;
@@ -1096,10 +1095,10 @@ calc_time_diff(const MYSQL_TIME *l_time1, const MYSQL_TIME *l_time2,
   }
 
   microseconds= ((longlong)days * SECONDS_IN_24H +
-                 (longlong)(l_time1->hour*3600L +
+                 (longlong)(l_time1->hour*3600LL +
                             l_time1->minute*60L +
                             l_time1->second) -
-                 l_sign*(longlong)(l_time2->hour*3600L +
+                 l_sign*(longlong)(l_time2->hour*3600LL +
                                    l_time2->minute*60L +
                                    l_time2->second)) * 1000000LL +
                 (longlong)l_time1->second_part -
@@ -1111,17 +1110,17 @@ calc_time_diff(const MYSQL_TIME *l_time1, const MYSQL_TIME *l_time2,
     microseconds= -microseconds;
     neg= 1;
   }
-  *seconds_out= microseconds/1000000L;
-  *microseconds_out= (long) (microseconds%1000000L);
+  *seconds_out= (ulonglong) microseconds/1000000L;
+  *microseconds_out= (ulong) (microseconds%1000000L);
   return neg;
 }
 
 
 bool calc_time_diff(const MYSQL_TIME *l_time1, const MYSQL_TIME *l_time2,
-                    int l_sign, MYSQL_TIME *l_time3, ulonglong fuzzydate)
+                    int l_sign, MYSQL_TIME *l_time3, date_mode_t fuzzydate)
 {
-  longlong seconds;
-  long microseconds;
+  ulonglong seconds;
+  ulong microseconds;
   bzero((char *) l_time3, sizeof(*l_time3));
   l_time3->neg= calc_time_diff(l_time1, l_time2, l_sign,
 			       &seconds, &microseconds);
@@ -1140,7 +1139,7 @@ bool calc_time_diff(const MYSQL_TIME *l_time1, const MYSQL_TIME *l_time2,
     ("invalid" means > TIME_MAX_SECOND)
   */
   set_if_smaller(seconds, INT_MAX32);
-  calc_time_from_sec(l_time3, (long) seconds, microseconds);
+  calc_time_from_sec(l_time3, (ulong) seconds, microseconds);
   return ((fuzzydate & TIME_NO_ZERO_DATE) && (seconds == 0) &&
           (microseconds == 0));
 }
@@ -1197,56 +1196,6 @@ bool time_to_datetime(MYSQL_TIME *ltime)
 }
 
 
-/**
-  Return a valid DATE or DATETIME value from an arbitrary MYSQL_TIME.
-  If ltime is TIME, it's first converted to DATETIME.
-  If ts_type is DATE, hhmmss is set to zero.
-  The date part of the result is checked against fuzzy_date.
-
-  @param   ltime       The value to convert.
-  @param   fuzzy_date  Flags to check date.
-  @param   ts_type     The type to convert to.
-  @return  false on success, true of error (negative time).*/
-bool
-make_date_with_warn(MYSQL_TIME *ltime, ulonglong fuzzy_date,
-                    timestamp_type ts_type)
-{
-  DBUG_ASSERT(ts_type == MYSQL_TIMESTAMP_DATE ||
-              ts_type == MYSQL_TIMESTAMP_DATETIME);
-  if (ltime->time_type == MYSQL_TIMESTAMP_TIME && time_to_datetime(ltime))
-  {
-    /* e.g. negative time */
-    ErrConvTime str(ltime);
-    make_truncated_value_warning(current_thd, Sql_condition::WARN_LEVEL_WARN,
-                                 &str, ts_type, NULL, NULL, NULL);
-    return true;
-  }
-  if ((ltime->time_type= ts_type) == MYSQL_TIMESTAMP_DATE)
-    ltime->hour= ltime->minute= ltime->second= ltime->second_part= 0;
-  return check_date_with_warn(ltime, fuzzy_date, ts_type);
-}
-
-
-/*
-  Convert a TIME value to DAY-TIME interval, e.g. for extraction:
-    EXTRACT(DAY FROM x), EXTRACT(HOUR FROM x), etc.
-  Moves full days from ltime->hour to ltime->day.
-  Note, time_type is set to MYSQL_TIMESTAMP_NONE, to make sure that
-  the structure is not used for anything else other than extraction:
-  non-extraction TIME functions expect zero day value!
-*/
-void time_to_daytime_interval(MYSQL_TIME *ltime)
-{
-  DBUG_ASSERT(ltime->time_type == MYSQL_TIMESTAMP_TIME);
-  DBUG_ASSERT(ltime->year == 0);
-  DBUG_ASSERT(ltime->month == 0);
-  DBUG_ASSERT(ltime->day == 0);
-  ltime->day= ltime->hour / 24;
-  ltime->hour%= 24;
-  ltime->time_type= MYSQL_TIMESTAMP_NONE;
-}
-
-
 /*** Conversion from TIME to DATETIME ***/
 
 /*
@@ -1274,8 +1223,8 @@ mix_date_and_time_complex(MYSQL_TIME *ldate, const MYSQL_TIME *ltime)
 {
   DBUG_ASSERT(ldate->time_type == MYSQL_TIMESTAMP_DATE ||
               ldate->time_type == MYSQL_TIMESTAMP_DATETIME);
-  longlong seconds;
-  long days, useconds;
+  ulonglong seconds;
+  ulong days, useconds;
   int sign= ltime->neg ? 1 : -1;
   ldate->neg= calc_time_diff(ldate, ltime, sign, &seconds, &useconds);
 
@@ -1365,7 +1314,7 @@ time_to_datetime(THD *thd, const MYSQL_TIME *from, MYSQL_TIME *to)
 bool
 time_to_datetime_with_warn(THD *thd,
                            const MYSQL_TIME *from, MYSQL_TIME *to,
-                           ulonglong fuzzydate)
+                           date_conv_mode_t fuzzydate)
 {
   int warn= 0;
   DBUG_ASSERT(from->time_type == MYSQL_TIMESTAMP_TIME);
@@ -1381,29 +1330,51 @@ time_to_datetime_with_warn(THD *thd,
         check_date(to, fuzzydate, &warn)))
   {
     ErrConvTime str(from);
-    make_truncated_value_warning(thd, Sql_condition::WARN_LEVEL_WARN,
-                                 &str, MYSQL_TIMESTAMP_DATETIME, NULL, NULL, NULL); 
+    thd->push_warning_truncated_wrong_value("datetime", str.ptr());
     return true;
   }
   return false;
 }
 
 
-bool datetime_to_time_with_warn(THD *thd, const MYSQL_TIME *dt,
-                                MYSQL_TIME *tm, uint dec)
+longlong pack_time(const MYSQL_TIME *my_time)
 {
-  if (thd->variables.old_behavior & OLD_MODE_ZERO_DATE_TIME_CAST)
-  {
-    *tm= *dt;
-    datetime_to_time(tm);
-    return false;
+  return  ((((((my_time->year     * 13ULL +
+               my_time->month)    * 32ULL +
+               my_time->day)      * 24ULL +
+               my_time->hour)     * 60ULL +
+               my_time->minute)   * 60ULL +
+               my_time->second)   * 1000000ULL +
+               my_time->second_part) * (my_time->neg ? -1 : 1);
+}
+
+#define get_one(WHERE, FACTOR) WHERE= (ulong)(packed % FACTOR); packed/= FACTOR
+
+void unpack_time(longlong packed, MYSQL_TIME *my_time,
+                     enum_mysql_timestamp_type ts_type)
+{
+  if ((my_time->neg= packed < 0))
+    packed= -packed;
+  get_one(my_time->second_part, 1000000ULL);
+  get_one(my_time->second,           60U);
+  get_one(my_time->minute,           60U);
+  get_one(my_time->hour,             24U);
+  get_one(my_time->day,              32U);
+  get_one(my_time->month,            13U);
+  my_time->year= (uint)packed;
+  my_time->time_type= ts_type;
+  switch (ts_type) {
+  case MYSQL_TIMESTAMP_TIME:
+    my_time->hour+= (my_time->month * 32 + my_time->day) * 24;
+    my_time->month= my_time->day= 0;
+    break;
+  case MYSQL_TIMESTAMP_DATE:
+    my_time->hour= my_time->minute= my_time->second= my_time->second_part= 0;
+    break;
+  case MYSQL_TIMESTAMP_NONE:
+  case MYSQL_TIMESTAMP_ERROR:
+    DBUG_ASSERT(0);
+  case MYSQL_TIMESTAMP_DATETIME:
+    break;
   }
-  else /* new mode */
-  {
-    MYSQL_TIME current_date;
-    set_current_date(thd, &current_date);
-    calc_time_diff(dt, &current_date, 1, tm, 0);
-  }
-  int warnings= 0;
-  return check_time_range(tm, dec, &warnings);
 }
