@@ -1,4 +1,4 @@
-/* Copyright (C) 2012, 2020, MariaDB
+/* Copyright (C) 2012, 2019, MariaDB Corporation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -23,6 +23,8 @@
 #include <sql_audit.h>
 #include <debug_sync.h>
 #include <threadpool.h>
+#include <my_counter.h>
+
 #ifdef WITH_WSREP
 #include "wsrep_trans_observer.h"
 #endif /* WITH_WSREP */
@@ -38,6 +40,8 @@ uint threadpool_max_threads;
 uint threadpool_oversubscribe;
 uint threadpool_mode;
 uint threadpool_prio_kickup_timer;
+my_bool threadpool_exact_stats;
+my_bool threadpool_dedicated_listener;
 
 /* Stats */
 TP_STATISTICS tp_stats;
@@ -47,6 +51,7 @@ static void  threadpool_remove_connection(THD *thd);
 static int   threadpool_process_request(THD *thd);
 static THD*  threadpool_add_connection(CONNECT *connect, void *scheduler_data);
 
+extern "C" pthread_key(struct st_my_thread_var*, THR_KEY_mysys);
 extern bool do_command(THD*);
 
 static inline TP_connection *get_TP_connection(THD *thd)
@@ -85,14 +90,14 @@ struct Worker_thread_context
 
   void save()
   {
-    psi_thread= PSI_CALL_get_thread();
-    mysys_var= my_thread_var;
+    psi_thread = PSI_CALL_get_thread();
+    mysys_var= (st_my_thread_var *)pthread_getspecific(THR_KEY_mysys);
   }
 
   void restore()
   {
     PSI_CALL_set_thread(psi_thread);
-    set_mysys_var(mysys_var);
+    pthread_setspecific(THR_KEY_mysys,mysys_var);
     pthread_setspecific(THR_THD, 0);
   }
 };
@@ -143,7 +148,7 @@ static void thread_attach(THD* thd)
      attaching the thd. */
   wsrep_wait_rollback_complete_and_acquire_ownership(thd);
 #endif /* WITH_WSREP */
-  set_mysys_var(thd->mysys_var);
+  pthread_setspecific(THR_KEY_mysys,thd->mysys_var);
   thd->thread_stack=(char*)&thd;
   thd->store_globals();
   PSI_CALL_set_thread(thd->event_scheduler.m_psi);
@@ -159,9 +164,8 @@ static TP_PRIORITY get_priority(TP_connection *c)
   DBUG_ASSERT(c->thd == current_thd);
   TP_PRIORITY prio= (TP_PRIORITY)c->thd->variables.threadpool_priority;
   if (prio == TP_PRIORITY_AUTO)
-  {
-    return c->thd->transaction.is_active() ? TP_PRIORITY_HIGH : TP_PRIORITY_LOW;
-  }
+    prio= c->thd->transaction.is_active() ? TP_PRIORITY_HIGH : TP_PRIORITY_LOW;
+
   return prio;
 }
 
@@ -191,7 +195,7 @@ void tp_callback(TP_connection *c)
   }
   else if (threadpool_process_request(thd))
   {
-    /* QUIT or an error occurred. */
+    /* QUIT or an error occured. */
     goto error;
   }
 
@@ -228,9 +232,9 @@ static THD* threadpool_add_connection(CONNECT *connect, void *scheduler_data)
     Store them in THD.
   */
 
-  set_mysys_var(NULL);
+  pthread_setspecific(THR_KEY_mysys, 0);
   my_thread_init();
-  st_my_thread_var* mysys_var= my_thread_var;
+  st_my_thread_var* mysys_var= (st_my_thread_var *)pthread_getspecific(THR_KEY_mysys);
   if (!mysys_var ||!(thd= connect->create_thd(NULL)))
   {
     /* Out of memory? */
@@ -323,16 +327,6 @@ static void handle_wait_timeout(THD *thd)
   thd->net.error= 2;
 }
 
-/** Check if some client data is cached in thd->net or thd->net.vio */
-static bool has_unread_data(THD* thd)
-{
-  NET *net= &thd->net;
-  if (net->compress && net->remain_in_buf)
-    return true;
-  Vio *vio= net->vio;
-  return vio->has_data(vio);
-}
-
 
 /**
  Process a single client request or a single batch.
@@ -367,6 +361,7 @@ static int threadpool_process_request(THD *thd)
   */
   for(;;)
   {
+    Vio *vio;
     thd->net.reading_or_writing= 0;
     if (mysql_audit_release_required(thd))
       mysql_audit_release(thd);
@@ -382,7 +377,8 @@ static int threadpool_process_request(THD *thd)
 
     set_thd_idle(thd);
 
-    if (!has_unread_data(thd))
+    vio= thd->net.vio;
+    if (!vio->has_data(vio))
     { 
       /* More info on this debug sync is in sql_parse.cc*/
       DEBUG_SYNC(thd, "before_do_command_net_read");
@@ -395,19 +391,6 @@ end:
 }
 
 
-
-/* Dummy functions, do nothing */
-
-static bool tp_init_new_connection_thread()
-{
-  return 0;
-}
-
-static bool tp_end_thread(THD *, bool)
-{
-  return 0;
-}
-
 static TP_pool *pool;
 
 static bool tp_init()
@@ -483,34 +466,25 @@ void tp_timeout_handler(TP_connection *c)
 {
   if (c->state != TP_STATE_IDLE)
     return;
-  THD *thd= c->thd;
+  THD *thd=c->thd;
   mysql_mutex_lock(&thd->LOCK_thd_kill);
-  Vio *vio= thd->net.vio;
-  if (vio && (vio_pending(vio) > 0 || vio->has_data(vio)) &&
-      c->state == TP_STATE_IDLE)
-  {
-    /*
-     There is some data on that connection, i.e
-     i.e there was no inactivity timeout.
-     Don't kill.
-    */
-    c->state= TP_STATE_PENDING;
-  }
-  else if (c->state == TP_STATE_IDLE)
-  {
-    thd->set_killed_no_mutex(KILL_WAIT_TIMEOUT);
-    c->priority= TP_PRIORITY_HIGH;
-    post_kill_notification(thd);
-  }
+  thd->set_killed_no_mutex(KILL_WAIT_TIMEOUT);
+  c->priority= TP_PRIORITY_HIGH;
+  post_kill_notification(thd);
   mysql_mutex_unlock(&thd->LOCK_thd_kill);
 }
 
+MY_ALIGNED(CPU_LEVEL1_DCACHE_LINESIZE) Atomic_counter<unsigned long long> tp_waits[THD_WAIT_LAST];
 
 static void tp_wait_begin(THD *thd, int type)
 {
   TP_connection *c = get_TP_connection(thd);
   if (c)
+  {
+    DBUG_ASSERT(type > 0 && type < THD_WAIT_LAST);
+    tp_waits[type]++;
     c->wait_begin(type);
+  }
 }
 
 
@@ -541,18 +515,16 @@ static scheduler_functions tp_scheduler_functions=
   NULL,
   NULL,
   tp_init,                            // init
-  tp_init_new_connection_thread,      // init_new_connection_thread
   tp_add_connection,                  // add_connection
   tp_wait_begin,                      // thd_wait_begin
   tp_wait_end,                        // thd_wait_end
   tp_post_kill_notification,          // post kill notification
-  tp_end_thread,                      // Dummy function
   tp_end                              // end
 };
 
 void pool_of_threads_scheduler(struct scheduler_functions *func,
     ulong *arg_max_connections,
-    uint *arg_connection_count)
+    Atomic_counter<uint> *arg_connection_count)
 {
   *func = tp_scheduler_functions;
   func->max_threads= threadpool_max_threads;
