@@ -25,7 +25,7 @@
     str is a (long) to record position where 0 is the first position.
 */
 
-#include <my_global.h>
+#include "mariadb.h"
 #include "sql_priv.h"
 #include "unireg.h"
 #include "sql_partition.h"                      // struct partition_info
@@ -39,7 +39,7 @@
 /* threshold for safe_alloca */
 #define ALLOCA_THRESHOLD       2048
 
-static uint pack_keys(uchar *,uint, KEY *, ulong);
+static uint pack_keys(uchar *,uint, KEY *, ulong, uint);
 static bool pack_header(THD *, uchar *, List<Create_field> &, HA_CREATE_INFO *,
                         ulong, handler *);
 static bool pack_vcols(String *, List<Create_field> &, List<Virtual_column_info> *);
@@ -72,22 +72,161 @@ static uchar *extra2_write_len(uchar *pos, size_t len)
   return pos;
 }
 
+static uchar* extra2_write_str(uchar *pos, const LEX_CSTRING &str)
+{
+  pos= extra2_write_len(pos, str.length);
+  memcpy(pos, str.str, str.length);
+  return pos + str.length;
+}
+
 static uchar *extra2_write(uchar *pos, enum extra2_frm_value_type type,
-                           LEX_STRING *str)
+                           const LEX_CSTRING &str)
 {
   *pos++ = type;
-  pos= extra2_write_len(pos, str->length);
-  memcpy(pos, str->str, str->length);
-  return pos + str->length;
+  return extra2_write_str(pos, str);
 }
 
 static uchar *extra2_write(uchar *pos, enum extra2_frm_value_type type,
-                           LEX_CUSTRING *str)
+                           const LEX_CUSTRING &str)
 {
-  return extra2_write(pos, type, reinterpret_cast<LEX_STRING *>(str));
+  return extra2_write(pos, type, *reinterpret_cast<const LEX_CSTRING*>(&str));
 }
 
-/*
+static uchar *extra2_write_field_properties(uchar *pos,
+                   List<Create_field> &create_fields)
+{
+  List_iterator<Create_field> it(create_fields);
+  *pos++= EXTRA2_FIELD_FLAGS;
+  /*
+   always first 2  for field visibility
+  */
+  pos= extra2_write_len(pos, create_fields.elements);
+  while (Create_field *cf= it++)
+  {
+    uchar flags= cf->invisible;
+    if (cf->flags & VERS_UPDATE_UNVERSIONED_FLAG)
+      flags|= VERS_OPTIMIZED_UPDATE;
+    *pos++= flags;
+  }
+  return pos;
+}
+
+static uint16
+get_fieldno_by_name(HA_CREATE_INFO *create_info, List<Create_field> &create_fields,
+                    const Lex_ident &field_name)
+{
+  List_iterator<Create_field> it(create_fields);
+  Create_field *sql_field = NULL;
+
+  DBUG_ASSERT(field_name);
+
+  for (unsigned field_no = 0; (sql_field = it++); ++field_no)
+  {
+    if (field_name.streq(sql_field->field_name))
+    {
+      DBUG_ASSERT(field_no <= uint16(~0U));
+      return uint16(field_no);
+    }
+  }
+
+  DBUG_ASSERT(0); /* Not Reachable */
+  return 0;
+}
+
+static inline
+bool has_extra2_field_flags(List<Create_field> &create_fields)
+{
+  List_iterator<Create_field> it(create_fields);
+  while (Create_field *f= it++)
+  {
+    if (f->invisible)
+      return true;
+    if (f->flags & VERS_UPDATE_UNVERSIONED_FLAG)
+      return true;
+  }
+  return false;
+}
+
+static size_t extra2_str_size(size_t len)
+{
+  return (len > 255 ? 3 : 1) + len;
+}
+
+
+static uint gis_field_options_image(uchar *buff,
+                                    List<Create_field> &create_fields)
+{
+  uint image_size= 0;
+  List_iterator<Create_field> it(create_fields);
+  Create_field *field;
+  while ((field= it++))
+  {
+    if (field->real_field_type() != MYSQL_TYPE_GEOMETRY)
+      continue;
+    uchar *cbuf= buff ? buff + image_size : NULL;
+    image_size+= field->type_handler()->
+                   Column_definition_gis_options_image(cbuf, *field);
+  }
+  return image_size;
+}
+
+
+class Field_data_type_info_image: public BinaryStringBuffer<512>
+{
+  static uchar *store_length(uchar *pos, ulonglong length)
+  {
+    return net_store_length(pos, length);
+  }
+  static uchar *store_string(uchar *pos, const LEX_CSTRING &str)
+  {
+    pos= store_length(pos, str.length);
+    memcpy(pos, str.str, str.length);
+    return pos + str.length;
+  }
+  static uint store_length_required_length(ulonglong length)
+  {
+    return net_length_size(length);
+  }
+public:
+  Field_data_type_info_image() { }
+  bool append(uint fieldnr, const Column_definition &def)
+  {
+    BinaryStringBuffer<64> type_info;
+    if (def.type_handler()->
+              Column_definition_data_type_info_image(&type_info, def) ||
+        type_info.length() > 0xFFFF/*Some reasonable limit*/)
+      return true; // Error
+    if (!type_info.length())
+      return false;
+    size_t need_length= store_length_required_length(fieldnr) +
+                        store_length_required_length(type_info.length()) +
+                        type_info.length();
+    if (reserve(need_length))
+      return true; // Error
+    uchar *pos= (uchar *) end();
+    pos= store_length(pos, fieldnr);
+    pos= store_string(pos, type_info.lex_cstring());
+    size_t new_length= (const char *) pos - ptr();
+    DBUG_ASSERT(new_length < alloced_length());
+    length((uint32) new_length);
+    return false;
+  }
+  bool append(List<Create_field> &fields)
+  {
+    uint fieldnr= 0;
+    Create_field *field;
+    List_iterator<Create_field> it(fields);
+    for (field= it++; field; field= it++, fieldnr++)
+    {
+      if (append(fieldnr, *field))
+        return true; // Error
+    }
+    return false;
+  }
+};
+
+
+/**
   Create a frm (table definition) file
 
   @param thd                    Thread handler
@@ -102,23 +241,31 @@ static uchar *extra2_write(uchar *pos, enum extra2_frm_value_type type,
   or null LEX_CUSTRING (str==0) in case of an error.
 */
 
-LEX_CUSTRING build_frm_image(THD *thd, const char *table,
+LEX_CUSTRING build_frm_image(THD *thd, const LEX_CSTRING &table,
                               HA_CREATE_INFO *create_info,
                               List<Create_field> &create_fields,
                               uint keys, KEY *key_info, handler *db_file)
 {
-  LEX_STRING str_db_type;
+  LEX_CSTRING str_db_type;
   uint reclength, key_info_length, i;
   ulong key_buff_length;
-  ulong filepos, data_offset;
+  size_t filepos;
+  ulong data_offset;
   uint options_len;
   uint gis_extra2_len= 0;
+  size_t period_info_len= create_info->period_info.name
+                          ? extra2_str_size(create_info->period_info.name.length)
+                            + extra2_str_size(create_info->period_info.constr->name.length)
+                            + 2 * frm_fieldno_size
+                          : 0;
+  uint e_unique_hash_extra_parts= 0;
   uchar fileinfo[FRM_HEADER_SIZE],forminfo[FRM_FORMINFO_SIZE];
   const partition_info *part_info= IF_PARTITIONING(thd->work_part_info, 0);
-  int error;
+  bool error;
   uchar *frm_ptr, *pos;
   LEX_CUSTRING frm= {0,0};
   StringBuffer<MAX_FIELD_WIDTH> vcols;
+  Field_data_type_info_image field_data_type_info_image;
   DBUG_ENTER("build_frm_image");
 
  /* If fixed row records, we need one bit to check for deleted rows */
@@ -131,15 +278,15 @@ LEX_CUSTRING build_frm_image(THD *thd, const char *table,
   error= pack_vcols(&vcols, create_fields, create_info->check_constraint_list);
   thd->variables.sql_mode= save_sql_mode;
 
-  if (error)
+  if (unlikely(error))
     DBUG_RETURN(frm);
 
   if (vcols.length())
     create_info->expression_length= vcols.length() + FRM_VCOL_NEW_BASE_SIZE;
 
   error= pack_header(thd, forminfo, create_fields, create_info,
-                     data_offset, db_file);
-  if (error)
+                     (ulong)data_offset, db_file);
+  if (unlikely(error))
     DBUG_RETURN(frm);
 
   reclength= uint2korr(forminfo+266);
@@ -147,7 +294,7 @@ LEX_CUSTRING build_frm_image(THD *thd, const char *table,
   /* Calculate extra data segment length */
   str_db_type= *hton_name(create_info->db_type);
   /* str_db_type */
-  create_info->extra_size= (2 + str_db_type.length +
+  create_info->extra_size= (uint)(2 + str_db_type.length +
                             2 + create_info->connect_string.length);
   /*
     Partition:
@@ -158,25 +305,38 @@ LEX_CUSTRING build_frm_image(THD *thd, const char *table,
   */
   create_info->extra_size+= 6;
   if (part_info)
-    create_info->extra_size+= part_info->part_info_len;
+    create_info->extra_size+= (uint)part_info->part_info_len;
 
   for (i= 0; i < keys; i++)
   {
     if (key_info[i].parser_name)
-      create_info->extra_size+= key_info[i].parser_name->length + 1;
+      create_info->extra_size+= (uint)key_info[i].parser_name->length + 1;
   }
 
   options_len= engine_table_options_frm_length(create_info->option_list,
                                                create_fields,
                                                keys, key_info);
-#ifdef HAVE_SPATIAL
   gis_extra2_len= gis_field_options_image(NULL, create_fields);
-#endif /*HAVE_SPATIAL*/
   DBUG_PRINT("info", ("Options length: %u", options_len));
 
+  if (field_data_type_info_image.append(create_fields))
+  {
+    my_printf_error(ER_CANT_CREATE_TABLE,
+                    "Cannot create table %`s: "
+                    "Building the field data type info image failed.",
+                    MYF(0), table.str);
+    DBUG_RETURN(frm);
+  }
+  DBUG_PRINT("info", ("Field data type info length: %u",
+                      (uint) field_data_type_info_image.length()));
+  DBUG_EXECUTE_IF("frm_data_type_info",
+                  push_warning_printf(thd, Sql_condition::WARN_LEVEL_NOTE,
+                  ER_UNKNOWN_ERROR,
+                  "build_frm_image: Field data type info length: %u",
+                  (uint) field_data_type_info_image.length()););
+
   if (validate_comment_length(thd, &create_info->comment, TABLE_COMMENT_MAXLEN,
-                              ER_TOO_LONG_TABLE_COMMENT,
-                              table))
+                              ER_TOO_LONG_TABLE_COMMENT, table.str))
      DBUG_RETURN(frm);
   /*
     If table comment is longer than TABLE_COMMENT_INLINE_MAXLEN bytes,
@@ -186,7 +346,7 @@ LEX_CUSTRING build_frm_image(THD *thd, const char *table,
   if (create_info->comment.length > TABLE_COMMENT_INLINE_MAXLEN)
   {
     forminfo[46]=255;
-    create_info->extra_size+= 2 + create_info->comment.length;
+    create_info->extra_size+= 2 + (uint)create_info->comment.length;
   }
   else
   {
@@ -210,17 +370,38 @@ LEX_CUSTRING build_frm_image(THD *thd, const char *table,
   prepare_frm_header(thd, reclength, fileinfo, create_info, keys, key_info);
 
   /* one byte for a type, one or three for a length */
-  uint extra2_size= 1 + 1 + create_info->tabledef_version.length;
+  size_t extra2_size= 1 + extra2_str_size(create_info->tabledef_version.length);
   if (options_len)
-    extra2_size+= 1 + (options_len > 255 ? 3 : 1) + options_len;
+    extra2_size+= 1 + extra2_str_size(options_len);
 
   if (part_info)
-    extra2_size+= 1 + 1 + hton_name(part_info->default_engine_type)->length;
+    extra2_size+= 1 + extra2_str_size(hton_name(part_info->default_engine_type)->length);
 
   if (gis_extra2_len)
-    extra2_size+= 1 + (gis_extra2_len > 255 ? 3 : 1) + gis_extra2_len;
+    extra2_size+= 1 + extra2_str_size(gis_extra2_len);
 
+  if (field_data_type_info_image.length())
+    extra2_size+= 1 + extra2_str_size(field_data_type_info_image.length());
 
+  if (create_info->versioned())
+  {
+    extra2_size+= 1 + extra2_str_size(2 * frm_fieldno_size);
+  }
+
+  if (create_info->period_info.name)
+  {
+    extra2_size+= 1 + extra2_str_size(period_info_len);
+  }
+
+  bool has_extra2_field_flags_= has_extra2_field_flags(create_fields);
+  if (has_extra2_field_flags_)
+  {
+    extra2_size+= 1 + extra2_str_size(create_fields.elements);
+  }
+
+  for (i= 0; i < keys; i++)
+    if (key_info[i].algorithm == HA_KEY_ALG_LONG_HASH)
+      e_unique_hash_extra_parts++;
   key_buff_length= uint4korr(fileinfo+47);
 
   frm.length= FRM_HEADER_SIZE;                  // fileinfo;
@@ -240,7 +421,7 @@ LEX_CUSTRING build_frm_image(THD *thd, const char *table,
   if (frm.length > FRM_MAX_SIZE ||
       create_info->expression_length > UINT_MAX32)
   {
-    my_error(ER_TABLE_DEFINITION_TOO_BIG, MYF(0), table);
+    my_error(ER_TABLE_DEFINITION_TOO_BIG, MYF(0), table.str);
     DBUG_RETURN(frm);
   }
 
@@ -253,11 +434,11 @@ LEX_CUSTRING build_frm_image(THD *thd, const char *table,
   pos = frm_ptr + 64;
   compile_time_assert(EXTRA2_TABLEDEF_VERSION != '/');
   pos= extra2_write(pos, EXTRA2_TABLEDEF_VERSION,
-                    &create_info->tabledef_version);
+                    create_info->tabledef_version);
 
   if (part_info)
     pos= extra2_write(pos, EXTRA2_DEFAULT_PART_ENGINE,
-                      hton_name(part_info->default_engine_type));
+                      *hton_name(part_info->default_engine_type));
 
   if (options_len)
   {
@@ -267,26 +448,71 @@ LEX_CUSTRING build_frm_image(THD *thd, const char *table,
                                         create_fields, keys, key_info);
   }
 
-#ifdef HAVE_SPATIAL
   if (gis_extra2_len)
   {
     *pos= EXTRA2_GIS;
     pos= extra2_write_len(pos+1, gis_extra2_len);
     pos+= gis_field_options_image(pos, create_fields);
   }
-#endif /*HAVE_SPATIAL*/
+
+  if (field_data_type_info_image.length())
+  {
+    if (field_data_type_info_image.length() > 0xFFFF)
+    {
+      my_printf_error(ER_CANT_CREATE_TABLE,
+                      "Cannot create table %`s: "
+                      "field data type info image is too large. "
+                      "Decrease the number of columns with "
+                      "extended data types.",
+                      MYF(0), table.str);
+      goto err;
+    }
+    *pos= EXTRA2_FIELD_DATA_TYPE_INFO;
+    pos= extra2_write_str(pos + 1, field_data_type_info_image.lex_cstring());
+  }
+
+  // PERIOD
+  if (create_info->period_info.is_set())
+  {
+    *pos++= EXTRA2_APPLICATION_TIME_PERIOD;
+    pos= extra2_write_len(pos, period_info_len);
+    pos= extra2_write_str(pos, create_info->period_info.name);
+    pos= extra2_write_str(pos, create_info->period_info.constr->name);
+
+    store_frm_fieldno(pos, get_fieldno_by_name(create_info, create_fields,
+                                       create_info->period_info.period.start));
+    pos+= frm_fieldno_size;
+    store_frm_fieldno(pos, get_fieldno_by_name(create_info, create_fields,
+                                       create_info->period_info.period.end));
+    pos+= frm_fieldno_size;
+  }
+
+  if (create_info->versioned())
+  {
+    *pos++= EXTRA2_PERIOD_FOR_SYSTEM_TIME;
+    *pos++= 2 * frm_fieldno_size;
+    store_frm_fieldno(pos, get_fieldno_by_name(create_info, create_fields,
+                                       create_info->vers_info.as_row.start));
+    pos+= frm_fieldno_size;
+    store_frm_fieldno(pos, get_fieldno_by_name(create_info, create_fields,
+                                       create_info->vers_info.as_row.end));
+    pos+= frm_fieldno_size;
+  }
+
+  if (has_extra2_field_flags_)
+    pos= extra2_write_field_properties(pos, create_fields);
 
   int4store(pos, filepos); // end of the extra2 segment
   pos+= 4;
 
   DBUG_ASSERT(pos == frm_ptr + uint2korr(fileinfo+6));
-  key_info_length= pack_keys(pos, keys, key_info, data_offset);
+  key_info_length= pack_keys(pos, keys, key_info, data_offset, e_unique_hash_extra_parts);
   if (key_info_length > UINT_MAX16)
   {
     my_printf_error(ER_CANT_CREATE_TABLE,
                     "Cannot create table %`s: index information is too long. "
                     "Decrease number of indexes or use shorter index names or shorter comments.",
-                    MYF(0), table);
+                    MYF(0), table.str);
     goto err;
   }
 
@@ -312,8 +538,7 @@ LEX_CUSTRING build_frm_image(THD *thd, const char *table,
   pos+= reclength;
   int2store(pos, create_info->connect_string.length);
   pos+= 2;
-  if (create_info->connect_string.length)
-    memcpy(pos, create_info->connect_string.str, create_info->connect_string.length);
+  memcpy(pos, create_info->connect_string.str, create_info->connect_string.length);
   pos+= create_info->connect_string.length;
   int2store(pos, str_db_type.length);
   pos+= 2;
@@ -390,60 +615,10 @@ err:
 }
 
 
-/**
-  Create a frm (table definition) file and the tables
-
-  @param thd           Thread handler
-  @param frm           Binary frm image of the table to create
-  @param path          Name of file (including database, without .frm)
-  @param db            Data base name
-  @param table_name    Table name
-  @param create_info   create info parameters
-  @param file          Handler to use or NULL if only frm needs to be created
-
-  @retval 0   ok
-  @retval 1   error
-*/
-
-int rea_create_table(THD *thd, LEX_CUSTRING *frm,
-                     const char *path, const char *db, const char *table_name,
-                     HA_CREATE_INFO *create_info, handler *file,
-                     bool no_ha_create_table)
-{
-  DBUG_ENTER("rea_create_table");
-
-  if (no_ha_create_table)
-  {
-    if (writefrm(path, db, table_name, true, frm->str, frm->length))
-      goto err_frm;
-  }
-
-  if (thd->variables.keep_files_on_create)
-    create_info->options|= HA_CREATE_KEEP_FILES;
-
-  if (file->ha_create_partitioning_metadata(path, NULL, CHF_CREATE_FLAG))
-    goto err_part;
-
-  if (!no_ha_create_table)
-  {
-    if (ha_create_table(thd, path, db, table_name, create_info, frm))
-      goto err_part;
-  }
-
-  DBUG_RETURN(0);
-
-err_part:
-  file->ha_create_partitioning_metadata(path, NULL, CHF_DELETE_FLAG);
-err_frm:
-  deletefrm(path);
-  DBUG_RETURN(1);
-} /* rea_create_table */
-
-
 /* Pack keyinfo and keynames to keybuff for save in form-file. */
 
 static uint pack_keys(uchar *keybuff, uint key_count, KEY *keyinfo,
-                      ulong data_offset)
+                      ulong data_offset, uint e_unique_hash_extra_parts)
 {
   uint key_parts,length;
   uchar *pos, *keyname_pos;
@@ -488,7 +663,7 @@ static uint pack_keys(uchar *keybuff, uint key_count, KEY *keyinfo,
   *pos++=(uchar) NAMES_SEP_CHAR;
   for (key=keyinfo ; key != end ; key++)
   {
-    uchar *tmp=(uchar*) strmov((char*) pos,key->name);
+    uchar *tmp=(uchar*) strmov((char*) pos,key->name.str);
     *tmp++= (uchar) NAMES_SEP_CHAR;
     *tmp=0;
     pos=tmp;
@@ -505,6 +680,7 @@ static uint pack_keys(uchar *keybuff, uint key_count, KEY *keyinfo,
     }
   }
 
+  key_parts+= e_unique_hash_extra_parts;
   if (key_count > 127 || key_parts > 127)
   {
     keybuff[0]= (key_count & 0x7f) | 0x80;
@@ -549,7 +725,7 @@ static bool pack_expression(String *buf, Virtual_column_info *vcol,
   size_t len_off= buf->length();
   buf->q_append2b(0); // to be added later
   buf->q_append((char)vcol->name.length);
-  buf->q_append(vcol->name.str, vcol->name.length);
+  buf->q_append(&vcol->name);
   size_t expr_start= buf->length();
   vcol->print(buf);
   size_t expr_len= buf->length() - expr_start;
@@ -571,7 +747,7 @@ static bool pack_vcols(String *buf, List<Create_field> &create_fields,
 
   for (uint field_nr=0; (field= it++); field_nr++)
   {
-    if (field->vcol_info)
+    if (field->vcol_info && field->vcol_info->expr)
       if (pack_expression(buf, field->vcol_info, field_nr,
                           field->vcol_info->stored_in_db
                           ? VCOL_GENERATED_STORED : VCOL_GENERATED_VIRTUAL))
@@ -594,18 +770,6 @@ static bool pack_vcols(String *buf, List<Create_field> &create_fields,
 }
 
 
-static uint typelib_values_packed_length(const TYPELIB *t)
-{
-  uint length= 0;
-  for (uint i= 0; t->type_names[i]; i++)
-  {
-    length+= t->type_lengths[i];
-    length++; /* Separator */
-  }
-  return length;
-}
-
-
 /* Make formheader */
 
 static bool pack_header(THD *thd, uchar *forminfo,
@@ -638,7 +802,8 @@ static bool pack_header(THD *thd, uchar *forminfo,
   while ((field=it++))
   {
     if (validate_comment_length(thd, &field->comment, COLUMN_COMMENT_MAXLEN,
-                                ER_TOO_LONG_FIELD_COMMENT, field->field_name))
+                                ER_TOO_LONG_FIELD_COMMENT,
+                                field->field_name.str))
        DBUG_RETURN(1);
 
     totlength+= (size_t)field->length;
@@ -647,14 +812,14 @@ static bool pack_header(THD *thd, uchar *forminfo,
       We mark first TIMESTAMP field with NOW() in DEFAULT or ON UPDATE
       as auto-update field.
     */
-    if (field->sql_type == MYSQL_TYPE_TIMESTAMP &&
+    if (field->real_field_type() == MYSQL_TYPE_TIMESTAMP &&
         MTYP_TYPENR(field->unireg_check) != Field::NONE &&
 	!time_stamp_pos)
       time_stamp_pos= (uint) field->offset+ (uint) data_offset + 1;
     length=field->pack_length;
     if ((uint) field->offset+ (uint) data_offset+ length > reclength)
       reclength=(uint) (field->offset+ data_offset + length);
-    n_length+= (ulong) strlen(field->field_name)+1;
+    n_length+= field->field_name.length + 1;
     field->interval_id=0;
     field->save_interval= 0;
     if (field->interval)
@@ -663,6 +828,7 @@ static bool pack_header(THD *thd, uchar *forminfo,
 
       if (field->charset->mbminlen > 1)
       {
+        TYPELIB *tmpint;
         /* 
           Escape UCS2 intervals using HEX notation to avoid
           problems with delimiters between enum elements.
@@ -671,26 +837,26 @@ static bool pack_header(THD *thd, uchar *forminfo,
           filled with default values it is saved in save_interval
           The HEX representation is created from this copy.
         */
+        uint count= field->interval->count;
         field->save_interval= field->interval;
-        field->interval= (TYPELIB*) thd->alloc(sizeof(TYPELIB));
-        *field->interval= *field->save_interval; 
-        field->interval->type_names= 
-          (const char **) thd->alloc(sizeof(char*) * 
-                                     (field->interval->count+1));
-        field->interval->type_names[field->interval->count]= 0;
-        field->interval->type_lengths=
-          (uint *) thd->alloc(sizeof(uint) * field->interval->count);
- 
+        field->interval= tmpint= (TYPELIB*) thd->alloc(sizeof(TYPELIB));
+        *tmpint= *field->save_interval;
+        tmpint->type_names=
+          (const char **) thd->alloc(sizeof(char*) *
+                                     (count + 1));
+        tmpint->type_lengths= (uint *) thd->alloc(sizeof(uint) * (count + 1));
+        tmpint->type_names[count]= 0;
+        tmpint->type_lengths[count]= 0;
+
         for (uint pos= 0; pos < field->interval->count; pos++)
         {
           char *dst;
           const char *src= field->save_interval->type_names[pos];
-          uint hex_length;
+          size_t hex_length;
           length= field->save_interval->type_lengths[pos];
           hex_length= length * 2;
-          field->interval->type_lengths[pos]= hex_length;
-          field->interval->type_names[pos]= dst=
-            (char*) thd->alloc(hex_length + 1);
+          tmpint->type_lengths[pos]= (uint) hex_length;
+          tmpint->type_names[pos]= dst= (char*) thd->alloc(hex_length + 1);
           octet2hex(dst, src, length);
         }
       }
@@ -698,8 +864,9 @@ static bool pack_header(THD *thd, uchar *forminfo,
       field->interval_id=get_interval_id(&int_count,create_fields,field);
       if (old_int_count != int_count)
       {
-        int_length+= typelib_values_packed_length(field->interval);
-        int_parts+= field->interval->count + 1;
+	for (const char **pos=field->interval->type_names ; *pos ; pos++)
+	  int_length+=(uint) strlen(*pos)+1;	// field + suffix prefix
+	int_parts+=field->interval->count+1;
       }
     }
     if (f_maybe_null(field->pack_flag))
@@ -753,7 +920,7 @@ static uint get_interval_id(uint *int_count,List<Create_field> &create_fields,
 {
   List_iterator<Create_field> it(create_fields);
   Create_field *field;
-  TYPELIB *interval=last_field->interval;
+  const TYPELIB *interval= last_field->interval;
 
   while ((field=it++) != last_field)
   {
@@ -788,12 +955,16 @@ static size_t packed_fields_length(List<Create_field> &create_fields)
     {
       int_count= field->interval_id;
       length++;
-      length+= typelib_values_packed_length(field->interval);
+      for (int i=0; field->interval->type_names[i]; i++)
+      {
+        length+= field->interval->type_lengths[i];
+        length++;
+      }
       length++;
     }
 
     length+= FCOMP;
-    length+= strlen(field->field_name)+1;
+    length+= field->field_name.length + 1;
     length+= field->comment.length;
   }
   length+= 2;
@@ -807,7 +978,8 @@ static bool pack_fields(uchar **buff_arg, List<Create_field> &create_fields,
                         ulong data_offset)
 {
   uchar *buff= *buff_arg;
-  uint int_count, comment_length= 0;
+  uint int_count;
+  size_t comment_length= 0;
   Create_field *field;
   DBUG_ENTER("pack_fields");
 
@@ -817,32 +989,12 @@ static bool pack_fields(uchar **buff_arg, List<Create_field> &create_fields,
   while ((field=it++))
   {
     uint recpos;
-    int2store(buff+3, field->length);
     /* The +1 is here becasue the col offset in .frm file have offset 1 */
     recpos= field->offset+1 + (uint) data_offset;
     int3store(buff+5,recpos);
-    int2store(buff+8,field->pack_flag);
-    buff[10]= (uchar) field->unireg_check;
     buff[12]= (uchar) field->interval_id;
-    buff[13]= (uchar) field->sql_type;
-    if (field->sql_type == MYSQL_TYPE_GEOMETRY)
-    {
-      buff[11]= 0;
-      buff[14]= (uchar) field->geom_type;
-#ifndef HAVE_SPATIAL
-      DBUG_ASSERT(0);                           // Should newer happen
-#endif
-    }
-    else if (field->charset) 
-    {
-      buff[11]= (uchar) (field->charset->number >> 8);
-      buff[14]= (uchar) field->charset->number;
-    }
-    else
-    {
-      buff[11]= buff[14]= 0;			// Numerical
-    }
-
+    buff[13]= (uchar) field->type_handler()->real_field_type();
+    field->type_handler()->Column_definition_attributes_frm_pack(field, buff);
     int2store(buff+15, field->comment.length);
     comment_length+= field->comment.length;
     set_if_bigger(int_count,field->interval_id);
@@ -854,7 +1006,7 @@ static bool pack_fields(uchar **buff_arg, List<Create_field> &create_fields,
   it.rewind();
   while ((field=it++))
   {
-    buff= (uchar*)strmov((char*) buff, field->field_name);
+    buff= (uchar*)strmov((char*) buff, field->field_name.str);
     *buff++=NAMES_SEP_CHAR;
   }
   *buff++= 0;
@@ -929,13 +1081,36 @@ static bool pack_fields(uchar **buff_arg, List<Create_field> &create_fields,
   DBUG_RETURN(0);
 }
 
+
+static bool make_empty_rec_store_default(THD *thd, Field *regfield,
+                                         Virtual_column_info *default_value)
+{
+  if (default_value && !default_value->flags)
+  {
+    Item *expr= default_value->expr;
+    // may be already fixed if ALTER TABLE
+    if (expr->fix_fields_if_needed(thd, &expr))
+      return true;
+    DBUG_ASSERT(expr == default_value->expr); // Should not change
+    if (regfield->make_empty_rec_store_default_value(thd, expr))
+    {
+      my_error(ER_INVALID_DEFAULT, MYF(0), regfield->field_name.str);
+      return true;
+    }
+    return false;
+  }
+  regfield->make_empty_rec_reset(thd);
+  return false;
+}
+
+
 /* save an empty record on start of formfile */
 
 static bool make_empty_rec(THD *thd, uchar *buff, uint table_options,
 			   List<Create_field> &create_fields,
 			   uint reclength, ulong data_offset)
 {
-  int error= 0;
+  int error= false;
   uint null_count;
   uchar *null_pos;
   TABLE table;
@@ -963,23 +1138,19 @@ static bool make_empty_rec(THD *thd, uchar *buff, uint table_options,
   thd->count_cuted_fields= CHECK_FIELD_WARN;    // To find wrong default values
   while ((field=it++))
   {
+    Record_addr addr(buff + field->offset + data_offset,
+                     null_pos + null_count / 8, null_count & 7);
+    Column_definition_attributes tmp(*field);
+    tmp.interval= field->save_interval ?
+                  field->save_interval : field->interval;
     /* regfield don't have to be deleted as it's allocated on THD::mem_root */
-    Field *regfield= make_field(&share, thd->mem_root,
-                                buff+field->offset + data_offset,
-                                (uint32)field->length,
-                                null_pos + null_count / 8,
-                                null_count & 7,
-                                field->pack_flag,
-                                field->sql_type,
-                                field->charset,
-                                field->geom_type, field->srid,
-                                field->unireg_check,
-                                field->save_interval ? field->save_interval
-                                                     : field->interval,
-                                field->field_name);
+    Field *regfield= tmp.make_field(&share, thd->mem_root, &addr,
+                                    field->type_handler(),
+                                    &field->field_name,
+                                    field->flags);
     if (!regfield)
     {
-      error= 1;
+      error= true;
       goto err;                                 // End of memory
     }
 
@@ -992,39 +1163,14 @@ static bool make_empty_rec(THD *thd, uchar *buff, uint table_options,
       null_count++;
     }
 
-    if (field->sql_type == MYSQL_TYPE_BIT && !f_bit_as_char(field->pack_flag))
+    if (field->real_field_type() == MYSQL_TYPE_BIT &&
+        !f_bit_as_char(field->pack_flag))
       null_count+= field->length & 7;
 
-    if (field->default_value && !field->default_value->flags &&
-        (!(field->flags & BLOB_FLAG) || field->sql_type == MYSQL_TYPE_GEOMETRY))
-    {
-      Item *expr= field->default_value->expr;
-
-      int res= !expr->fixed && // may be already fixed if ALTER TABLE
-                expr->fix_fields(thd, &expr);
-      if (!res)
-        res= expr->save_in_field(regfield, 1);
-      if (!res && (field->flags & BLOB_FLAG))
-        regfield->reset();
-
-      /* If not ok or warning of level 'note' */
-      if (res != 0 && res != 3)
-      {
-        my_error(ER_INVALID_DEFAULT, MYF(0), regfield->field_name);
-        error= 1;
-        delete regfield; //To avoid memory leak
-        goto err;
-      }
-      delete regfield; //To avoid memory leak
-    }
-    else if (regfield->real_type() == MYSQL_TYPE_ENUM &&
-	     (field->flags & NOT_NULL_FLAG))
-    {
-      regfield->set_notnull();
-      regfield->store((longlong) 1, TRUE);
-    }
-    else
-      regfield->reset();
+    error= make_empty_rec_store_default(thd, regfield, field->default_value);
+    delete regfield; // Avoid memory leaks
+    if (error)
+      goto err;
   }
   DBUG_ASSERT(data_offset == ((null_count + 7) / 8));
 
