@@ -1,4 +1,4 @@
-/* Copyright (C) 2014, 2020, MariaDB
+/* Copyright (C) 2014 SkySQL Ab.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,21 +13,27 @@
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1335 USA. */
 
-#include <my_global.h>
+#include "mariadb.h"
 #include <sql_class.h>
 #include <mysql/service_wsrep.h>
 
-my_bool wsrep_thd_is_BF(THD *, my_bool)
+my_bool wsrep_thd_is_BF(const THD *, my_bool)
 { return 0; }
 
-int wsrep_trx_order_before(THD *, THD *)
+int wsrep_is_wsrep_xid(const void* xid)
 { return 0; }
 
-enum wsrep_conflict_state wsrep_thd_conflict_state(THD *, my_bool)
-{ return NO_CONFLICT; }
+long long wsrep_xid_seqno(const XID* x)
+{ return -1; }
 
-int wsrep_is_wsrep_xid(const XID*)
-{ return 0; }
+const unsigned char* wsrep_xid_uuid(const XID*)
+{
+    static const unsigned char uuid[16]= {0};
+    return uuid;
+}
+
+bool wsrep_prepare_key_for_innodb(THD* thd, const uchar*, size_t, const uchar*, size_t, struct wsrep_buf*, size_t*)
+{ return 1; }
 
 bool wsrep_prepare_key(const uchar*, size_t, const uchar*, size_t, struct wsrep_buf*, size_t*)
 { return 0; }
@@ -35,94 +41,64 @@ bool wsrep_prepare_key(const uchar*, size_t, const uchar*, size_t, struct wsrep_
 struct wsrep *get_wsrep()
 { return 0; }
 
-my_bool get_wsrep_certify_nonPK()
-{ return 0; }
-
-my_bool get_wsrep_debug()
-{ return 0; }
-
-my_bool get_wsrep_drupal_282555_workaround()
-{ return 0; }
-
-my_bool get_wsrep_load_data_splitting()
-{ return 0; }
-
 my_bool get_wsrep_recovery()
 { return 0; }
 
-my_bool get_wsrep_log_conflicts()
-{ return 0; }
-
-long get_wsrep_protocol_version()
-{ return 0; }
-
-my_bool wsrep_aborting_thd_contains(THD *)
-{ return 0; }
-
-void wsrep_aborting_thd_enqueue(THD *)
-{ }
-
 bool wsrep_consistency_check(THD *)
 { return 0; }
 
 void wsrep_lock_rollback()
 { }
 
-int wsrep_on(THD *thd)
+my_bool wsrep_on(const THD *)
 { return 0; }
 
-void wsrep_post_commit(THD*, bool)
+void wsrep_thd_LOCK(const THD *)
 { }
 
-enum wsrep_trx_status wsrep_run_wsrep_commit(THD *, bool)
-{ return WSREP_TRX_ERROR; }
-
-void wsrep_thd_LOCK(THD *)
-{ }
-
-void wsrep_thd_UNLOCK(THD *)
-{ }
-
-void wsrep_thd_awake(THD *, my_bool)
+void wsrep_thd_UNLOCK(const THD *)
 { }
 
 const char *wsrep_thd_conflict_state_str(THD *)
 { return 0; }
 
-enum wsrep_exec_mode wsrep_thd_exec_mode(THD *)
-{ return LOCAL_STATE; }
-
 const char *wsrep_thd_exec_mode_str(THD *)
 { return NULL; }
 
-enum wsrep_conflict_state wsrep_thd_get_conflict_state(THD *)
-{ return NO_CONFLICT; }
-
-my_bool wsrep_thd_is_wsrep(THD *)
-{ return 0; }
-
-const char *wsrep_thd_query(THD *)
+const char *wsrep_thd_query(const THD *)
 { return "NULL"; }
 
-enum wsrep_query_state wsrep_thd_query_state(THD *)
-{ return QUERY_IDLE; }
-
 const char *wsrep_thd_query_state_str(THD *)
 { return 0; }
 
-int wsrep_thd_retry_counter(THD *)
+int wsrep_thd_retry_counter(const THD *)
 { return 0; }
 
-void wsrep_thd_set_conflict_state(THD *, enum wsrep_conflict_state)
-{ }
-
 bool wsrep_thd_ignore_table(THD *)
 { return 0; }
 
-longlong wsrep_thd_trx_seqno(THD *)
+long long wsrep_thd_trx_seqno(const THD *)
 { return -1; }
 
-struct wsrep_ws_handle* wsrep_thd_ws_handle(THD *)
+my_bool wsrep_thd_is_aborting(const THD *)
+{ return 0; }
+
+void wsrep_set_data_home_dir(const char *)
+{ }
+
+my_bool wsrep_thd_is_local(const THD *)
+{ return 0; }
+
+void wsrep_thd_self_abort(THD *)
+{ }
+
+int wsrep_thd_append_key(THD *, const struct wsrep_key*, int, enum Wsrep_service_key_type)
+{ return 0; }
+
+const char* wsrep_thd_client_state_str(const THD*)
+{ return 0; }
+
+const char* wsrep_thd_client_mode_str(const THD*)
 { return 0; }
 
 void wsrep_thd_auto_increment_variables(THD *thd,
@@ -133,28 +109,38 @@ void wsrep_thd_auto_increment_variables(THD *thd,
   *increment= thd->variables.auto_increment_increment;
 }
 
-void wsrep_set_load_multi_commit(THD *thd, bool split)
-{ }
-
-bool wsrep_is_load_multi_commit(THD *thd)
-{ return false; }
-
-int wsrep_trx_is_aborting(THD *)
+const char* wsrep_thd_transaction_state_str(const THD*)
 { return 0; }
 
-void wsrep_unlock_rollback()
+query_id_t wsrep_thd_transaction_id(const THD *)
+{ return 0; }
+
+my_bool wsrep_thd_bf_abort(const THD *, THD *, my_bool)
+{ return 0; }
+
+my_bool wsrep_thd_order_before(const THD*, const THD *)
+{ return 0; }
+
+void wsrep_handle_SR_rollback(THD*, THD*)
 { }
 
-void wsrep_set_data_home_dir(const char *)
+my_bool wsrep_thd_skip_locking(const THD*)
+{ return 0;}
+
+const char* wsrep_get_sr_table_name()
+{ return 0; }
+
+my_bool wsrep_get_debug()
+{ return 0;}
+
+void wsrep_commit_ordered(THD* )
 { }
 
-void wsrep_log(void (*)(const char *, ...), const char *, ...)
-{
-}
+my_bool wsrep_thd_is_applying(const THD*)
+{ return 0;}
 
-my_bool wsrep_thd_is_applier(MYSQL_THD thd)
-{ return false; }
+my_bool wsrep_thd_has_ignored_error(const THD*)
+{ return 0;}
 
-void wsrep_report_bf_lock_wait(MYSQL_THD thd,
-                               unsigned long long id)
-{}
+void wsrep_thd_set_ignored_error(THD*, my_bool)
+{ }
