@@ -46,6 +46,7 @@
 #include <cstdlib>
 #include <string>
 #include "log_event.h"
+#include <slave.h>
 
 #include <sstream>
 
@@ -150,10 +151,6 @@ mysql_mutex_t LOCK_wsrep_config_state;
 mysql_mutex_t LOCK_wsrep_group_commit;
 mysql_mutex_t LOCK_wsrep_SR_pool;
 mysql_mutex_t LOCK_wsrep_SR_store;
-mysql_mutex_t LOCK_wsrep_joiner_monitor;
-mysql_mutex_t LOCK_wsrep_donor_monitor;
-mysql_cond_t  COND_wsrep_joiner_monitor;
-mysql_cond_t  COND_wsrep_donor_monitor;
 
 int wsrep_replaying= 0;
 ulong  wsrep_running_threads = 0; // # of currently running wsrep
@@ -164,7 +161,7 @@ ulong  wsrep_running_rollbacker_threads = 0; // # of running
 ulong  my_bind_addr;
 
 #ifdef HAVE_PSI_INTERFACE
-PSI_mutex_key
+PSI_mutex_key 
   key_LOCK_wsrep_replaying, key_LOCK_wsrep_ready, key_LOCK_wsrep_sst,
   key_LOCK_wsrep_sst_thread, key_LOCK_wsrep_sst_init,
   key_LOCK_wsrep_slave_threads, key_LOCK_wsrep_desync,
@@ -172,15 +169,13 @@ PSI_mutex_key
   key_LOCK_wsrep_group_commit,
   key_LOCK_wsrep_SR_pool,
   key_LOCK_wsrep_SR_store,
-  key_LOCK_wsrep_thd_queue,
-  key_LOCK_wsrep_joiner_monitor,
-  key_LOCK_wsrep_donor_monitor;
+  key_LOCK_wsrep_thd_queue;
 
 PSI_cond_key key_COND_wsrep_thd,
   key_COND_wsrep_replaying, key_COND_wsrep_ready, key_COND_wsrep_sst,
   key_COND_wsrep_sst_init, key_COND_wsrep_sst_thread,
-  key_COND_wsrep_thd_queue, key_COND_wsrep_slave_threads,
-  key_COND_wsrep_joiner_monitor, key_COND_wsrep_donor_monitor;
+  key_COND_wsrep_thd_queue, key_COND_wsrep_slave_threads;
+  
 
 PSI_file_key key_file_wsrep_gra_log;
 
@@ -198,9 +193,7 @@ static PSI_mutex_info wsrep_mutexes[]=
   { &key_LOCK_wsrep_config_state, "LOCK_wsrep_config_state", PSI_FLAG_GLOBAL},
   { &key_LOCK_wsrep_group_commit, "LOCK_wsrep_group_commit", PSI_FLAG_GLOBAL},
   { &key_LOCK_wsrep_SR_pool, "LOCK_wsrep_SR_pool", PSI_FLAG_GLOBAL},
-  { &key_LOCK_wsrep_SR_store, "LOCK_wsrep_SR_store", PSI_FLAG_GLOBAL},
-  { &key_LOCK_wsrep_joiner_monitor, "LOCK_wsrep_joiner_monitor", PSI_FLAG_GLOBAL},
-  { &key_LOCK_wsrep_donor_monitor, "LOCK_wsrep_donor_monitor", PSI_FLAG_GLOBAL}
+  { &key_LOCK_wsrep_SR_store, "LOCK_wsrep_SR_store", PSI_FLAG_GLOBAL}
 };
 
 static PSI_cond_info wsrep_conds[]=
@@ -211,9 +204,7 @@ static PSI_cond_info wsrep_conds[]=
   { &key_COND_wsrep_sst_thread, "wsrep_sst_thread", 0},
   { &key_COND_wsrep_thd, "THD::COND_wsrep_thd", 0},
   { &key_COND_wsrep_replaying, "COND_wsrep_replaying", PSI_FLAG_GLOBAL},
-  { &key_COND_wsrep_slave_threads, "COND_wsrep_wsrep_slave_threads", PSI_FLAG_GLOBAL},
-  { &key_COND_wsrep_joiner_monitor, "COND_wsrep_joiner_monitor", PSI_FLAG_GLOBAL},
-  { &key_COND_wsrep_donor_monitor, "COND_wsrep_donor_monitor", PSI_FLAG_GLOBAL}
+  { &key_COND_wsrep_slave_threads, "COND_wsrep_wsrep_slave_threads", PSI_FLAG_GLOBAL}
 };
 
 static PSI_file_info wsrep_files[]=
@@ -222,17 +213,14 @@ static PSI_file_info wsrep_files[]=
 };
 
 PSI_thread_key key_wsrep_sst_joiner, key_wsrep_sst_donor,
-  key_wsrep_rollbacker, key_wsrep_applier,
-  key_wsrep_sst_joiner_monitor, key_wsrep_sst_donor_monitor;
+  key_wsrep_rollbacker, key_wsrep_applier;
 
 static PSI_thread_info wsrep_threads[]=
 {
  {&key_wsrep_sst_joiner, "wsrep_sst_joiner_thread", PSI_FLAG_GLOBAL},
  {&key_wsrep_sst_donor, "wsrep_sst_donor_thread", PSI_FLAG_GLOBAL},
  {&key_wsrep_rollbacker, "wsrep_rollbacker_thread", PSI_FLAG_GLOBAL},
- {&key_wsrep_applier, "wsrep_applier_thread", PSI_FLAG_GLOBAL},
- {&key_wsrep_sst_joiner_monitor, "wsrep_sst_joiner_monitor", PSI_FLAG_GLOBAL},
- {&key_wsrep_sst_donor_monitor, "wsrep_sst_donor_monitor", PSI_FLAG_GLOBAL}
+ {&key_wsrep_applier, "wsrep_applier_thread", PSI_FLAG_GLOBAL}
 };
 
 #endif /* HAVE_PSI_INTERFACE */
@@ -267,45 +255,6 @@ char* wsrep_cluster_capabilities    = NULL;
 
 wsp::Config_state *wsrep_config_state;
 
-void WSREP_LOG(void (*fun)(const char* fmt, ...), const char* fmt, ...)
-{
-  /* Allocate short buffer from stack. If the vsnprintf() return value
-     indicates that the message was truncated, a new buffer will be allocated
-     dynamically and the message will be reprinted. */
-  char msg[128] = {'\0'};
-  va_list arglist;
-  va_start(arglist, fmt);
-  int n= vsnprintf(msg, sizeof(msg) - 1, fmt, arglist);
-  va_end(arglist);
-  if (n < 0)
-  {
-    sql_print_warning("WSREP: Printing message failed");
-  }
-  else if (n < (int)sizeof(msg))
-  {
-    fun("WSREP: %s", msg);
-  }
-  else
-  {
-    size_t dynbuf_size= std::max(n, 4096);
-    char* dynbuf= (char*) my_malloc(dynbuf_size, MYF(0));
-    if (dynbuf)
-    {
-      va_start(arglist, fmt);
-      (void)vsnprintf(&dynbuf[0], dynbuf_size - 1, fmt, arglist);
-      va_end(arglist);
-      dynbuf[dynbuf_size - 1] = '\0';
-      fun("WSREP: %s", &dynbuf[0]);
-      my_free(dynbuf);
-    }
-    else
-    {
-      /* Memory allocation for vector failed, print truncated message. */
-      fun("WSREP: %s", msg);
-    }
-  }
-}
-
 
 wsrep_uuid_t               local_uuid       = WSREP_UUID_UNDEFINED;
 wsrep_seqno_t              local_seqno      = WSREP_SEQNO_UNDEFINED;
@@ -331,7 +280,7 @@ static void wsrep_log_cb(wsrep::log::level level, const char *msg)
       sql_print_warning("WSREP: %s", msg);
       break;
     case wsrep::log::error:
-    sql_print_error("WSREP: %s", msg);
+      sql_print_error("WSREP: %s", msg);
     break;
     case wsrep::log::debug:
       if (wsrep_debug) sql_print_information ("[Debug] WSREP: %s", msg);
@@ -601,11 +550,11 @@ static std::string wsrep_server_incoming_address()
     {
       if (node_addr.size())
       {
-        size_t const ip_len_mdb= wsrep_host_len(node_addr.c_str(), node_addr.size());
-        if (ip_len_mdb + 7 /* :55555\0 */ < inc_addr_max)
+        size_t const ip_len= wsrep_host_len(node_addr.c_str(), node_addr.size());
+        if (ip_len + 7 /* :55555\0 */ < inc_addr_max)
         {
-          memcpy (inc_addr, node_addr.c_str(), ip_len_mdb);
-          snprintf(inc_addr + ip_len_mdb, inc_addr_max - ip_len_mdb, ":%u",
+          memcpy (inc_addr, node_addr.c_str(), ip_len);
+          snprintf(inc_addr + ip_len, inc_addr_max - ip_len, ":%u",
                    (int)mysqld_port);
         }
         else
@@ -732,7 +681,6 @@ void wsrep_init_globals()
 {
   wsrep_init_sidno(Wsrep_server_state::instance().connected_gtid().id());
   wsrep_init_schema();
-
   if (WSREP_ON)
   {
     Wsrep_server_state::instance().initialized();
@@ -770,8 +718,6 @@ int wsrep_init()
 
   global_system_variables.wsrep_on= 1;
 
-  WSREP_ON_= wsrep_provider && strcmp(wsrep_provider, WSREP_NONE);
-
   if (wsrep_gtid_mode && opt_bin_log && !opt_log_slave_updates)
   {
     WSREP_ERROR("Option --log-slave-updates is required if "
@@ -843,13 +789,6 @@ void wsrep_thr_init()
                    &LOCK_wsrep_SR_pool, MY_MUTEX_INIT_FAST);
   mysql_mutex_init(key_LOCK_wsrep_SR_store,
                    &LOCK_wsrep_SR_store, MY_MUTEX_INIT_FAST);
-  mysql_mutex_init(key_LOCK_wsrep_joiner_monitor,
-                   &LOCK_wsrep_joiner_monitor, MY_MUTEX_INIT_FAST);
-  mysql_mutex_init(key_LOCK_wsrep_donor_monitor,
-                   &LOCK_wsrep_donor_monitor, MY_MUTEX_INIT_FAST);
-  mysql_cond_init(key_COND_wsrep_joiner_monitor, &COND_wsrep_joiner_monitor, NULL);
-  mysql_cond_init(key_COND_wsrep_donor_monitor, &COND_wsrep_donor_monitor, NULL);
-
   DBUG_VOID_RETURN;
 }
 
@@ -953,10 +892,6 @@ void wsrep_thr_deinit()
   mysql_mutex_destroy(&LOCK_wsrep_group_commit);
   mysql_mutex_destroy(&LOCK_wsrep_SR_pool);
   mysql_mutex_destroy(&LOCK_wsrep_SR_store);
-  mysql_mutex_destroy(&LOCK_wsrep_joiner_monitor);
-  mysql_mutex_destroy(&LOCK_wsrep_donor_monitor);
-  mysql_cond_destroy(&COND_wsrep_joiner_monitor);
-  mysql_cond_destroy(&COND_wsrep_donor_monitor);
 
   delete wsrep_config_state;
   wsrep_config_state= 0;                        // Safety
@@ -1711,39 +1646,6 @@ static bool wsrep_can_run_in_toi(THD *thd, const char *db, const char *table,
     {
       return false;
     }
-    /*
-      If mariadb master has replicated a CTAS, we should not replicate the create table
-      part separately as TOI, but to replicate both create table and following inserts
-      as one write set.
-      Howver, if CTAS creates empty table, we should replicate the create table alone
-      as TOI. We have to do relay log event lookup to see if row events follow the
-      create table event.
-    */
-    if (thd->slave_thread && !(thd->rgi_slave->gtid_ev_flags2 & Gtid_log_event::FL_STANDALONE))
-    {
-      /* this is CTAS, either empty or populated table */
-      ulonglong event_size = 0;
-      enum Log_event_type ev_type= wsrep_peak_event(thd->rgi_slave, &event_size);
-      switch (ev_type)
-      {
-      case QUERY_EVENT:
-        /* CTAS with empty table, we replicate create table as TOI */
-        break;
-
-      case TABLE_MAP_EVENT:
-        WSREP_DEBUG("replicating CTAS of empty table as TOI");
-        // fall through
-      case WRITE_ROWS_EVENT:
-        /* CTAS with populated table, we replicate later at commit time */
-        WSREP_DEBUG("skipping create table of CTAS replication");
-        return false;
-
-      default:
-        WSREP_WARN("unexpected async replication event: %d", ev_type);
-      }
-      return true;
-    }
-    /* no next async replication event */
     return true;
 
   case SQLCOM_CREATE_VIEW:
@@ -1882,13 +1784,16 @@ static void wsrep_TOI_begin_failed(THD* thd, const wsrep_buf_t* /* const err */)
     if (wsrep_emulate_bin_log) wsrep_thd_binlog_trx_reset(thd);
     if (wsrep_write_dummy_event(thd, "TOI begin failed")) { goto fail; }
     wsrep::client_state& cs(thd->wsrep_cs());
-    int const ret= cs.leave_toi_local(wsrep::mutable_buffer());
+    std::string const err(wsrep::to_c_string(cs.current_error()));
+    wsrep::mutable_buffer err_buf;
+    err_buf.push_back(err);
+    int const ret= cs.leave_toi_local(err_buf);
     if (ret)
     {
       WSREP_ERROR("Leaving critical section for failed TOI failed: thd: %lld, "
                   "schema: %s, SQL: %s, rcode: %d wsrep_error: %s",
                   (long long)thd->real_id, thd->db.str,
-                  thd->query(), ret, wsrep::to_c_string(cs.current_error()));
+                  thd->query(), ret, err.c_str());
       goto fail;
     }
   }
@@ -1951,7 +1856,9 @@ static int wsrep_TOI_begin(THD *thd, const char *db, const char *table,
 
   wsrep::client_state& cs(thd->wsrep_cs());
   int ret= cs.enter_toi_local(key_array,
-                              wsrep::const_buffer(buff.ptr, buff.len));
+                              wsrep::const_buffer(buff.ptr, buff.len),
+                              wsrep::provider::flag::start_transaction |
+                              wsrep::provider::flag::commit);
 
   if (ret)
   {
@@ -2007,7 +1914,12 @@ static void wsrep_TOI_end(THD *thd) {
   if (wsrep_thd_is_local_toi(thd))
   {
     wsrep_set_SE_checkpoint(client_state.toi_meta().gtid());
-    int ret= client_state.leave_toi_local(wsrep::mutable_buffer());
+    wsrep::mutable_buffer err;
+    if (thd->is_error() && !wsrep_must_ignore_error(thd))
+    {
+        wsrep_store_error(thd, err);
+    }
+    int const ret= client_state.leave_toi_local(err);
     if (!ret)
     {
       WSREP_DEBUG("TO END: %lld", client_state.toi_meta().seqno().get());
@@ -2070,10 +1982,10 @@ int wsrep_to_isolation_begin(THD *thd, const char *db_, const char *table_,
   DBUG_ASSERT(wsrep_thd_is_local(thd));
   DBUG_ASSERT(thd->wsrep_trx().ws_meta().seqno().is_undefined());
 
-  if (Wsrep_server_state::instance().desynced_on_pause())
+  if (thd->global_read_lock.is_acquired())
   {
-    my_message(ER_UNKNOWN_COM_ERROR,
-               "Aborting TOI: Global Read-Lock (FTWRL) in place.", MYF(0));
+    WSREP_DEBUG("Aborting TOI: Global Read-Lock (FTWRL) in place: %s %llu",
+                WSREP_QUERY(thd), thd->thread_id);
     return -1;
   }
 
@@ -2197,18 +2109,12 @@ void wsrep_handle_mdl_conflict(MDL_context *requestor_ctx,
     if (wsrep_thd_is_toi(granted_thd) ||
         wsrep_thd_is_applying(granted_thd))
     {
-      if (wsrep_thd_is_aborting(granted_thd))
-      {
-        WSREP_DEBUG("BF thread waiting for SR in aborting state");
-        ticket->wsrep_report(wsrep_debug);
-        mysql_mutex_unlock(&granted_thd->LOCK_thd_data);
-      }
-      else if (wsrep_thd_is_SR(granted_thd) && !wsrep_thd_is_SR(request_thd))
+      if (wsrep_thd_is_SR(granted_thd) && !wsrep_thd_is_SR(request_thd))
       {
         WSREP_MDL_LOG(INFO, "MDL conflict, DDL vs SR", 
                       schema, schema_len, request_thd, granted_thd);
         mysql_mutex_unlock(&granted_thd->LOCK_thd_data);
-        wsrep_abort_thd(request_thd, granted_thd, 1);
+        wsrep_abort_thd((void*)request_thd, (void*)granted_thd, 1);
       }
       else
       {
@@ -2232,7 +2138,7 @@ void wsrep_handle_mdl_conflict(MDL_context *requestor_ctx,
                   wsrep_thd_transaction_state_str(granted_thd));
       ticket->wsrep_report(wsrep_debug);
       mysql_mutex_unlock(&granted_thd->LOCK_thd_data);
-      wsrep_abort_thd(request_thd, granted_thd, 1);
+      wsrep_abort_thd((void*)request_thd, (void*)granted_thd, 1);
     }
     else
     {
@@ -2354,7 +2260,6 @@ int wsrep_wait_committing_connections_close(int wait_time)
 {
   int sleep_time= 100;
 
-  WSREP_DEBUG("wait for committing transaction to close: %d sleep: %d", wait_time, sleep_time);
   while (server_threads.iterate(have_committing_connections) && wait_time > 0)
   {
     WSREP_DEBUG("wait for committing transaction to close: %d", wait_time);
@@ -2507,7 +2412,7 @@ int wsrep_must_ignore_error(THD* thd)
   const uint flags= sql_command_flags[thd->lex->sql_command];
 
   DBUG_ASSERT(error);
-  DBUG_ASSERT(wsrep_thd_is_toi(thd) || wsrep_thd_is_applying(thd));
+  DBUG_ASSERT(wsrep_thd_is_toi(thd));
 
   if ((wsrep_ignore_apply_errors & WSREP_IGNORE_ERRORS_ON_DDL))
     goto ignore_error;
@@ -2725,13 +2630,6 @@ void* start_wsrep_THD(void *arg)
 
   mysql_thread_set_psi_id(thd->thread_id);
   thd->thr_create_utime=  microsecond_interval_timer();
-  if (MYSQL_CALLBACK_ELSE(thread_scheduler, init_new_connection_thread, (), 0))
-  {
-    close_connection(thd, ER_OUT_OF_RESOURCES);
-    statistic_increment(aborted_connects,&LOCK_status);
-    MYSQL_CALLBACK(thread_scheduler, end_thread, (thd, 0));
-    goto error;
-  }
 
 // </5.1.17>
   /*
@@ -2752,7 +2650,7 @@ void* start_wsrep_THD(void *arg)
   {
     close_connection(thd, ER_OUT_OF_RESOURCES);
     statistic_increment(aborted_connects,&LOCK_status);
-    MYSQL_CALLBACK(thread_scheduler, end_thread, (thd, 0));
+    unlink_thd(thd);
     delete thd;
     delete thd_args;
     goto error;
@@ -2828,7 +2726,7 @@ void* start_wsrep_THD(void *arg)
   if (plugins_are_initialized)
   {
     net_end(&thd->net);
-    MYSQL_CALLBACK(thread_scheduler, end_thread, (thd, 1));
+    unlink_thd(thd);
   }
   else
   {
@@ -2837,9 +2735,9 @@ void* start_wsrep_THD(void *arg)
       'Error in my_thread_global_end(): 2 threads didn't exit'
       at server shutdown
     */
+    server_threads.erase(thd);
   }
 
-  server_threads.erase(thd);
   delete thd;
   my_thread_end();
   return(NULL);
@@ -2878,3 +2776,54 @@ bool wsrep_consistency_check(THD *thd)
 {
   return thd->wsrep_consistency_check == CONSISTENCY_CHECK_RUNNING;
 }
+
+
+/*
+  Commit an empty transaction.
+
+  If the transaction is real and the wsrep transaction is still active,
+  the transaction did not generate any rows or keys and is committed
+  as empty. Here the wsrep transaction is rolled back and after statement
+  step is performed to leave the wsrep transaction in the state as it
+  never existed.
+
+  This should not be an inline functions as it requires a lot of stack space
+  because of WSREP_DBUG() usage.  It's also not a function that is
+  frequently called.
+*/
+
+void wsrep_commit_empty(THD* thd, bool all)
+{
+  DBUG_ENTER("wsrep_commit_empty");
+  WSREP_DEBUG("wsrep_commit_empty(%llu)", thd->thread_id);
+  if (wsrep_is_real(thd, all) &&
+      wsrep_thd_is_local(thd) &&
+      thd->wsrep_trx().active() &&
+      thd->wsrep_trx().state() != wsrep::transaction::s_committed)
+  {
+    /* @todo CTAS with STATEMENT binlog format and empty result set
+       seems to be committing empty. Figure out why and try to fix
+       elsewhere. */
+    DBUG_ASSERT(!wsrep_has_changes(thd) ||
+                (thd->lex->sql_command == SQLCOM_CREATE_TABLE &&
+                 !thd->is_current_stmt_binlog_format_row()));
+    bool have_error= wsrep_current_error(thd);
+    int ret= wsrep_before_rollback(thd, all) ||
+      wsrep_after_rollback(thd, all) ||
+      wsrep_after_statement(thd);
+    /* The committing transaction was empty but it held some locks and
+       got BF aborted. As there were no certified changes in the
+       data, we ignore the deadlock error and rely on error reporting
+       by storage engine/server. */
+    if (!ret && !have_error && wsrep_current_error(thd))
+    {
+      DBUG_ASSERT(wsrep_current_error(thd) == wsrep::e_deadlock_error);
+      thd->wsrep_cs().reset_error();
+    }
+    if (ret)
+    {
+      WSREP_DEBUG("wsrep_commit_empty failed: %d", wsrep_current_error(thd));
+    }
+  }
+  DBUG_VOID_RETURN;
+}
