@@ -1,11 +1,11 @@
 /************* TabDos C++ Program Source Code File (.CPP) **************/
 /* PROGRAM NAME: TABDOS                                                */
 /* -------------                                                       */
-/*  Version 4.9.5                                                      */
+/*  Version 4.9.4                                                      */
 /*                                                                     */
 /* COPYRIGHT:                                                          */
 /* ----------                                                          */
-/*  (C) Copyright to the author Olivier BERTRAND          1998-2020    */
+/*  (C) Copyright to the author Olivier BERTRAND          1998-2019    */
 /*                                                                     */
 /* WHAT THIS PROGRAM DOES:                                             */
 /* -----------------------                                             */
@@ -45,7 +45,7 @@
 #include "global.h"
 #include "osutil.h"
 #include "plgdbsem.h"
-//#include "catalog.h"
+#include "catalog.h"
 #include "mycat.h"
 #include "xindex.h"
 #include "filamap.h"
@@ -161,12 +161,7 @@ bool DOSDEF::DefineAM(PGLOBAL g, LPCSTR am, int)
 //Last = GetIntCatInfo("Last", 0);
   Ending = GetIntCatInfo("Ending", CRLF);
 
-	if (Ending <= 0) {
-		Ending = (Recfm == RECFM_BIN || Recfm == RECFM_VCT) ? 0 : CRLF;
-		SetIntCatInfo("Ending", Ending);
-	} // endif ending
-
-	if (Recfm == RECFM_FIX || Recfm == RECFM_BIN) {
+  if (Recfm == RECFM_FIX || Recfm == RECFM_BIN) {
     Huge = GetBoolCatInfo("Huge", Cat->GetDefHuge());
     Padded = GetBoolCatInfo("Padded", false);
     Blksize = GetIntCatInfo("Blksize", 0);
@@ -196,8 +191,7 @@ bool DOSDEF::GetOptFileName(PGLOBAL g, char *filename)
     case RECFM_FIX: ftype = ".fop"; break;
     case RECFM_BIN: ftype = ".bop"; break;
     case RECFM_VCT: ftype = ".vop"; break;
-		case RECFM_CSV: ftype = ".cop"; break;
-		case RECFM_DBF: ftype = ".dbp"; break;
+    case RECFM_DBF: ftype = ".dbp"; break;
     default:
       sprintf(g->Message, MSG(INVALID_FTYPE), Recfm);
       return true;
@@ -267,8 +261,7 @@ bool DOSDEF::DeleteIndexFile(PGLOBAL g, PIXDEF pxdf)
     case RECFM_FIX: ftype = ".fnx"; break;
     case RECFM_BIN: ftype = ".bnx"; break;
     case RECFM_VCT: ftype = ".vnx"; break;
-		case RECFM_CSV: ftype = ".cnx"; break;
-		case RECFM_DBF: ftype = ".dbx"; break;
+    case RECFM_DBF: ftype = ".dbx"; break;
     default:
       sprintf(g->Message, MSG(BAD_RECFM_VAL), Recfm);
       return true;
@@ -359,26 +352,7 @@ PTDB DOSDEF::GetTable(PGLOBAL g, MODE mode)
   /*  Allocate table and file processing class of the proper type.     */
   /*  Column blocks will be allocated only when needed.                */
   /*********************************************************************/
-	if (Recfm == RECFM_DBF) {
-		if (Catfunc == FNC_NO) {
-			if (Zipped) {
-				if (mode == MODE_READ || mode == MODE_ANY || mode == MODE_ALTER) {
-					txfp = new(g) UZDFAM(this);
-				}	else {
-					strcpy(g->Message, "Zipped DBF tables are read only");
-					return NULL;
-				}	// endif's mode
-
-			} else if (map)
-				txfp = new(g) DBMFAM(this);
-			else
-				txfp = new(g) DBFFAM(this);
-
-			tdbp = new(g) TDBFIX(this, txfp);
-		} else
-			tdbp = new(g) TDBDCL(this);    // Catfunc should be 'C'
-
-	} else if (Zipped) {
+	if (Zipped) {
 #if defined(ZIP_SUPPORT)
 		if (Recfm == RECFM_VAR) {
 			if (mode == MODE_READ || mode == MODE_ANY || mode == MODE_ALTER) {
@@ -408,6 +382,17 @@ PTDB DOSDEF::GetTable(PGLOBAL g, MODE mode)
 		sprintf(g->Message, MSG(NO_FEAT_SUPPORT), "ZIP");
 		return NULL;
 #endif  // !ZIP_SUPPORT
+	} else if (Recfm == RECFM_DBF) {
+    if (Catfunc == FNC_NO) {
+      if (map)
+        txfp = new(g) DBMFAM(this);
+      else
+        txfp = new(g) DBFFAM(this);
+
+      tdbp = new(g) TDBFIX(this, txfp);
+    } else                   // Catfunc should be 'C'
+      tdbp = new(g) TDBDCL(this);
+
   } else if (Recfm != RECFM_VAR && Compressed < 2) {
     if (Huge)
       txfp = new(g) BGXFAM(this);
@@ -648,8 +633,9 @@ int TDBDOS::MakeBlockValues(PGLOBAL g)
   PDOSDEF    defp = (PDOSDEF)To_Def;
   PDOSCOL    colp = NULL;
   PDBUSER    dup = PlgGetUser(g);
-  PCATLG     cat __attribute__((unused))= defp->GetCat();
 //void      *memp = cat->GetDescp();
+  (void) defp->GetCat();                        // XXX Should be removed?
+
 
   if ((nrec = defp->GetElemt()) < 2) {
     if (!To_Def->Partitioned()) {
@@ -1018,8 +1004,10 @@ bool TDBDOS::GetBlockValues(PGLOBAL g)
   FILE      *opfile;
   PCOLDEF    cdp;
   PDOSDEF    defp = (PDOSDEF)To_Def;
-  PCATLG     cat __attribute__((unused))= defp->GetCat();
-	PDBUSER    dup = PlgGetUser(g);
+  PDBUSER    dup = PlgGetUser(g);
+
+  (void) defp->GetCat();                        // XXX Should be removed?
+
 
 #if 0
   if (Mode == MODE_INSERT && Txfp->GetAmType() == TYPE_AM_DOS)
@@ -1429,7 +1417,9 @@ PBF TDBDOS::CheckBlockFilari(PGLOBAL g, PXOB *arg, int op, bool *cnv)
   bool    conv = false, xdb2 = false;
   PXOB    xp[2];
   PCOL    colp;
-  PBF     bfp = NULL;
+//LSTVAL *vlp = NULL;
+//SFROW  *sfr[2];
+  PBF    bfp = NULL;
 
   for (i = 0; i < 2; i++) {
     switch (arg[i]->GetType()) {
@@ -2269,7 +2259,7 @@ int TDBDOS::ReadDB(PGLOBAL g)
 /***********************************************************************/
 bool TDBDOS::PrepareWriting(PGLOBAL)
   {
-  if (Ftype == RECFM_VAR && (Mode == MODE_INSERT || Txfp->GetUseTemp())) {
+  if (!Ftype && (Mode == MODE_INSERT || Txfp->GetUseTemp())) {
     char *p;
 
     /*******************************************************************/
@@ -2554,8 +2544,7 @@ void DOSCOL::ReadColumn(PGLOBAL g)
   /*********************************************************************/
   /*  For a variable length file, check if the field exists.           */
   /*********************************************************************/
-  if ((tdbp->Ftype == RECFM_VAR || tdbp->Ftype == RECFM_CSV)
-				&& strlen(tdbp->To_Line) < (unsigned)Deplac)
+  if (tdbp->Ftype == RECFM_VAR && strlen(tdbp->To_Line) < (unsigned)Deplac)
     field = 0;
   else if (Dsp)
     for(i = 0; i < field; i++)
@@ -2565,8 +2554,7 @@ void DOSCOL::ReadColumn(PGLOBAL g)
   switch (tdbp->Ftype) {
     case RECFM_VAR:
     case RECFM_FIX:            // Fixed length text file
-		case RECFM_CSV:            // Variable length CSV or FMT file
-		case RECFM_DBF:            // Fixed length DBase file
+    case RECFM_DBF:            // Fixed length DBase file
       if (Nod) switch (Buf_Type) {
         case TYPE_INT:
         case TYPE_SHORT:
