@@ -52,10 +52,19 @@
 /*  External functions.                                                */
 /***********************************************************************/
 USETEMP UseTemp(void);
-bool    JsonAllPath(void);
-int     GetDefaultDepth(void);
 char   *GetJsonNull(void);
 
+//typedef struct _jncol {
+//  struct _jncol *Next;
+//  char *Name;
+//  char *Fmt;
+//  int   Type;
+//  int   Len;
+//  int   Scale;
+//  bool  Cbn;
+//  bool  Found;
+//} JCOL, *PJCL;
+
 /***********************************************************************/
 /* JSONColumns: construct the result blocks containing the description */
 /* of all the columns of a table contained inside a JSON file.         */
@@ -158,20 +167,23 @@ JSONDISC::JSONDISC(PGLOBAL g, uint *lg)
   jsp = NULL;
   row = NULL;
   sep = NULL;
-  i = n = bf = ncol = lvl = sz = 0;
-  all = strfy = false;
+  i = n = bf = ncol = lvl = 0;
+  all = false;
 } // end of JSONDISC constructor
 
 int JSONDISC::GetColumns(PGLOBAL g, PCSZ db, PCSZ dsn, PTOS topt)
 {
   char filename[_MAX_PATH];
   bool mgo = (GetTypeID(topt->type) == TAB_MONGO);
+  PCSZ level = GetStringTableOption(g, topt, "Level", NULL);
 
-	lvl = GetIntegerTableOption(g, topt, "Level", GetDefaultDepth());
-	lvl = GetIntegerTableOption(g, topt, "Depth", lvl);
-	sep = GetStringTableOption(g, topt, "Separator", ".");
-	sz = GetIntegerTableOption(g, topt, "Jsize", 1024);
-	strfy = GetBooleanTableOption(g, topt, "Stringify", false);
+  if (level) {
+    lvl = atoi(level);
+    lvl = (lvl > 16) ? 16 : lvl;
+  } else
+    lvl = 0;
+
+  sep = GetStringTableOption(g, topt, "Separator", ".");
 
   /*********************************************************************/
   /*  Open the input file.                                             */
@@ -242,14 +254,14 @@ int JSONDISC::GetColumns(PGLOBAL g, PCSZ db, PCSZ dsn, PTOS topt)
 
     jsp = (tjsp->GetDoc()) ? tjsp->GetDoc()->GetValue(0) : NULL;
   } else {
-		if (!((tdp->Lrecl = GetIntegerTableOption(g, topt, "Lrecl", 0)))) {
-			if (!mgo) {
-				sprintf(g->Message, "LRECL must be specified for pretty=%d", tdp->Pretty);
-				return 0;
-			} else
-				tdp->Lrecl = 8192;       // Should be enough
-
-		} // endif Lrecl
+    if (!(tdp->Lrecl = GetIntegerTableOption(g, topt, "Lrecl", 0)))
+    {
+      if (!mgo) {
+        sprintf(g->Message, "LRECL must be specified for pretty=%d", tdp->Pretty);
+        return 0;
+      } else
+        tdp->Lrecl = 8192;       // Should be enough
+    }
 
     tdp->Ending = GetIntegerTableOption(g, topt, "Ending", CRLF);
 
@@ -294,7 +306,7 @@ int JSONDISC::GetColumns(PGLOBAL g, PCSZ db, PCSZ dsn, PTOS topt)
     // Allocate the parse work memory
     PGLOBAL G = (PGLOBAL)PlugSubAlloc(g, NULL, sizeof(GLOBAL));
     memset(G, 0, sizeof(GLOBAL));
-    G->Sarea_Size = (size_t)tdp->Lrecl * 10;
+    G->Sarea_Size = tdp->Lrecl * 10;
     G->Sarea = PlugSubAlloc(g, NULL, G->Sarea_Size);
     PlugSubSet(G->Sarea, G->Sarea_Size);
     G->jump_level = 0;
@@ -384,17 +396,13 @@ err:
 
 bool JSONDISC::Find(PGLOBAL g, PJVAL jvp, PCSZ key, int j)
 {
-  char  *p, *pc = colname + strlen(colname);
-  int    ars;
-	size_t n;
-  PJOB   job;
-  PJAR   jar;
+  char *p, *pc = colname + strlen(colname);
+  int   ars;
+  PJOB  job;
+  PJAR  jar;
 
   if ((valp = jvp ? jvp->GetValue() : NULL)) {
-		if (JsonAllPath() && !fmt[bf])
-			strcat(fmt, colname);
-
-		jcol.Type = valp->GetType();
+    jcol.Type = valp->GetType();
     jcol.Len = valp->GetValLen();
     jcol.Scale = valp->GetValPrec();
     jcol.Cbn = valp->IsNull();
@@ -417,10 +425,8 @@ bool JSONDISC::Find(PGLOBAL g, PJVAL jvp, PCSZ key, int j)
           PCSZ k = jrp->GetKey();
 
           if (*k != '$') {
-						n = sizeof(fmt) - strlen(fmt) -1;
-						strncat(strncat(fmt, sep, n), k, n - strlen(sep));
-						n = sizeof(colname) - strlen(colname) - 1;
-						strncat(strncat(colname, "_", n), k, n - 1);
+            strncat(strncat(fmt, sep, 128), k, 128);
+            strncat(strncat(colname, "_", 64), k, 64);
           } // endif Key
 
           if (Find(g, jrp->GetVal(), k, j + 1))
@@ -439,26 +445,19 @@ bool JSONDISC::Find(PGLOBAL g, PJVAL jvp, PCSZ key, int j)
           ars = MY_MIN(jar->GetSize(false), 1);
 
         for (int k = 0; k < ars; k++) {
-					n = sizeof(fmt) - (strlen(fmt) + 1);
-
-					if (!tdp->Xcol || stricmp(tdp->Xcol, key)) {
+          if (!tdp->Xcol || stricmp(tdp->Xcol, key)) {
             sprintf(buf, "%d", k);
 
-						if (tdp->Uri) {
-							strncat(strncat(fmt, sep, n), buf, n - strlen(sep));
-						} else {
-							strncat(strncat(fmt, "[", n), buf, n - 1);
-							strncat(fmt, "]", n - (strlen(buf) + 1));
-						} // endif uri
+            if (tdp->Uri)
+              strncat(strncat(fmt, sep, 128), buf, 128);
+            else
+              strncat(strncat(strncat(fmt, "[", 128), buf, 128), "]", 128);
 
-						if (all) {
-							n = sizeof(colname) - (strlen(colname) + 1);
-							strncat(strncat(colname, "_", n), buf, n - 1);
-						} // endif all
+            if (all)
+              strncat(strncat(colname, "_", 64), buf, 64);
 
-					} else {
-						strncat(fmt, (tdp->Uri ? sep : "[*]"), n);
-					}
+          } else
+            strncat(fmt, (tdp->Uri ? sep : "[*]"), 128);
 
           if (Find(g, jar->GetValue(k), "", j))
             return true;
@@ -473,16 +472,8 @@ bool JSONDISC::Find(PGLOBAL g, PJVAL jvp, PCSZ key, int j)
     } // endswitch Type
 
   } else if (lvl >= 0) {
-		if (strfy) {
-			if (!fmt[bf])
-				strcat(fmt, colname);
-
-			strcat(fmt, ".*");
-		}	else if (JsonAllPath() && !fmt[bf])
-			strcat(fmt, colname);
-
-		jcol.Type = TYPE_STRING;
-    jcol.Len = sz;
+    jcol.Type = TYPE_STRING;
+    jcol.Len = 256;
     jcol.Scale = 0;
     jcol.Cbn = true;
   } else
@@ -740,7 +731,6 @@ PTDB JSONDEF::GetTable(PGLOBAL g, MODE m)
 /***********************************************************************/
 TDBJSN::TDBJSN(PJDEF tdp, PTXF txfp) : TDBDOS(tdp, txfp)
   {
-	G = NULL;
   Top = NULL;
   Row = NULL;
   Val = NULL;
@@ -1487,18 +1477,7 @@ PVAL JSONCOL::MakeJson(PGLOBAL g, PJSON jsp)
   if (Value->IsTypeNum()) {
     strcpy(g->Message, "Cannot make Json for a numeric column");
     Value->Reset();
-	} else if (Value->GetType() == TYPE_BIN) {
-		if ((unsigned)Value->GetClen() >= sizeof(BSON)) {
-			ulong len = Tjp->Lrecl ? Tjp->Lrecl : 500;
-			PBSON bsp = JbinAlloc(g, NULL, len, jsp);
-
-			strcat(bsp->Msg, " column");
-			((BINVAL*)Value)->SetBinValue(bsp, sizeof(BSON));
-		} else {
-			strcpy(g->Message, "Column size too small");
-			Value->SetValue_char(NULL, 0);
-		} // endif Clen
-	}	else
+  } else
     Value->SetValue_psz(Serialize(g, jsp, NULL, 0));
 
   return Value;
@@ -1994,9 +1973,8 @@ int TDBJSON::MakeNewDoc(PGLOBAL g)
 /***********************************************************************/
 int TDBJSON::MakeDocument(PGLOBAL g)
   {
-  char   *p, *p1, *p2, *memory, *objpath, *key = NULL;
+  char   *p, *memory, *objpath, *key = NULL;
   int     len, i = 0;
-	my_bool a;
   MODE    mode = Mode;
   PJSON   jsp;
   PJOB    objp = NULL;
@@ -2039,39 +2017,22 @@ int TDBJSON::MakeDocument(PGLOBAL g)
   if ((objpath = PlugDup(g, Objname))) {
     if (*objpath == '$') objpath++;
     if (*objpath == '.') objpath++;
-		p1 = (*objpath == '[') ? objpath++ : NULL;
 
     /*********************************************************************/
     /*  Find the table in the tree structure.                            */
     /*********************************************************************/
-    for (p = objpath; jsp && p; p = (p2 ? p2 : NULL)) {
-			a = (p1 != NULL);
-			p1 = strchr(p, '[');
-			p2 = strchr(p, '.');
+    for (; jsp && objpath; objpath = p) {
+      if ((p = strchr(objpath, Sep)))
+        *p++ = 0;
 
-			if (!p2)
-				p2 = p1;
-			else if (p1) {
-				if (p1 < p2)
-					p2 = p1;
-				else if (p1 == p2 + 1)
-					*p2++ = 0;		 // Old syntax .[
-				else
-					p1 = NULL;
-
-			}	// endif p1
-
-			if (p2)
-				*p2++ = 0;
-
-      if (!a && *p && *p != '[' && !IsNum(p)) {
-        // obj is a key
+      if (*objpath != '[' && !IsNum(objpath)) {
+        // objpass is a key
         if (jsp->GetType() != TYPE_JOB) {
           strcpy(g->Message, "Table path does not match the json file");
           return RC_FX;
         } // endif Type
 
-        key = p;
+        key = objpath;
         objp = jsp->GetObject();
         arp = NULL;
         val = objp->GetValue(key);
@@ -2082,15 +2043,15 @@ int TDBJSON::MakeDocument(PGLOBAL g)
         } // endif val
 
       } else {
-        if (*p == '[') {
+        if (*objpath == '[') {
           // Old style
-          if (p[strlen(p) - 1] != ']') {
-            sprintf(g->Message, "Invalid Table path near %s", p);
+          if (objpath[strlen(objpath) - 1] != ']') {
+            sprintf(g->Message, "Invalid Table path %s", Objname);
             return RC_FX;
           } else
-            p++;
+            objpath++;
 
-        } // endif p
+        } // endif objpath
 
         if (jsp->GetType() != TYPE_JAR) {
           strcpy(g->Message, "Table path does not match the json file");
@@ -2099,7 +2060,7 @@ int TDBJSON::MakeDocument(PGLOBAL g)
 
         arp = jsp->GetArray();
         objp = NULL;
-        i = atoi(p) - B;
+        i = atoi(objpath) - B;
         val = arp->GetValue(i);
 
         if (!val) {
@@ -2110,7 +2071,7 @@ int TDBJSON::MakeDocument(PGLOBAL g)
       } // endif
 
       jsp = val->GetJson();
-    } // endfor p
+    } // endfor objpath
 
   } // endif objpath
 
@@ -2148,15 +2109,15 @@ int TDBJSON::Cardinality(PGLOBAL g)
   {
   if (!g)
     return (Xcol || Multiple) ? 0 : 1;
-	else if (Cardinal < 0) {
-		if (!Multiple) {
-			if (MakeDocument(g) == RC_OK)
-				Cardinal = Doc->size();
+  else if (Cardinal < 0)
+  {
+    if (!Multiple) {
+      if (MakeDocument(g) == RC_OK)
+        Cardinal = Doc->size();
 
-		} else
-			return 10;
-
-	} // endif Cardinal
+    } else
+      return 10;
+  }
 
   return Cardinal;
   } // end of Cardinality
