@@ -218,10 +218,8 @@ PQRYRES XMLColumns(PGLOBAL g, char *db, char *tab, PTOS topt, bool info)
 
     while (true) {
       if (!vp->atp &&
-          !(node = (vp->nl) ? vp->nl->GetItem(g, vp->k++, tdp->Usedom ?
-                                              node : NULL)
-            : NULL))
-      {
+        !(node = (vp->nl) ? vp->nl->GetItem(g, vp->k++, tdp->Usedom ? node : NULL)
+                          : NULL)) {
         if (j) {
           vp = lvlp[--j];
 
@@ -237,20 +235,18 @@ PQRYRES XMLColumns(PGLOBAL g, char *db, char *tab, PTOS topt, bool info)
         } else
           break;
       }
+
       xcol->Name[vp->n] = 0;
       fmt[vp->m] = 0;
 
      more:
       if (vp->atp) {
-				size_t z = sizeof(colname) - 1;
-        strncpy(colname, vp->atp->GetName(g), z);
-				colname[z] = 0;
+        strncpy(colname, vp->atp->GetName(g), sizeof(colname));
 				strncat(xcol->Name, colname, XLEN(xcol->Name));
 
         switch (vp->atp->GetText(g, buf, sizeof(buf))) {
           case RC_INFO:
             PushWarning(g, txmp);
-            /* falls through */
           case RC_OK:
             strncat(fmt, "@", XLEN(fmt));
             break;
@@ -311,7 +307,6 @@ PQRYRES XMLColumns(PGLOBAL g, char *db, char *tab, PTOS topt, bool info)
         switch (node->GetContent(g, buf, sizeof(buf))) {
           case RC_INFO:
             PushWarning(g, txmp);
-            /* falls through */
           case RC_OK:
 						xcol->Cbn = !strlen(buf);
             break;
@@ -700,7 +695,7 @@ PTDB TDBXML::Clone(PTABS t)
 /***********************************************************************/
 const CHARSET_INFO *TDBXML::data_charset()
 {
-	return &my_charset_utf8_general_ci;
+	return &my_charset_utf8mb3_general_ci;
 }	// end of data_charset
 
 /***********************************************************************/
@@ -817,15 +812,12 @@ bool TDBXML::Initialize(PGLOBAL g)
   if (Void)
     return false;
 
-  if (Columns) {
+  if (Columns && !Bufdone) {
     // Allocate the buffers that will contain node values
     for (colp = (PXMLCOL)Columns; colp; colp = (PXMLCOL)colp->GetNext())
-			if (!colp->IsSpecial()) {            // Not a pseudo column
-				if (!Bufdone && colp->AllocBuf(g, Mode == MODE_INSERT))
-					return true;
-
-				colp->Nx = colp->Sx = -1;
-			} // endif Special
+      if (!colp->IsSpecial())            // Not a pseudo column
+        if (colp->AllocBuf(g, Mode == MODE_INSERT))
+          return true;
 
     Bufdone = true;
     } // endif Bufdone
@@ -1277,8 +1269,7 @@ int TDBXML::ReadDB(PGLOBAL g)
 /***********************************************************************/
 bool TDBXML::CheckRow(PGLOBAL g, bool b)
   {
-  if (NewRow && Mode == MODE_INSERT)
-  {
+  if (NewRow && Mode == MODE_INSERT) {
     if (Rowname) {
       TabNode->AddText(g, "\n\t");
       RowNode = TabNode->AddChildNode(g, Rowname, RowNode);
@@ -1533,8 +1524,7 @@ bool XMLCOL::ParseXpath(PGLOBAL g, bool mode)
 
     // Analyze the Xpath for this column
     for (i = 0, p = pbuf; (p2 = strchr(p, '/')); i++, p = p2 + 1) {
-      if (Tdbp->Mulnode && !strncmp(p, Tdbp->Mulnode, p2 - p))
-      {
+      if (Tdbp->Mulnode && !strncmp(p, Tdbp->Mulnode, p2 - p)) {
         if (!Tdbp->Xpand && mode) {
           strcpy(g->Message, MSG(CONCAT_SUBNODE));
           return true;
@@ -1786,8 +1776,7 @@ void XMLCOL::WriteColumn(PGLOBAL g)
       else
         break;
 
-    if (ColNode)
-    {
+    if (ColNode) {
       if (Type)
         ValNode = ColNode->SelectSingleNode(g, Xname, Vxnp);
       else
@@ -1805,8 +1794,7 @@ void XMLCOL::WriteColumn(PGLOBAL g)
   /*  Create missing nodes.                                            */
   /*********************************************************************/
   if (ColNode == NULL) {
-    if (TopNode == NULL)
-    {
+    if (TopNode == NULL) {
       if (Tdbp->Clist) {
         Tdbp->RowNode->AddText(g, "\n\t\t");
         ColNode = Tdbp->RowNode->AddChildNode(g, Tdbp->Colname);
@@ -2028,8 +2016,7 @@ void XMULCOL::WriteColumn(PGLOBAL g)
       TopNode = ColNode;
       } // endfor k
 
-    if (ColNode)
-    {
+    if (ColNode) {
       if (Inod == Nod) {
         /***************************************************************/
         /*  The node value can be multiple.                            */
@@ -2047,12 +2034,10 @@ void XMULCOL::WriteColumn(PGLOBAL g)
           ValNode = Nlx->GetItem(g, Tdbp->Nsub, Vxnp);
 
       } else  // Inod != Nod
-      {
         if (Type)
           ValNode = ColNode->SelectSingleNode(g, Xname, Vxnp);
         else
           AttNode = ColNode->GetAttribute(g, Xname, Vxap);
-      }
     }
     if (TopNode || ValNode || AttNode)
       break;                     // We found the good column
@@ -2065,8 +2050,7 @@ void XMULCOL::WriteColumn(PGLOBAL g)
   /*  Create missing nodes.                                            */
   /*********************************************************************/
   if (ColNode == NULL) {
-    if (TopNode == NULL)
-    {
+    if (TopNode == NULL) {
       if (Tdbp->Clist) {
         Tdbp->RowNode->AddText(g, "\n\t\t");
         ColNode = Tdbp->RowNode->AddChildNode(g, Tdbp->Colname);
@@ -2075,7 +2059,6 @@ void XMULCOL::WriteColumn(PGLOBAL g)
       } else
         TopNode = Tdbp->RowNode;
     }
-
     for (; k < Nod && TopNode; k++) {
       if (!done) {
         TopNode->AddText(g, "\n\t\t");
