@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 2014, 2019, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2017, 2020, MariaDB Corporation.
+Copyright (c) 2017, 2019, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -29,11 +29,10 @@ Created 03/11/2014 Shaohua Wang
 #include "btr0cur.h"
 #include "btr0pcur.h"
 #include "ibuf0ibuf.h"
-#include "page0page.h"
 #include "trx0trx.h"
 
 /** Innodb B-tree index fill factor for bulk load. */
-long	innobase_fill_factor;
+uint	innobase_fill_factor;
 /** whether to reduce redo logging during ALTER TABLE */
 my_bool	innodb_log_optimize_ddl;
 
@@ -57,7 +56,7 @@ PageBulk::init()
 		m_mtr.set_log_mode(MTR_LOG_NO_REDO);
 		m_mtr.set_flush_observer(m_flush_observer);
 	} else {
-		m_mtr.set_named_space(m_index->space);
+		m_index->set_modified(m_mtr);
 	}
 
 	if (m_page_no == FIL_NULL) {
@@ -68,11 +67,12 @@ PageBulk::init()
 		the allocation order, and we will always generate redo log
 		for page allocation, even when creating a new tablespace. */
 		alloc_mtr.start();
-		alloc_mtr.set_named_space(m_index->space);
+		m_index->set_modified(alloc_mtr);
 
 		ulint	n_reserved;
 		bool	success;
-		success = fsp_reserve_free_extents(&n_reserved, m_index->space,
+		success = fsp_reserve_free_extents(&n_reserved,
+						   m_index->table->space,
 						   1, FSP_NORMAL, &alloc_mtr);
 		if (!success) {
 			alloc_mtr.commit();
@@ -84,10 +84,7 @@ PageBulk::init()
 		new_block = btr_page_alloc(m_index, 0, FSP_UP, m_level,
 					   &alloc_mtr, &m_mtr);
 
-		if (n_reserved > 0) {
-			fil_space_release_free_extents(m_index->space,
-						       n_reserved);
-		}
+		m_index->table->space->release_free_extents(n_reserved);
 
 		alloc_mtr.commit();
 
@@ -97,7 +94,7 @@ PageBulk::init()
 
 		if (new_page_zip) {
 			page_create_zip(new_block, m_index, m_level, 0,
-					NULL, &m_mtr);
+					&m_mtr);
 			memset(FIL_PAGE_PREV + new_page, 0xff, 8);
 			page_zip_write_header(new_page_zip,
 					      FIL_PAGE_PREV + new_page,
@@ -110,23 +107,20 @@ PageBulk::init()
 		} else {
 			ut_ad(!dict_index_is_spatial(m_index));
 			page_create(new_block, &m_mtr,
-				    dict_table_is_comp(m_index->table),
+				    m_index->table->not_redundant(),
 				    false);
-			mlog_write_ulint(FIL_PAGE_PREV + new_page, FIL_NULL,
-					 MLOG_4BYTES, &m_mtr);
-			mlog_write_ulint(FIL_PAGE_NEXT + new_page, FIL_NULL,
-					 MLOG_4BYTES, &m_mtr);
+			compile_time_assert(FIL_PAGE_NEXT
+					    == FIL_PAGE_PREV + 4);
+			compile_time_assert(FIL_NULL == 0xffffffff);
+			mlog_memset(new_block, FIL_PAGE_PREV, 8, 0xff, &m_mtr);
 			mlog_write_ulint(PAGE_HEADER + PAGE_LEVEL + new_page,
 					 m_level, MLOG_2BYTES, &m_mtr);
 			mlog_write_ull(PAGE_HEADER + PAGE_INDEX_ID + new_page,
 				       m_index->id, &m_mtr);
 		}
 	} else {
-		page_id_t	page_id(dict_index_get_space(m_index), m_page_no);
-		page_size_t	page_size(dict_table_page_size(m_index->table));
-
-		new_block = btr_block_get(page_id, page_size,
-					  RW_X_LATCH, m_index, &m_mtr);
+		new_block = btr_block_get(*m_index, m_page_no, RW_X_LATCH,
+					  false, &m_mtr);
 
 		new_page = buf_block_get_frame(new_block);
 		new_page_zip = buf_block_get_page_zip(new_block);
@@ -144,6 +138,7 @@ PageBulk::init()
 	}
 
 	m_block = new_block;
+	m_block->skip_flush_check = true;
 	m_page = new_page;
 	m_page_zip = new_page_zip;
 	m_page_no = new_page_no;
@@ -156,16 +151,14 @@ PageBulk::init()
 		m_reserved_space = dict_index_get_space_reserve();
 	} else {
 		m_reserved_space =
-			UNIV_PAGE_SIZE * (100 - innobase_fill_factor) / 100;
+			srv_page_size * (100 - innobase_fill_factor) / 100;
 	}
 
 	m_padding_space =
-		UNIV_PAGE_SIZE - dict_index_zip_pad_optimal_page_size(m_index);
+		srv_page_size - dict_index_zip_pad_optimal_page_size(m_index);
 	m_heap_top = page_header_get_ptr(new_page, PAGE_HEAP_TOP);
 	m_rec_no = page_header_get_field(new_page, PAGE_N_RECS);
-	ut_ad(page_header_get_field(m_page, PAGE_DIRECTION)
-	      == PAGE_NO_DIRECTION);
-	page_header_set_field(m_page, NULL, PAGE_DIRECTION, 0);
+
 	ut_d(m_total_data = 0);
 	/* See page_copy_rec_list_end_to_created_page() */
 	ut_d(page_header_set_field(m_page, NULL, PAGE_HEAP_TOP,
@@ -180,20 +173,21 @@ PageBulk::init()
 void
 PageBulk::insert(
 	const rec_t*		rec,
-	rec_offs*		offsets)
+	ulint*			offsets)
 {
 	ulint		rec_size;
 
 	ut_ad(m_heap != NULL);
 
 	rec_size = rec_offs_size(offsets);
+	ut_d(const bool is_leaf = page_rec_is_leaf(m_cur_rec));
 
 #ifdef UNIV_DEBUG
 	/* Check whether records are in order. */
-	if (!page_rec_is_infimum_low(page_offset(m_cur_rec))) {
+	if (!page_rec_is_infimum(m_cur_rec)) {
 		rec_t*	old_rec = m_cur_rec;
-		rec_offs* old_offsets = rec_get_offsets(
-			old_rec, m_index, NULL,	page_rec_is_leaf(old_rec),
+		ulint*	old_offsets = rec_get_offsets(
+			old_rec, m_index, NULL,	is_leaf,
 			ULINT_UNDEFINED, &m_heap);
 
 		ut_ad(cmp_rec_rec(rec, old_rec, offsets, old_offsets, m_index)
@@ -205,24 +199,21 @@ PageBulk::insert(
 
 	/* 1. Copy the record to page. */
 	rec_t*	insert_rec = rec_copy(m_heap_top, rec, offsets);
-	rec_offs_make_valid(insert_rec, m_index, offsets);
+	rec_offs_make_valid(insert_rec, m_index, is_leaf, offsets);
 
 	/* 2. Insert the record in the linked list. */
+	rec_t*	next_rec = page_rec_get_next(m_cur_rec);
+
+	page_rec_set_next(insert_rec, next_rec);
+	page_rec_set_next(m_cur_rec, insert_rec);
+
 	/* 3. Set the n_owned field in the inserted record to zero,
 	and set the heap_no field. */
 	if (m_is_comp) {
-		ulint next_offs = rec_get_next_offs(m_cur_rec, TRUE);
-		rec_set_next_offs_new(insert_rec, next_offs);
-		rec_set_next_offs_new(m_cur_rec, page_offset(insert_rec));
-
 		rec_set_n_owned_new(insert_rec, NULL, 0);
 		rec_set_heap_no_new(insert_rec,
 				    PAGE_HEAP_NO_USER_LOW + m_rec_no);
 	} else {
-		ulint next_offs = rec_get_next_offs(m_cur_rec, FALSE);
-		rec_set_next_offs_old(insert_rec, next_offs);
-		rec_set_next_offs_old(m_cur_rec, page_offset(insert_rec));
-
 		rec_set_n_owned_old(insert_rec, 0);
 		rec_set_heap_no_old(insert_rec,
 				    PAGE_HEAP_NO_USER_LOW + m_rec_no);
@@ -234,7 +225,7 @@ PageBulk::insert(
 		- page_dir_calc_reserved_space(m_rec_no);
 
 	ut_ad(m_free_space >= rec_size + slot_size);
-	ut_ad(m_heap_top + rec_size < m_page + UNIV_PAGE_SIZE);
+	ut_ad(m_heap_top + rec_size < m_page + srv_page_size);
 
 	m_free_space -= rec_size + slot_size;
 	m_heap_top += rec_size;
@@ -250,54 +241,17 @@ PageBulk::insert(
 	m_cur_rec = insert_rec;
 }
 
-inline bool PageBulk::needs_finish() const
-{
-  ut_ad(page_align(m_cur_rec) == m_block->frame);
-  ut_ad(m_page == m_block->frame);
-  ulint n_heap= page_header_get_field(m_page, PAGE_N_HEAP);
-  if (!n_heap || !page_header_get_field(m_page, PAGE_DIRECTION))
-    return true;
-  ulint heap_no;
-  if (n_heap & 0x8000)
-  {
-    n_heap&= 0x7fff;
-    heap_no= rec_get_heap_no_new(m_cur_rec);
-    if (heap_no == PAGE_HEAP_NO_INFIMUM &&
-	page_header_get_field(m_page, PAGE_HEAP_TOP) == PAGE_NEW_SUPREMUM_END)
-      return false;
-  }
-  else
-  {
-    heap_no= rec_get_heap_no_old(m_cur_rec);
-    if (heap_no == PAGE_HEAP_NO_INFIMUM &&
-	page_header_get_field(m_page, PAGE_HEAP_TOP) == PAGE_OLD_SUPREMUM_END)
-      return false;
-  }
-  return heap_no != n_heap - 1;
-}
-
 /** Mark end of insertion to the page. Scan all records to set page dirs,
 and set page header members.
 Note: we refer to page_copy_rec_list_end_to_created_page. */
 void
 PageBulk::finish()
 {
-	ut_ad(!dict_index_is_spatial(m_index));
-
-	if (!needs_finish()) {
-		return;
-	}
-
+	ut_ad(m_rec_no > 0);
 	ut_ad(m_total_data + page_dir_calc_reserved_space(m_rec_no)
 	      <= page_get_free_space_of_empty(m_is_comp));
-#ifdef UNIV_DEBUG
 	/* See page_copy_rec_list_end_to_created_page() */
-	if (m_rec_no) {
-		page_dir_set_n_slots(m_page, NULL, srv_page_size / 2);
-	}
-	mach_write_to_2(PAGE_HEADER + PAGE_HEAP_TOP + m_page,
-			ulint(m_heap_top - m_page));
-#endif
+	ut_d(page_dir_set_n_slots(m_page, NULL, srv_page_size / 2));
 
 	ulint	count = 0;
 	ulint	n_recs = 0;
@@ -306,7 +260,8 @@ PageBulk::finish()
 	page_dir_slot_t* slot = NULL;
 
 	/* Set owner & dir. */
-	while (!page_rec_is_supremum(insert_rec)) {
+	do {
+
 		count++;
 		n_recs++;
 
@@ -323,7 +278,7 @@ PageBulk::finish()
 		}
 
 		insert_rec = page_rec_get_next(insert_rec);
-	}
+	} while (!page_rec_is_supremum(insert_rec));
 
 	if (slot_index > 0
 	    && (count + 1 + (PAGE_DIR_SLOT_MAX_N_OWNED + 1) / 2
@@ -346,10 +301,10 @@ PageBulk::finish()
 	page_dir_slot_set_rec(slot, page_get_supremum_rec(m_page));
 	page_dir_slot_set_n_owned(slot, NULL, count + 1);
 
-	if (!m_rec_no) {
-		page_header_set_field(m_page, NULL, PAGE_DIRECTION,
-				      PAGE_NO_DIRECTION);
-	} else if (!m_flush_observer && !m_page_zip) {
+	ut_ad(!dict_index_is_spatial(m_index));
+	ut_ad(!page_get_instant(m_page));
+
+	if (!m_flush_observer && !m_page_zip) {
 		mlog_write_ulint(PAGE_HEADER + PAGE_N_DIR_SLOTS + m_page,
 				 2 + slot_index, MLOG_2BYTES, &m_mtr);
 		mlog_write_ulint(PAGE_HEADER + PAGE_HEAP_TOP + m_page,
@@ -364,7 +319,7 @@ PageBulk::finish()
 		mlog_write_ulint(PAGE_HEADER + PAGE_LAST_INSERT + m_page,
 				 ulint(m_cur_rec - m_page),
 				 MLOG_2BYTES, &m_mtr);
-		mlog_write_ulint(PAGE_HEADER + PAGE_DIRECTION + m_page,
+		mlog_write_ulint(PAGE_HEADER + PAGE_DIRECTION_B - 1 + m_page,
 				 PAGE_RIGHT, MLOG_2BYTES, &m_mtr);
 		mlog_write_ulint(PAGE_HEADER + PAGE_N_DIRECTION + m_page, 0,
 				 MLOG_2BYTES, &m_mtr);
@@ -381,23 +336,31 @@ PageBulk::finish()
 		mach_write_to_2(PAGE_HEADER + PAGE_N_RECS + m_page, m_rec_no);
 		mach_write_to_2(PAGE_HEADER + PAGE_LAST_INSERT + m_page,
 				ulint(m_cur_rec - m_page));
-		mach_write_to_2(PAGE_HEADER + PAGE_DIRECTION + m_page,
+		mach_write_to_2(PAGE_HEADER + PAGE_DIRECTION_B - 1 + m_page,
 				PAGE_RIGHT);
 		mach_write_to_2(PAGE_HEADER + PAGE_N_DIRECTION + m_page, 0);
 	}
 
-	ut_ad(!needs_finish());
-	ut_ad(page_validate(m_page, m_index));
+	m_block->skip_flush_check = false;
 }
 
 /** Commit inserts done to the page
 @param[in]	success		Flag whether all inserts succeed. */
-void PageBulk::commit(bool success)
+void
+PageBulk::commit(
+	bool	success)
 {
-  finish();
-  if (success && !dict_index_is_clust(m_index) && page_is_leaf(m_page))
-    ibuf_set_bitmap_for_bulk_load(m_block, innobase_fill_factor == 100);
-  m_mtr.commit();
+	if (success) {
+		ut_ad(page_validate(m_page, m_index));
+
+		/* Set no free space left and no buffered changes in ibuf. */
+		if (!dict_index_is_clust(m_index) && page_is_leaf(m_page)) {
+			ibuf_set_bitmap_for_bulk_load(
+				m_block, innobase_fill_factor == 100);
+		}
+	}
+
+	m_mtr.commit();
 }
 
 /** Compress a page of compressed table
@@ -408,8 +371,7 @@ PageBulk::compress()
 {
 	ut_ad(m_page_zip != NULL);
 
-	return(page_zip_compress(m_page_zip, m_page, m_index,
-				 page_zip_level, NULL, &m_mtr));
+	return page_zip_compress(m_block, m_index, page_zip_level, &m_mtr);
 }
 
 /** Get node pointer
@@ -436,7 +398,7 @@ rec_t*
 PageBulk::getSplitRec()
 {
 	rec_t*		rec;
-	rec_offs*	offsets;
+	ulint*		offsets;
 	ulint		total_used_size;
 	ulint		total_recs_size;
 	ulint		n_recs;
@@ -482,7 +444,7 @@ PageBulk::copyIn(
 {
 
 	rec_t*		rec = split_rec;
-	rec_offs*	offsets = NULL;
+	ulint*		offsets = NULL;
 
 	ut_ad(m_rec_no == 0);
 	ut_ad(page_rec_is_user_rec(rec));
@@ -528,7 +490,7 @@ PageBulk::copyOut(
 	ut_ad(n > 0);
 
 	/* Set last record's next in page */
-	rec_offs*	offsets = NULL;
+	ulint*		offsets = NULL;
 	rec = page_rec_get_prev(split_rec);
 	offsets = rec_get_offsets(rec, m_index, offsets,
 				  page_rec_is_leaf(split_rec),
@@ -543,15 +505,14 @@ PageBulk::copyOut(
 				  page_rec_is_leaf(split_rec),
 				  ULINT_UNDEFINED, &m_heap);
 
-	m_free_space += rec_get_end(last_rec, offsets)
-		- m_heap_top
+	m_free_space += ulint(rec_get_end(last_rec, offsets) - m_heap_top)
 		+ page_dir_calc_reserved_space(m_rec_no)
 		- page_dir_calc_reserved_space(n);
-	ut_ad(m_free_space > 0);
+	ut_ad(lint(m_free_space) > 0);
 	m_rec_no = n;
 
 #ifdef UNIV_DEBUG
-	m_total_data -= rec_get_end(last_rec, offsets) - m_heap_top;
+	m_total_data -= ulint(rec_get_end(last_rec, offsets) - m_heap_top);
 #endif /* UNIV_DEBUG */
 }
 
@@ -625,8 +586,9 @@ PageBulk::needExt(
 	const dtuple_t*		tuple,
 	ulint			rec_size)
 {
-	return(page_zip_rec_needs_ext(rec_size, m_is_comp,
-		dtuple_get_n_fields(tuple), m_block->page.size));
+	return page_zip_rec_needs_ext(rec_size, m_is_comp,
+				      dtuple_get_n_fields(tuple),
+				      m_block->zip_size());
 }
 
 /** Store external record
@@ -638,11 +600,9 @@ the blob data is logged first, then the record is logged in bulk mode.
 dberr_t
 PageBulk::storeExt(
 	const big_rec_t*	big_rec,
-	rec_offs*		offsets)
+	ulint*			offsets)
 {
-	finish();
-
-	/* Note: not all fields are initialized in btr_pcur. */
+	/* Note: not all fileds are initialized in btr_pcur. */
 	btr_pcur_t	btr_pcur;
 	btr_pcur.pos_state = BTR_PCUR_IS_POSITIONED;
 	btr_pcur.latch_mode = BTR_MODIFY_LEAF;
@@ -671,7 +631,7 @@ Note: log_free_check requires holding no lock/latch in current thread. */
 void
 PageBulk::release()
 {
-	finish();
+	ut_ad(!dict_index_is_spatial(m_index));
 
 	/* We fix the block because we will re-pin it soon. */
 	buf_block_buf_fix_inc(m_block, __FILE__, __LINE__);
@@ -692,17 +652,15 @@ PageBulk::latch()
 		m_mtr.set_log_mode(MTR_LOG_NO_REDO);
 		m_mtr.set_flush_observer(m_flush_observer);
 	} else {
-		m_mtr.set_named_space(m_index->space);
+		m_index->set_modified(m_mtr);
 	}
 
-	ut_ad(m_block->page.buf_fix_count);
-
 	/* In case the block is S-latched by page_cleaner. */
 	if (!buf_page_optimistic_get(RW_X_LATCH, m_block, m_modify_clock,
 				     __FILE__, __LINE__, &m_mtr)) {
-		m_block = buf_page_get_gen(page_id_t(m_index->space,
+		m_block = buf_page_get_gen(page_id_t(m_index->table->space_id,
 						     m_page_no),
-					   univ_page_size, RW_X_LATCH,
+					   0, RW_X_LATCH,
 					   m_block, BUF_GET_IF_IN_POOL,
 					   __FILE__, __LINE__, &m_mtr, &m_err);
 
@@ -715,8 +673,6 @@ PageBulk::latch()
 
 	buf_block_buf_fix_dec(m_block);
 
-	ut_ad(m_block->page.buf_fix_count);
-
 	ut_ad(m_cur_rec > m_page && m_cur_rec < m_heap_top);
 
 	return (m_err);
@@ -733,11 +689,12 @@ BtrBulk::pageSplit(
 {
 	ut_ad(page_bulk->getPageZip() != NULL);
 
+	/* 1. Check if we have only one user record on the page. */
 	if (page_bulk->getRecNo() <= 1) {
 		return(DB_TOO_BIG_RECORD);
 	}
 
-	/* Initialize a new page */
+	/* 2. create a new page. */
 	PageBulk new_page_bulk(m_index, m_trx->id, FIL_NULL,
 			       page_bulk->getLevel(), m_flush_observer);
 	dberr_t	err = new_page_bulk.init();
@@ -745,18 +702,19 @@ BtrBulk::pageSplit(
 		return(err);
 	}
 
-	/* Copy the upper half to the new page. */
+	/* 3. copy the upper half to new page. */
 	rec_t*	split_rec = page_bulk->getSplitRec();
 	new_page_bulk.copyIn(split_rec);
 	page_bulk->copyOut(split_rec);
 
-	/* Commit the pages after split. */
+	/* 4. commit the splitted page. */
 	err = pageCommit(page_bulk, &new_page_bulk, true);
 	if (err != DB_SUCCESS) {
 		pageAbort(&new_page_bulk);
 		return(err);
 	}
 
+	/* 5. commit the new page. */
 	err = pageCommit(&new_page_bulk, next_page_bulk, true);
 	if (err != DB_SUCCESS) {
 		pageAbort(&new_page_bulk);
@@ -822,7 +780,7 @@ BtrBulk::pageCommit(
 /** Log free check */
 inline void BtrBulk::logFreeCheck()
 {
-	if (log_sys->check_flush_or_checkpoint) {
+	if (log_sys.check_flush_or_checkpoint) {
 		release();
 
 		log_free_check();
@@ -902,7 +860,7 @@ BtrBulk::insert(
 	ulint		rec_size = rec_get_converted_size(m_index, tuple, n_ext);
 	big_rec_t*	big_rec = NULL;
 	rec_t*		rec = NULL;
-	rec_offs*	offsets = NULL;
+	ulint*		offsets = NULL;
 
 	if (page_bulk->needExt(tuple, rec_size)) {
 		/* The record is so big that we have to store some fields
@@ -982,9 +940,11 @@ BtrBulk::insert(
 		ut_ad(page_bulk->getLevel() == 0);
 		ut_ad(page_bulk == m_page_bulks.at(0));
 
-		/* Release all pages above the leaf level */
+		/* Release all latched but leaf node. */
 		for (ulint level = 1; level <= m_root_level; level++) {
-			m_page_bulks.at(level)->release();
+			PageBulk*    page_bulk = m_page_bulks.at(level);
+
+			page_bulk->release();
 		}
 
 		err = page_bulk->storeExt(big_rec, offsets);
@@ -1014,7 +974,7 @@ BtrBulk::finish(dberr_t	err)
 {
 	ulint		last_page_no = FIL_NULL;
 
-	ut_ad(!dict_table_is_temporary(m_index->table));
+	ut_ad(!m_index->table->is_temporary());
 
 	if (m_page_bulks.size() == 0) {
 		/* The table is empty. The root page of the index tree
@@ -1046,34 +1006,28 @@ BtrBulk::finish(dberr_t	err)
 		rec_t*		first_rec;
 		mtr_t		mtr;
 		buf_block_t*	last_block;
-		page_t*		last_page;
-		page_id_t	page_id(dict_index_get_space(m_index),
-					last_page_no);
-		page_size_t	page_size(dict_table_page_size(m_index->table));
-		ulint		root_page_no = dict_index_get_page(m_index);
 		PageBulk	root_page_bulk(m_index, m_trx->id,
-					       root_page_no, m_root_level,
+					       m_index->page, m_root_level,
 					       m_flush_observer);
 
-		mtr_start(&mtr);
-		mtr.set_named_space(dict_index_get_space(m_index));
-		mtr_x_lock(dict_index_get_lock(m_index), &mtr);
+		mtr.start();
+		m_index->set_modified(mtr);
+		mtr_x_lock_index(m_index, &mtr);
 
 		ut_ad(last_page_no != FIL_NULL);
-		last_block = btr_block_get(page_id, page_size,
-					   RW_X_LATCH, m_index, &mtr);
-		last_page = buf_block_get_frame(last_block);
-		first_rec = page_rec_get_next(page_get_infimum_rec(last_page));
+		last_block = btr_block_get(*m_index, last_page_no, RW_X_LATCH,
+					   false, &mtr);
+		first_rec = page_rec_get_next(
+			page_get_infimum_rec(last_block->frame));
 		ut_ad(page_rec_is_user_rec(first_rec));
 
 		/* Copy last page to root page. */
 		err = root_page_bulk.init();
 		if (err != DB_SUCCESS) {
-			mtr_commit(&mtr);
+			mtr.commit();
 			return(err);
 		}
 		root_page_bulk.copyIn(first_rec);
-		root_page_bulk.finish();
 
 		/* Remove last page. */
 		btr_page_free(m_index, last_block, &mtr);
@@ -1081,7 +1035,7 @@ BtrBulk::finish(dberr_t	err)
 		/* Do not flush the last page. */
 		last_block->page.flush_observer = NULL;
 
-		mtr_commit(&mtr);
+		mtr.commit();
 
 		err = pageCommit(&root_page_bulk, NULL, false);
 		ut_ad(err == DB_SUCCESS);
@@ -1090,6 +1044,6 @@ BtrBulk::finish(dberr_t	err)
 	ut_ad(!sync_check_iterate(dict_sync_check()));
 
 	ut_ad(err != DB_SUCCESS
-	      || btr_validate_index(m_index, NULL, false) == DB_SUCCESS);
+	      || btr_validate_index(m_index, NULL) == DB_SUCCESS);
 	return(err);
 }
