@@ -2,7 +2,7 @@
 
 Copyright (c) 1996, 2016, Oracle and/or its affiliates. All Rights Reserved.
 Copyright (c) 2008, Google Inc.
-Copyright (c) 2017, 2020, MariaDB Corporation.
+Copyright (c) 2017, 2019, MariaDB Corporation.
 
 Portions of this file contain modifications contributed and copyrighted by
 Google, Inc. Those modifications are gratefully acknowledged and are described
@@ -97,7 +97,7 @@ static inline
 ulint
 rec_fold(
 	const rec_t*	rec,
-	const rec_offs*	offsets,
+	const ulint*	offsets,
 	ulint		n_fields,
 	ulint		n_bytes,
 	index_id_t	tree_id)
@@ -191,32 +191,35 @@ static
 void
 btr_search_check_free_space_in_heap(const dict_index_t* index)
 {
+	hash_table_t*	table;
+	mem_heap_t*	heap;
+
+	ut_ad(!btr_search_own_any(RW_LOCK_S));
+	ut_ad(!btr_search_own_any(RW_LOCK_X));
+
+	table = btr_get_search_table(index);
+
+	heap = table->heap;
+
 	/* Note that we peek the value of heap->free_block without reserving
 	the latch: this is ok, because we will not guarantee that there will
 	be enough free space in the hash table. */
 
-	buf_block_t*	block = buf_block_alloc(NULL);
-	rw_lock_t*	latch = btr_get_search_latch(index);
-	hash_table_t*	table;
-	mem_heap_t*	heap;
-
-	rw_lock_x_lock(latch);
-
-	if (!btr_search_enabled) {
-		goto func_exit;
-	}
-
-	table = btr_get_search_table(index);
-	heap = table->heap;
-
 	if (heap->free_block == NULL) {
-		heap->free_block = block;
-	} else {
-func_exit:
-		buf_block_free(block);
-	}
+		buf_block_t*	block = buf_block_alloc(NULL);
+		rw_lock_t*	ahi_latch = btr_get_search_latch(index);
 
-	rw_lock_x_unlock(latch);
+		rw_lock_x_lock(ahi_latch);
+
+		if (btr_search_enabled
+		    && heap->free_block == NULL) {
+			heap->free_block = block;
+		} else {
+			buf_block_free(block);
+		}
+
+		rw_lock_x_unlock(ahi_latch);
+	}
 }
 
 /** Creates and initializes the adaptive search system at a database start.
@@ -244,47 +247,90 @@ void btr_search_sys_create(ulint hash_size)
 	btr_search_sys = reinterpret_cast<btr_search_sys_t*>(
 		ut_malloc(sizeof(btr_search_sys_t), mem_key_ahi));
 
-	btr_search_sys->hash_tables = NULL;
+	btr_search_sys->hash_tables = reinterpret_cast<hash_table_t**>(
+		ut_malloc(sizeof(hash_table_t*) * btr_ahi_parts, mem_key_ahi));
+
+	for (ulint i = 0; i < btr_ahi_parts; ++i) {
+
+		btr_search_sys->hash_tables[i] =
+			ib_create((hash_size / btr_ahi_parts),
+				  LATCH_ID_HASH_TABLE_MUTEX,
+				  0, MEM_HEAP_FOR_BTR_SEARCH);
+
+#if defined UNIV_AHI_DEBUG || defined UNIV_DEBUG
+		btr_search_sys->hash_tables[i]->adaptive = TRUE;
+#endif /* UNIV_AHI_DEBUG || UNIV_DEBUG */
+	}
+}
+
+/** Resize hash index hash table.
+@param[in]	hash_size	hash index hash table size */
+void btr_search_sys_resize(ulint hash_size)
+{
+	/* Step-1: Lock all search latches in exclusive mode. */
+	btr_search_x_lock_all();
 
 	if (btr_search_enabled) {
-		btr_search_enable();
+
+		btr_search_x_unlock_all();
+
+		ib::error() << "btr_search_sys_resize failed because"
+			" hash index hash table is not empty.";
+		ut_ad(0);
+		return;
 	}
+
+	/* Step-2: Recreate hash tables with new size. */
+	for (ulint i = 0; i < btr_ahi_parts; ++i) {
+
+		mem_heap_free(btr_search_sys->hash_tables[i]->heap);
+		hash_table_free(btr_search_sys->hash_tables[i]);
+
+		btr_search_sys->hash_tables[i] =
+			ib_create((hash_size / btr_ahi_parts),
+				  LATCH_ID_HASH_TABLE_MUTEX,
+				  0, MEM_HEAP_FOR_BTR_SEARCH);
+
+#if defined UNIV_AHI_DEBUG || defined UNIV_DEBUG
+		btr_search_sys->hash_tables[i]->adaptive = TRUE;
+#endif /* UNIV_AHI_DEBUG || UNIV_DEBUG */
+	}
+
+	/* Step-3: Unlock all search latches from exclusive mode. */
+	btr_search_x_unlock_all();
 }
 
 /** Frees the adaptive search system at a database shutdown. */
 void btr_search_sys_free()
 {
-  if (!btr_search_sys)
-  {
-    ut_ad(!btr_search_latches);
-    return;
-  }
+	if (!btr_search_sys) {
+		ut_ad(!btr_search_latches);
+		return;
+	}
 
-  ut_ad(btr_search_sys);
-  ut_ad(btr_search_latches);
+	ut_ad(btr_search_latches);
 
-  if (btr_search_sys->hash_tables)
-  {
-    for (ulint i= 0; i < btr_ahi_parts; ++i)
-    {
-      mem_heap_free(btr_search_sys->hash_tables[i]->heap);
-      hash_table_free(btr_search_sys->hash_tables[i]);
-    }
-    ut_free(btr_search_sys->hash_tables);
-  }
+	/* Step-1: Release the hash tables. */
+	for (ulint i = 0; i < btr_ahi_parts; ++i) {
 
-  ut_free(btr_search_sys);
-  btr_search_sys= NULL;
+		mem_heap_free(btr_search_sys->hash_tables[i]->heap);
+		hash_table_free(btr_search_sys->hash_tables[i]);
 
-  /* Free all latches. */
-  for (ulint i= 0; i < btr_ahi_parts; ++i)
-  {
-    rw_lock_free(btr_search_latches[i]);
-    ut_free(btr_search_latches[i]);
-  }
+	}
 
-  ut_free(btr_search_latches);
-  btr_search_latches= NULL;
+	ut_free(btr_search_sys->hash_tables);
+	ut_free(btr_search_sys);
+	btr_search_sys = NULL;
+
+	/* Step-2: Release all allocates latches. */
+	for (ulint i = 0; i < btr_ahi_parts; ++i) {
+
+		rw_lock_free(btr_search_latches[i]);
+		ut_free(btr_search_latches[i]);
+	}
+
+	ut_free(btr_search_latches);
+	btr_search_latches = NULL;
 }
 
 /** Set index->ref_count = 0 on all indexes of a table.
@@ -296,6 +342,8 @@ btr_search_disable_ref_count(
 {
 	dict_index_t*	index;
 
+	ut_ad(mutex_own(&dict_sys.mutex));
+
 	for (index = dict_table_get_first_index(table);
 	     index != NULL;
 	     index = dict_table_get_next_index(index)) {
@@ -303,92 +351,24 @@ btr_search_disable_ref_count(
 	}
 }
 
-/** Lazily free detached metadata when removing the last reference. */
-ATTRIBUTE_COLD static void btr_search_lazy_free(dict_index_t *index)
-{
-  ut_ad(index->freed());
-  dict_table_t *table= index->table;
-  /* Perform the skipped steps of dict_index_remove_from_cache_low(). */
-  UT_LIST_REMOVE(table->freed_indexes, index);
-  rw_lock_free(&index->lock);
-  dict_mem_index_free(index);
-
-  if (!UT_LIST_GET_LEN(table->freed_indexes) &&
-      !UT_LIST_GET_LEN(table->indexes))
-  {
-    ut_ad(table->id == 0);
-    dict_mem_table_free(table);
-  }
-}
-
-/** Clear the adaptive hash index on all pages in the buffer pool. */
-static void buf_pool_clear_hash_index()
-{
-  ut_ad(btr_search_own_all(RW_LOCK_X));
-  ut_ad(!btr_search_enabled);
-
-  std::set<dict_index_t*> garbage;
-
-  for (ulong p = 0; p < srv_buf_pool_instances; p++)
-  {
-    buf_pool_t *buf_pool= buf_pool_from_array(p);
-    buf_chunk_t *chunks= buf_pool->chunks;
-    buf_chunk_t *chunk= chunks + buf_pool->n_chunks;
-
-    while (--chunk >= chunks)
-    {
-      buf_block_t *block= chunk->blocks;
-      for (ulint i= chunk->size; i--; block++)
-      {
-        dict_index_t *index= block->index;
-        assert_block_ahi_valid(block);
-
-        /* We can clear block->index and block->n_pointers when
-        btr_search_own_all(RW_LOCK_X); see the comments in buf0buf.h */
-
-        if (!index)
-        {
-# if defined UNIV_AHI_DEBUG || defined UNIV_DEBUG
-          ut_a(!block->n_pointers);
-# endif /* UNIV_AHI_DEBUG || UNIV_DEBUG */
-          continue;
-        }
-
-        ut_d(buf_page_state state= buf_block_get_state(block));
-        /* Another thread may have set the state to
-        BUF_BLOCK_REMOVE_HASH in buf_LRU_block_remove_hashed().
-
-        The state change in buf_page_realloc() is not observable here,
-        because in that case we would have !block->index.
-
-        In the end, the entire adaptive hash index will be removed. */
-        ut_ad(state == BUF_BLOCK_FILE_PAGE || state == BUF_BLOCK_REMOVE_HASH);
-# if defined UNIV_AHI_DEBUG || defined UNIV_DEBUG
-        block->n_pointers= 0;
-# endif /* UNIV_AHI_DEBUG || UNIV_DEBUG */
-        if (index->freed())
-          garbage.insert(index);
-        block->index= NULL;
-      }
-    }
-  }
-
-  for (std::set<dict_index_t*>::iterator i= garbage.begin();
-       i != garbage.end(); i++)
-    btr_search_lazy_free(*i);
-}
-
-/** Disable the adaptive hash search system and empty the index. */
-void btr_search_disable()
+/** Disable the adaptive hash search system and empty the index.
+@param[in]	need_mutex	need to acquire dict_sys.mutex */
+void btr_search_disable(bool need_mutex)
 {
 	dict_table_t*	table;
 
-	mutex_enter(&dict_sys->mutex);
+	if (need_mutex) {
+		mutex_enter(&dict_sys.mutex);
+	}
 
+	ut_ad(mutex_own(&dict_sys.mutex));
 	btr_search_x_lock_all();
 
 	if (!btr_search_enabled) {
-		mutex_exit(&dict_sys->mutex);
+		if (need_mutex) {
+			mutex_exit(&dict_sys.mutex);
+		}
+
 		btr_search_x_unlock_all();
 		return;
 	}
@@ -397,72 +377,74 @@ void btr_search_disable()
 
 	/* Clear the index->search_info->ref_count of every index in
 	the data dictionary cache. */
-	for (table = UT_LIST_GET_FIRST(dict_sys->table_LRU); table;
+	for (table = UT_LIST_GET_FIRST(dict_sys.table_LRU); table;
 	     table = UT_LIST_GET_NEXT(table_LRU, table)) {
 
 		btr_search_disable_ref_count(table);
 	}
 
-	for (table = UT_LIST_GET_FIRST(dict_sys->table_non_LRU); table;
+	for (table = UT_LIST_GET_FIRST(dict_sys.table_non_LRU); table;
 	     table = UT_LIST_GET_NEXT(table_LRU, table)) {
 
 		btr_search_disable_ref_count(table);
 	}
 
-	mutex_exit(&dict_sys->mutex);
+	if (need_mutex) {
+		mutex_exit(&dict_sys.mutex);
+	}
 
 	/* Set all block->index = NULL. */
 	buf_pool_clear_hash_index();
 
 	/* Clear the adaptive hash index. */
 	for (ulint i = 0; i < btr_ahi_parts; ++i) {
-		mem_heap_free(btr_search_sys->hash_tables[i]->heap);
-		hash_table_free(btr_search_sys->hash_tables[i]);
+		hash_table_clear(btr_search_sys->hash_tables[i]);
+		mem_heap_empty(btr_search_sys->hash_tables[i]->heap);
 	}
-	ut_free(btr_search_sys->hash_tables);
-	btr_search_sys->hash_tables = NULL;
 
 	btr_search_x_unlock_all();
 }
 
-/** Enable the adaptive hash search system.
-@param resize whether buf_pool_resize() is the caller */
-void btr_search_enable(bool resize)
+/** Enable the adaptive hash search system. */
+void btr_search_enable()
 {
-	if (!resize) {
-		buf_pool_mutex_enter_all();
-		if (srv_buf_pool_old_size != srv_buf_pool_size) {
-			buf_pool_mutex_exit_all();
-			return;
-		}
+	buf_pool_mutex_enter_all();
+	if (srv_buf_pool_old_size != srv_buf_pool_size) {
 		buf_pool_mutex_exit_all();
-	}
-
-	ulint hash_size = buf_pool_get_curr_size() / sizeof(void *) / 64;
-	btr_search_x_lock_all();
-
-	if (btr_search_sys->hash_tables) {
-		ut_ad(btr_search_enabled);
-		btr_search_x_unlock_all();
 		return;
 	}
+	buf_pool_mutex_exit_all();
 
-	btr_search_sys->hash_tables = reinterpret_cast<hash_table_t**>(
-		ut_malloc(sizeof(hash_table_t*) * btr_ahi_parts, mem_key_ahi));
-	for (ulint i = 0; i < btr_ahi_parts; ++i) {
-		btr_search_sys->hash_tables[i] =
-			ib_create((hash_size / btr_ahi_parts),
-				  LATCH_ID_HASH_TABLE_MUTEX,
-				  0, MEM_HEAP_FOR_BTR_SEARCH);
-#if defined UNIV_AHI_DEBUG || defined UNIV_DEBUG
-                btr_search_sys->hash_tables[i]->adaptive = TRUE;
-#endif /* UNIV_AHI_DEBUG || UNIV_DEBUG */
-	}
-
+	btr_search_x_lock_all();
 	btr_search_enabled = true;
 	btr_search_x_unlock_all();
 }
 
+/** Returns the value of ref_count. The value is protected by latch.
+@param[in]	info		search info
+@param[in]	index		index identifier
+@return ref_count value. */
+ulint
+btr_search_info_get_ref_count(
+	btr_search_t*	info,
+	dict_index_t*	index)
+{
+	ulint ret = 0;
+
+	if (!btr_search_enabled) {
+		return(ret);
+	}
+
+	ut_ad(info);
+
+	rw_lock_t* ahi_latch = btr_get_search_latch(index);
+	rw_lock_s_lock(ahi_latch);
+	ret = info->ref_count;
+	rw_lock_s_unlock(ahi_latch);
+
+	return(ret);
+}
+
 /** Updates the search info of an index about hash successes. NOTE that info
 is NOT protected by any semaphore, to save CPU time! Do not assume its fields
 are consistent.
@@ -659,48 +641,49 @@ btr_search_update_hash_ref(
 	buf_block_t*		block,
 	const btr_cur_t*	cursor)
 {
-	ut_ad(cursor->flag == BTR_CUR_HASH_FAIL);
+	dict_index_t*	index;
+	ulint		fold;
+	rec_t*		rec;
 
+	ut_ad(cursor->flag == BTR_CUR_HASH_FAIL);
+	ut_ad(rw_lock_own(btr_get_search_latch(cursor->index), RW_LOCK_X));
 	ut_ad(rw_lock_own_flagged(&block->lock,
 				  RW_LOCK_FLAG_X | RW_LOCK_FLAG_S));
 	ut_ad(page_align(btr_cur_get_rec(cursor)) == block->frame);
 	ut_ad(page_is_leaf(block->frame));
 	assert_block_ahi_valid(block);
 
-	dict_index_t* index = block->index;
+	index = block->index;
+
+	if (!index) {
 
-	if (!index || !info->n_hash_potential) {
 		return;
 	}
 
 	ut_ad(block->page.id.space() == index->table->space_id);
 	ut_ad(index == cursor->index);
 	ut_ad(!dict_index_is_ibuf(index));
-	rw_lock_t* const latch = btr_get_search_latch(index);
-	rw_lock_x_lock(latch);
-	ut_ad(!block->index || block->index == index);
 
-	if (block->index
+	if ((info->n_hash_potential > 0)
 	    && (block->curr_n_fields == info->n_fields)
 	    && (block->curr_n_bytes == info->n_bytes)
-	    && (block->curr_left_side == info->left_side)
-	    && btr_search_enabled) {
+	    && (block->curr_left_side == info->left_side)) {
 		mem_heap_t*	heap		= NULL;
-		rec_offs	offsets_[REC_OFFS_NORMAL_SIZE];
+		ulint		offsets_[REC_OFFS_NORMAL_SIZE];
 		rec_offs_init(offsets_);
 
-		const rec_t* rec = btr_cur_get_rec(cursor);
+		rec = btr_cur_get_rec(cursor);
 
 		if (!page_rec_is_user_rec(rec)) {
-			goto func_exit;
+
+			return;
 		}
 
-		ulint fold = rec_fold(
-			rec,
-			rec_get_offsets(rec, index, offsets_, true,
-					ULINT_UNDEFINED, &heap),
-			block->curr_n_fields,
-			block->curr_n_bytes, index->id);
+		fold = rec_fold(rec,
+				rec_get_offsets(rec, index, offsets_, true,
+						ULINT_UNDEFINED, &heap),
+				block->curr_n_fields,
+				block->curr_n_bytes, index->id);
 		if (UNIV_LIKELY_NULL(heap)) {
 			mem_heap_free(heap);
 		}
@@ -710,9 +693,6 @@ btr_search_update_hash_ref(
 
 		MONITOR_INC(MONITOR_ADAPTIVE_HASH_ROW_ADDED);
 	}
-
-func_exit:
-	rw_lock_x_unlock(latch);
 }
 
 /** Checks if a guessed position for a tree cursor is right. Note that if
@@ -742,8 +722,8 @@ btr_search_check_guess(
 	ulint		match;
 	int		cmp;
 	mem_heap_t*	heap		= NULL;
-	rec_offs	offsets_[REC_OFFS_NORMAL_SIZE];
-	rec_offs*	offsets		= offsets_;
+	ulint		offsets_[REC_OFFS_NORMAL_SIZE];
+	ulint*		offsets		= offsets_;
 	ibool		success		= FALSE;
 	rec_offs_init(offsets_);
 
@@ -893,24 +873,24 @@ btr_search_guess_on_hash(
 	rw_lock_t*	ahi_latch,
 	mtr_t*		mtr)
 {
+	const rec_t*	rec;
 	ulint		fold;
 	index_id_t	index_id;
-#ifdef notdefined
-	btr_cur_t	cursor2;
-	btr_pcur_t	pcur;
-#endif
+
+	ut_ad(mtr->is_active());
 	ut_ad(!ahi_latch || rw_lock_own_flagged(
 		      ahi_latch, RW_LOCK_FLAG_X | RW_LOCK_FLAG_S));
 
 	if (!btr_search_enabled) {
-		return false;
+		return(FALSE);
 	}
 
-	ut_ad(index && info && tuple && cursor && mtr);
-	ut_ad(!dict_index_is_ibuf(index));
+	ut_ad(!index->is_ibuf());
 	ut_ad(!ahi_latch || ahi_latch == btr_get_search_latch(index));
 	ut_ad((latch_mode == BTR_SEARCH_LEAF)
 	      || (latch_mode == BTR_MODIFY_LEAF));
+	compile_time_assert(ulint{BTR_SEARCH_LEAF} == ulint{RW_S_LATCH});
+	compile_time_assert(ulint{BTR_MODIFY_LEAF} == ulint{RW_X_LATCH});
 
 	/* Not supported for spatial index */
 	ut_ad(!dict_index_is_spatial(index));
@@ -919,14 +899,16 @@ btr_search_guess_on_hash(
 	any latch here! */
 
 	if (info->n_hash_potential == 0) {
-		return false;
+
+		return(FALSE);
 	}
 
 	cursor->n_fields = info->n_fields;
 	cursor->n_bytes = info->n_bytes;
 
 	if (dtuple_get_n_fields(tuple) < btr_search_get_n_fields(cursor)) {
-		return false;
+
+		return(FALSE);
 	}
 
 	index_id = index->id;
@@ -940,7 +922,6 @@ btr_search_guess_on_hash(
 	cursor->flag = BTR_CUR_HASH;
 
 	rw_lock_t* use_latch = ahi_latch ? NULL : btr_get_search_latch(index);
-	const rec_t* rec;
 
 	if (use_latch) {
 		rw_lock_s_lock(use_latch);
@@ -953,54 +934,78 @@ btr_search_guess_on_hash(
 		ut_ad(rw_lock_own(ahi_latch, RW_LOCK_S));
 	}
 
-	rec = static_cast<const rec_t*>(
-		ha_search_and_get_data(btr_get_search_table(index), fold));
+	rec = (rec_t*) ha_search_and_get_data(
+		btr_get_search_table(index), fold);
 
-	if (!rec) {
+	if (rec == NULL) {
 		if (use_latch) {
 fail:
 			rw_lock_s_unlock(use_latch);
 		}
 
 		btr_search_failure(info, cursor);
-		return false;
+
+		return(FALSE);
 	}
 
-	buf_block_t*	block = buf_block_from_ahi(rec);
+	buf_block_t* block = buf_block_from_ahi(rec);
+	buf_pool_t* buf_pool = buf_pool_from_block(block);
 
 	if (use_latch) {
-		if (!buf_page_get_known_nowait(
-			latch_mode, block, BUF_MAKE_YOUNG,
-			__FILE__, __LINE__, mtr)) {
+		mutex_enter(&block->mutex);
+
+		if (buf_block_get_state(block) == BUF_BLOCK_REMOVE_HASH) {
+			/* Another thread is just freeing the block
+			from the LRU list of the buffer pool: do not
+			try to access this page. */
+			mutex_exit(&block->mutex);
 			goto fail;
 		}
 
-		const bool fail = index != block->index
-			&& index_id == block->index->id;
-		ut_a(!fail || block->index->freed());
+		ut_ad(buf_block_get_state(block) == BUF_BLOCK_FILE_PAGE);
+		ut_ad(!block->page.file_page_was_freed);
+		buf_page_set_accessed(&block->page);
+		buf_block_buf_fix_inc(block, __FILE__, __LINE__);
+		mutex_exit(&block->mutex);
+
+		buf_page_make_young_if_needed(buf_pool, &block->page);
+		mtr_memo_type_t	fix_type;
+		if (latch_mode == BTR_SEARCH_LEAF) {
+			if (!rw_lock_s_lock_nowait(&block->lock,
+						   __FILE__, __LINE__)) {
+got_no_latch:
+				buf_block_buf_fix_dec(block);
+				goto fail;
+			}
+			fix_type = MTR_MEMO_PAGE_S_FIX;
+		} else {
+			if (!rw_lock_x_lock_func_nowait_inline(
+				    &block->lock, __FILE__, __LINE__)) {
+				goto got_no_latch;
+			}
+			fix_type = MTR_MEMO_PAGE_X_FIX;
+		}
+		mtr->memo_push(block, fix_type);
+
+		buf_pool->stat.n_page_gets++;
+
 		rw_lock_s_unlock(use_latch);
 
 		buf_block_dbg_add_level(block, SYNC_TREE_NODE_FROM_HASH);
-		if (UNIV_UNLIKELY(fail)) {
-			goto fail_and_release_page;
-		}
-	} else if (UNIV_UNLIKELY(index != block->index
-				 && index_id == block->index->id)) {
-		ut_a(block->index->freed());
-		goto fail_and_release_page;
 	}
 
 	if (buf_block_get_state(block) != BUF_BLOCK_FILE_PAGE) {
 
 		ut_ad(buf_block_get_state(block) == BUF_BLOCK_REMOVE_HASH);
 
-fail_and_release_page:
 		if (!ahi_latch) {
+
 			btr_leaf_page_release(block, latch_mode, mtr);
 		}
 
 		btr_search_failure(info, cursor);
-		return false;
+
+		return(FALSE);
 	}
 
 	ut_ad(page_rec_is_user_rec(rec));
@@ -1016,7 +1021,14 @@ fail_and_release_page:
 	right. */
 	if (index_id != btr_page_get_index_id(block->frame)
 	    || !btr_search_check_guess(cursor, !!ahi_latch, tuple, mode)) {
-		goto fail_and_release_page;
+
+		if (!ahi_latch) {
+			btr_leaf_page_release(block, latch_mode, mtr);
+		}
+
+		btr_search_failure(info, cursor);
+
+		return(FALSE);
 	}
 
 	if (info->n_hash_potential < BTR_SEARCH_BUILD_LIMIT + 5) {
@@ -1064,17 +1076,12 @@ fail_and_release_page:
 #ifdef UNIV_SEARCH_PERF_STAT
 	btr_search_n_succ++;
 #endif
-	if (!ahi_latch && buf_page_peek_if_too_old(&block->page)) {
-
-		buf_page_make_young(&block->page);
-	}
-
 	/* Increment the page get statistics though we did not really
 	fix the page: for user info only */
-	{
-		buf_pool_t*	buf_pool = buf_pool_from_bpage(&block->page);
+	++buf_pool->stat.n_page_gets;
 
-		++buf_pool->stat.n_page_gets;
+	if (!ahi_latch) {
+		buf_page_make_young_if_needed(buf_pool, &block->page);
 	}
 
 	return true;
@@ -1099,29 +1106,33 @@ void btr_search_drop_page_hash_index(buf_block_t* block)
 	ulint*			folds;
 	ulint			i;
 	mem_heap_t*		heap;
-	rec_offs*		offsets;
+	const dict_index_t*	index;
+	ulint*			offsets;
 	rw_lock_t*		latch;
+	btr_search_t*		info;
 
 retry:
+	/* Do a dirty check on block->index, return if the block is
+	not in the adaptive hash index. */
+	index = block->index;
 	/* This debug check uses a dirty read that could theoretically cause
 	false positives while buf_pool_clear_hash_index() is executing. */
 	assert_block_ahi_valid(block);
 	ut_ad(!btr_search_own_any(RW_LOCK_S));
 	ut_ad(!btr_search_own_any(RW_LOCK_X));
 
-	if (!block->index) {
+	if (index == NULL) {
 		return;
 	}
 
 	ut_ad(block->page.buf_fix_count == 0
 	      || buf_block_get_state(block) == BUF_BLOCK_REMOVE_HASH
 	      || rw_lock_own_flagged(&block->lock,
-				     RW_LOCK_FLAG_X | RW_LOCK_FLAG_S
-				     | RW_LOCK_FLAG_SX));
+				     RW_LOCK_FLAG_X | RW_LOCK_FLAG_S));
 	ut_ad(page_is_leaf(block->frame));
 
-	/* We must not dereference block->index here, because it could be freed
-	if (index->table->n_ref_count == 0 && !mutex_own(&dict_sys->mutex)).
+	/* We must not dereference index here, because it could be freed
+	if (index->table->n_ref_count == 0 && !mutex_own(&dict_sys.mutex)).
 	Determine the ahi_slot based on the block contents. */
 
 	const index_id_t	index_id
@@ -1135,12 +1146,18 @@ retry:
 	rw_lock_s_lock(latch);
 	assert_block_ahi_valid(block);
 
-	if (!block->index || !btr_search_enabled) {
+	if (block->index == NULL) {
 		rw_lock_s_unlock(latch);
 		return;
 	}
 
-	dict_index_t* index = block->index;
+	/* The index associated with a block must remain the
+	same, because we are holding block->lock or the block is
+	not accessible by other threads (BUF_BLOCK_REMOVE_HASH),
+	or the index is not accessible to other threads
+	(buf_fix_count == 0 when DROP TABLE or similar is executing
+	buf_LRU_drop_page_hash_for_tablespace()). */
+	ut_a(index == block->index);
 #ifdef MYSQL_INDEX_DISABLE_AHI
 	ut_ad(!index->disable_ahi);
 #endif
@@ -1148,7 +1165,27 @@ retry:
 
 	ut_ad(block->page.id.space() == index->table->space_id);
 	ut_a(index_id == index->id);
-	ut_ad(!dict_index_is_ibuf(index));
+	ut_a(!dict_index_is_ibuf(index));
+#ifdef UNIV_DEBUG
+	switch (dict_index_get_online_status(index)) {
+	case ONLINE_INDEX_CREATION:
+		/* The index is being created (bulk loaded). */
+	case ONLINE_INDEX_COMPLETE:
+		/* The index has been published. */
+	case ONLINE_INDEX_ABORTED:
+		/* Either the index creation was aborted due to an
+		error observed by InnoDB (in which case there should
+		not be any adaptive hash index entries), or it was
+		completed and then flagged aborted in
+		rollback_inplace_alter_table(). */
+		break;
+	case ONLINE_INDEX_ABORTED_DROPPED:
+		/* The index should have been dropped from the tablespace
+		already, and the adaptive hash index entries should have
+		been dropped as well. */
+		ut_error;
+	}
+#endif /* UNIV_DEBUG */
 
 	n_fields = block->curr_n_fields;
 	n_bytes = block->curr_n_bytes;
@@ -1172,7 +1209,7 @@ retry:
 
 	rec = page_get_infimum_rec(page);
 	rec = page_rec_get_next_low(rec, page_is_comp(page));
-	if (rec_is_metadata(rec, index)) {
+	if (rec_is_metadata(rec, *index)) {
 		rec = page_rec_get_next_low(rec, page_is_comp(page));
 	}
 
@@ -1236,14 +1273,9 @@ next_rec:
 			folds[i], page);
 	}
 
-	switch (index->search_info->ref_count--) {
-	case 0:
-		ut_error;
-	case 1:
-		if (index->freed()) {
-			btr_search_lazy_free(index);
-		}
-	}
+	info = btr_search_get_info(block->index);
+	ut_a(info->ref_count > 0);
+	info->ref_count--;
 
 	block->index = NULL;
 
@@ -1274,7 +1306,7 @@ void btr_search_drop_page_hash_when_freed(const page_id_t page_id)
 	are possibly holding, we cannot s-latch the page, but must
 	(recursively) x-latch it, even though we are only reading. */
 
-	block = buf_page_get_gen(page_id, univ_page_size, RW_X_LATCH, NULL,
+	block = buf_page_get_gen(page_id, 0, RW_X_LATCH, NULL,
 				 BUF_PEEK_IF_IN_POOL, __FILE__, __LINE__,
 				 &mtr, &err);
 
@@ -1292,7 +1324,7 @@ void btr_search_drop_page_hash_when_freed(const page_id_t page_id)
 			be open, or we should be in the process of
 			dropping the table (preventing eviction). */
 			ut_ad(index->table->get_ref_count() > 0
-			      || mutex_own(&dict_sys->mutex));
+			      || mutex_own(&dict_sys.mutex));
 			btr_search_drop_page_hash_index(block);
 		}
 	}
@@ -1328,9 +1360,10 @@ btr_search_build_page_hash_index(
 	ulint		n_recs;
 	ulint*		folds;
 	const rec_t**	recs;
+	ulint		i;
 	mem_heap_t*	heap		= NULL;
-	rec_offs	offsets_[REC_OFFS_NORMAL_SIZE];
-	rec_offs*	offsets		= offsets_;
+	ulint		offsets_[REC_OFFS_NORMAL_SIZE];
+	ulint*		offsets		= offsets_;
 
 #ifdef MYSQL_INDEX_DISABLE_AHI
 	if (index->disable_ahi) return;
@@ -1343,27 +1376,21 @@ btr_search_build_page_hash_index(
 	ut_ad(ahi_latch == btr_get_search_latch(index));
 	ut_ad(index);
 	ut_ad(block->page.id.space() == index->table->space_id);
-	ut_ad(!dict_index_is_ibuf(index));
+	ut_a(!dict_index_is_ibuf(index));
 	ut_ad(page_is_leaf(block->frame));
 
 	ut_ad(rw_lock_own_flagged(&block->lock,
 				  RW_LOCK_FLAG_X | RW_LOCK_FLAG_S));
-	ut_ad(block->page.id.page_no() >= 3);
 
 	rw_lock_s_lock(ahi_latch);
 
-	const bool enabled = btr_search_enabled;
-	const bool rebuild = enabled && block->index
+	const bool rebuild = block->index
 		&& (block->curr_n_fields != n_fields
 		    || block->curr_n_bytes != n_bytes
 		    || block->curr_left_side != left_side);
 
 	rw_lock_s_unlock(ahi_latch);
 
-	if (!enabled) {
-		return;
-	}
-
 	if (rebuild) {
 		btr_search_drop_page_hash_index(block);
 	}
@@ -1390,7 +1417,7 @@ btr_search_build_page_hash_index(
 
 	rec = page_rec_get_next_const(page_get_infimum_rec(page));
 
-	if (rec_is_metadata(rec, index)) {
+	if (rec_is_metadata(rec, *index)) {
 		rec = page_rec_get_next_const(rec);
 		if (!--n_recs) return;
 	}
@@ -1464,12 +1491,19 @@ btr_search_build_page_hash_index(
 
 	btr_search_check_free_space_in_heap(index);
 
+	hash_table_t*	table	= btr_get_search_table(index);
 	rw_lock_x_lock(ahi_latch);
 
 	if (!btr_search_enabled) {
 		goto exit_func;
 	}
 
+	if (block->index && ((block->curr_n_fields != n_fields)
+			     || (block->curr_n_bytes != n_bytes)
+			     || (block->curr_left_side != left_side))) {
+		goto exit_func;
+	}
+
 	/* This counter is decremented every time we drop page
 	hash index entries and is incremented here. Since we can
 	rebuild hash index for a page that is already hashed, we
@@ -1478,10 +1512,6 @@ btr_search_build_page_hash_index(
 	if (!block->index) {
 		assert_block_ahi_empty(block);
 		index->search_info->ref_count++;
-	} else if (block->curr_n_fields != n_fields
-		   || block->curr_n_bytes != n_bytes
-		   || block->curr_left_side != left_side) {
-		goto exit_func;
 	}
 
 	block->n_hash_helps = 0;
@@ -1491,11 +1521,9 @@ btr_search_build_page_hash_index(
 	block->curr_left_side = unsigned(left_side);
 	block->index = index;
 
-	{
-		hash_table_t*	table = btr_get_search_table(index);
-		for (ulint i = 0; i < n_cached; i++) {
-			ha_insert_for_fold(table, folds[i], block, recs[i]);
-		}
+	for (i = 0; i < n_cached; i++) {
+
+		ha_insert_for_fold(table, folds[i], block, recs[i]);
 	}
 
 	MONITOR_INC(MONITOR_ADAPTIVE_HASH_PAGE_ADDED);
@@ -1545,7 +1573,11 @@ btr_search_info_update_slow(btr_search_t* info, btr_cur_t* cursor)
 		btr_search_n_hash_fail++;
 #endif /* UNIV_SEARCH_PERF_STAT */
 
+		rw_lock_x_lock(ahi_latch);
+
 		btr_search_update_hash_ref(info, block, cursor);
+
+		rw_lock_x_unlock(ahi_latch);
 	}
 
 	if (build_index) {
@@ -1630,11 +1662,12 @@ btr_search_move_or_delete_hash_entries(
 			using btr_cur_search_, the record is not yet deleted.*/
 void btr_search_update_hash_on_delete(btr_cur_t* cursor)
 {
+	hash_table_t*	table;
 	buf_block_t*	block;
 	const rec_t*	rec;
 	ulint		fold;
 	dict_index_t*	index;
-	rec_offs	offsets_[REC_OFFS_NORMAL_SIZE];
+	ulint		offsets_[REC_OFFS_NORMAL_SIZE];
 	mem_heap_t*	heap		= NULL;
 	rec_offs_init(offsets_);
 
@@ -1662,7 +1695,9 @@ void btr_search_update_hash_on_delete(btr_cur_t* cursor)
 	ut_ad(block->page.id.space() == index->table->space_id);
 	ut_a(index == cursor->index);
 	ut_a(block->curr_n_fields > 0 || block->curr_n_bytes > 0);
-	ut_ad(!dict_index_is_ibuf(index));
+	ut_a(!dict_index_is_ibuf(index));
+
+	table = btr_get_search_table(index);
 
 	rec = btr_cur_get_rec(cursor);
 
@@ -1673,24 +1708,22 @@ void btr_search_update_hash_on_delete(btr_cur_t* cursor)
 		mem_heap_free(heap);
 	}
 
-	rw_lock_t* ahi_latch = btr_get_search_latch(index);
+	rw_lock_t*	ahi_latch = btr_get_search_latch(index);
 
 	rw_lock_x_lock(ahi_latch);
 	assert_block_ahi_valid(block);
 
-	if (btr_search_enabled) {
-		hash_table_t* table = btr_get_search_table(index);
-		if (block->index) {
-			ut_a(block->index == index);
+	if (block->index) {
+		ut_a(block->index == index);
 
-			if (ha_search_and_delete_if_found(table, fold, rec)) {
-				MONITOR_INC(MONITOR_ADAPTIVE_HASH_ROW_REMOVED);
-			} else {
-				MONITOR_INC(MONITOR_ADAPTIVE_HASH_ROW_REMOVE_NOT_FOUND);
-			}
-
-			assert_block_ahi_valid(block);
+		if (ha_search_and_delete_if_found(table, fold, rec)) {
+			MONITOR_INC(MONITOR_ADAPTIVE_HASH_ROW_REMOVED);
+		} else {
+			MONITOR_INC(
+				MONITOR_ADAPTIVE_HASH_ROW_REMOVE_NOT_FOUND);
 		}
+
+		assert_block_ahi_valid(block);
 	}
 
 	rw_lock_x_unlock(ahi_latch);
@@ -1733,10 +1766,10 @@ btr_search_update_hash_node_on_insert(btr_cur_t* cursor, rw_lock_t* ahi_latch)
 	}
 
 	ut_a(cursor->index == index);
-	ut_ad(!dict_index_is_ibuf(index));
+	ut_a(!dict_index_is_ibuf(index));
 	rw_lock_x_lock(ahi_latch);
 
-	if (!block->index || !btr_search_enabled) {
+	if (!block->index) {
 
 		goto func_exit;
 	}
@@ -1775,6 +1808,7 @@ func_exit:
 void
 btr_search_update_hash_on_insert(btr_cur_t* cursor, rw_lock_t* ahi_latch)
 {
+	hash_table_t*	table;
 	buf_block_t*	block;
 	dict_index_t*	index;
 	const rec_t*	rec;
@@ -1785,9 +1819,11 @@ btr_search_update_hash_on_insert(btr_cur_t* cursor, rw_lock_t* ahi_latch)
 	ulint		next_fold = 0; /* remove warning (??? bug ???) */
 	ulint		n_fields;
 	ulint		n_bytes;
+	ibool		left_side;
+	bool		locked		= false;
 	mem_heap_t*	heap		= NULL;
-	rec_offs	offsets_[REC_OFFS_NORMAL_SIZE];
-	rec_offs*	offsets		= offsets_;
+	ulint		offsets_[REC_OFFS_NORMAL_SIZE];
+	ulint*		offsets		= offsets_;
 	rec_offs_init(offsets_);
 
 	ut_ad(ahi_latch == btr_get_search_latch(cursor->index));
@@ -1816,17 +1852,19 @@ btr_search_update_hash_on_insert(btr_cur_t* cursor, rw_lock_t* ahi_latch)
 	ut_ad(block->page.id.space() == index->table->space_id);
 	btr_search_check_free_space_in_heap(index);
 
+	table = btr_get_search_table(index);
+
 	rec = btr_cur_get_rec(cursor);
 
 #ifdef MYSQL_INDEX_DISABLE_AHI
 	ut_a(!index->disable_ahi);
 #endif
 	ut_a(index == cursor->index);
-	ut_ad(!dict_index_is_ibuf(index));
+	ut_a(!dict_index_is_ibuf(index));
 
 	n_fields = block->curr_n_fields;
 	n_bytes = block->curr_n_bytes;
-	const bool left_side = block->curr_left_side;
+	left_side = block->curr_left_side;
 
 	ins_rec = page_rec_get_next_const(rec);
 	next_rec = page_rec_get_next_const(ins_rec);
@@ -1843,11 +1881,7 @@ btr_search_update_hash_on_insert(btr_cur_t* cursor, rw_lock_t* ahi_latch)
 				     n_bytes, index->id);
 	}
 
-	/* We must not look up "table" before acquiring ahi_latch. */
-	hash_table_t* table = NULL;
-	bool locked = false;
-
-	if (!page_rec_is_infimum(rec) && !rec_is_metadata(rec, index)) {
+	if (!page_rec_is_infimum(rec) && !rec_is_metadata(rec, *index)) {
 		offsets = rec_get_offsets(
 			rec, index, offsets, true,
 			btr_search_get_n_fields(n_fields, n_bytes), &heap);
@@ -1857,11 +1891,10 @@ btr_search_update_hash_on_insert(btr_cur_t* cursor, rw_lock_t* ahi_latch)
 			locked = true;
 			rw_lock_x_lock(ahi_latch);
 
-			if (!btr_search_enabled || !block->index) {
+			if (!btr_search_enabled) {
 				goto function_exit;
 			}
 
-			table = btr_get_search_table(index);
 			ha_insert_for_fold(table, ins_fold, block, ins_rec);
 		}
 
@@ -1874,11 +1907,9 @@ btr_search_update_hash_on_insert(btr_cur_t* cursor, rw_lock_t* ahi_latch)
 			locked = true;
 			rw_lock_x_lock(ahi_latch);
 
-			if (!btr_search_enabled || !block->index) {
+			if (!btr_search_enabled) {
 				goto function_exit;
 			}
-
-			table = btr_get_search_table(index);
 		}
 
 		if (!left_side) {
@@ -1896,11 +1927,9 @@ check_next_rec:
 				locked = true;
 				rw_lock_x_lock(ahi_latch);
 
-				if (!btr_search_enabled || !block->index) {
+				if (!btr_search_enabled) {
 					goto function_exit;
 				}
-
-				table = btr_get_search_table(index);
 			}
 
 			ha_insert_for_fold(table, ins_fold, block, ins_rec);
@@ -1910,15 +1939,14 @@ check_next_rec:
 	}
 
 	if (ins_fold != next_fold) {
+
 		if (!locked) {
 			locked = true;
 			rw_lock_x_lock(ahi_latch);
 
-			if (!btr_search_enabled || !block->index) {
+			if (!btr_search_enabled) {
 				goto function_exit;
 			}
-
-			table = btr_get_search_table(index);
 		}
 
 		if (!left_side) {
@@ -1952,12 +1980,10 @@ btr_search_hash_table_validate(ulint hash_table_id)
 	ulint		i;
 	ulint		cell_count;
 	mem_heap_t*	heap		= NULL;
-	rec_offs	offsets_[REC_OFFS_NORMAL_SIZE];
-	rec_offs*	offsets		= offsets_;
+	ulint		offsets_[REC_OFFS_NORMAL_SIZE];
+	ulint*		offsets		= offsets_;
 
-	btr_search_x_lock_all();
 	if (!btr_search_enabled) {
-		btr_search_x_unlock_all();
 		return(TRUE);
 	}
 
@@ -1967,6 +1993,7 @@ btr_search_hash_table_validate(ulint hash_table_id)
 
 	rec_offs_init(offsets_);
 
+	btr_search_x_lock_all();
 	buf_pool_mutex_enter_all();
 
 	cell_count = hash_get_n_cells(
@@ -1983,12 +2010,6 @@ btr_search_hash_table_validate(ulint hash_table_id)
 			os_thread_yield();
 
 			btr_search_x_lock_all();
-
-			if (!btr_search_enabled) {
-				ok = true;
-				goto func_exit;
-			}
-
 			buf_pool_mutex_enter_all();
 
 			ulint	curr_cell_count = hash_get_n_cells(
@@ -2048,7 +2069,7 @@ btr_search_hash_table_validate(ulint hash_table_id)
 				     == BUF_BLOCK_REMOVE_HASH);
 			}
 
-			ut_ad(!dict_index_is_ibuf(block->index));
+			ut_a(!dict_index_is_ibuf(block->index));
 			ut_ad(block->page.id.space()
 			      == block->index->table->space_id);
 
@@ -2107,12 +2128,6 @@ btr_search_hash_table_validate(ulint hash_table_id)
 			os_thread_yield();
 
 			btr_search_x_lock_all();
-
-			if (!btr_search_enabled) {
-				ok = true;
-				goto func_exit;
-			}
-
 			buf_pool_mutex_enter_all();
 
 			ulint	curr_cell_count = hash_get_n_cells(
@@ -2137,7 +2152,6 @@ btr_search_hash_table_validate(ulint hash_table_id)
 	}
 
 	buf_pool_mutex_exit_all();
-func_exit:
 	btr_search_x_unlock_all();
 
 	if (UNIV_LIKELY_NULL(heap)) {
