@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1995, 2017, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2013, 2020, MariaDB Corporation.
+Copyright (c) 2013, 2019, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -34,8 +34,6 @@ Created 2011/12/19
 #include "fil0crypt.h"
 #include "fil0pagecompress.h"
 
-using st_::span;
-
 /** The doublewrite buffer */
 buf_dblwr_t*	buf_dblwr = NULL;
 
@@ -101,10 +99,6 @@ void
 buf_dblwr_sync_datafiles()
 /*======================*/
 {
-	/* Wake possible simulated aio thread to actually post the
-	writes to the operating system */
-	os_aio_simulated_wake_handler_threads();
-
 	/* Wait that all async writes to tablespaces have been posted to
 	the OS */
 	os_aio_wait_until_no_pending_writes();
@@ -530,54 +524,30 @@ buf_dblwr_init_or_load_pages(
 void
 buf_dblwr_process()
 {
-	ut_ad(recv_sys.parse_start_lsn);
-
 	ulint		page_no_dblwr	= 0;
 	byte*		read_buf;
+	byte*		unaligned_read_buf;
 	recv_dblwr_t&	recv_dblwr	= recv_sys.dblwr;
 
 	if (!buf_dblwr) {
 		return;
 	}
 
+	unaligned_read_buf = static_cast<byte*>(
+		ut_malloc_nokey(3U << srv_page_size_shift));
+
 	read_buf = static_cast<byte*>(
-		aligned_malloc(3 * srv_page_size, srv_page_size));
+		ut_align(unaligned_read_buf, srv_page_size));
 	byte* const buf = read_buf + srv_page_size;
 
 	for (recv_dblwr_t::list::iterator i = recv_dblwr.pages.begin();
 	     i != recv_dblwr.pages.end();
 	     ++i, ++page_no_dblwr) {
-		byte* page = *i;
-		const ulint page_no = page_get_page_no(page);
+		byte*	page		= *i;
+		ulint	space_id	= page_get_space_id(page);
+		fil_space_t*	space = fil_space_get(space_id);
 
-		if (!page_no) {
-			/* page 0 should have been recovered
-			already via Datafile::restore_from_doublewrite() */
-			continue;
-		}
-
-		const ulint space_id = page_get_space_id(page);
-		const lsn_t lsn = mach_read_from_8(page + FIL_PAGE_LSN);
-
-		if (recv_sys.parse_start_lsn > lsn) {
-			/* Pages written before the checkpoint are
-			not useful for recovery. */
-			continue;
-		}
-
-		const page_id_t page_id(space_id, page_no);
-
-		if (recv_sys.scanned_lsn < lsn) {
-			ib::warn() << "Ignoring a doublewrite copy of page "
-				   << page_id
-				   << " with future log sequence number "
-				   << lsn;
-			continue;
-		}
-
-		fil_space_t* space = fil_space_acquire_for_io(space_id);
-
-		if (!space) {
+		if (space == NULL) {
 			/* Maybe we have dropped the tablespace
 			and this page once belonged to it: do nothing */
 			continue;
@@ -585,26 +555,25 @@ buf_dblwr_process()
 
 		fil_space_open_if_needed(space);
 
-		if (UNIV_UNLIKELY(page_no >= space->size)) {
+		const ulint		page_no	= page_get_page_no(page);
+		const page_id_t		page_id(space_id, page_no);
+
+		if (page_no >= space->size) {
 
 			/* Do not report the warning for undo
 			tablespaces, because they can be truncated in place. */
 			if (!srv_is_undo_tablespace(space_id)) {
-				ib::warn() << "A copy of page " << page_no
+				ib::warn() << "A copy of page " << page_id
 					<< " in the doublewrite buffer slot "
 					<< page_no_dblwr
-					<< " is beyond the end of tablespace "
-					<< space->name
-					<< " (" << space->size << " pages)";
+					<< " is not within space bounds";
 			}
-next_page:
-			space->release_for_io();
 			continue;
 		}
 
 		const ulint physical_size = space->physical_size();
 		const ulint zip_size = space->zip_size();
-		ut_ad(!buf_is_zeroes(span<const byte>(page, physical_size)));
+		ut_ad(!buf_page_is_zeroes(page, physical_size));
 
 		/* We want to ensure that for partial reads the
 		unread portion of the page is NUL. */
@@ -620,33 +589,92 @@ next_page:
 			page_id, zip_size,
 			0, physical_size, read_buf, NULL);
 
-		if (UNIV_UNLIKELY(err != DB_SUCCESS)) {
+		if (err != DB_SUCCESS) {
 			ib::warn()
 				<< "Double write buffer recovery: "
 				<< page_id << " read failed with "
-				<< "error: " << err;
+				<< "error: " << ut_strerr(err);
 		}
 
-		if (buf_is_zeroes(span<const byte>(read_buf, physical_size))) {
+		const bool is_all_zero = buf_page_is_zeroes(
+			read_buf, physical_size);
+		const bool expect_encrypted = space->crypt_data
+			&& space->crypt_data->type != CRYPT_SCHEME_UNENCRYPTED;
+		bool is_corrupted = false;
+
+		if (is_all_zero) {
 			/* We will check if the copy in the
 			doublewrite buffer is valid. If not, we will
 			ignore this page (there should be redo log
 			records to initialize it). */
-		} else if (recv_dblwr.validate_page(
-				page_id, read_buf, space, buf)) {
-			goto next_page;
 		} else {
+			/* Decompress the page before
+			validating the checksum. */
+			ulint decomp = fil_page_decompress(buf, read_buf,
+							   space->flags);
+			if (!decomp || (zip_size && decomp != srv_page_size)) {
+				goto bad;
+			}
+
+			if (expect_encrypted
+			    && buf_page_get_key_version(read_buf, space->flags)) {
+				is_corrupted = !buf_page_verify_crypt_checksum(
+							read_buf, space->flags);
+			} else {
+				is_corrupted = buf_page_is_corrupted(
+					true, read_buf, space->flags);
+			}
+
+			if (!is_corrupted) {
+				/* The page is good; there is no need
+				to consult the doublewrite buffer. */
+				continue;
+			}
+
+bad:
 			/* We intentionally skip this message for
-			all-zero pages. */
+			is_all_zero pages. */
 			ib::info()
 				<< "Trying to recover page " << page_id
 				<< " from the doublewrite buffer.";
 		}
 
-		page = recv_dblwr.find_page(page_id, space, buf);
+		ulint decomp = fil_page_decompress(buf, page, space->flags);
+		if (!decomp || (zip_size && decomp != srv_page_size)) {
+			continue;
+		}
 
-		if (!page) {
-			goto next_page;
+		if (expect_encrypted
+		    && buf_page_get_key_version(read_buf, space->flags)) {
+			is_corrupted = !buf_page_verify_crypt_checksum(
+						page, space->flags);
+		} else {
+			is_corrupted = buf_page_is_corrupted(
+					true, page, space->flags);
+		}
+
+		if (is_corrupted) {
+			/* Theoretically we could have another good
+			copy for this page in the doublewrite
+			buffer. If not, we will report a fatal error
+			for a corrupted page somewhere else if that
+			page was truly needed. */
+			continue;
+		}
+
+		if (page_no == 0) {
+			/* Check the FSP_SPACE_FLAGS. */
+			ulint flags = fsp_header_get_flags(page);
+			if (!fil_space_t::is_valid_flags(flags, space_id)
+			    && fsp_flags_convert_from_101(flags)
+			    == ULINT_UNDEFINED) {
+				ib::warn() << "Ignoring a doublewrite copy"
+					" of page " << page_id
+					<< " due to invalid flags "
+					<< ib::hex(flags);
+				continue;
+			}
+			/* The flags on the page should be converted later. */
 		}
 
 		/* Write the good page from the doublewrite buffer to
@@ -655,18 +683,17 @@ next_page:
 		IORequest	write_request(IORequest::WRITE);
 
 		fil_io(write_request, true, page_id, zip_size,
-		       0, physical_size, page, nullptr);
+		       0, physical_size,
+				const_cast<byte*>(page), NULL);
 
 		ib::info() << "Recovered page " << page_id
 			<< " from the doublewrite buffer.";
-
-		goto next_page;
 	}
 
 	recv_dblwr.pages.clear();
 
 	fil_flush_file_spaces(FIL_TYPE_TABLESPACE);
-	aligned_free(read_buf);
+	ut_free(unaligned_read_buf);
 }
 
 /****************************************************************//**
@@ -825,6 +852,10 @@ buf_dblwr_check_block(
 {
 	ut_ad(buf_block_get_state(block) == BUF_BLOCK_FILE_PAGE);
 
+	if (block->skip_flush_check) {
+		return;
+	}
+
 	switch (fil_page_get_type(block->frame)) {
 	case FIL_PAGE_INDEX:
 	case FIL_PAGE_TYPE_INSTANT:
@@ -879,11 +910,6 @@ buf_dblwr_write_block_to_datafile(
 	ut_a(buf_page_in_file(bpage));
 
 	ulint	type = IORequest::WRITE;
-
-	if (sync) {
-		type |= IORequest::DO_NOT_WAKE;
-	}
-
 	IORequest	request(type, const_cast<buf_page_t*>(bpage));
 
 	/* We request frame here to get correct buffer in case of
@@ -915,9 +941,8 @@ buf_dblwr_write_block_to_datafile(
 }
 
 /********************************************************************//**
-Flushes possible buffered writes from the doublewrite memory buffer to disk,
-and also wakes up the aio thread if simulated aio is used. It is very
-important to call this function after a batch of writes has been posted,
+Flushes possible buffered writes from the doublewrite memory buffer to disk.
+It is very important to call this function after a batch of writes has been posted,
 and also when we may have to wait for a page latch! Otherwise a deadlock
 of threads can occur. */
 void
@@ -947,13 +972,6 @@ try_again:
 	if (buf_dblwr->first_free == 0) {
 
 		mutex_exit(&buf_dblwr->mutex);
-
-		/* Wake possible simulated aio thread as there could be
-		system temporary tablespace pages active for flushing.
-		Note: system temporary tablespace pages are not scheduled
-		for doublewrite. */
-		os_aio_simulated_wake_handler_threads();
-
 		return;
 	}
 
@@ -963,7 +981,6 @@ try_again:
 		int64_t	sig_count = os_event_reset(buf_dblwr->b_event);
 		mutex_exit(&buf_dblwr->mutex);
 
-		os_aio_simulated_wake_handler_threads();
 		os_event_wait_low(buf_dblwr->b_event, sig_count);
 		goto try_again;
 	}
@@ -1056,12 +1073,6 @@ flush:
 		buf_dblwr_write_block_to_datafile(
 			buf_dblwr->buf_block_arr[i], false);
 	}
-
-	/* Wake possible simulated aio thread to actually post the
-	writes to the operating system. We don't flush the files
-	at this point. We leave it to the IO helper thread to flush
-	datafiles when the whole batch has been processed. */
-	os_aio_simulated_wake_handler_threads();
 }
 
 /********************************************************************//**
@@ -1090,7 +1101,6 @@ try_again:
 		checkpoint. */
 		int64_t	sig_count = os_event_reset(buf_dblwr->b_event);
 		mutex_exit(&buf_dblwr->mutex);
-		os_aio_simulated_wake_handler_threads();
 
 		os_event_wait_low(buf_dblwr->b_event, sig_count);
 		goto try_again;
@@ -1112,13 +1122,14 @@ try_again:
 	void * frame = buf_page_get_frame(bpage);
 
 	if (auto zip_size = bpage->zip_size()) {
-		MEM_CHECK_DEFINED(bpage->zip.data, zip_size);
+		UNIV_MEM_ASSERT_RW(bpage->zip.data, zip_size);
 		/* Copy the compressed page and clear the rest. */
 		memcpy(p, frame, zip_size);
 		memset(p + zip_size, 0x0, srv_page_size - zip_size);
 	} else {
 		ut_a(buf_page_get_state(bpage) == BUF_BLOCK_FILE_PAGE);
-		MEM_CHECK_DEFINED(frame, srv_page_size);
+
+		UNIV_MEM_ASSERT_RW(frame, srv_page_size);
 		memcpy(p, frame, srv_page_size);
 	}
 
