@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1996, 2016, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2017, 2020, MariaDB Corporation.
+Copyright (c) 2017, 2019, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -26,9 +26,7 @@ Created 1/8/1996 Heikki Tuuri
 
 #include "dict0crea.h"
 #include "btr0pcur.h"
-#ifdef BTR_CUR_HASH_ADAPT
-# include "btr0sea.h"
-#endif /* BTR_CUR_HASH_ADAPT */
+#include "btr0btr.h"
 #include "page0page.h"
 #include "mach0data.h"
 #include "dict0boot.h"
@@ -58,7 +56,6 @@ dict_create_sys_tables_tuple(
 					which the memory for the built
 					tuple is allocated */
 {
-	dict_table_t*	sys_tables;
 	dtuple_t*	entry;
 	dfield_t*	dfield;
 	byte*		ptr;
@@ -69,11 +66,9 @@ dict_create_sys_tables_tuple(
 	ut_ad(heap);
 	ut_ad(table->n_cols >= DATA_N_SYS_COLS);
 
-	sys_tables = dict_sys->sys_tables;
-
 	entry = dtuple_create(heap, 8 + DATA_N_SYS_COLS);
 
-	dict_table_copy_types(entry, sys_tables);
+	dict_table_copy_types(entry, dict_sys.sys_tables);
 
 	/* 0: NAME -----------------------------*/
 	dfield = dtuple_get_nth_field(
@@ -171,7 +166,6 @@ dict_create_sys_columns_tuple(
 					which the memory for the built
 					tuple is allocated */
 {
-	dict_table_t*		sys_columns;
 	dtuple_t*		entry;
 	const dict_col_t*	column;
 	dfield_t*		dfield;
@@ -195,11 +189,9 @@ dict_create_sys_columns_tuple(
 		ut_ad(!column->is_virtual());
 	}
 
-	sys_columns = dict_sys->sys_columns;
-
 	entry = dtuple_create(heap, 7 + DATA_N_SYS_COLS);
 
-	dict_table_copy_types(entry, sys_columns);
+	dict_table_copy_types(entry, dict_sys.sys_columns);
 
 	/* 0: TABLE_ID -----------------------*/
 	dfield = dtuple_get_nth_field(entry, DICT_COL__SYS_COLUMNS__TABLE_ID);
@@ -236,7 +228,7 @@ dict_create_sys_columns_tuple(
 		col_name = dict_table_get_col_name(table, i);
 	}
 
-	dfield_set_data(dfield, col_name, ut_strlen(col_name));
+	dfield_set_data(dfield, col_name, strlen(col_name));
 
 	/* 5: MTYPE --------------------------*/
 	dfield = dtuple_get_nth_field(entry, DICT_COL__SYS_COLUMNS__MTYPE);
@@ -290,7 +282,6 @@ dict_create_sys_virtual_tuple(
 	ulint			b_col_n,
 	mem_heap_t*		heap)
 {
-	dict_table_t*		sys_virtual;
 	dtuple_t*		entry;
 	const dict_col_t*	base_column;
 	dfield_t*		dfield;
@@ -303,12 +294,10 @@ dict_create_sys_virtual_tuple(
 	dict_v_col_t*	v_col = dict_table_get_nth_v_col(table, v_col_n);
 	base_column = v_col->base_col[b_col_n];
 
-	sys_virtual = dict_sys->sys_virtual;
-
 	entry = dtuple_create(heap, DICT_NUM_COLS__SYS_VIRTUAL
 			      + DATA_N_SYS_COLS);
 
-	dict_table_copy_types(entry, sys_virtual);
+	dict_table_copy_types(entry, dict_sys.sys_virtual);
 
 	/* 0: TABLE_ID -----------------------*/
 	dfield = dtuple_get_nth_field(entry, DICT_COL__SYS_VIRTUAL__TABLE_ID);
@@ -352,12 +341,14 @@ dict_build_table_def_step(
 	que_thr_t*	thr,	/*!< in: query thread */
 	tab_node_t*	node)	/*!< in: table create node */
 {
-	ut_ad(mutex_own(&dict_sys->mutex));
+	ut_ad(mutex_own(&dict_sys.mutex));
 	dict_table_t*	table = node->table;
+	trx_t* trx = thr_get_trx(thr);
 	ut_ad(!table->is_temporary());
 	ut_ad(!table->space);
 	ut_ad(table->space_id == ULINT_UNDEFINED);
-	dict_table_assign_new_id(table, thr_get_trx(thr));
+	dict_hdr_get_new_id(&table->id, NULL, NULL);
+	trx->table_id = table->id;
 
 	/* Always set this bit for all new created tables */
 	DICT_TF2_FLAG_SET(table, DICT_TF2_FTS_AUX_HEX_NAME);
@@ -370,8 +361,6 @@ dict_build_table_def_step(
 
 		ut_ad(DICT_TF_GET_ZIP_SSIZE(table->flags) == 0
 		      || dict_table_has_atomic_blobs(table));
-		trx_t* trx = thr_get_trx(thr);
-		ut_ad(trx->table_id);
 		mtr_t mtr;
 		trx_undo_t* undo = trx->rsegs.m_redo.undo;
 		if (undo && !undo->table_id
@@ -399,7 +388,7 @@ dict_build_table_def_step(
 		}
 		/* Get a new tablespace ID */
 		ulint space_id;
-		dict_hdr_get_new_id(NULL, NULL, &space_id, table, false);
+		dict_hdr_get_new_id(NULL, NULL, &space_id);
 
 		DBUG_EXECUTE_IF(
 			"ib_create_table_fail_out_of_space_ids",
@@ -486,24 +475,21 @@ dict_create_sys_indexes_tuple(
 					which the memory for the built
 					tuple is allocated */
 {
-	dict_table_t*	sys_indexes;
 	dtuple_t*	entry;
 	dfield_t*	dfield;
 	byte*		ptr;
 
-	ut_ad(mutex_own(&dict_sys->mutex));
+	ut_ad(mutex_own(&dict_sys.mutex));
 	ut_ad(index);
 	ut_ad(index->table->space || index->table->file_unreadable);
 	ut_ad(!index->table->space
 	      || index->table->space->id == index->table->space_id);
 	ut_ad(heap);
 
-	sys_indexes = dict_sys->sys_indexes;
-
 	entry = dtuple_create(
 		heap, DICT_NUM_COLS__SYS_INDEXES + DATA_N_SYS_COLS);
 
-	dict_table_copy_types(entry, sys_indexes);
+	dict_table_copy_types(entry, dict_sys.sys_indexes);
 
 	/* 0: TABLE_ID -----------------------*/
 	dfield = dtuple_get_nth_field(
@@ -607,7 +593,6 @@ dict_create_sys_fields_tuple(
 					which the memory for the built
 					tuple is allocated */
 {
-	dict_table_t*	sys_fields;
 	dtuple_t*	entry;
 	dict_field_t*	field;
 	dfield_t*	dfield;
@@ -627,11 +612,9 @@ dict_create_sys_fields_tuple(
 
 	field = dict_index_get_nth_field(index, fld_no);
 
-	sys_fields = dict_sys->sys_fields;
-
 	entry = dtuple_create(heap, 3 + DATA_N_SYS_COLS);
 
-	dict_table_copy_types(entry, sys_fields);
+	dict_table_copy_types(entry, dict_sys.sys_fields);
 
 	/* 0: INDEX_ID -----------------------*/
 	dfield = dtuple_get_nth_field(entry, DICT_COL__SYS_FIELDS__INDEX_ID);
@@ -668,8 +651,7 @@ dict_create_sys_fields_tuple(
 	/* 4: COL_NAME -------------------------*/
 	dfield = dtuple_get_nth_field(entry, DICT_COL__SYS_FIELDS__COL_NAME);
 
-	dfield_set_data(dfield, field->name,
-			ut_strlen(field->name));
+	dfield_set_data(dfield, field->name, strlen(field->name));
 	/*---------------------------------*/
 
 	return(entry);
@@ -726,7 +708,7 @@ dict_build_index_def_step(
 	dtuple_t*	row;
 	trx_t*		trx;
 
-	ut_ad(mutex_own(&dict_sys->mutex));
+	ut_ad(mutex_own(&dict_sys.mutex));
 
 	trx = thr_get_trx(thr);
 
@@ -747,7 +729,7 @@ dict_build_index_def_step(
 	ut_ad((UT_LIST_GET_LEN(table->indexes) > 0)
 	      || dict_index_is_clust(index));
 
-	dict_hdr_get_new_id(NULL, &index->id, NULL, table, false);
+	dict_hdr_get_new_id(NULL, &index->id, NULL);
 
 	/* Inherit the space id from the table; we store all indexes of a
 	table in the same tablespace */
@@ -777,7 +759,7 @@ dict_build_index_def(
 	dict_index_t*		index,	/*!< in/out: index */
 	trx_t*			trx)	/*!< in/out: InnoDB transaction handle */
 {
-	ut_ad(mutex_own(&dict_sys->mutex));
+	ut_ad(mutex_own(&dict_sys.mutex));
 
 	if (trx->table_id == 0) {
 		/* Record only the first table id. */
@@ -787,7 +769,7 @@ dict_build_index_def(
 	ut_ad((UT_LIST_GET_LEN(table->indexes) > 0)
 	      || dict_index_is_clust(index));
 
-	dict_hdr_get_new_id(NULL, &index->id, NULL, table, false);
+	dict_hdr_get_new_id(NULL, &index->id, NULL);
 
 	/* Note that the index was created by this transaction. */
 	index->trx_id = trx->id;
@@ -823,15 +805,12 @@ dict_create_index_tree_step(
 	mtr_t		mtr;
 	btr_pcur_t	pcur;
 	dict_index_t*	index;
-	dict_table_t*	sys_indexes;
 	dtuple_t*	search_tuple;
 
-	ut_ad(mutex_own(&dict_sys->mutex));
+	ut_ad(mutex_own(&dict_sys.mutex));
 
 	index = node->index;
 
-	sys_indexes = dict_sys->sys_indexes;
-
 	if (index->type == DICT_FTS) {
 		/* FTS index does not need an index tree */
 		return(DB_SUCCESS);
@@ -845,7 +824,7 @@ dict_create_index_tree_step(
 
 	search_tuple = dict_create_search_tuple(node->ind_row, node->heap);
 
-	btr_pcur_open(UT_LIST_GET_FIRST(sys_indexes->indexes),
+	btr_pcur_open(UT_LIST_GET_FIRST(dict_sys.sys_indexes->indexes),
 		      search_tuple, PAGE_CUR_L, BTR_MODIFY_LEAF,
 		      &pcur, &mtr);
 
@@ -861,7 +840,7 @@ dict_create_index_tree_step(
 
 		node->page_no = btr_create(
 			index->type, index->table->space,
-			index->id, index, NULL, &mtr);
+			index->id, index, &mtr);
 
 		if (node->page_no == FIL_NULL) {
 			err = DB_OUT_OF_FILE_SPACE;
@@ -898,7 +877,7 @@ dict_create_index_tree_in_mem(
 {
 	mtr_t		mtr;
 
-	ut_ad(mutex_own(&dict_sys->mutex));
+	ut_ad(mutex_own(&dict_sys.mutex));
 	ut_ad(!(index->type & DICT_FTS));
 
 	mtr_start(&mtr);
@@ -910,7 +889,7 @@ dict_create_index_tree_in_mem(
 	ut_ad(!(index->table->flags2 & DICT_TF2_DISCARDED));
 
 	index->page = btr_create(index->type, index->table->space,
-				 index->id, index, NULL, &mtr);
+				 index->id, index, &mtr);
 	mtr_commit(&mtr);
 
 	index->trx_id = trx->id;
@@ -922,16 +901,14 @@ dict_create_index_tree_in_mem(
 @param[in,out]	rec	SYS_INDEXES record
 @param[in,out]	pcur	persistent cursor on rec
 @param[in,out]	trx	dictionary transaction
-@param[in,out]	mtr	mini-transaction
-@return	whether freeing the B-tree was attempted */
-bool dict_drop_index_tree(rec_t* rec, btr_pcur_t* pcur, trx_t* trx, mtr_t* mtr)
+@param[in,out]	mtr	mini-transaction */
+void dict_drop_index_tree(rec_t* rec, btr_pcur_t* pcur, trx_t* trx, mtr_t* mtr)
 {
-	const byte*	ptr;
-	ulint		len;
-	ulint		root_page_no;
+	byte*	ptr;
+	ulint	len;
 
-	ut_ad(mutex_own(&dict_sys->mutex));
-	ut_a(!dict_table_is_comp(dict_sys->sys_indexes));
+	ut_ad(mutex_own(&dict_sys.mutex));
+	ut_a(!dict_table_is_comp(dict_sys.sys_indexes));
 
 	ptr = rec_get_nth_field_old(rec, DICT_FLD__SYS_INDEXES__PAGE_NO, &len);
 
@@ -939,15 +916,15 @@ bool dict_drop_index_tree(rec_t* rec, btr_pcur_t* pcur, trx_t* trx, mtr_t* mtr)
 
 	btr_pcur_store_position(pcur, mtr);
 
-	root_page_no = mtr_read_ulint(ptr, MLOG_4BYTES, mtr);
+	const uint32_t root_page_no = mach_read_from_4(ptr);
 
 	if (root_page_no == FIL_NULL) {
 		/* The tree has already been freed */
-
-		return(false);
+		return;
 	}
 
-	mlog_write_ulint(const_cast<byte*>(ptr), FIL_NULL, MLOG_4BYTES, mtr);
+	compile_time_assert(FIL_NULL == 0xffffffff);
+	mlog_memset(ptr, 4, 0xff, mtr);
 
 	ptr = rec_get_nth_field_old(
 		rec, DICT_FLD__SYS_INDEXES__SPACE, &len);
@@ -960,7 +937,7 @@ bool dict_drop_index_tree(rec_t* rec, btr_pcur_t* pcur, trx_t* trx, mtr_t* mtr)
 	    && trx_get_dict_operation(trx) == TRX_DICT_OP_TABLE) {
 		/* We are about to delete the entire .ibd file;
 		do not bother to free pages inside it. */
-		return false;
+		return;
 	}
 
 	ptr = rec_get_nth_field_old(
@@ -968,109 +945,16 @@ bool dict_drop_index_tree(rec_t* rec, btr_pcur_t* pcur, trx_t* trx, mtr_t* mtr)
 
 	ut_ad(len == 8);
 
-	bool			found;
-	const page_size_t	page_size(fil_space_get_page_size(space_id,
-								  &found));
-
-	if (!found) {
-		/* It is a single table tablespace and the .ibd file is
-		missing: do nothing */
-
-		return(false);
-	}
-
-	/* If tablespace is scheduled for truncate, do not try to drop
-	the indexes in that tablespace. There is a truncate fixup action
-	which will take care of it. */
-	if (srv_is_tablespace_truncated(space_id)) {
-		return(false);
-	}
-
-	btr_free_if_exists(page_id_t(space_id, root_page_no), page_size,
-			   mach_read_from_8(ptr), mtr);
-
-	return(true);
-}
-
-/*******************************************************************//**
-Recreate the index tree associated with a row in SYS_INDEXES table.
-@return	new root page number, or FIL_NULL on failure */
-ulint
-dict_recreate_index_tree(
-/*=====================*/
-	const dict_table_t*
-			table,	/*!< in/out: the table the index belongs to */
-	btr_pcur_t*	pcur,	/*!< in/out: persistent cursor pointing to
-				record in the clustered index of
-				SYS_INDEXES table. The cursor may be
-				repositioned in this call. */
-	mtr_t*		mtr)	/*!< in/out: mtr having the latch
-				on the record page. */
-{
-	ut_ad(mutex_own(&dict_sys->mutex));
-	ut_a(!dict_table_is_comp(dict_sys->sys_indexes));
-	ut_ad(!table->space || table->space->id == table->space_id);
-
-	ulint		len;
-	const rec_t*	rec = btr_pcur_get_rec(pcur);
-
-	const byte*	ptr = rec_get_nth_field_old(
-		rec, DICT_FLD__SYS_INDEXES__PAGE_NO, &len);
-
-	ut_ad(len == 4);
-
-	ut_ad(table->space_id == mach_read_from_4(
-		      rec_get_nth_field_old(rec, DICT_FLD__SYS_INDEXES__SPACE,
-					    &len)));
-	ut_ad(len == 4);
-
-	if (!table->space) {
-		/* It is a single table tablespae and the .ibd file is
-		missing: do nothing. */
-
-		ib::warn()
-			<< "Trying to TRUNCATE a missing .ibd file of table "
-			<< table->name << "!";
-
-		return(FIL_NULL);
-	}
-
-	ptr = rec_get_nth_field_old(rec, DICT_FLD__SYS_INDEXES__TYPE, &len);
-	ut_ad(len == 4);
-	ulint	type = mach_read_from_4(ptr);
-
-	ptr = rec_get_nth_field_old(rec, DICT_FLD__SYS_INDEXES__ID, &len);
-	ut_ad(len == 8);
-	index_id_t	index_id = mach_read_from_8(ptr);
-
-	/* We will need to commit the mini-transaction in order to avoid
-	deadlocks in the btr_create() call, because otherwise we would
-	be freeing and allocating pages in the same mini-transaction. */
-	btr_pcur_store_position(pcur, mtr);
-	mtr_commit(mtr);
-
-	mtr_start(mtr);
-	mtr->set_named_space(table->space);
-	btr_pcur_restore_position(BTR_MODIFY_LEAF, pcur, mtr);
-
-	/* Find the index corresponding to this SYS_INDEXES record. */
-	for (dict_index_t* index = UT_LIST_GET_FIRST(table->indexes);
-	     index != NULL;
-	     index = UT_LIST_GET_NEXT(indexes, index)) {
-		if (index->id == index_id) {
-			ulint root_page_no = (index->type & DICT_FTS)
-				? FIL_NULL
-				: btr_create(type, table->space,
-					     index_id, index, NULL, mtr);
-			index->page = unsigned(root_page_no);
-			return root_page_no;
+	if (fil_space_t* s = fil_space_acquire_silent(space_id)) {
+		/* Ensure that the tablespace file exists
+		in order to avoid a crash in buf_page_get_gen(). */
+		if (s->size || fil_space_get_size(space_id)) {
+			btr_free_if_exists(page_id_t(space_id, root_page_no),
+					   s->zip_size(),
+					   mach_read_from_8(ptr), mtr);
 		}
+		s->release();
 	}
-
-	ib::error() << "Failed to create index with index id " << index_id
-		<< " of table " << table->name;
-
-	return(FIL_NULL);
 }
 
 /*********************************************************************//**
@@ -1099,15 +983,15 @@ tab_create_graph_create(
 	node->mode = mode;
 	node->key_id = key_id;
 
-	node->tab_def = ins_node_create(INS_DIRECT, dict_sys->sys_tables,
+	node->tab_def = ins_node_create(INS_DIRECT, dict_sys.sys_tables,
 					heap);
 	node->tab_def->common.parent = node;
 
-	node->col_def = ins_node_create(INS_DIRECT, dict_sys->sys_columns,
+	node->col_def = ins_node_create(INS_DIRECT, dict_sys.sys_columns,
 					heap);
 	node->col_def->common.parent = node;
 
-	node->v_col_def = ins_node_create(INS_DIRECT, dict_sys->sys_virtual,
+	node->v_col_def = ins_node_create(INS_DIRECT, dict_sys.sys_virtual,
                                           heap);
 	node->v_col_def->common.parent = node;
 
@@ -1146,11 +1030,11 @@ ind_create_graph_create(
 	node->heap = mem_heap_create(256);
 
 	node->ind_def = ins_node_create(INS_DIRECT,
-					dict_sys->sys_indexes, heap);
+					dict_sys.sys_indexes, heap);
 	node->ind_def->common.parent = node;
 
 	node->field_def = ins_node_create(INS_DIRECT,
-					  dict_sys->sys_fields, heap);
+					  dict_sys.sys_fields, heap);
 	node->field_def->common.parent = node;
 
 	return(node);
@@ -1169,7 +1053,7 @@ dict_create_table_step(
 	trx_t*		trx;
 
 	ut_ad(thr);
-	ut_ad(mutex_own(&dict_sys->mutex));
+	ut_ad(mutex_own(&dict_sys.mutex));
 
 	trx = thr_get_trx(thr);
 
@@ -1250,7 +1134,8 @@ dict_create_table_step(
 				ut_ad(node->col_no == v_col->v_pos);
 				dict_build_v_col_def_step(node);
 
-				if (node->base_col_no < v_col->num_base - 1) {
+				if (node->base_col_no
+				    < unsigned{v_col->num_base} - 1) {
 					/* move on to next base column */
 					node->base_col_no++;
 				} else {
@@ -1311,7 +1196,7 @@ dict_create_index_step(
 	trx_t*		trx;
 
 	ut_ad(thr);
-	ut_ad(mutex_own(&dict_sys->mutex));
+	ut_ad(mutex_own(&dict_sys.mutex));
 
 	trx = thr_get_trx(thr);
 
@@ -1414,9 +1299,6 @@ dict_create_index_step(
 					&node->table->fts->cache->init_lock);
 			}
 
-#ifdef BTR_CUR_HASH_ADAPT
-			ut_ad(!node->index->search_info->ref_count);
-#endif /* BTR_CUR_HASH_ADAPT */
 			dict_index_remove_from_cache(node->table, node->index);
 			node->index = NULL;
 
@@ -1470,9 +1352,9 @@ dict_check_if_system_table_exists(
 	dict_table_t*	sys_table;
 	dberr_t		error = DB_SUCCESS;
 
-	ut_a(srv_get_active_thread_type() == SRV_NONE);
+	ut_ad(!srv_any_background_activity());
 
-	mutex_enter(&dict_sys->mutex);
+	mutex_enter(&dict_sys.mutex);
 
 	sys_table = dict_table_get_low(tablename);
 
@@ -1490,7 +1372,7 @@ dict_check_if_system_table_exists(
 		dict_table_prevent_eviction(sys_table);
 	}
 
-	mutex_exit(&dict_sys->mutex);
+	mutex_exit(&dict_sys.mutex);
 
 	return(error);
 }
@@ -1510,7 +1392,7 @@ dict_create_or_check_foreign_constraint_tables(void)
 	dberr_t		sys_foreign_err;
 	dberr_t		sys_foreign_cols_err;
 
-	ut_a(srv_get_active_thread_type() == SRV_NONE);
+	ut_ad(!srv_any_background_activity());
 
 	/* Note: The master thread has not been started at this point. */
 
@@ -1603,9 +1485,10 @@ dict_create_or_check_foreign_constraint_tables(void)
 		"END;\n",
 		FALSE, trx);
 
-	if (UNIV_UNLIKELY(err != DB_SUCCESS)) {
+	if (err != DB_SUCCESS) {
+
 		ib::error() << "Creation of SYS_FOREIGN and SYS_FOREIGN_COLS"
-			" failed: " << err << ". Tablespace is"
+			" failed: " << ut_strerr(err) << ". Tablespace is"
 			" full. Dropping incompletely created tables.";
 
 		ut_ad(err == DB_OUT_OF_FILE_SPACE
@@ -1651,16 +1534,16 @@ dict_create_or_check_sys_virtual()
 	my_bool		srv_file_per_table_backup;
 	dberr_t		err;
 
-	ut_a(srv_get_active_thread_type() == SRV_NONE);
+	ut_ad(!srv_any_background_activity());
 
 	/* Note: The master thread has not been started at this point. */
 	err = dict_check_if_system_table_exists(
 		"SYS_VIRTUAL", DICT_NUM_FIELDS__SYS_VIRTUAL + 1, 1);
 
 	if (err == DB_SUCCESS) {
-		mutex_enter(&dict_sys->mutex);
-		dict_sys->sys_virtual = dict_table_get_low("SYS_VIRTUAL");
-		mutex_exit(&dict_sys->mutex);
+		mutex_enter(&dict_sys.mutex);
+		dict_sys.sys_virtual = dict_table_get_low("SYS_VIRTUAL");
+		mutex_exit(&dict_sys.mutex);
 		return(DB_SUCCESS);
 	}
 
@@ -1704,9 +1587,10 @@ dict_create_or_check_sys_virtual()
 		"END;\n",
 		FALSE, trx);
 
-	if (UNIV_UNLIKELY(err != DB_SUCCESS)) {
+	if (err != DB_SUCCESS) {
+
 		ib::error() << "Creation of SYS_VIRTUAL"
-			" failed: " << err << ". Tablespace is"
+			" failed: " << ut_strerr(err) << ". Tablespace is"
 			" full or too many transactions."
 			" Dropping incompletely created tables.";
 
@@ -1733,9 +1617,9 @@ dict_create_or_check_sys_virtual()
 	dberr_t sys_virtual_err = dict_check_if_system_table_exists(
 		"SYS_VIRTUAL", DICT_NUM_FIELDS__SYS_VIRTUAL + 1, 1);
 	ut_a(sys_virtual_err == DB_SUCCESS);
-	mutex_enter(&dict_sys->mutex);
-	dict_sys->sys_virtual = dict_table_get_low("SYS_VIRTUAL");
-	mutex_exit(&dict_sys->mutex);
+	mutex_enter(&dict_sys.mutex);
+	dict_sys.sys_virtual = dict_table_get_low("SYS_VIRTUAL");
+	mutex_exit(&dict_sys.mutex);
 
 	return(err);
 }
@@ -1784,9 +1668,9 @@ dict_foreign_eval_sql(
 		return(error);
 	}
 
-	if (UNIV_UNLIKELY(error != DB_SUCCESS)) {
+	if (error != DB_SUCCESS) {
 		ib::error() << "Foreign key constraint creation failed: "
-			<< error;
+			<< ut_strerr(error);
 
 		mutex_enter(&dict_foreign_err_mutex);
 		ut_print_timestamp(ef);
@@ -2132,7 +2016,7 @@ dict_create_add_foreigns_to_dictionary(
 	dict_foreign_t*	foreign;
 	dberr_t		error;
 
-	ut_ad(mutex_own(&dict_sys->mutex));
+	ut_ad(mutex_own(&dict_sys.mutex));
 
 	if (NULL == dict_table_get_low("SYS_FOREIGN")) {
 
@@ -2142,6 +2026,8 @@ dict_create_add_foreigns_to_dictionary(
 		return(DB_ERROR);
 	}
 
+	error = DB_SUCCESS;
+
 	for (dict_foreign_set::const_iterator it = local_fk_set.begin();
 	     it != local_fk_set.end();
 	     ++it) {
@@ -2153,12 +2039,11 @@ dict_create_add_foreigns_to_dictionary(
 			table->name.m_name, foreign, trx);
 
 		if (error != DB_SUCCESS) {
-
-			return(error);
+			break;
 		}
 	}
 
-	return(DB_SUCCESS);
+	return error;
 }
 
 /****************************************************************//**
@@ -2176,7 +2061,7 @@ dict_create_or_check_sys_tablespace(void)
 	dberr_t		sys_tablespaces_err;
 	dberr_t		sys_datafiles_err;
 
-	ut_a(srv_get_active_thread_type() == SRV_NONE);
+	ut_ad(!srv_any_background_activity());
 
 	/* Note: The master thread has not been started at this point. */
 
@@ -2236,9 +2121,10 @@ dict_create_or_check_sys_tablespace(void)
 		"END;\n",
 		FALSE, trx);
 
-	if (UNIV_UNLIKELY(err != DB_SUCCESS)) {
+	if (err != DB_SUCCESS) {
+
 		ib::error() << "Creation of SYS_TABLESPACES and SYS_DATAFILES"
-			" has failed with error " << err
+			" has failed with error " << ut_strerr(err)
 			<< ". Dropping incompletely created tables.";
 
 		ut_a(err == DB_OUT_OF_FILE_SPACE
@@ -2347,14 +2233,44 @@ dict_replace_tablespace_in_dictionary(
 	return(error);
 }
 
-/** Assign a new table ID and put it into the table cache and the transaction.
-@param[in,out]	table	Table that needs an ID
-@param[in,out]	trx	Transaction */
-void
-dict_table_assign_new_id(
-	dict_table_t*	table,
+/** Delete records from SYS_TABLESPACES and SYS_DATAFILES associated
+with a particular tablespace ID.
+@param[in]	space	Tablespace ID
+@param[in,out]	trx	Current transaction
+@return DB_SUCCESS if OK, dberr_t if the operation failed */
+
+dberr_t
+dict_delete_tablespace_and_datafiles(
+	ulint		space,
 	trx_t*		trx)
 {
-	dict_hdr_get_new_id(&table->id, NULL, NULL, table, false);
-	trx->table_id = table->id;
+	dberr_t		err = DB_SUCCESS;
+
+	ut_d(dict_sys.assert_locked());
+	ut_ad(srv_sys_tablespaces_open);
+
+	trx->op_info = "delete tablespace and datafiles from dictionary";
+
+	pars_info_t*	info = pars_info_create();
+	ut_a(!is_system_tablespace(space));
+	pars_info_add_int4_literal(info, "space", space);
+
+	err = que_eval_sql(info,
+			   "PROCEDURE P () IS\n"
+			   "BEGIN\n"
+			   "DELETE FROM SYS_TABLESPACES\n"
+			   "WHERE SPACE = :space;\n"
+			   "DELETE FROM SYS_DATAFILES\n"
+			   "WHERE SPACE = :space;\n"
+			   "END;\n",
+			   FALSE, trx);
+
+	if (err != DB_SUCCESS) {
+		ib::warn() << "Could not delete space_id "
+			<< space << " from data dictionary";
+	}
+
+	trx->op_info = "";
+
+	return(err);
 }
