@@ -2,7 +2,7 @@
 
 Copyright (c) 1996, 2018, Oracle and/or its affiliates. All Rights Reserved.
 Copyright (c) 2012, Facebook Inc.
-Copyright (c) 2013, 2020, MariaDB Corporation.
+Copyright (c) 2013, 2019, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -148,9 +148,6 @@ dict_mem_table_create(
 	lock_table_lock_list_init(&table->locks);
 
 	UT_LIST_INIT(table->indexes, &dict_index_t::indexes);
-#ifdef BTR_CUR_HASH_ADAPT
-	UT_LIST_INIT(table->freed_indexes, &dict_index_t::indexes);
-#endif /* BTR_CUR_HASH_ADAPT */
 
 	table->heap = heap;
 
@@ -170,17 +167,13 @@ dict_mem_table_create(
 		mem_heap_alloc(heap, table->n_cols * sizeof(dict_col_t)));
 	table->v_cols = static_cast<dict_v_col_t*>(
 		mem_heap_alloc(heap, n_v_cols * sizeof(*table->v_cols)));
-
-	/* true means that the stats latch will be enabled -
-	dict_table_stats_lock() will not be noop. */
-	dict_table_stats_latch_create(table, true);
+	for (ulint i = n_v_cols; i--; ) {
+		new (&table->v_cols[i]) dict_v_col_t();
+	}
 
 	table->autoinc_lock = static_cast<ib_lock_t*>(
 		mem_heap_alloc(heap, lock_get_size()));
 
-	/* lazy creation of table autoinc latch */
-	dict_table_autoinc_create_lazy(table);
-
 	/* If the table has an FTS index or we are in the process
 	of building one, create the table->fts */
 	if (dict_table_has_fts_index(table)
@@ -188,13 +181,14 @@ dict_mem_table_create(
 	    || DICT_TF2_FLAG_IS_SET(table, DICT_TF2_FTS_ADD_DOC_ID)) {
 		table->fts = fts_create(table);
 		table->fts->cache = fts_cache_create(table);
-	} else {
-		table->fts = NULL;
 	}
 
 	new(&table->foreign_set) dict_foreign_set();
 	new(&table->referenced_set) dict_foreign_set();
 
+	rw_lock_create(dict_table_stats_key, &table->stats_latch,
+		       SYNC_INDEX_TREE);
+
 	return(table);
 }
 
@@ -207,10 +201,6 @@ dict_mem_table_free(
 {
 	ut_ad(table);
 	ut_ad(table->magic_n == DICT_TABLE_MAGIC_N);
-	ut_ad(UT_LIST_GET_LEN(table->indexes) == 0);
-#ifdef BTR_CUR_HASH_ADAPT
-	ut_ad(UT_LIST_GET_LEN(table->freed_indexes) == 0);
-#endif /* BTR_CUR_HASH_ADAPT */
 	ut_d(table->cached = FALSE);
 
 	if (dict_table_has_fts_index(table)
@@ -223,9 +213,7 @@ dict_mem_table_free(
 		}
 	}
 
-	dict_table_autoinc_destroy(table);
 	dict_mem_table_free_foreign_vcol_set(table);
-	dict_table_stats_latch_destroy(table);
 
 	table->foreign_set.~dict_foreign_set();
 	table->referenced_set.~dict_foreign_set();
@@ -236,15 +224,12 @@ dict_mem_table_free(
 	/* Clean up virtual index info structures that are registered
 	with virtual columns */
 	for (ulint i = 0; i < table->n_v_def; i++) {
-		dict_v_col_t*	vcol
-			= dict_table_get_nth_v_col(table, i);
-
-		UT_DELETE(vcol->v_indexes);
+		dict_table_get_nth_v_col(table, i)->~dict_v_col_t();
 	}
 
-	if (table->s_cols != NULL) {
-		UT_DELETE(table->s_cols);
-	}
+	UT_DELETE(table->s_cols);
+
+	rw_lock_free(&table->stats_latch);
 
 	mem_heap_free(table->heap);
 }
@@ -423,7 +408,8 @@ dict_mem_table_add_v_col(
 	v_col->num_base = num_base;
 
 	/* Initialize the index list for virtual columns */
-	v_col->v_indexes = UT_NEW_NOKEY(dict_v_idx_list());
+	ut_ad(v_col->v_indexes.empty());
+	v_col->n_v_indexes = 0;
 
 	return(v_col);
 }
@@ -457,7 +443,7 @@ dict_mem_table_add_s_col(
 	}
 
 	s_col.num_base = num_base;
-	table->s_cols->push_back(s_col);
+	table->s_cols->push_front(s_col);
 }
 
 /**********************************************************************//**
@@ -533,6 +519,14 @@ dict_mem_table_col_rename_low(
 					= dict_index_get_nth_field(
 						index, i);
 
+				ut_ad(!field->name
+				      == field->col->is_dropped());
+				if (!field->name) {
+					/* dropped columns lack a name */
+					ut_ad(index->is_instant());
+					continue;
+				}
+
 				/* if is_virtual and that in field->col does
 				not match, continue */
 				if ((!is_virtual) !=
@@ -599,14 +593,15 @@ dict_mem_table_col_rename_low(
 				}
 			}
 
-			/* New index can be null if InnoDB already dropped
-			the foreign index when FOREIGN_KEY_CHECKS is
-			disabled */
-			foreign->foreign_index = dict_foreign_find_index(
+			dict_index_t* new_index = dict_foreign_find_index(
 				foreign->foreign_table, NULL,
 				foreign->foreign_col_names,
 				foreign->n_fields, NULL, true, false,
 				NULL, NULL, NULL);
+			/* There must be an equivalent index in this case. */
+			ut_ad(new_index != NULL);
+
+			foreign->foreign_index = new_index;
 
 		} else {
 
@@ -629,41 +624,7 @@ dict_mem_table_col_rename_low(
 
 		foreign = *it;
 
-		if (!foreign->referenced_index) {
-			/* Referenced index could have been dropped
-			when foreign_key_checks is disabled. In that case,
-			rename the corresponding referenced_col_names and
-			find the equivalent referenced index also */
-			for (unsigned f = 0; f < foreign->n_fields; f++) {
-
-				const char*& rc =
-					foreign->referenced_col_names[f];
-				if (strcmp(rc, from)) {
-					continue;
-				}
-
-				if (to_len <= strlen(rc)) {
-					memcpy(const_cast<char*>(rc), to,
-					       to_len + 1);
-				} else {
-					rc = static_cast<char*>(
-						mem_heap_dup(
-							foreign->heap,
-							to, to_len + 1));
-				}
-			}
-
-			/* New index can be null if InnoDB already dropped
-			the referenced index when FOREIGN_KEY_CHECKS is
-			disabled */
-			foreign->referenced_index = dict_foreign_find_index(
-				foreign->referenced_table, NULL,
-				foreign->referenced_col_names,
-				foreign->n_fields, NULL, true, false,
-				NULL, NULL, NULL);
-			return;
-		}
-
+		ut_ad(foreign->referenced_index != NULL);
 
 		for (unsigned f = 0; f < foreign->n_fields; f++) {
 			/* foreign->referenced_col_names[] need to be
@@ -750,6 +711,7 @@ dict_mem_fill_column_struct(
 	column->mbmaxlen = mbmaxlen;
 	column->def_val.data = NULL;
 	column->def_val.len = UNIV_SQL_DEFAULT;
+	ut_ad(!column->is_dropped());
 }
 
 /**********************************************************************//**
@@ -778,16 +740,15 @@ dict_mem_index_create(
 
 	dict_mem_fill_index_struct(index, heap, index_name, type, n_fields);
 
-	dict_index_zip_pad_mutex_create_lazy(index);
+	mutex_create(LATCH_ID_ZIP_PAD_MUTEX, &index->zip_pad.mutex);
 
 	if (type & DICT_SPATIAL) {
-		index->rtr_track = static_cast<rtr_info_track_t*>(
-					mem_heap_alloc(
-						heap,
-						sizeof(*index->rtr_track)));
+		mutex_create(LATCH_ID_RTR_SSN_MUTEX, &index->rtr_ssn.mutex);
+		index->rtr_track = new
+			(mem_heap_alloc(heap, sizeof *index->rtr_track))
+			rtr_info_track_t();
 		mutex_create(LATCH_ID_RTR_ACTIVE_MUTEX,
 			     &index->rtr_track->rtr_active_mutex);
-		index->rtr_track->rtr_active = UT_NEW_NOKEY(rtr_info_active());
 	}
 
 	return(index);
@@ -895,11 +856,7 @@ dict_mem_fill_vcol_has_index(
 			continue;
 		}
 
-		dict_v_idx_list::iterator it;
-		for (it = v_col->v_indexes->begin();
-		     it != v_col->v_indexes->end(); ++it) {
-			dict_v_idx_t	v_idx = *it;
-
+		for (const auto& v_idx : v_col->v_indexes) {
 			if (v_idx.index != index) {
 				continue;
 			}
@@ -972,7 +929,7 @@ dict_mem_fill_vcol_set_for_base_col(
 			continue;
 		}
 
-		for (ulint j = 0; j < v_col->num_base; j++) {
+		for (ulint j = 0; j < unsigned{v_col->num_base}; j++) {
 			if (strcmp(col_name, dict_table_get_col_name(
 					table,
 					v_col->base_col[j]->ind)) == 0) {
@@ -1093,24 +1050,18 @@ dict_mem_index_free(
 	ut_ad(index);
 	ut_ad(index->magic_n == DICT_INDEX_MAGIC_N);
 
-	dict_index_zip_pad_mutex_destroy(index);
+	mutex_free(&index->zip_pad.mutex);
 
 	if (dict_index_is_spatial(index)) {
-		rtr_info_active::iterator	it;
-		rtr_info_t*			rtr_info;
-
-		for (it = index->rtr_track->rtr_active->begin();
-		     it != index->rtr_track->rtr_active->end(); ++it) {
-			rtr_info = *it;
-
+		for (auto& rtr_info : index->rtr_track->rtr_active) {
 			rtr_info->index = NULL;
 		}
-
+		mutex_destroy(&index->rtr_ssn.mutex);
 		mutex_destroy(&index->rtr_track->rtr_active_mutex);
-		UT_DELETE(index->rtr_track->rtr_active);
+		index->rtr_track->~rtr_info_track_t();
 	}
 
-	dict_index_remove_from_v_col_list(index);
+	index->detach_columns();
 	mem_heap_free(index->heap);
 }
 
@@ -1197,293 +1148,122 @@ operator<< (std::ostream& out, const dict_foreign_set& fk_set)
 	return(out);
 }
 
-/** Adjust clustered index metadata for instant ADD COLUMN.
-@param[in]	clustered index definition after instant ADD COLUMN */
-inline void dict_index_t::instant_add_field(const dict_index_t& instant)
+/** Reconstruct the clustered index fields. */
+inline void dict_index_t::reconstruct_fields()
 {
 	DBUG_ASSERT(is_primary());
-	DBUG_ASSERT(instant.is_primary());
-	DBUG_ASSERT(!instant.is_instant());
-	DBUG_ASSERT(n_def == n_fields);
-	DBUG_ASSERT(instant.n_def == instant.n_fields);
 
-	DBUG_ASSERT(type == instant.type);
-	DBUG_ASSERT(trx_id_offset == instant.trx_id_offset);
-	DBUG_ASSERT(n_user_defined_cols == instant.n_user_defined_cols);
-	DBUG_ASSERT(n_uniq == instant.n_uniq);
-	DBUG_ASSERT(instant.n_fields > n_fields);
-	DBUG_ASSERT(instant.n_def > n_def);
-	DBUG_ASSERT(instant.n_nullable >= n_nullable);
-	DBUG_ASSERT(instant.n_core_fields >= n_core_fields);
-	DBUG_ASSERT(instant.n_core_null_bytes >= n_core_null_bytes);
+	n_fields += table->instant->n_dropped;
+	n_def += table->instant->n_dropped;
 
-	n_fields = instant.n_fields;
-	n_def = instant.n_def;
-	n_nullable = instant.n_nullable;
-	fields = static_cast<dict_field_t*>(
-		mem_heap_dup(heap, instant.fields, n_fields * sizeof *fields));
+	const unsigned n_first = first_user_field();
 
-	ut_d(unsigned n_null = 0);
+	dict_field_t* tfields = static_cast<dict_field_t*>(
+		mem_heap_zalloc(heap, n_fields * sizeof *fields));
 
-	for (unsigned i = 0; i < n_fields; i++) {
-		DBUG_ASSERT(fields[i].same(instant.fields[i]));
-		const dict_col_t* icol = instant.fields[i].col;
-		DBUG_ASSERT(!icol->is_virtual());
-		dict_col_t* col = fields[i].col = &table->cols[
-			icol - instant.table->cols];
-		fields[i].name = col->name(*table);
-		ut_d(n_null += col->is_nullable());
-	}
+	memcpy(tfields, fields, n_first * sizeof *fields);
 
-	ut_ad(n_null == n_nullable);
-}
-
-/** Adjust metadata for instant ADD COLUMN.
-@param[in]	table	table definition after instant ADD COLUMN */
-void dict_table_t::instant_add_column(const dict_table_t& table)
-{
-	DBUG_ASSERT(!table.cached);
-	DBUG_ASSERT(table.n_def == table.n_cols);
-	DBUG_ASSERT(table.n_t_def == table.n_t_cols);
-	DBUG_ASSERT(n_def == n_cols);
-	DBUG_ASSERT(n_t_def == n_t_cols);
-	DBUG_ASSERT(table.n_cols > n_cols);
-	ut_ad(mutex_own(&dict_sys->mutex));
-
-	const char* end = table.col_names;
-	for (unsigned i = table.n_cols; i--; ) end += strlen(end) + 1;
-
-	col_names = static_cast<char*>(
-		mem_heap_dup(heap, table.col_names,
-			     ulint(end - table.col_names)));
-	const dict_col_t* const old_cols = cols;
-	const dict_col_t* const old_cols_end = cols + n_cols;
-	cols = static_cast<dict_col_t*>(mem_heap_dup(heap, table.cols,
-						     table.n_cols
-						     * sizeof *cols));
-
-	/* Preserve the default values of previously instantly
-	added columns. */
-	for (unsigned i = unsigned(n_cols) - DATA_N_SYS_COLS; i--; ) {
-		cols[i].def_val = old_cols[i].def_val;
-	}
-
-	/* Copy the new default values to this->heap. */
-	for (unsigned i = n_cols; i < table.n_cols; i++) {
-		dict_col_t& c = cols[i - DATA_N_SYS_COLS];
-		DBUG_ASSERT(c.is_instant());
-		if (c.def_val.len == 0) {
-			c.def_val.data = field_ref_zero;
-		} else if (const void*& d = c.def_val.data) {
-			d = mem_heap_dup(heap, d, c.def_val.len);
+	n_nullable = 0;
+	ulint n_core_null = 0;
+	const bool comp = dict_table_is_comp(table);
+	const auto* field_map_it = table->instant->field_map;
+	for (unsigned i = n_first, j = 0; i < n_fields; ) {
+		dict_field_t& f = tfields[i++];
+		auto c = *field_map_it++;
+		if (c.is_dropped()) {
+			f.col = &table->instant->dropped[j++];
+			DBUG_ASSERT(f.col->is_dropped());
+			f.fixed_len = dict_col_get_fixed_size(f.col, comp);
 		} else {
-			DBUG_ASSERT(c.def_val.len == UNIV_SQL_NULL);
+			DBUG_ASSERT(!c.is_not_null());
+			const auto old = std::find_if(
+				fields + n_first, fields + n_fields,
+				[c](const dict_field_t& o)
+				{ return o.col->ind == c.ind(); });
+			ut_ad(old >= &fields[n_first]);
+			ut_ad(old < &fields[n_fields]);
+			DBUG_ASSERT(!old->prefix_len);
+			DBUG_ASSERT(old->col == &table->cols[c.ind()]);
+			f = *old;
+		}
+
+		f.col->clear_instant();
+		if (f.col->is_nullable()) {
+			n_nullable++;
+			n_core_null += i <= n_core_fields;
 		}
 	}
 
-	const unsigned old_n_cols = n_cols;
-	const unsigned n_add = unsigned(table.n_cols - n_cols);
-
-	n_t_def += n_add;
-	n_t_cols += n_add;
-	n_cols = table.n_cols;
-	n_def = n_cols;
-
-	for (unsigned i = n_v_def; i--; ) {
-		const dict_v_col_t& v = v_cols[i];
-		for (ulint n = v.num_base; n--; ) {
-			dict_col_t*& base = v.base_col[n];
-			if (!base->is_virtual()) {
-				DBUG_ASSERT(base >= old_cols);
-				size_t n = size_t(base - old_cols);
-				DBUG_ASSERT(n + DATA_N_SYS_COLS < old_n_cols);
-				base = &cols[n];
-			}
-		}
-	}
-
-	dict_index_t* index = dict_table_get_first_index(this);
-
-	index->instant_add_field(*dict_table_get_first_index(&table));
-
-	while ((index = dict_table_get_next_index(index)) != NULL) {
-		for (unsigned i = 0; i < index->n_fields; i++) {
-			dict_field_t& field = index->fields[i];
-			if (field.col < old_cols
-			    || field.col >= old_cols_end) {
-				DBUG_ASSERT(field.col->is_virtual());
-			} else {
-				/* Secondary indexes may contain user
-				columns and DB_ROW_ID (if there is
-				GEN_CLUST_INDEX instead of PRIMARY KEY),
-				but not DB_TRX_ID,DB_ROLL_PTR. */
-				DBUG_ASSERT(field.col >= old_cols);
-				size_t n = size_t(field.col - old_cols);
-				DBUG_ASSERT(n + DATA_N_SYS_COLS <= old_n_cols);
-				if (n + DATA_N_SYS_COLS >= old_n_cols) {
-					/* Replace DB_ROW_ID */
-					n += n_add;
-				}
-				field.col = &cols[n];
-				DBUG_ASSERT(!field.col->is_virtual());
-				field.name = field.col->name(*this);
-			}
-		}
-	}
+	fields = tfields;
+	n_core_null_bytes = UT_BITS_IN_BYTES(n_core_null);
 }
 
-/** Roll back instant_add_column().
-@param[in]	old_n_cols	original n_cols
-@param[in]	old_cols	original cols
-@param[in]	old_col_names	original col_names */
-void
-dict_table_t::rollback_instant(
-	unsigned	old_n_cols,
-	dict_col_t*	old_cols,
-	const char*	old_col_names)
+/** Reconstruct dropped or reordered columns.
+@param[in]	metadata	data from serialise_columns()
+@param[in]	len		length of the metadata, in bytes
+@return whether parsing the metadata failed */
+bool dict_table_t::deserialise_columns(const byte* metadata, ulint len)
 {
-	ut_ad(mutex_own(&dict_sys->mutex));
-	dict_index_t* index = indexes.start;
-	/* index->is_instant() does not necessarily hold here, because
-	the table may have been emptied */
-	DBUG_ASSERT(old_n_cols >= DATA_N_SYS_COLS);
-	DBUG_ASSERT(n_cols >= old_n_cols);
-	DBUG_ASSERT(n_cols == n_def);
-	DBUG_ASSERT(index->n_def == index->n_fields);
+	DBUG_ASSERT(!instant);
 
-	const unsigned n_remove = n_cols - old_n_cols;
+	unsigned num_non_pk_fields = mach_read_from_4(metadata);
+	metadata += 4;
 
-	for (unsigned i = index->n_fields - n_remove; i < index->n_fields;
-	     i++) {
-		if (index->fields[i].col->is_nullable()) {
-			index->n_nullable--;
-		}
+	if (num_non_pk_fields >= REC_MAX_N_FIELDS - 3) {
+		return true;
 	}
 
-	index->n_fields -= n_remove;
-	index->n_def = index->n_fields;
-	if (index->n_core_fields > index->n_fields) {
-		index->n_core_fields = index->n_fields;
-		index->n_core_null_bytes
-			= UT_BITS_IN_BYTES(unsigned(index->n_nullable));
+	dict_index_t* index = UT_LIST_GET_FIRST(indexes);
+
+	if (num_non_pk_fields < unsigned(index->n_fields)
+	    - index->first_user_field()) {
+		return true;
 	}
 
-	const dict_col_t* const new_cols = cols;
-	const dict_col_t* const new_cols_end = cols + n_cols;
+	field_map_element_t* field_map = static_cast<field_map_element_t*>(
+		mem_heap_alloc(heap,
+			       num_non_pk_fields * sizeof *field_map));
 
-	cols = old_cols;
-	col_names = old_col_names;
-	n_cols = old_n_cols;
-	n_def = old_n_cols;
-	n_t_def -= n_remove;
-	n_t_cols -= n_remove;
+	unsigned n_dropped_cols = 0;
 
-	for (unsigned i = n_v_def; i--; ) {
-		const dict_v_col_t& v = v_cols[i];
-		for (ulint n = v.num_base; n--; ) {
-			dict_col_t*& base = v.base_col[n];
-			if (!base->is_virtual()) {
-				base = &cols[base - new_cols];
+	for (unsigned i = 0; i < num_non_pk_fields; i++) {
+		auto c = field_map[i] = mach_read_from_2(metadata);
+		metadata += 2;
+
+		if (field_map[i].is_dropped()) {
+			if (c.ind() > DICT_MAX_FIXED_COL_LEN + 1) {
+				return true;
 			}
+			n_dropped_cols++;
+		} else if (c >= n_cols) {
+			return true;
 		}
 	}
 
-	do {
-		for (unsigned i = 0; i < index->n_fields; i++) {
-			dict_field_t& field = index->fields[i];
-			if (field.col < new_cols
-			    || field.col >= new_cols_end) {
-				DBUG_ASSERT(field.col->is_virtual());
-			} else {
-				DBUG_ASSERT(field.col >= new_cols);
-				size_t n = size_t(field.col - new_cols);
-				DBUG_ASSERT(n <= n_cols);
-				if (n + DATA_N_SYS_COLS >= n_cols) {
-					n -= n_remove;
-				}
-				field.col = &cols[n];
-				DBUG_ASSERT(!field.col->is_virtual());
-				field.name = field.col->name(*this);
-			}
+	dict_col_t* dropped_cols = static_cast<dict_col_t*>(mem_heap_zalloc(
+		heap, n_dropped_cols * sizeof(dict_col_t)));
+	instant = new (mem_heap_alloc(heap, sizeof *instant)) dict_instant_t();
+	instant->n_dropped = n_dropped_cols;
+	instant->dropped = dropped_cols;
+	instant->field_map = field_map;
+
+	dict_col_t* col = dropped_cols;
+	for (unsigned i = 0; i < num_non_pk_fields; i++) {
+		if (field_map[i].is_dropped()) {
+			auto fixed_len = field_map[i].ind();
+			DBUG_ASSERT(fixed_len <= DICT_MAX_FIXED_COL_LEN + 1);
+			(col++)->set_dropped(field_map[i].is_not_null(),
+					     fixed_len == 1,
+					     fixed_len > 1 ? fixed_len - 1
+					     : 0);
 		}
-	} while ((index = dict_table_get_next_index(index)) != NULL);
+	}
+	DBUG_ASSERT(col == &dropped_cols[n_dropped_cols]);
+
+	UT_LIST_GET_FIRST(indexes)->reconstruct_fields();
+	return false;
 }
 
-/** Trim the instantly added columns when an insert into SYS_COLUMNS
-is rolled back during ALTER TABLE or recovery.
-@param[in]	n	number of surviving non-system columns */
-void dict_table_t::rollback_instant(unsigned n)
-{
-	ut_ad(mutex_own(&dict_sys->mutex));
-	dict_index_t* index = indexes.start;
-	DBUG_ASSERT(index->is_instant());
-	DBUG_ASSERT(index->n_def == index->n_fields);
-	DBUG_ASSERT(n_cols == n_def);
-	DBUG_ASSERT(n >= index->n_uniq);
-	DBUG_ASSERT(n_cols > n + DATA_N_SYS_COLS);
-	const unsigned n_remove = n_cols - n - DATA_N_SYS_COLS;
-
-	char* names = const_cast<char*>(dict_table_get_col_name(this, n));
-	const char* sys = names;
-	for (unsigned i = n_remove; i--; ) {
-		sys += strlen(sys) + 1;
-	}
-	static const char system[] = "DB_ROW_ID\0DB_TRX_ID\0DB_ROLL_PTR";
-	DBUG_ASSERT(!memcmp(sys, system, sizeof system));
-	for (unsigned i = index->n_fields - n_remove; i < index->n_fields;
-	     i++) {
-		if (index->fields[i].col->is_nullable()) {
-			index->n_nullable--;
-		}
-	}
-	index->n_fields -= n_remove;
-	index->n_def = index->n_fields;
-	memmove(names, sys, sizeof system);
-	memmove(cols + n, cols + n_cols - DATA_N_SYS_COLS,
-		DATA_N_SYS_COLS * sizeof *cols);
-	n_cols -= n_remove;
-	n_def = n_cols;
-	n_t_cols -= n_remove;
-	n_t_def -= n_remove;
-
-	for (unsigned i = DATA_N_SYS_COLS; i--; ) {
-		cols[n_cols - i].ind--;
-	}
-
-	if (dict_index_is_auto_gen_clust(index)) {
-		DBUG_ASSERT(index->n_uniq == 1);
-		dict_field_t* field = index->fields;
-		field->name = sys;
-		field->col = dict_table_get_sys_col(this, DATA_ROW_ID);
-		field++;
-		field->name = sys + sizeof "DB_ROW_ID";
-		field->col = dict_table_get_sys_col(this, DATA_TRX_ID);
-		field++;
-		field->name = sys + sizeof "DB_ROW_ID\0DB_TRX_ID";
-		field->col = dict_table_get_sys_col(this, DATA_ROLL_PTR);
-
-		/* Replace the DB_ROW_ID column in secondary indexes. */
-		while ((index = dict_table_get_next_index(index)) != NULL) {
-			field = &index->fields[index->n_fields - 1];
-			DBUG_ASSERT(field->col->mtype == DATA_SYS);
-			DBUG_ASSERT(field->col->prtype
-				    == DATA_NOT_NULL + DATA_TRX_ID);
-			field->col--;
-			field->name = sys;
-		}
-
-		return;
-	}
-
-	dict_field_t* field = &index->fields[index->n_uniq];
-	field->name = sys + sizeof "DB_ROW_ID";
-	field->col = dict_table_get_sys_col(this, DATA_TRX_ID);
-	field++;
-	field->name = sys + sizeof "DB_ROW_ID\0DB_TRX_ID";
-	field->col = dict_table_get_sys_col(this, DATA_ROLL_PTR);
-}
-
-
 /** Check if record in clustered index is historical row.
 @param[in]	rec	clustered row
 @param[in]	offsets	offsets
@@ -1491,7 +1271,7 @@ void dict_table_t::rollback_instant(unsigned n)
 bool
 dict_index_t::vers_history_row(
 	const rec_t*		rec,
-	const rec_offs*		offsets)
+	const ulint*		offsets)
 {
 	ut_ad(is_primary());
 
@@ -1522,8 +1302,8 @@ dict_index_t::vers_history_row(
 	bool error = false;
 	mem_heap_t* heap = NULL;
 	dict_index_t* clust_index = NULL;
-	rec_offs offsets_[REC_OFFS_NORMAL_SIZE];
-	rec_offs* offsets = offsets_;
+	ulint offsets_[REC_OFFS_NORMAL_SIZE];
+	ulint* offsets = offsets_;
 	rec_offs_init(offsets_);
 
 	mtr_t mtr;
