@@ -23,16 +23,16 @@ Created            Jonas Oreland Google
 Modified           Jan Lindström jan.lindstrom@mariadb.com
 *******************************************************/
 
-#include "fil0fil.h"
+#include "fil0crypt.h"
 #include "mtr0types.h"
 #include "mach0data.h"
-#include "page0size.h"
 #include "page0zip.h"
-#ifndef UNIV_INNOCHECKSUM
-#include "fil0crypt.h"
+#include "buf0checksum.h"
+#ifdef UNIV_INNOCHECKSUM
+# include "buf0buf.h"
+#else
 #include "srv0srv.h"
 #include "srv0start.h"
-#include "log0recv.h"
 #include "mtr0mtr.h"
 #include "mtr0log.h"
 #include "ut0ut.h"
@@ -274,16 +274,14 @@ fil_space_merge_crypt_data(
 }
 
 /** Initialize encryption parameters from a tablespace header page.
-@param[in]	page_size	page size of the tablespace
+@param[in]	zip_size	ROW_FORMAT=COMPRESSED page size, or 0
 @param[in]	page		first page of the tablespace
 @return crypt data from page 0
 @retval	NULL	if not present or not valid */
-UNIV_INTERN
-fil_space_crypt_t*
-fil_space_read_crypt_data(const page_size_t& page_size, const byte* page)
+fil_space_crypt_t* fil_space_read_crypt_data(ulint zip_size, const byte* page)
 {
 	const ulint offset = FSP_HEADER_OFFSET
-		+ fsp_header_get_encryption_offset(page_size);
+		+ fsp_header_get_encryption_offset(zip_size);
 
 	if (memcmp(page + offset, CRYPT_MAGIC, MAGIC_SZ) != 0) {
 		/* Crypt data is not stored. */
@@ -367,7 +365,8 @@ fil_space_crypt_t::fill_page0(
 {
 	const uint len = sizeof(iv);
 	const ulint offset = FSP_HEADER_OFFSET
-		+ fsp_header_get_encryption_offset(page_size_t(flags));
+		+ fsp_header_get_encryption_offset(
+			fil_space_t::zip_size(flags));
 	page0_offset = offset;
 
 	memcpy(page + offset, CRYPT_MAGIC, MAGIC_SZ);
@@ -398,7 +397,7 @@ fil_space_crypt_t::write_page0(
 	ut_ad(this == space->crypt_data);
 	const uint len = sizeof(iv);
 	const ulint offset = FSP_HEADER_OFFSET
-		+ fsp_header_get_encryption_offset(page_size_t(space->flags));
+		+ fsp_header_get_encryption_offset(space->zip_size());
 	page0_offset = offset;
 
 	/*
@@ -452,9 +451,9 @@ Parse a MLOG_FILE_WRITE_CRYPT_DATA log entry
 @param[in]	block		buffer block
 @return position on log buffer */
 UNIV_INTERN
-byte*
+const byte*
 fil_parse_write_crypt_data(
-	byte*			ptr,
+	const byte*		ptr,
 	const byte*		end_ptr,
 	dberr_t*		err)
 {
@@ -537,29 +536,27 @@ fil_parse_write_crypt_data(
 	return ptr;
 }
 
-/** Encrypt a buffer.
-@param[in,out]		crypt_data	Crypt data
-@param[in]		space		space_id
-@param[in]		offset		Page offset
-@param[in]		lsn		Log sequence number
-@param[in]		src_frame	Page to encrypt
-@param[in]		page_size	Page size
-@param[in,out]		dst_frame	Output buffer
+/** Encrypt a buffer for non full checksum.
+@param[in,out]		crypt_data		Crypt data
+@param[in]		space			space_id
+@param[in]		offset			Page offset
+@param[in]		lsn			Log sequence number
+@param[in]		src_frame		Page to encrypt
+@param[in]		zip_size		ROW_FORMAT=COMPRESSED
+						page size, or 0
+@param[in,out]		dst_frame		Output buffer
 @return encrypted buffer or NULL */
-UNIV_INTERN
-byte*
-fil_encrypt_buf(
+static byte* fil_encrypt_buf_for_non_full_checksum(
 	fil_space_crypt_t*	crypt_data,
 	ulint			space,
 	ulint			offset,
 	lsn_t			lsn,
 	const byte*		src_frame,
-	const page_size_t&	page_size,
+	ulint			zip_size,
 	byte*			dst_frame)
 {
-	uint size = uint(page_size.physical());
+	uint size = uint(zip_size ? zip_size : srv_page_size);
 	uint key_version = fil_crypt_get_latest_key_version(crypt_data);
-
 	ut_a(key_version != ENCRYPTION_KEY_VERSION_INVALID);
 
 	ulint orig_page_type = mach_read_from_2(src_frame+FIL_PAGE_TYPE);
@@ -567,21 +564,21 @@ fil_encrypt_buf(
 	uint header_len = FIL_PAGE_DATA;
 
 	if (page_compressed) {
-		header_len += (FIL_PAGE_COMPRESSED_SIZE + FIL_PAGE_COMPRESSION_METHOD_SIZE);
+		header_len += FIL_PAGE_ENCRYPT_COMP_METADATA_LEN;
 	}
 
 	/* FIL page header is not encrypted */
 	memcpy(dst_frame, src_frame, header_len);
-
-	/* Store key version */
-	mach_write_to_4(dst_frame + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION, key_version);
+	mach_write_to_4(dst_frame + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION,
+			key_version);
 
 	/* Calculate the start offset in a page */
-	uint unencrypted_bytes = header_len + FIL_PAGE_DATA_END;
-	uint srclen = size - unencrypted_bytes;
-	const byte* src = src_frame + header_len;
-	byte* dst = dst_frame + header_len;
-	uint32 dstlen = 0;
+	uint		unencrypted_bytes = header_len + FIL_PAGE_DATA_END;
+	uint		srclen = size - unencrypted_bytes;
+	const byte*	src = src_frame + header_len;
+	byte*		dst = dst_frame + header_len;
+	uint32		dstlen = 0;
+	ib_uint32_t	checksum = 0;
 
 	if (page_compressed) {
 		srclen = mach_read_from_2(src_frame + FIL_PAGE_DATA);
@@ -599,54 +596,151 @@ fil_encrypt_buf(
 	to sector boundary is written. */
 	if (!page_compressed) {
 		/* FIL page trailer is also not encrypted */
-		memcpy(dst_frame + page_size.physical() - FIL_PAGE_DATA_END,
-			src_frame + page_size.physical() - FIL_PAGE_DATA_END,
+		memcpy(dst_frame + size - FIL_PAGE_DATA_END,
+			src_frame + size - FIL_PAGE_DATA_END,
 			FIL_PAGE_DATA_END);
 	} else {
 		/* Clean up rest of buffer */
 		memset(dst_frame+header_len+srclen, 0,
-		       page_size.physical() - (header_len + srclen));
+		       size - (header_len + srclen));
 	}
 
-	/* handle post encryption checksum */
-	ib_uint32_t checksum = 0;
+	checksum = fil_crypt_calculate_checksum(zip_size, dst_frame);
 
-	checksum = fil_crypt_calculate_checksum(page_size, dst_frame);
+	/* store the post-encryption checksum after the key-version */
+	mach_write_to_4(dst_frame + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION + 4,
+			checksum);
 
-	// store the post-encryption checksum after the key-version
-	mach_write_to_4(dst_frame + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION + 4, checksum);
-
-	ut_ad(fil_space_verify_crypt_checksum(dst_frame, page_size));
+	ut_ad(fil_space_verify_crypt_checksum(dst_frame, zip_size));
 
 	srv_stats.pages_encrypted.inc();
 
 	return dst_frame;
 }
 
+/** Encrypt a buffer for full checksum format.
+@param[in,out]		crypt_data		Crypt data
+@param[in]		space			space_id
+@param[in]		offset			Page offset
+@param[in]		lsn			Log sequence number
+@param[in]		src_frame		Page to encrypt
+@param[in,out]		dst_frame		Output buffer
+@return encrypted buffer or NULL */
+static byte* fil_encrypt_buf_for_full_crc32(
+	fil_space_crypt_t*	crypt_data,
+	ulint			space,
+	ulint			offset,
+	lsn_t			lsn,
+	const byte*		src_frame,
+	byte*			dst_frame)
+{
+	uint key_version = fil_crypt_get_latest_key_version(crypt_data);
+	ut_d(bool corrupted = false);
+	const uint size = buf_page_full_crc32_size(src_frame, NULL,
+#ifdef UNIV_DEBUG
+						   &corrupted
+#else
+						   NULL
+#endif
+						   );
+	ut_ad(!corrupted);
+	uint srclen = size - (FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION
+			      + FIL_PAGE_FCRC32_CHECKSUM);
+	const byte* src = src_frame + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION;
+	byte* dst = dst_frame + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION;
+	uint dstlen = 0;
+
+	ut_a(key_version != ENCRYPTION_KEY_VERSION_INVALID);
+
+	/* Till FIL_PAGE_LSN, page is not encrypted */
+	memcpy(dst_frame, src_frame, FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION);
+
+	/* Write key version to the page. */
+	mach_write_to_4(dst_frame + FIL_PAGE_FCRC32_KEY_VERSION, key_version);
+
+	int rc = encryption_scheme_encrypt(src, srclen, dst, &dstlen,
+					   crypt_data, key_version,
+					   uint(space), uint(offset), lsn);
+	ut_a(rc == MY_AES_OK);
+	ut_a(dstlen == srclen);
+
+	const ulint payload = size - FIL_PAGE_FCRC32_CHECKSUM;
+	mach_write_to_4(dst_frame + payload, ut_crc32(dst_frame, payload));
+	/* Clean the rest of the buffer. FIXME: Punch holes when writing! */
+	memset(dst_frame + (payload + 4), 0, srv_page_size - (payload + 4));
+
+	srv_stats.pages_encrypted.inc();
+
+	return dst_frame;
+}
+
+/** Encrypt a buffer.
+@param[in,out]		crypt_data		Crypt data
+@param[in]		space			space_id
+@param[in]		offset			Page offset
+@param[in]		src_frame		Page to encrypt
+@param[in]		zip_size		ROW_FORMAT=COMPRESSED
+						page size, or 0
+@param[in,out]		dst_frame		Output buffer
+@param[in]		use_full_checksum	full crc32 algo is used
+@return encrypted buffer or NULL */
+byte* fil_encrypt_buf(
+	fil_space_crypt_t*	crypt_data,
+	ulint			space,
+	ulint			offset,
+	const byte*		src_frame,
+	ulint			zip_size,
+	byte*			dst_frame,
+	bool			use_full_checksum)
+{
+	const lsn_t lsn = mach_read_from_8(src_frame + FIL_PAGE_LSN);
+
+	if (use_full_checksum) {
+		ut_ad(!zip_size);
+		return fil_encrypt_buf_for_full_crc32(
+			crypt_data, space, offset,
+			lsn, src_frame, dst_frame);
+	}
+
+	return fil_encrypt_buf_for_non_full_checksum(
+		crypt_data, space, offset, lsn,
+		src_frame, zip_size, dst_frame);
+}
+
+/** Check whether these page types are allowed to encrypt.
+@param[in]	space		tablespace object
+@param[in]	src_frame	source page
+@return true if it is valid page type */
+static bool fil_space_encrypt_valid_page_type(
+	const fil_space_t*	space,
+	byte*			src_frame)
+{
+	switch (mach_read_from_2(src_frame+FIL_PAGE_TYPE)) {
+	case FIL_PAGE_RTREE:
+		return space->full_crc32();
+	case FIL_PAGE_TYPE_FSP_HDR:
+	case FIL_PAGE_TYPE_XDES:
+		return false;
+	}
+
+	return true;
+}
+
 /******************************************************************
 Encrypt a page
 
 @param[in]		space		Tablespace
 @param[in]		offset		Page offset
-@param[in]		lsn		Log sequence number
 @param[in]		src_frame	Page to encrypt
 @param[in,out]		dst_frame	Output buffer
 @return encrypted buffer or NULL */
-UNIV_INTERN
-byte*
-fil_space_encrypt(
+byte* fil_space_encrypt(
 	const fil_space_t*	space,
 	ulint			offset,
-	lsn_t			lsn,
 	byte*			src_frame,
 	byte*			dst_frame)
 {
-	switch (mach_read_from_2(src_frame+FIL_PAGE_TYPE)) {
-	case FIL_PAGE_TYPE_FSP_HDR:
-	case FIL_PAGE_TYPE_XDES:
-	case FIL_PAGE_RTREE:
-		/* File space header, extent descriptor or spatial index
-		are not encrypted. */
+	if (!fil_space_encrypt_valid_page_type(space, src_frame)) {
 		return src_frame;
 	}
 
@@ -655,73 +749,167 @@ fil_space_encrypt(
 	}
 
 	fil_space_crypt_t* crypt_data = space->crypt_data;
-	const page_size_t	page_size(space->flags);
+	const ulint zip_size = space->zip_size();
 	ut_ad(space->pending_io());
-	byte* tmp = fil_encrypt_buf(crypt_data, space->id, offset, lsn,
-				    src_frame, page_size, dst_frame);
+
+	const bool full_crc32 = space->full_crc32();
+
+	byte* tmp = fil_encrypt_buf(crypt_data, space->id, offset,
+				    src_frame, zip_size, dst_frame,
+				    full_crc32);
 
 #ifdef UNIV_DEBUG
 	if (tmp) {
 		/* Verify that encrypted buffer is not corrupted */
 		dberr_t err = DB_SUCCESS;
 		byte* src = src_frame;
-		bool page_compressed_encrypted = (mach_read_from_2(tmp+FIL_PAGE_TYPE) == FIL_PAGE_PAGE_COMPRESSED_ENCRYPTED);
-		byte uncomp_mem[UNIV_PAGE_SIZE_MAX];
 		byte tmp_mem[UNIV_PAGE_SIZE_MAX];
 
-		if (page_compressed_encrypted) {
-			memcpy(uncomp_mem, src, srv_page_size);
-			ulint unzipped1 = fil_page_decompress(
-				tmp_mem, uncomp_mem);
-			ut_ad(unzipped1);
-			if (unzipped1 != srv_page_size) {
-				src = uncomp_mem;
+		if (full_crc32) {
+			bool compressed = false, corrupted = false;
+			uint size = buf_page_full_crc32_size(
+				tmp, &compressed, &corrupted);
+			ut_ad(!corrupted);
+			ut_ad(!compressed == (size == srv_page_size));
+			ut_ad(fil_space_decrypt(space->id, crypt_data, tmp_mem,
+						size, space->flags, tmp,
+						&err));
+			ut_ad(err == DB_SUCCESS);
+			memcpy(tmp_mem, src, FIL_PAGE_OFFSET);
+			ut_ad(!memcmp(src, tmp_mem,
+				      (size - FIL_PAGE_FCRC32_CHECKSUM)));
+		} else {
+			bool page_compressed_encrypted =
+				(mach_read_from_2(tmp+FIL_PAGE_TYPE)
+				 == FIL_PAGE_PAGE_COMPRESSED_ENCRYPTED);
+			byte uncomp_mem[UNIV_PAGE_SIZE_MAX];
+
+			if (page_compressed_encrypted) {
+				memcpy(uncomp_mem, src, srv_page_size);
+				ulint unzipped1 = fil_page_decompress(
+					tmp_mem, uncomp_mem, space->flags);
+				ut_ad(unzipped1);
+				if (unzipped1 != srv_page_size) {
+					src = uncomp_mem;
+				}
 			}
+
+			ut_ad(!buf_page_is_corrupted(true, src, space->flags));
+
+			ut_ad(fil_space_decrypt(space->id, crypt_data, tmp_mem,
+						space->physical_size(),
+						space->flags, tmp, &err));
+			ut_ad(err == DB_SUCCESS);
+
+			if (page_compressed_encrypted) {
+				memcpy(tmp_mem, uncomp_mem, srv_page_size);
+				ulint unzipped2 = fil_page_decompress(
+					uncomp_mem, tmp_mem, space->flags);
+				ut_ad(unzipped2);
+			}
+
+			memcpy(tmp_mem + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION,
+			       src + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION, 8);
+			ut_ad(!memcmp(src, tmp_mem, space->physical_size()));
 		}
-
-		ut_ad(!buf_page_is_corrupted(true, src, page_size, space));
-		ut_ad(fil_space_decrypt(crypt_data, tmp_mem, page_size, tmp,
-					&err));
-		ut_ad(err == DB_SUCCESS);
-
-		/* Need to decompress the page if it was also compressed */
-		if (page_compressed_encrypted) {
-			byte buf[UNIV_PAGE_SIZE_MAX];
-			memcpy(buf, tmp_mem, srv_page_size);
-			ulint unzipped2 = fil_page_decompress(tmp_mem, buf);
-			ut_ad(unzipped2);
-		}
-
-		memcpy(tmp_mem + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION,
-		       src + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION, 8);
-		ut_ad(!memcmp(src, tmp_mem, page_size.physical()));
 	}
 #endif /* UNIV_DEBUG */
 
 	return tmp;
 }
 
-/** Decrypt a page.
+/** Decrypt a page for full checksum format.
+@param[in]	space			space id
 @param[in]	crypt_data		crypt_data
 @param[in]	tmp_frame		Temporary buffer
-@param[in]	page_size		Page size
 @param[in,out]	src_frame		Page to decrypt
 @param[out]	err			DB_SUCCESS or DB_DECRYPTION_FAILED
 @return true if page decrypted, false if not.*/
-UNIV_INTERN
-bool
-fil_space_decrypt(
+static bool fil_space_decrypt_full_crc32(
+	ulint			space,
 	fil_space_crypt_t*	crypt_data,
 	byte*			tmp_frame,
-	const page_size_t&	page_size,
+	byte*			src_frame,
+	dberr_t*		err)
+{
+	uint key_version = mach_read_from_4(
+		src_frame + FIL_PAGE_FCRC32_KEY_VERSION);
+	lsn_t lsn = mach_read_from_8(src_frame + FIL_PAGE_LSN);
+	uint offset = mach_read_from_4(src_frame + FIL_PAGE_OFFSET);
+	*err = DB_SUCCESS;
+
+	if (key_version == ENCRYPTION_KEY_NOT_ENCRYPTED) {
+		return false;
+	}
+
+	ut_ad(crypt_data);
+	ut_ad(crypt_data->is_encrypted());
+
+	memcpy(tmp_frame, src_frame, FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION);
+
+	/* Calculate the offset where decryption starts */
+	const byte* src = src_frame + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION;
+	byte* dst = tmp_frame + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION;
+	uint dstlen = 0;
+	bool corrupted = false;
+	uint size = buf_page_full_crc32_size(src_frame, NULL, &corrupted);
+	if (UNIV_UNLIKELY(corrupted)) {
+fail:
+		*err = DB_DECRYPTION_FAILED;
+		return false;
+	}
+
+	uint srclen = size - (FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION
+			      + FIL_PAGE_FCRC32_CHECKSUM);
+
+	int rc = encryption_scheme_decrypt(src, srclen, dst, &dstlen,
+					   crypt_data, key_version,
+					   (uint) space, offset, lsn);
+
+	if (rc != MY_AES_OK || dstlen != srclen) {
+		if (rc == -1) {
+			goto fail;
+		}
+
+		ib::fatal() << "Unable to decrypt data-block "
+			    << " src: " << src << "srclen: "
+			    << srclen << " buf: " << dst << "buflen: "
+			    << dstlen << " return-code: " << rc
+			    << " Can't continue!";
+	}
+
+	/* Copy only checksum part in the trailer */
+	memcpy(tmp_frame + srv_page_size - FIL_PAGE_FCRC32_CHECKSUM,
+	       src_frame + srv_page_size - FIL_PAGE_FCRC32_CHECKSUM,
+	       FIL_PAGE_FCRC32_CHECKSUM);
+
+	srv_stats.pages_decrypted.inc();
+
+	return true; /* page was decrypted */
+}
+
+/** Decrypt a page for non full checksum format.
+@param[in]	crypt_data		crypt_data
+@param[in]	tmp_frame		Temporary buffer
+@param[in]	physical_size		page size
+@param[in,out]	src_frame		Page to decrypt
+@param[out]	err			DB_SUCCESS or DB_DECRYPTION_FAILED
+@return true if page decrypted, false if not.*/
+static bool fil_space_decrypt_for_non_full_checksum(
+	fil_space_crypt_t*	crypt_data,
+	byte*			tmp_frame,
+	ulint			physical_size,
 	byte*			src_frame,
 	dberr_t*		err)
 {
 	ulint page_type = mach_read_from_2(src_frame+FIL_PAGE_TYPE);
-	uint key_version = mach_read_from_4(src_frame + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION);
-	bool page_compressed = (page_type == FIL_PAGE_PAGE_COMPRESSED_ENCRYPTED);
+	uint key_version = mach_read_from_4(
+			src_frame + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION);
+	bool page_compressed = (page_type
+				== FIL_PAGE_PAGE_COMPRESSED_ENCRYPTED);
 	uint offset = mach_read_from_4(src_frame + FIL_PAGE_OFFSET);
-	uint space = mach_read_from_4(src_frame + FIL_PAGE_ARCH_LOG_NO_OR_SPACE_ID);
+	uint space = mach_read_from_4(
+			src_frame + FIL_PAGE_ARCH_LOG_NO_OR_SPACE_ID);
 	ib_uint64_t lsn = mach_read_from_8(src_frame + FIL_PAGE_LSN);
 
 	*err = DB_SUCCESS;
@@ -736,7 +924,7 @@ fil_space_decrypt(
 	uint header_len = FIL_PAGE_DATA;
 
 	if (page_compressed) {
-		header_len += (FIL_PAGE_COMPRESSED_SIZE + FIL_PAGE_COMPRESSION_METHOD_SIZE);
+		header_len += FIL_PAGE_ENCRYPT_COMP_METADATA_LEN;
 	}
 
 	/* Copy FIL page header, it is not encrypted */
@@ -746,8 +934,7 @@ fil_space_decrypt(
 	const byte* src = src_frame + header_len;
 	byte* dst = tmp_frame + header_len;
 	uint32 dstlen = 0;
-	uint srclen = uint(page_size.physical())
-		- header_len - FIL_PAGE_DATA_END;
+	uint srclen = uint(physical_size) - header_len - FIL_PAGE_DATA_END;
 
 	if (page_compressed) {
 		srclen = mach_read_from_2(src_frame + FIL_PAGE_DATA);
@@ -777,8 +964,8 @@ fil_space_decrypt(
 	to sector boundary is written. */
 	if (!page_compressed) {
 		/* Copy FIL trailer */
-		memcpy(tmp_frame + page_size.physical() - FIL_PAGE_DATA_END,
-		       src_frame + page_size.physical() - FIL_PAGE_DATA_END,
+		memcpy(tmp_frame + physical_size - FIL_PAGE_DATA_END,
+		       src_frame + physical_size - FIL_PAGE_DATA_END,
 		       FIL_PAGE_DATA_END);
 	}
 
@@ -787,12 +974,41 @@ fil_space_decrypt(
 	return true; /* page was decrypted */
 }
 
+/** Decrypt a page.
+@param[in]	space_id		tablespace id
+@param[in]	crypt_data		crypt_data
+@param[in]	tmp_frame		Temporary buffer
+@param[in]	physical_size		page size
+@param[in]	fsp_flags		Tablespace flags
+@param[in,out]	src_frame		Page to decrypt
+@param[out]	err			DB_SUCCESS or DB_DECRYPTION_FAILED
+@return true if page decrypted, false if not.*/
+UNIV_INTERN
+bool
+fil_space_decrypt(
+	ulint			space_id,
+	fil_space_crypt_t*	crypt_data,
+	byte*			tmp_frame,
+	ulint			physical_size,
+	ulint			fsp_flags,
+	byte*			src_frame,
+	dberr_t*		err)
+{
+	if (fil_space_t::full_crc32(fsp_flags)) {
+		return fil_space_decrypt_full_crc32(
+			space_id, crypt_data, tmp_frame, src_frame, err);
+	}
+
+	return fil_space_decrypt_for_non_full_checksum(crypt_data, tmp_frame,
+						       physical_size, src_frame,
+						       err);
+}
+
 /**
 Decrypt a page.
 @param[in]	space			Tablespace
 @param[in]	tmp_frame		Temporary buffer used for decrypting
 @param[in,out]	src_frame		Page to decrypt
-@param[out]	decrypted		true if page was decrypted
 @return decrypted page, or original not encrypted page if decryption is
 not needed.*/
 UNIV_INTERN
@@ -800,26 +1016,25 @@ byte*
 fil_space_decrypt(
 	const fil_space_t* space,
 	byte*		tmp_frame,
-	byte*		src_frame,
-	bool*		decrypted)
+	byte*		src_frame)
 {
 	dberr_t err = DB_SUCCESS;
 	byte* res = NULL;
-	const page_size_t page_size(space->flags);
-	*decrypted = false;
+	const ulint physical_size = space->physical_size();
 
 	ut_ad(space->crypt_data != NULL && space->crypt_data->is_encrypted());
 	ut_ad(space->pending_io());
 
-	bool encrypted = fil_space_decrypt(space->crypt_data, tmp_frame,
-					   page_size, src_frame, &err);
+	bool encrypted = fil_space_decrypt(space->id, space->crypt_data,
+					   tmp_frame, physical_size,
+					   space->flags,
+					   src_frame, &err);
 
 	if (err == DB_SUCCESS) {
 		if (encrypted) {
-			*decrypted = true;
 			/* Copy the decrypted page back to page buffer, not
 			really any other options. */
-			memcpy(src_frame, tmp_frame, page_size.physical());
+			memcpy(src_frame, tmp_frame, physical_size);
 		}
 
 		res = src_frame;
@@ -828,21 +1043,18 @@ fil_space_decrypt(
 	return res;
 }
 
-/******************************************************************
+/**
 Calculate post encryption checksum
-@param[in]	page_size	page size
+@param[in]	zip_size	ROW_FORMAT=COMPRESSED page size, or 0
 @param[in]	dst_frame	Block where checksum is calculated
 @return page checksum
 not needed. */
-UNIV_INTERN
 uint32_t
-fil_crypt_calculate_checksum(
-	const page_size_t&	page_size,
-	const byte*		dst_frame)
+fil_crypt_calculate_checksum(ulint zip_size, const byte* dst_frame)
 {
 	/* For encrypted tables we use only crc32 and strict_crc32 */
-	return page_size.is_compressed()
-		? page_zip_calc_checksum(dst_frame, page_size.physical(),
+	return zip_size
+		? page_zip_calc_checksum(dst_frame, zip_size,
 					 SRV_CHECKSUM_ALGORITHM_CRC32)
 		: buf_calc_page_crc32(dst_frame);
 }
@@ -954,15 +1166,15 @@ fil_crypt_read_crypt_data(fil_space_t* space)
 		return;
 	}
 
-	const page_size_t page_size(space->flags);
+	const ulint zip_size = space->zip_size();
 	mtr_t	mtr;
 	mtr.start();
 	if (buf_block_t* block = buf_page_get(page_id_t(space->id, 0),
-					      page_size, RW_S_LATCH, &mtr)) {
+					      zip_size, RW_S_LATCH, &mtr)) {
 		mutex_enter(&fil_system.mutex);
 		if (!space->crypt_data) {
 			space->crypt_data = fil_space_read_crypt_data(
-				page_size, block->frame);
+				zip_size, block->frame);
 		}
 		mutex_exit(&fil_system.mutex);
 	}
@@ -1033,7 +1245,7 @@ static bool fil_crypt_start_encrypting_space(fil_space_t* space)
 		/* 2 - get page 0 */
 		dberr_t err = DB_SUCCESS;
 		buf_block_t* block = buf_page_get_gen(
-			page_id_t(space->id, 0), page_size_t(space->flags),
+			page_id_t(space->id, 0), space->zip_size(),
 			RW_X_LATCH, NULL, BUF_GET,
 			__FILE__, __LINE__,
 			&mtr, &err);
@@ -1618,7 +1830,7 @@ fil_crypt_get_page_throttle_func(
 	unsigned		line)
 {
 	fil_space_t* space = state->space;
-	const page_size_t page_size = page_size_t(space->flags);
+	const ulint zip_size = space->zip_size();
 	const page_id_t page_id(space->id, offset);
 	ut_ad(space->referenced());
 
@@ -1629,7 +1841,7 @@ fil_crypt_get_page_throttle_func(
 	}
 
 	dberr_t err = DB_SUCCESS;
-	buf_block_t* block = buf_page_get_gen(page_id, page_size, RW_X_LATCH,
+	buf_block_t* block = buf_page_get_gen(page_id, zip_size, RW_X_LATCH,
 					      NULL,
 					      BUF_PEEK_IF_IN_POOL, file, line,
 					      mtr, &err);
@@ -1646,7 +1858,7 @@ fil_crypt_get_page_throttle_func(
 	state->crypt_stat.pages_read_from_disk++;
 
 	const ulonglong start = my_interval_timer();
-	block = buf_page_get_gen(page_id, page_size,
+	block = buf_page_get_gen(page_id, zip_size,
 				 RW_X_LATCH,
 				 NULL, BUF_GET_POSSIBLY_FREED,
 				file, line, mtr, &err);
@@ -1770,8 +1982,6 @@ fil_crypt_rotate_page(
 		return;
 	}
 
-	ut_d(const bool was_free = fseg_page_is_free(space, (uint32_t)offset));
-
 	mtr_t mtr;
 	mtr.start();
 	if (buf_block_t* block = fil_crypt_get_page_throttle(state,
@@ -1779,16 +1989,16 @@ fil_crypt_rotate_page(
 							     &sleeptime_ms)) {
 		bool modified = false;
 		int needs_scrubbing = BTR_SCRUB_SKIP_PAGE;
-		lsn_t block_lsn = block->page.newest_modification;
 		byte* frame = buf_block_get_frame(block);
-		uint kv =  mach_read_from_4(frame+FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION);
+		const lsn_t block_lsn = mach_read_from_8(FIL_PAGE_LSN + frame);
+		uint kv = buf_page_get_key_version(frame, space->flags);
 
 		if (space->is_stopping()) {
 			/* The tablespace is closing (in DROP TABLE or
 			TRUNCATE TABLE or similar): avoid further access */
-		} else if (!*reinterpret_cast<uint32_t*>(FIL_PAGE_OFFSET
-							 + frame)) {
-			/* It looks like this page was never
+		} else if (!kv && !*reinterpret_cast<uint16_t*>
+			   (&frame[FIL_PAGE_TYPE])) {
+			/* It looks like this page is not
 			allocated. Because key rotation is accessing
 			pages in a pattern that is unlike the normal
 			B-tree and undo log access pattern, we cannot
@@ -1798,9 +2008,20 @@ fil_crypt_rotate_page(
 			tablespace latch before acquiring block->lock,
 			then the fseg_page_is_free() information
 			could be stale already. */
-			ut_ad(was_free);
-			ut_ad(kv == 0);
-			ut_ad(page_get_space_id(frame) == 0);
+
+			/* If the data file was originally created
+			before MariaDB 10.0 or MySQL 5.6, some
+			allocated data pages could carry 0 in
+			FIL_PAGE_TYPE. The FIL_PAGE_TYPE on those
+			pages will be updated in
+			buf_flush_init_for_writing() when the page
+			is modified the next time.
+
+			Also, when the doublewrite buffer pages are
+			allocated on bootstrap in a non-debug build,
+			some dummy pages will be allocated, with 0 in
+			the FIL_PAGE_TYPE. Those pages should be
+			skipped from key rotation forever. */
 		} else if (fil_crypt_needs_rotation(
 				crypt_data,
 				kv,
@@ -2014,7 +2235,7 @@ fil_crypt_flush_space(
 	dberr_t err;
 
 	if (buf_block_t* block = buf_page_get_gen(
-		    page_id_t(space->id, 0), page_size_t(space->flags),
+		    page_id_t(space->id, 0), space->zip_size(),
 		    RW_X_LATCH, NULL, BUF_GET,
 		    __FILE__, __LINE__, &mtr, &err)) {
 		mtr.set_named_space(space);
@@ -2296,8 +2517,8 @@ static void fil_crypt_rotation_list_fill()
 			release fil_system.mutex. */
 			space->n_pending_ops++;
 #ifndef DBUG_OFF
-			fil_space_t* s= fil_system.read_page0(
-				space->id);
+			ut_d(const fil_space_t* s=)
+			        fil_system.read_page0(space->id);
 			ut_ad(!s || s == space);
 #endif
 			space->n_pending_ops--;
@@ -2591,10 +2812,9 @@ calculated checksum as if it does page could be valid unencrypted,
 encrypted, or corrupted.
 
 @param[in,out]	page		page frame (checksum is temporarily modified)
-@param[in]	page_size	page size
-@return whether the encrypted page is OK */
-bool
-fil_space_verify_crypt_checksum(const byte* page, const page_size_t& page_size)
+@param[in]	zip_size	ROW_FORMAT=COMPRESSED page size, or 0
+@return true if page is encrypted AND OK, false otherwise */
+bool fil_space_verify_crypt_checksum(const byte* page, ulint zip_size)
 {
 	ut_ad(mach_read_from_4(page + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION));
 
@@ -2614,24 +2834,14 @@ fil_space_verify_crypt_checksum(const byte* page, const page_size_t& page_size)
 	page is not corrupted. */
 
 	switch (srv_checksum_algorithm_t(srv_checksum_algorithm)) {
+	case SRV_CHECKSUM_ALGORITHM_STRICT_FULL_CRC32:
 	case SRV_CHECKSUM_ALGORITHM_STRICT_CRC32:
-		if (page_size.is_compressed()) {
+		if (zip_size) {
 			return checksum == page_zip_calc_checksum(
-				page, page_size.physical(),
-				SRV_CHECKSUM_ALGORITHM_CRC32)
-#ifdef INNODB_BUG_ENDIAN_CRC32
-				|| checksum == page_zip_calc_checksum(
-					page, page_size.physical(),
-					SRV_CHECKSUM_ALGORITHM_CRC32, true)
-#endif
-				;
+				page, zip_size, SRV_CHECKSUM_ALGORITHM_CRC32);
 		}
 
-		return checksum == buf_calc_page_crc32(page)
-#ifdef INNODB_BUG_ENDIAN_CRC32
-			|| checksum == buf_calc_page_crc32(page, true)
-#endif
-			;
+		return checksum == buf_calc_page_crc32(page);
 	case SRV_CHECKSUM_ALGORITHM_STRICT_NONE:
 		/* Starting with MariaDB 10.1.25, 10.2.7, 10.3.1,
 		due to MDEV-12114, fil_crypt_calculate_checksum()
@@ -2646,27 +2856,20 @@ fil_space_verify_crypt_checksum(const byte* page, const page_size_t& page_size)
 		Due to this, we must treat "strict_innodb" as "innodb". */
 	case SRV_CHECKSUM_ALGORITHM_INNODB:
 	case SRV_CHECKSUM_ALGORITHM_CRC32:
+	case SRV_CHECKSUM_ALGORITHM_FULL_CRC32:
 		if (checksum == BUF_NO_CHECKSUM_MAGIC) {
 			return true;
 		}
-		if (page_size.is_compressed()) {
+		if (zip_size) {
 			return checksum == page_zip_calc_checksum(
-				page, page_size.physical(),
+				page, zip_size,
 				SRV_CHECKSUM_ALGORITHM_CRC32)
-#ifdef INNODB_BUG_ENDIAN_CRC32
 				|| checksum == page_zip_calc_checksum(
-					page, page_size.physical(),
-					SRV_CHECKSUM_ALGORITHM_CRC32, true)
-#endif
-				|| checksum == page_zip_calc_checksum(
-					page, page_size.physical(),
+					page, zip_size,
 					SRV_CHECKSUM_ALGORITHM_INNODB);
 		}
 
 		return checksum == buf_calc_page_crc32(page)
-#ifdef INNODB_BUG_ENDIAN_CRC32
-			|| checksum == buf_calc_page_crc32(page, true)
-#endif
 			|| checksum == buf_calc_page_new_checksum(page);
 	}
 
