@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 2007, 2018, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2016, 2020, MariaDB Corporation.
+Copyright (c) 2016, 2019, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -39,13 +39,16 @@ Completed 2011/7/10 Sunny and Jimmy Yang
 
 /** The FTS optimize thread's work queue. */
 ib_wqueue_t* fts_optimize_wq;
+static void fts_optimize_callback(void *);
+static void timer_callback(void*);
+static tpool::timer* timer;
+static tpool::task_group task_group(1);
+static tpool::task task(fts_optimize_callback,0, &task_group);
+
 
 /** The FTS vector to store fts_slot_t */
 static ib_vector_t*  fts_slots;
 
-/** Time to wait for a message. */
-static const ulint FTS_QUEUE_WAIT_IN_USECS = 5000000;
-
 /** Default optimize interval in secs. */
 static const ulint FTS_OPTIMIZE_INTERVAL_IN_SECS = 300;
 
@@ -525,7 +528,7 @@ fts_index_fetch_nodes(
 	for (;;) {
 		error = fts_eval_sql(trx, *graph);
 
-		if (UNIV_LIKELY(error == DB_SUCCESS)) {
+		if (error == DB_SUCCESS) {
 			fts_sql_commit(trx);
 
 			break;				/* Exit the loop. */
@@ -538,7 +541,7 @@ fts_index_fetch_nodes(
 
 				trx->error_state = DB_SUCCESS;
 			} else {
-				ib::error() << "(" << error
+				ib::error() << "(" << ut_strerr(error)
 					<< ") while reading FTS index.";
 
 				break;			/* Exit the loop. */
@@ -860,7 +863,7 @@ fts_index_fetch_words(
 				error = fts_eval_sql(optim->trx, graph);
 			}
 
-			if (UNIV_LIKELY(error == DB_SUCCESS)) {
+			if (error == DB_SUCCESS) {
 				//FIXME fts_sql_commit(optim->trx);
 				break;
 			} else {
@@ -877,7 +880,7 @@ fts_index_fetch_words(
 
 					optim->trx->error_state = DB_SUCCESS;
 				} else {
-					ib::error() << "(" << error
+					ib::error() << "(" << ut_strerr(error)
 						<< ") while reading document.";
 
 					break;	/* Exit the loop. */
@@ -920,7 +923,7 @@ fts_fetch_doc_ids(
 	int		i = 0;
 	sel_node_t*	sel_node = static_cast<sel_node_t*>(row);
 	fts_doc_ids_t*	fts_doc_ids = static_cast<fts_doc_ids_t*>(user_arg);
-	doc_id_t*	update = static_cast<doc_id_t*>(
+	fts_update_t*	update = static_cast<fts_update_t*>(
 		ib_vector_push(fts_doc_ids->doc_ids, NULL));
 
 	for (exp = sel_node->select_list;
@@ -936,7 +939,8 @@ fts_fetch_doc_ids(
 		/* Note: The column numbers below must match the SELECT. */
 		switch (i) {
 		case 0: /* DOC_ID */
-			*update = fts_read_doc_id(
+			update->fts_indexes = NULL;
+			update->doc_id = fts_read_doc_id(
 				static_cast<byte*>(data));
 			break;
 
@@ -1004,7 +1008,7 @@ fts_table_fetch_doc_ids(
 	mutex_exit(&dict_sys.mutex);
 
 	if (error == DB_SUCCESS) {
-		ib_vector_sort(doc_ids->doc_ids, fts_doc_id_cmp);
+		ib_vector_sort(doc_ids->doc_ids, fts_update_doc_id_cmp);
 	}
 
 	if (alloc_bk_trx) {
@@ -1021,7 +1025,7 @@ Do a binary search for a doc id in the array
 int
 fts_bsearch(
 /*========*/
-	doc_id_t*	array,	/*!< in: array to sort */
+	fts_update_t*	array,	/*!< in: array to sort */
 	int		lower,	/*!< in: the array lower bound */
 	int		upper,	/*!< in: the array upper bound */
 	doc_id_t	doc_id)	/*!< in: the doc id to search for */
@@ -1035,9 +1039,9 @@ fts_bsearch(
 		while (lower < upper) {
 			int	i = (lower + upper) >> 1;
 
-			if (doc_id > array[i]) {
+			if (doc_id > array[i].doc_id) {
 				lower = i + 1;
-			} else if (doc_id < array[i]) {
+			} else if (doc_id < array[i].doc_id) {
 				upper = i - 1;
 			} else {
 				return(i); /* Found. */
@@ -1046,7 +1050,7 @@ fts_bsearch(
 	}
 
 	if (lower == upper && lower < orig_size) {
-		if (doc_id == array[lower]) {
+		if (doc_id == array[lower].doc_id) {
 			return(lower);
 		} else if (lower == 0) {
 			return(-1);
@@ -1073,7 +1077,7 @@ fts_optimize_lookup(
 {
 	int		pos;
 	int		upper = static_cast<int>(ib_vector_size(doc_ids));
-	doc_id_t*	array = (doc_id_t*) doc_ids->data;
+	fts_update_t*	array = (fts_update_t*) doc_ids->data;
 
 	pos = fts_bsearch(array, static_cast<int>(lower), upper, first_doc_id);
 
@@ -1086,10 +1090,10 @@ fts_optimize_lookup(
 		/* If i is 1, it could be first_doc_id is less than
 		either the first or second array item, do a
 		double check */
-		if (i == 1 && array[0] <= last_doc_id
-		    && first_doc_id < array[0]) {
+		if (i == 1 && array[0].doc_id <= last_doc_id
+		    && first_doc_id < array[0].doc_id) {
 			pos = 0;
-		} else if (i < upper && array[i] <= last_doc_id) {
+		} else if (i < upper && array[i].doc_id <= last_doc_id) {
 
 			/* Check if the "next" doc id is within the
 			first & last doc id of the node. */
@@ -1228,12 +1232,12 @@ test_again:
 		delta for decoding the entries following this document's
 		entries. */
 		if (*del_pos >= 0 && *del_pos < (int) ib_vector_size(del_vec)) {
-			doc_id_t*	update;
+			fts_update_t*	update;
 
-			update = (doc_id_t*) ib_vector_get(
+			update = (fts_update_t*) ib_vector_get(
 				del_vec, ulint(*del_pos));
 
-			del_doc_id = *update;
+			del_doc_id = update->doc_id;
 		}
 
 		if (enc->src_ilist_ptr == src_node->ilist && doc_id == 0) {
@@ -1356,6 +1360,12 @@ fts_optimize_word(
 	enc.src_last_doc_id = 0;
 	enc.src_ilist_ptr = NULL;
 
+	if (fts_enable_diag_print) {
+		word->text.f_str[word->text.f_len] = 0;
+		ib::info() << "FTS_OPTIMIZE: optimize \"" << word->text.f_str
+			<< "\"";
+	}
+
 	while (i < size) {
 		ulint		copied;
 		fts_node_t*	src_node;
@@ -1433,6 +1443,11 @@ fts_optimize_write_word(
 
 	ut_ad(fts_table->charset);
 
+	if (fts_enable_diag_print) {
+		ib::info() << "FTS_OPTIMIZE: processed \"" << word->f_str
+			<< "\"";
+	}
+
 	pars_info_bind_varchar_literal(
 		info, "word", word->f_str, word->f_len);
 
@@ -1450,8 +1465,8 @@ fts_optimize_write_word(
 
 	error = fts_eval_sql(trx, graph);
 
-	if (UNIV_UNLIKELY(error != DB_SUCCESS)) {
-		ib::error() << "(" << error << ") during optimize,"
+	if (error != DB_SUCCESS) {
+		ib::error() << "(" << ut_strerr(error) << ") during optimize,"
 			" when deleting a word from the FTS index.";
 	}
 
@@ -1474,8 +1489,8 @@ fts_optimize_write_word(
 			error = fts_write_node(
 				trx, &graph, fts_table, word, node);
 
-			if (UNIV_UNLIKELY(error != DB_SUCCESS)) {
-				ib::error() << "(" << error << ")"
+			if (error != DB_SUCCESS) {
+				ib::error() << "(" << ut_strerr(error) << ")"
 					" during optimize, while adding a"
 					" word to the FTS index.";
 			}
@@ -1864,8 +1879,9 @@ fts_optimize_index_completed(
 	error = fts_config_set_index_value(
 		optim->trx, index, FTS_LAST_OPTIMIZED_WORD, &word);
 
-	if (UNIV_UNLIKELY(error != DB_SUCCESS)) {
-		ib::error() << "(" << error << ") while updating"
+	if (error != DB_SUCCESS) {
+
+		ib::error() << "(" << ut_strerr(error) << ") while updating"
 			" last optimized word!";
 	}
 
@@ -2007,7 +2023,7 @@ fts_optimize_purge_deleted_doc_ids(
 	ulint		i;
 	pars_info_t*	info;
 	que_t*		graph;
-	doc_id_t*	update;
+	fts_update_t*	update;
 	doc_id_t	write_doc_id;
 	dberr_t		error = DB_SUCCESS;
 	char		deleted[MAX_FULL_NAME_LEN];
@@ -2017,11 +2033,11 @@ fts_optimize_purge_deleted_doc_ids(
 
 	ut_a(ib_vector_size(optim->to_delete->doc_ids) > 0);
 
-	update = static_cast<doc_id_t*>(
+	update = static_cast<fts_update_t*>(
 		ib_vector_get(optim->to_delete->doc_ids, 0));
 
 	/* Convert to "storage" byte order. */
-	fts_write_doc_id((byte*) &write_doc_id, *update);
+	fts_write_doc_id((byte*) &write_doc_id, update->doc_id);
 
 	/* This is required for the SQL parser to work. It must be able
 	to find the following variables. So we do it twice. */
@@ -2043,11 +2059,11 @@ fts_optimize_purge_deleted_doc_ids(
 	/* Delete the doc ids that were copied at the start. */
 	for (i = 0; i < ib_vector_size(optim->to_delete->doc_ids); ++i) {
 
-		update = static_cast<doc_id_t*>(ib_vector_get(
+		update = static_cast<fts_update_t*>(ib_vector_get(
 			optim->to_delete->doc_ids, i));
 
 		/* Convert to "storage" byte order. */
-		fts_write_doc_id((byte*) &write_doc_id, *update);
+		fts_write_doc_id((byte*) &write_doc_id, update->doc_id);
 
 		fts_bind_doc_id(info, "doc_id1", &write_doc_id);
 
@@ -2417,7 +2433,7 @@ fts_optimize_table(
 	fts_optimize_t*	optim = NULL;
 	fts_t*		fts = table->fts;
 
-	if (UNIV_UNLIKELY(fts_enable_diag_print)) {
+	if (fts_enable_diag_print) {
 		ib::info() << "FTS start optimize " << table->name;
 	}
 
@@ -2469,7 +2485,7 @@ fts_optimize_table(
 		if (error == DB_SUCCESS
 		    && optim->n_completed == ib_vector_size(fts->indexes)) {
 
-			if (UNIV_UNLIKELY(fts_enable_diag_print)) {
+			if (fts_enable_diag_print) {
 				ib::info() << "FTS_OPTIMIZE: Completed"
 					" Optimize, cleanup DELETED table";
 			}
@@ -2492,7 +2508,7 @@ fts_optimize_table(
 
 	fts_optimize_free(optim);
 
-	if (UNIV_UNLIKELY(fts_enable_diag_print)) {
+	if (fts_enable_diag_print) {
 		ib::info() << "FTS end optimize " << table->name;
 	}
 
@@ -2522,6 +2538,22 @@ fts_optimize_create_msg(
 	return(msg);
 }
 
+/** Add message to wqueue, signal thread pool*/
+static void add_msg(fts_msg_t *msg, bool wq_locked= false)
+{
+  ib_wqueue_add(fts_optimize_wq, msg, msg->heap, wq_locked);
+  srv_thread_pool->submit_task(&task);
+}
+
+/**
+Called by "idle" timer. Submits optimize task, which
+will only recalculate is_sync_needed, in case the queue is empty.
+*/
+static void timer_callback(void*)
+{
+  srv_thread_pool->submit_task(&task);
+}
+
 /** Add the table to add to the OPTIMIZER's list.
 @param[in]	table	table to add */
 void fts_optimize_add_table(dict_table_t* table)
@@ -2545,7 +2577,7 @@ void fts_optimize_add_table(dict_table_t* table)
 
 	mutex_enter(&fts_optimize_wq->mutex);
 
-	ib_wqueue_add(fts_optimize_wq, msg, msg->heap, true);
+	add_msg(msg, true);
 
 	table->fts->in_queue = true;
 
@@ -2595,7 +2627,7 @@ fts_optimize_remove_table(
 	remove->event = event;
 	msg->ptr = remove;
 
-	ib_wqueue_add(fts_optimize_wq, msg, msg->heap, true);
+	add_msg(msg, true);
 
 	mutex_exit(&fts_optimize_wq->mutex);
 
@@ -2630,7 +2662,9 @@ fts_optimize_request_sync_table(
 
 	mutex_enter(&fts_optimize_wq->mutex);
 
-	ib_wqueue_add(fts_optimize_wq, msg, msg->heap, true);
+	add_msg(msg, true);
+
+	table->fts->in_queue = true;
 
 	mutex_exit(&fts_optimize_wq->mutex);
 }
@@ -2677,7 +2711,7 @@ static bool fts_optimize_del_table(const dict_table_t* table)
 		slot = static_cast<fts_slot_t*>(ib_vector_get(fts_slots, i));
 
 		if (slot->table == table) {
-			if (UNIV_UNLIKELY(fts_enable_diag_print)) {
+			if (fts_enable_diag_print) {
 				ib::info() << "FTS Optimize Removing table "
 					<< table->name;
 			}
@@ -2770,32 +2804,27 @@ static void fts_optimize_sync_table(dict_table_t* table)
 /**********************************************************************//**
 Optimize all FTS tables.
 @return Dummy return */
-static
-os_thread_ret_t
-DECLARE_THREAD(fts_optimize_thread)(
-/*================*/
-	void*		arg)			/*!< in: work queue*/
+static void fts_optimize_callback(void *)
 {
-	ulint		current = 0;
-	ibool		done = FALSE;
-	ulint		n_tables = 0;
-	ulint		n_optimize = 0;
-	ib_wqueue_t*	wq = (ib_wqueue_t*) arg;
+	static ulint		current = 0;
+	static ibool		done = FALSE;
+	static ulint		n_tables = ib_vector_size(fts_slots);
+	static ulint		n_optimize = 0;
 
 	ut_ad(!srv_read_only_mode);
-	my_thread_init();
 
-	ut_ad(fts_slots);
+	if (!fts_optimize_wq) {
+		/* Possibly timer initiated callback, can come after FTS_MSG_STOP.*/
+		return;
+	}
 
-	/* Assign number of tables added in fts_slots_t to n_tables */
-	n_tables = ib_vector_size(fts_slots);
+	while (!done && srv_shutdown_state == SRV_SHUTDOWN_NONE) {
 
-	while (!done && srv_shutdown_state <= SRV_SHUTDOWN_INITIATED) {
 		/* If there is no message in the queue and we have tables
 		to optimize then optimize the tables. */
 
 		if (!done
-		    && ib_wqueue_is_empty(wq)
+		    && ib_wqueue_is_empty(fts_optimize_wq)
 		    && n_tables > 0
 		    && n_optimize > 0) {
 			fts_slot_t* slot = static_cast<fts_slot_t*>(
@@ -2813,19 +2842,18 @@ DECLARE_THREAD(fts_optimize_thread)(
 				current = 0;
 			}
 
-		} else if (n_optimize == 0 || !ib_wqueue_is_empty(wq)) {
-			fts_msg_t*	msg;
-
-			msg = static_cast<fts_msg_t*>(
-				ib_wqueue_timedwait(wq, FTS_QUEUE_WAIT_IN_USECS));
-
+		} else if (n_optimize == 0
+			   || !ib_wqueue_is_empty(fts_optimize_wq)) {
+			fts_msg_t* msg = static_cast<fts_msg_t*>
+				(ib_wqueue_nowait(fts_optimize_wq));
 			/* Timeout ? */
 			if (msg == NULL) {
 				if (fts_is_sync_needed()) {
 					fts_need_sync = true;
 				}
-
-				continue;
+				if (n_tables)
+					timer->set_time(5000, 0);
+				return;
 			}
 
 			switch (msg->type) {
@@ -2892,13 +2920,6 @@ DECLARE_THREAD(fts_optimize_thread)(
 	ib::info() << "FTS optimize thread exiting.";
 
 	os_event_set(fts_opt_shutdown_event);
-	my_thread_end();
-
-	/* We count the number of threads in os_thread_exit(). A created
-	thread should always use that to exit and not use return() to exit. */
-	os_thread_exit();
-
-	OS_THREAD_DUMMY_RETURN;
 }
 
 /**********************************************************************//**
@@ -2917,7 +2938,8 @@ fts_optimize_init(void)
 
 	/* Create FTS optimize work queue */
 	fts_optimize_wq = ib_wqueue_create();
-	ut_a(fts_optimize_wq != NULL);
+  ut_a(fts_optimize_wq != NULL);
+	timer = srv_thread_pool->create_timer(timer_callback);
 
 	/* Create FTS vector to store fts_slot_t */
 	heap = mem_heap_create(sizeof(dict_table_t*) * 64);
@@ -2946,8 +2968,6 @@ fts_optimize_init(void)
 
 	fts_opt_shutdown_event = os_event_create(0);
 	last_check_sync_time = time(NULL);
-
-	os_thread_create(fts_optimize_thread, fts_optimize_wq, NULL);
 }
 
 /** Shutdown fts optimize thread. */
@@ -2971,15 +2991,17 @@ fts_optimize_shutdown()
 	/* We tell the OPTIMIZE thread to switch to state done, we
 	can't delete the work queue here because the add thread needs
 	deregister the FTS tables. */
+	delete timer;
+	timer = NULL;
+	task_group.cancel_pending(&task);
 
 	msg = fts_optimize_create_msg(FTS_MSG_STOP, NULL);
 
-	ib_wqueue_add(fts_optimize_wq, msg, msg->heap);
+	add_msg(msg);
 
 	os_event_wait(fts_opt_shutdown_event);
 
 	os_event_destroy(fts_opt_shutdown_event);
-
 	ib_wqueue_free(fts_optimize_wq);
 	fts_optimize_wq = NULL;
 }
