@@ -603,16 +603,10 @@ update_mbr:
 		}
 	}
 
-#ifdef UNIV_DEBUG
-	ulint	left_page_no = btr_page_get_prev(page, mtr);
-
-	if (left_page_no == FIL_NULL) {
-
-		ut_a(REC_INFO_MIN_REC_FLAG & rec_get_info_bits(
-			page_rec_get_next(page_get_infimum_rec(page)),
-			page_is_comp(page)));
-	}
-#endif /* UNIV_DEBUG */
+	ut_ad(page_has_prev(page)
+	      || (REC_INFO_MIN_REC_FLAG & rec_get_info_bits(
+			  page_rec_get_next(page_get_infimum_rec(page)),
+			  page_is_comp(page))));
 
 	mem_heap_free(heap);
 
@@ -646,9 +640,6 @@ rtr_adjust_upper_level(
 	mem_heap_t*	heap;
 	ulint		level;
 	dtuple_t*	node_ptr_upper;
-	ulint		prev_page_no;
-	ulint		next_page_no;
-	ulint		space;
 	page_cur_t*	page_cursor;
 	lock_prdt_t	prdt;
 	lock_prdt_t	new_prdt;
@@ -749,38 +740,17 @@ rtr_adjust_upper_level(
 
 	mem_heap_free(heap);
 
-	/* Get the previous and next pages of page */
-	prev_page_no = btr_page_get_prev(page, mtr);
-	next_page_no = btr_page_get_next(page, mtr);
-	space = block->page.id.space();
-	const page_size_t&	page_size = dict_table_page_size(index->table);
+	ut_ad(block->zip_size() == index->table->space->zip_size());
 
-	/* Update page links of the level */
-	if (prev_page_no != FIL_NULL) {
-		page_id_t	prev_page_id(space, prev_page_no);
-
-		buf_block_t*	prev_block = btr_block_get(
-			prev_page_id, page_size, RW_X_LATCH, index, mtr);
-#ifdef UNIV_BTR_DEBUG
-		ut_a(page_is_comp(prev_block->frame) == page_is_comp(page));
-		ut_a(btr_page_get_next(prev_block->frame, mtr)
-		     == block->page.id.page_no());
-#endif /* UNIV_BTR_DEBUG */
-
-		btr_page_set_next(buf_block_get_frame(prev_block),
-				  buf_block_get_page_zip(prev_block),
-				  page_no, mtr);
-	}
+	const uint32_t next_page_no = btr_page_get_next(page);
 
 	if (next_page_no != FIL_NULL) {
-		page_id_t	next_page_id(space, next_page_no);
-
 		buf_block_t*	next_block = btr_block_get(
-			next_page_id, page_size, RW_X_LATCH, index, mtr);
+			*index, next_page_no, RW_X_LATCH, false, mtr);
 #ifdef UNIV_BTR_DEBUG
 		ut_a(page_is_comp(next_block->frame) == page_is_comp(page));
-		ut_a(btr_page_get_prev(next_block->frame, mtr)
-		     == page_get_page_no(page));
+		ut_a(btr_page_get_prev(next_block->frame)
+		     == block->page.id.page_no());
 #endif /* UNIV_BTR_DEBUG */
 
 		btr_page_set_prev(buf_block_get_frame(next_block),
@@ -788,7 +758,6 @@ rtr_adjust_upper_level(
 				  new_page_no, mtr);
 	}
 
-	btr_page_set_prev(page, page_zip, prev_page_no, mtr);
 	btr_page_set_next(page, page_zip, new_page_no, mtr);
 
 	btr_page_set_prev(new_page, new_page_zip, page_no, mtr);
@@ -920,8 +889,8 @@ rtr_split_page_move_rec_list(
 	if (new_page_zip) {
 		mtr_set_log_mode(mtr, log_mode);
 
-		if (!page_zip_compress(new_page_zip, new_page, index,
-				       page_zip_level, NULL, mtr)) {
+		if (!page_zip_compress(new_block, index,
+				       page_zip_level, mtr)) {
 			ulint	ret_pos;
 
 			/* Before trying to reorganize the page,
@@ -1128,7 +1097,7 @@ func_start:
 		as appropriate.  Deleting will always succeed. */
 		ut_a(new_page_zip);
 
-		page_zip_copy_recs(new_page_zip, new_page,
+		page_zip_copy_recs(new_block,
 				   page_zip, page, cursor->index, mtr);
 
 		page_cursor = btr_cur_get_page_cur(cursor);
@@ -1241,7 +1210,7 @@ func_start:
 	For compressed pages, page_cur_tuple_insert() will have
 	attempted this already. */
 	if (rec == NULL) {
-		if (!page_cur_get_page_zip(page_cursor)
+		if (!is_page_cur_get_page_zip(page_cursor)
 		    && btr_page_reorganize(page_cursor, cursor->index, mtr)) {
 			rec = page_cur_tuple_insert(page_cursor, tuple,
 						    cursor->index, offsets,
@@ -1875,18 +1844,19 @@ rtr_estimate_n_rows_in_range(
 
 	mtr.start();
 	index->set_modified(mtr);
-	mtr_s_lock(&index->lock, &mtr);
+	mtr_s_lock_index(index, &mtr);
 
-	buf_block_t* block = btr_block_get(
-		page_id_t(index->table->space_id, index->page),
-		page_size_t(index->table->space->flags),
-		RW_S_LATCH, index, &mtr);
+	buf_block_t* block = btr_root_block_get(index, RW_S_LATCH, &mtr);
+	if (!block) {
+err_exit:
+		mtr.commit();
+		return HA_POS_ERROR;
+	}
 	const page_t* page = buf_block_get_frame(block);
 	const unsigned n_recs = page_header_get_field(page, PAGE_N_RECS);
 
 	if (n_recs == 0) {
-		mtr.commit();
-		return(HA_POS_ERROR);
+		goto err_exit;
 	}
 
 	/* Scan records in root page and calculate area. */
