@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 2000, 2017, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2013, 2020, MariaDB Corporation.
+Copyright (c) 2013, 2019, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -211,7 +211,8 @@ public:
 	int delete_table(const char *name) override;
 
 	int rename_table(const char* from, const char* to) override;
-	inline int defragment_table(const char* name);
+	int defragment_table(const char* name, const char* index_name,
+						bool async);
 	int check(THD* thd, HA_CHECK_OPT* check_opt) override;
 	char* update_table_comment(const char* comment) override;
 
@@ -438,18 +439,6 @@ public:
 	can_convert_blob(const Field_blob* field,
 			 const Column_definition& new_field) const override;
 
-	/** @return whether innodb_strict_mode is active */
-	static bool is_innodb_strict_mode(THD* thd);
-
-	/** @return whether innodb_strict_mode is active */
-	bool is_innodb_strict_mode()
-	{ return is_innodb_strict_mode(m_user_thd); }
-	Compare_keys
-	compare_key_parts(const Field& old_field,
-			  const Column_definition& new_field,
-			  const KEY_PART_INFO& old_part,
-			  const KEY_PART_INFO& new_part) const override;
-
 protected:
 	dberr_t innobase_get_autoinc(ulonglong* value);
 	dberr_t innobase_lock_autoinc();
@@ -528,8 +517,13 @@ the definitions are bracketed with #ifdef INNODB_COMPATIBILITY_HOOKS */
 #error InnoDB needs MySQL to be built with #define INNODB_COMPATIBILITY_HOOKS
 #endif
 
+LEX_STRING* thd_query_string(MYSQL_THD thd);
+size_t thd_query_safe(MYSQL_THD thd, char *buf, size_t buflen);
+
 extern "C" {
 
+struct charset_info_st *thd_charset(MYSQL_THD thd);
+
 /** Check if a user thread is a replication slave thread
 @param thd user thread
 @retval 0 the user thread is not a replication slave thread
@@ -649,6 +643,9 @@ public:
 	/** Set m_tablespace_type. */
 	void set_tablespace_type(bool table_being_altered_is_file_per_table);
 
+	/** Create InnoDB foreign keys from MySQL alter_info. */
+	dberr_t create_foreign_keys();
+
 	/** Create the internal innodb table.
 	@param create_fk	whether to add FOREIGN KEY constraints */
 	int create_table(bool create_fk = true);
@@ -685,11 +682,9 @@ public:
 	void allocate_trx();
 
 	/** Checks that every index have sane size. Depends on strict mode */
-	bool row_size_is_acceptable(const dict_table_t& table,
-				    bool strict) const;
+	bool row_size_is_acceptable(const dict_table_t& table) const;
 	/** Checks that given index have sane size. Depends on strict mode */
-	bool row_size_is_acceptable(const dict_index_t& index,
-				    bool strict) const;
+	bool row_size_is_acceptable(const dict_index_t& index) const;
 
 	/** Determines InnoDB table flags.
 	If strict_mode=OFF, this will adjust the flags to what should be assumed.
@@ -964,15 +959,3 @@ ib_push_frm_error(
 @return true if index column length exceeds limit */
 MY_ATTRIBUTE((warn_unused_result))
 bool too_big_key_part_length(size_t max_field_len, const KEY& key);
-
-/** This function is used to rollback one X/Open XA distributed transaction
-which is in the prepared state
-
-@param[in] hton InnoDB handlerton
-@param[in] xid X/Open XA transaction identification
-
-@return 0 or error number */
-int innobase_rollback_by_xid(handlerton* hton, XID* xid);
-
-/** Free tablespace resources allocated. */
-void innobase_space_shutdown();
