@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1994, 2016, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2017, 2020, MariaDB Corporation.
+Copyright (c) 2017, 2018, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -61,86 +61,45 @@ The status is stored in the low-order bits. */
 /* Length of a B-tree node pointer, in bytes */
 #define REC_NODE_PTR_SIZE	4
 
-#ifndef UNIV_INNOCHECKSUM
 /** SQL null flag in a 1-byte offset of ROW_FORMAT=REDUNDANT records */
-static const rec_offs REC_1BYTE_SQL_NULL_MASK= 0x80;
+#define REC_1BYTE_SQL_NULL_MASK	0x80UL
 /** SQL null flag in a 2-byte offset of ROW_FORMAT=REDUNDANT records */
-static const rec_offs REC_2BYTE_SQL_NULL_MASK= 0x8000;
+#define REC_2BYTE_SQL_NULL_MASK	0x8000UL
 
 /** In a 2-byte offset of ROW_FORMAT=REDUNDANT records, the second most
 significant bit denotes that the tail of a field is stored off-page. */
-static const rec_offs REC_2BYTE_EXTERN_MASK= 0x4000;
+#define REC_2BYTE_EXTERN_MASK	0x4000UL
 
-static const size_t RECORD_OFFSET= 2;
-static const size_t INDEX_OFFSET=
-    RECORD_OFFSET + sizeof(rec_t *) / sizeof(rec_offs);
-#endif /* UNIV_INNOCHECKSUM */
-
-/* Length of the rec_get_offsets() header */
-static const size_t REC_OFFS_HEADER_SIZE=
 #ifdef UNIV_DEBUG
-#ifndef UNIV_INNOCHECKSUM
-    sizeof(rec_t *) / sizeof(rec_offs) +
-    sizeof(dict_index_t *) / sizeof(rec_offs) +
-#endif /* UNIV_INNOCHECKSUM */
+/* Length of the rec_get_offsets() header */
+# define REC_OFFS_HEADER_SIZE	4
+#else /* UNIV_DEBUG */
+/* Length of the rec_get_offsets() header */
+# define REC_OFFS_HEADER_SIZE	2
 #endif /* UNIV_DEBUG */
-    2;
 
 /* Number of elements that should be initially allocated for the
 offsets[] array, first passed to rec_get_offsets() */
-static const size_t REC_OFFS_NORMAL_SIZE= 300;
-static const size_t REC_OFFS_SMALL_SIZE= 18;
-static const size_t REC_OFFS_SEC_INDEX_SIZE=
-    /* PK max key parts */ 16 + /* sec idx max key parts */ 16 +
-    /* child page number for non-leaf pages */ 1;
+#define REC_OFFS_NORMAL_SIZE	OFFS_IN_REC_NORMAL_SIZE
+#define REC_OFFS_SMALL_SIZE	10
 
 /** Get the base address of offsets.  The extra_size is stored at
 this position, and following positions hold the end offsets of
 the fields. */
 #define rec_offs_base(offsets) (offsets + REC_OFFS_HEADER_SIZE)
 
-#ifndef UNIV_INNOCHECKSUM
-/* Offset consists of two parts: 2 upper bits is type and all other bits is
-value */
-
-/** Only 4 different values is possible! */
-enum field_type_t
-{
-  /** normal field */
-  STORED_IN_RECORD= 0 << 14,
-  /** this field is stored off-page */
-  STORED_OFFPAGE= 1 << 14,
-  /** just an SQL NULL */
-  SQL_NULL= 2 << 14,
-  /** instantly added field */
-  DEFAULT= 3 << 14,
-};
-
-/** without 2 upper bits */
-static const rec_offs DATA_MASK= 0x3fff;
-/** 2 upper bits */
-static const rec_offs TYPE_MASK= ~DATA_MASK;
-inline field_type_t get_type(rec_offs n)
-{
-  return static_cast<field_type_t>(n & TYPE_MASK);
-}
-inline void set_type(rec_offs &n, field_type_t type)
-{
-  n= (n & DATA_MASK) | static_cast<rec_offs>(type);
-}
-inline rec_offs get_value(rec_offs n) { return n & DATA_MASK; }
-inline rec_offs combine(rec_offs value, field_type_t type)
-{
-  return get_value(value) | static_cast<rec_offs>(type);
-}
-
 /** Compact flag ORed to the extra size returned by rec_get_offsets() */
-const rec_offs REC_OFFS_COMPACT= ~(rec_offs(~0) >> 1);
+const ulint REC_OFFS_COMPACT = ~(ulint(~0) >> 1);
+/** SQL NULL flag in offsets returned by rec_get_offsets() */
+const ulint REC_OFFS_SQL_NULL = REC_OFFS_COMPACT;
 /** External flag in offsets returned by rec_get_offsets() */
-const rec_offs REC_OFFS_EXTERNAL= REC_OFFS_COMPACT >> 1;
+const ulint REC_OFFS_EXTERNAL = REC_OFFS_COMPACT >> 1;
 /** Default value flag in offsets returned by rec_get_offsets() */
-const rec_offs REC_OFFS_DEFAULT= REC_OFFS_COMPACT >> 2;
-const rec_offs REC_OFFS_MASK= REC_OFFS_DEFAULT - 1;
+const ulint REC_OFFS_DEFAULT = REC_OFFS_COMPACT >> 2;
+/** Mask for offsets returned by rec_get_offsets() */
+const ulint REC_OFFS_MASK = REC_OFFS_DEFAULT - 1;
+
+#ifndef UNIV_INNOCHECKSUM
 /******************************************************//**
 The following function is used to get the pointer of the next chained record
 on the same page.
@@ -493,7 +452,7 @@ offsets form. If the field is SQL null, the flag is ORed in the returned
 value.
 @return offset of the start of the field, SQL null flag ORed */
 UNIV_INLINE
-uint8_t
+ulint
 rec_1_get_field_end_info(
 /*=====================*/
 	const rec_t*	rec,	/*!< in: record */
@@ -507,7 +466,7 @@ value.
 @return offset of the start of the field, SQL null flag and extern
 storage flag ORed */
 UNIV_INLINE
-uint16_t
+ulint
 rec_2_get_field_end_info(
 /*=====================*/
 	const rec_t*	rec,	/*!< in: record */
@@ -548,11 +507,11 @@ rec_get_n_extern_new(
 				(ULINT_UNDEFINED to compute all offsets)
 @param[in,out]	heap		memory heap
 @return the new offsets */
-rec_offs*
+ulint*
 rec_get_offsets_func(
 	const rec_t*		rec,
 	const dict_index_t*	index,
-	rec_offs*		offsets,
+	ulint*			offsets,
 	bool			leaf,
 	ulint			n_fields,
 #ifdef UNIV_DEBUG
@@ -587,7 +546,7 @@ rec_get_offsets_reverse(
 	const dict_index_t*	index,	/*!< in: record descriptor */
 	ulint			node_ptr,/*!< in: nonzero=node pointer,
 					0=leaf node */
-	rec_offs*		offsets)/*!< in/out: array consisting of
+	ulint*			offsets)/*!< in/out: array consisting of
 					offsets[0] allocated elements */
 	MY_ATTRIBUTE((nonnull));
 #ifdef UNIV_DEBUG
@@ -600,7 +559,7 @@ bool
 rec_offs_validate(
 	const rec_t*		rec,
 	const dict_index_t*	index,
-	const rec_offs*		offsets)
+	const ulint*		offsets)
 	MY_ATTRIBUTE((nonnull(3), warn_unused_result));
 /** Update debug data in offsets, in order to tame rec_offs_validate().
 @param[in]	rec	record
@@ -612,7 +571,7 @@ rec_offs_make_valid(
 	const rec_t*		rec,
 	const dict_index_t*	index,
 	bool			leaf,
-	rec_offs*		offsets)
+	ulint*			offsets)
 	MY_ATTRIBUTE((nonnull));
 #else
 # define rec_offs_make_valid(rec, index, leaf, offsets)
@@ -649,16 +608,17 @@ The following function is used to get an offset to the nth
 data field in a record.
 @return offset from the origin of rec */
 UNIV_INLINE
-rec_offs
+ulint
 rec_get_nth_field_offs(
 /*===================*/
-	const rec_offs*	offsets,/*!< in: array returned by rec_get_offsets() */
+	const ulint*	offsets,/*!< in: array returned by rec_get_offsets() */
 	ulint		n,	/*!< in: index of the field */
 	ulint*		len)	/*!< out: length of the field; UNIV_SQL_NULL
 				if SQL null */
 	MY_ATTRIBUTE((nonnull));
 #define rec_get_nth_field(rec, offsets, n, len) \
 ((rec) + rec_get_nth_field_offs(offsets, n, len))
+
 /******************************************************//**
 Determine if the offsets are for a record containing null BLOB pointers.
 @return first field containing a null BLOB pointer, or NULL if none found */
@@ -667,7 +627,7 @@ const byte*
 rec_offs_any_null_extern(
 /*=====================*/
 	const rec_t*	rec,		/*!< in: record */
-	const rec_offs*	offsets)	/*!< in: rec_get_offsets(rec) */
+	const ulint*	offsets)	/*!< in: rec_get_offsets(rec) */
 	MY_ATTRIBUTE((warn_unused_result));
 
 /** Mark the nth field as externally stored.
@@ -675,20 +635,22 @@ rec_offs_any_null_extern(
 @param[in]	n		nth field */
 void
 rec_offs_make_nth_extern(
-        rec_offs*	offsets,
+        ulint*		offsets,
         const ulint     n);
 
-MY_ATTRIBUTE((nonnull))
 /** Determine the number of allocated elements for an array of offsets.
 @param[in]	offsets		offsets after rec_offs_set_n_alloc()
 @return number of elements */
-inline ulint rec_offs_get_n_alloc(const rec_offs *offsets)
+inline
+ulint
+rec_offs_get_n_alloc(const ulint* offsets)
 {
-  ut_ad(offsets);
-  ulint n_alloc= offsets[0];
-  ut_ad(n_alloc > REC_OFFS_HEADER_SIZE);
-  MEM_CHECK_ADDRESSABLE(offsets, n_alloc * sizeof *offsets);
-  return n_alloc;
+	ulint	n_alloc;
+	ut_ad(offsets);
+	n_alloc = offsets[0];
+	ut_ad(n_alloc > REC_OFFS_HEADER_SIZE);
+	UNIV_MEM_ASSERT_W(offsets, n_alloc * sizeof *offsets);
+	return(n_alloc);
 }
 
 /** Determine the number of fields for which offsets have been initialized.
@@ -696,7 +658,7 @@ inline ulint rec_offs_get_n_alloc(const rec_offs *offsets)
 @return number of fields */
 inline
 ulint
-rec_offs_n_fields(const rec_offs* offsets)
+rec_offs_n_fields(const ulint* offsets)
 {
 	ulint	n_fields;
 	ut_ad(offsets);
@@ -712,12 +674,19 @@ rec_offs_n_fields(const rec_offs* offsets)
 @param[in]	offsets	rec_get_offsets()
 @param[in]	n	nth field
 @param[in]	flag	flag to extract
-@return	type of the record field */
-inline field_type_t rec_offs_nth_type(const rec_offs *offsets, ulint n)
+@return	the flag of the record field */
+inline
+ulint
+rec_offs_nth_flag(const ulint* offsets, ulint n, ulint flag)
 {
-  ut_ad(rec_offs_validate(NULL, NULL, offsets));
-  ut_ad(n < rec_offs_n_fields(offsets));
-  return get_type(rec_offs_base(offsets)[1 + n]);
+	ut_ad(rec_offs_validate(NULL, NULL, offsets));
+	ut_ad(n < rec_offs_n_fields(offsets));
+	/* The DEFAULT, NULL, EXTERNAL flags are mutually exclusive. */
+	ut_ad(ut_is_2pow(rec_offs_base(offsets)[1 + n]
+			 & (REC_OFFS_DEFAULT
+			    | REC_OFFS_SQL_NULL
+			    | REC_OFFS_EXTERNAL)));
+	return rec_offs_base(offsets)[1 + n] & flag;
 }
 
 /** Determine if a record field is missing
@@ -725,9 +694,11 @@ inline field_type_t rec_offs_nth_type(const rec_offs *offsets, ulint n)
 @param[in]	offsets	rec_get_offsets()
 @param[in]	n	nth field
 @return	nonzero if default bit is set */
-inline ulint rec_offs_nth_default(const rec_offs *offsets, ulint n)
+inline
+ulint
+rec_offs_nth_default(const ulint* offsets, ulint n)
 {
-  return rec_offs_nth_type(offsets, n) == DEFAULT;
+	return rec_offs_nth_flag(offsets, n, REC_OFFS_DEFAULT);
 }
 
 /** Determine if a record field is SQL NULL
@@ -735,9 +706,11 @@ inline ulint rec_offs_nth_default(const rec_offs *offsets, ulint n)
 @param[in]	offsets	rec_get_offsets()
 @param[in]	n	nth field
 @return	nonzero if SQL NULL set */
-inline ulint rec_offs_nth_sql_null(const rec_offs *offsets, ulint n)
+inline
+ulint
+rec_offs_nth_sql_null(const ulint* offsets, ulint n)
 {
-  return rec_offs_nth_type(offsets, n) == SQL_NULL;
+	return rec_offs_nth_flag(offsets, n, REC_OFFS_SQL_NULL);
 }
 
 /** Determine if a record field is stored off-page.
@@ -745,46 +718,54 @@ inline ulint rec_offs_nth_sql_null(const rec_offs *offsets, ulint n)
 @param[in]	n	nth field
 Returns nonzero if the extern bit is set in nth field of rec.
 @return nonzero if externally stored */
-inline ulint rec_offs_nth_extern(const rec_offs *offsets, ulint n)
+inline
+ulint
+rec_offs_nth_extern(const ulint* offsets, ulint n)
 {
-  return rec_offs_nth_type(offsets, n) == STORED_OFFPAGE;
+	return rec_offs_nth_flag(offsets, n, REC_OFFS_EXTERNAL);
 }
 
 /** Get a global flag of a record.
 @param[in]	offsets	rec_get_offsets()
 @param[in]	flag	flag to extract
 @return	the flag of the record field */
-inline ulint rec_offs_any_flag(const rec_offs *offsets, ulint flag)
+inline
+ulint
+rec_offs_any_flag(const ulint* offsets, ulint flag)
 {
-  ut_ad(rec_offs_validate(NULL, NULL, offsets));
-  return *rec_offs_base(offsets) & flag;
+	ut_ad(rec_offs_validate(NULL, NULL, offsets));
+	return *rec_offs_base(offsets) & flag;
 }
 
 /** Determine if the offsets are for a record containing off-page columns.
 @param[in]	offsets	rec_get_offsets()
 @return nonzero if any off-page columns exist */
-inline bool rec_offs_any_extern(const rec_offs *offsets)
+inline bool rec_offs_any_extern(const ulint* offsets)
 {
-  return rec_offs_any_flag(offsets, REC_OFFS_EXTERNAL);
+	return rec_offs_any_flag(offsets, REC_OFFS_EXTERNAL);
 }
 
 /** Determine if the offsets are for a record that is missing fields.
 @param[in]	offsets	rec_get_offsets()
 @return nonzero if any fields need to be replaced with
 		dict_index_t::instant_field_value() */
-inline ulint rec_offs_any_default(const rec_offs *offsets)
+inline
+ulint
+rec_offs_any_default(const ulint* offsets)
 {
-  return rec_offs_any_flag(offsets, REC_OFFS_DEFAULT);
+	return rec_offs_any_flag(offsets, REC_OFFS_DEFAULT);
 }
 
 /** Determine if the offsets are for other than ROW_FORMAT=REDUNDANT.
 @param[in]	offsets	rec_get_offsets()
 @return	nonzero	if ROW_FORMAT is COMPACT,DYNAMIC or COMPRESSED
 @retval	0	if ROW_FORMAT=REDUNDANT */
-inline ulint rec_offs_comp(const rec_offs *offsets)
+inline
+ulint
+rec_offs_comp(const ulint* offsets)
 {
-  ut_ad(rec_offs_validate(NULL, NULL, offsets));
-  return (*rec_offs_base(offsets) & REC_OFFS_COMPACT);
+	ut_ad(rec_offs_validate(NULL, NULL, offsets));
+	return(*rec_offs_base(offsets) & REC_OFFS_COMPACT);
 }
 
 /** Determine if the record is the metadata pseudo-record
@@ -883,15 +864,12 @@ const byte*
 rec_get_nth_cfield(
 	const rec_t*		rec,
 	const dict_index_t*	index,
-	const rec_offs*		offsets,
+	const ulint*		offsets,
 	ulint			n,
 	ulint*			len)
 {
-	/* Because this function may be invoked by innobase_rec_to_mysql()
-	for reporting a duplicate key during ALTER TABLE or
-	CREATE UNIQUE INDEX, and in that case the rec omit the fixed-size
-	header of 5 or 6 bytes, the check
-	rec_offs_validate(rec, index, offsets) must be avoided here. */
+	ut_ad(rec_offs_validate(rec, index, offsets));
+
 	if (!rec_offs_nth_default(offsets, n)) {
 		return rec_get_nth_field(rec, offsets, n, len);
 	}
@@ -905,7 +883,7 @@ UNIV_INLINE
 ulint
 rec_offs_nth_size(
 /*==============*/
-	const rec_offs*	offsets,/*!< in: array returned by rec_get_offsets() */
+	const ulint*	offsets,/*!< in: array returned by rec_get_offsets() */
 	ulint		n)	/*!< in: nth field */
 	MY_ATTRIBUTE((warn_unused_result));
 
@@ -916,7 +894,7 @@ UNIV_INLINE
 ulint
 rec_offs_n_extern(
 /*==============*/
-	const rec_offs*	offsets)/*!< in: array returned by rec_get_offsets() */
+	const ulint*	offsets)/*!< in: array returned by rec_get_offsets() */
 	MY_ATTRIBUTE((warn_unused_result));
 /***********************************************************//**
 This is used to modify the value of an already existing field in a record.
@@ -929,7 +907,7 @@ void
 rec_set_nth_field(
 /*==============*/
 	rec_t*		rec,	/*!< in: record */
-	const rec_offs*	offsets,/*!< in: array returned by rec_get_offsets() */
+	const ulint*	offsets,/*!< in: array returned by rec_get_offsets() */
 	ulint		n,	/*!< in: index number of the field */
 	const void*	data,	/*!< in: pointer to the data if not SQL null */
 	ulint		len)	/*!< in: length of the data or UNIV_SQL_NULL.
@@ -957,7 +935,7 @@ UNIV_INLINE
 void
 rec_offs_set_n_alloc(
 /*=================*/
-	rec_offs*offsets,	/*!< out: array for rec_get_offsets(),
+	ulint*	offsets,	/*!< out: array for rec_get_offsets(),
 				must be allocated */
 	ulint	n_alloc)	/*!< in: number of elements */
 	MY_ATTRIBUTE((nonnull));
@@ -973,7 +951,7 @@ UNIV_INLINE
 ulint
 rec_offs_data_size(
 /*===============*/
-	const rec_offs*	offsets)/*!< in: array returned by rec_get_offsets() */
+	const ulint*	offsets)/*!< in: array returned by rec_get_offsets() */
 	MY_ATTRIBUTE((warn_unused_result));
 /**********************************************************//**
 Returns the total size of record minus data size of record.
@@ -984,7 +962,7 @@ UNIV_INLINE
 ulint
 rec_offs_extra_size(
 /*================*/
-	const rec_offs*	offsets)/*!< in: array returned by rec_get_offsets() */
+	const ulint*	offsets)/*!< in: array returned by rec_get_offsets() */
 	MY_ATTRIBUTE((warn_unused_result));
 /**********************************************************//**
 Returns the total size of a physical record.
@@ -993,7 +971,7 @@ UNIV_INLINE
 ulint
 rec_offs_size(
 /*==========*/
-	const rec_offs*	offsets)/*!< in: array returned by rec_get_offsets() */
+	const ulint*	offsets)/*!< in: array returned by rec_get_offsets() */
 	MY_ATTRIBUTE((warn_unused_result));
 #ifdef UNIV_DEBUG
 /**********************************************************//**
@@ -1004,7 +982,7 @@ byte*
 rec_get_start(
 /*==========*/
 	const rec_t*	rec,	/*!< in: pointer to record */
-	const rec_offs*	offsets)/*!< in: array returned by rec_get_offsets() */
+	const ulint*	offsets)/*!< in: array returned by rec_get_offsets() */
 	MY_ATTRIBUTE((warn_unused_result));
 /**********************************************************//**
 Returns a pointer to the end of the record.
@@ -1014,7 +992,7 @@ byte*
 rec_get_end(
 /*========*/
 	const rec_t*	rec,	/*!< in: pointer to record */
-	const rec_offs*	offsets)/*!< in: array returned by rec_get_offsets() */
+	const ulint*	offsets)/*!< in: array returned by rec_get_offsets() */
 	MY_ATTRIBUTE((warn_unused_result));
 #else /* UNIV_DEBUG */
 # define rec_get_start(rec, offsets) ((rec) - rec_offs_extra_size(offsets))
@@ -1031,7 +1009,7 @@ rec_t*
 rec_copy(
 	void*		buf,
 	const rec_t*	rec,
-	const rec_offs*	offsets);
+	const ulint*	offsets);
 
 /** Determine the size of a data tuple prefix in a temporary file.
 @param[in]	index		clustered or secondary index
@@ -1060,7 +1038,7 @@ void
 rec_init_offsets_temp(
 	const rec_t*		rec,
 	const dict_index_t*	index,
-	rec_offs*		offsets,
+	ulint*			offsets,
 	ulint			n_core,
 	const dict_col_t::def_t*def_val,
 	rec_comp_status_t	status = REC_STATUS_ORDINARY)
@@ -1074,7 +1052,7 @@ void
 rec_init_offsets_temp(
 	const rec_t*		rec,
 	const dict_index_t*	index,
-	rec_offs*		offsets)
+	ulint*			offsets)
 	MY_ATTRIBUTE((nonnull));
 
 /** Convert a data tuple prefix to the temporary file format.
@@ -1194,7 +1172,7 @@ ibool
 rec_validate(
 /*=========*/
 	const rec_t*	rec,	/*!< in: physical record */
-	const rec_offs*	offsets)/*!< in: array returned by rec_get_offsets() */
+	const ulint*	offsets)/*!< in: array returned by rec_get_offsets() */
 	MY_ATTRIBUTE((nonnull));
 /***************************************************************//**
 Prints an old-style physical record. */
@@ -1211,7 +1189,7 @@ rec_print_mbr_rec(
 /*==========*/
 	FILE*		file,	/*!< in: file where to print */
 	const rec_t*	rec,	/*!< in: physical record */
-	const rec_offs*	offsets)/*!< in: array returned by rec_get_offsets() */
+	const ulint*	offsets)/*!< in: array returned by rec_get_offsets() */
 	MY_ATTRIBUTE((nonnull));
 /***************************************************************//**
 Prints a physical record. */
@@ -1220,7 +1198,7 @@ rec_print_new(
 /*==========*/
 	FILE*		file,	/*!< in: file where to print */
 	const rec_t*	rec,	/*!< in: physical record */
-	const rec_offs*	offsets)/*!< in: array returned by rec_get_offsets() */
+	const ulint*	offsets)/*!< in: array returned by rec_get_offsets() */
 	MY_ATTRIBUTE((nonnull));
 /***************************************************************//**
 Prints a physical record. */
@@ -1242,7 +1220,7 @@ rec_print(
 	std::ostream&	o,
 	const rec_t*	rec,
 	ulint		info,
-	const rec_offs*	offsets);
+	const ulint*	offsets);
 
 /** Wrapper for pretty-printing a record */
 struct rec_index_print
@@ -1269,32 +1247,31 @@ operator<<(std::ostream& o, const rec_index_print& r);
 struct rec_offsets_print
 {
 	/** Constructor */
-	rec_offsets_print(const rec_t* rec, const rec_offs* offsets) :
+	rec_offsets_print(const rec_t* rec, const ulint* offsets) :
 		m_rec(rec), m_offsets(offsets)
 	{}
 
 	/** Record */
 	const rec_t*		m_rec;
 	/** Offsets to each field */
-	const rec_offs*		m_offsets;
+	const ulint*		m_offsets;
 };
 
 /** Display a record.
 @param[in,out]	o	output stream
 @param[in]	r	record to display
 @return	the output stream */
-ATTRIBUTE_COLD
 std::ostream&
 operator<<(std::ostream& o, const rec_offsets_print& r);
 
+# ifndef DBUG_OFF
 /** Pretty-printer of records and tuples */
 class rec_printer : public std::ostringstream {
 public:
 	/** Construct a pretty-printed record.
 	@param rec	record with header
 	@param offsets	rec_get_offsets(rec, ...) */
-	ATTRIBUTE_COLD
-	rec_printer(const rec_t* rec, const rec_offs* offsets)
+	rec_printer(const rec_t* rec, const ulint* offsets)
 		:
 		std::ostringstream ()
 	{
@@ -1307,8 +1284,7 @@ public:
 	@param rec record, possibly lacking header
 	@param info rec_get_info_bits(rec)
 	@param offsets rec_get_offsets(rec, ...) */
-	ATTRIBUTE_COLD
-	rec_printer(const rec_t* rec, ulint info, const rec_offs* offsets)
+	rec_printer(const rec_t* rec, ulint info, const ulint* offsets)
 		:
 		std::ostringstream ()
 	{
@@ -1317,7 +1293,6 @@ public:
 
 	/** Construct a pretty-printed tuple.
 	@param tuple	data tuple */
-	ATTRIBUTE_COLD
 	rec_printer(const dtuple_t* tuple)
 		:
 		std::ostringstream ()
@@ -1328,7 +1303,6 @@ public:
 	/** Construct a pretty-printed tuple.
 	@param field	array of data tuple fields
 	@param n	number of fields */
-	ATTRIBUTE_COLD
 	rec_printer(const dfield_t* field, ulint n)
 		:
 		std::ostringstream ()
@@ -1345,7 +1319,7 @@ private:
 	/** Assignment operator */
 	rec_printer& operator=(const rec_printer& other);
 };
-
+# endif /* !DBUG_OFF */
 
 # ifdef UNIV_DEBUG
 /** Read the DB_TRX_ID of a clustered index record.
