@@ -2,7 +2,7 @@
 
 Copyright (c) 2010, 2015, Oracle and/or its affiliates. All Rights Reserved.
 Copyright (c) 2012, Facebook Inc.
-Copyright (c) 2013, 2020, MariaDB Corporation.
+Copyright (c) 2013, 2019, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
@@ -30,6 +30,7 @@ Created 12/15/2009	Jimmy Yang
 #define srv0mon_h
 
 #include "univ.i"
+#include "my_atomic.h"
 
 #ifndef __STDC_LIMIT_MACROS
 /* Required for FreeBSD so that INT64_MAX is defined. */
@@ -177,7 +178,6 @@ enum monitor_id_t {
 	MONITOR_OVLD_INDEX_PAGES_WRITTEN,
 	MONITOR_OVLD_NON_INDEX_PAGES_WRITTEN,
 	MONITOR_OVLD_PAGES_READ,
-	MONITOR_OVLD_PAGES0_READ,
 	MONITOR_OVLD_INDEX_SEC_REC_CLUSTER_READS,
 	MONITOR_OVLD_INDEX_SEC_REC_CLUSTER_READS_AVOIDED,
 	MONITOR_OVLD_BYTE_READ,
@@ -300,7 +300,6 @@ enum monitor_id_t {
 	MONITOR_TRX_COMMIT_UNDO,
 	MONITOR_TRX_ROLLBACK,
 	MONITOR_TRX_ROLLBACK_SAVEPOINT,
-	MONITOR_TRX_ROLLBACK_ACTIVE,
 	MONITOR_TRX_ACTIVE,
 	MONITOR_RSEG_HISTORY_LEN,
 	MONITOR_NUM_UNDO_SLOT_USED,
@@ -398,7 +397,6 @@ enum monitor_id_t {
 	MONITOR_MASTER_ACTIVE_LOOPS,
 	MONITOR_MASTER_IDLE_LOOPS,
 	MONITOR_SRV_BACKGROUND_DROP_TABLE_MICROSECOND,
-	MONITOR_SRV_IBUF_MERGE_MICROSECOND,
 	MONITOR_SRV_LOG_FLUSH_MICROSECOND,
 	MONITOR_SRV_MEM_VALIDATE_MICROSECOND,
 	MONITOR_SRV_PURGE_MICROSECOND,
@@ -503,18 +501,18 @@ extern ulint		monitor_set_tbl[(NUM_MONITOR + NUM_BITS_ULINT - 1) /
 
 /** Macros to turn on/off the control bit in monitor_set_tbl for a monitor
 counter option. */
-#define MONITOR_ON(monitor)				\
-	(monitor_set_tbl[monitor / NUM_BITS_ULINT] |=	\
-			((ulint)1 << (monitor % NUM_BITS_ULINT)))
+#define MONITOR_ON(monitor)					\
+	(monitor_set_tbl[unsigned(monitor) / NUM_BITS_ULINT] |=	\
+	 (ulint(1) << (unsigned(monitor) % NUM_BITS_ULINT)))
 
-#define MONITOR_OFF(monitor)				\
-	(monitor_set_tbl[monitor / NUM_BITS_ULINT] &=	\
-			~((ulint)1 << (monitor % NUM_BITS_ULINT)))
+#define MONITOR_OFF(monitor)					\
+	(monitor_set_tbl[unsigned(monitor) / NUM_BITS_ULINT] &=	\
+	 ~(ulint(1) << (unsigned(monitor) % NUM_BITS_ULINT)))
 
 /** Check whether the requested monitor is turned on/off */
-#define MONITOR_IS_ON(monitor)				\
-	(monitor_set_tbl[monitor / NUM_BITS_ULINT] &	\
-			((ulint)1 << (monitor % NUM_BITS_ULINT)))
+#define MONITOR_IS_ON(monitor)					\
+	(monitor_set_tbl[unsigned(monitor) / NUM_BITS_ULINT] &	\
+	 (ulint(1) << (unsigned(monitor) % NUM_BITS_ULINT)))
 
 /** The actual monitor counter array that records each monintor counter
 value */
@@ -608,8 +606,9 @@ Use MONITOR_INC if appropriate mutex protection exists.
 #define MONITOR_ATOMIC_INC_LOW(monitor, enabled)			\
 	if (enabled) {							\
 		ib_uint64_t	value;					\
-		value  = my_atomic_add64(				\
-			(int64*) &MONITOR_VALUE(monitor), 1) + 1;	\
+		value  = my_atomic_add64_explicit(			\
+			(int64*) &MONITOR_VALUE(monitor), 1,		\
+			MY_MEMORY_ORDER_RELAXED) + 1;			\
 		/* Note: This is not 100% accurate because of the	\
 		inherent race, we ignore it due to performance. */	\
 		if (value > (ib_uint64_t) MONITOR_MAX_VALUE(monitor)) {	\
@@ -624,8 +623,9 @@ Use MONITOR_DEC if appropriate mutex protection exists.
 #define MONITOR_ATOMIC_DEC_LOW(monitor, enabled)			\
 	if (enabled) {							\
 		ib_uint64_t	value;					\
-		value = my_atomic_add64(				\
-			(int64*) &MONITOR_VALUE(monitor), -1) - 1;	\
+		value = my_atomic_add64_explicit(			\
+			(int64*) &MONITOR_VALUE(monitor), -1,		\
+			MY_MEMORY_ORDER_RELAXED) - 1;			\
 		/* Note: This is not 100% accurate because of the	\
 		inherent race, we ignore it due to performance. */	\
 		if (value < (ib_uint64_t) MONITOR_MIN_VALUE(monitor)) {	\
@@ -652,14 +652,14 @@ Use MONITOR_DEC if appropriate mutex protection exists.
 		}							\
 	}
 
-#ifdef HAVE_valgrind_or_MSAN
+#ifdef UNIV_DEBUG_VALGRIND
 # define MONITOR_CHECK_DEFINED(value) do {	\
 	mon_type_t m = value;			\
-	MEM_CHECK_DEFINED(&m, sizeof m);	\
+	UNIV_MEM_ASSERT_RW(&m, sizeof m);	\
 } while (0)
-#else /* HAVE_valgrind_or_MSAN */
+#else /* UNIV_DEBUG_VALGRIND */
 # define MONITOR_CHECK_DEFINED(value) (void) 0
-#endif /* HAVE_valgrind_or_MSAN */
+#endif /* UNIV_DEBUG_VALGRIND */
 
 #define	MONITOR_INC_VALUE(monitor, value)				\
 	MONITOR_CHECK_DEFINED(value);					\
