@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1995, 2017, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2017, 2020, MariaDB Corporation.
+Copyright (c) 2017, 2019, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -28,9 +28,9 @@ Created 11/26/1995 Heikki Tuuri
 
 #include "buf0buf.h"
 #include "buf0flu.h"
+#include "fsp0sysspace.h"
 #include "page0types.h"
 #include "mtr0log.h"
-#include "row0trunc.h"
 #include "log0recv.h"
 
 /** Iterate over a memo block in reverse. */
@@ -126,7 +126,7 @@ struct Find {
 	mtr_memo_slot_t*m_slot;
 
 	/** Type of the object to look for */
-	ulint		m_type;
+	const ulint	m_type;
 
 	/** The object instance to look for */
 	const void*	m_object;
@@ -167,7 +167,7 @@ struct FindPage
 			slot->object);
 
 		if (m_ptr < block->frame
-		    || m_ptr >= block->frame + block->page.size.logical()) {
+		    || m_ptr >= block->frame + srv_page_size) {
 			return(true);
 		}
 
@@ -218,13 +218,6 @@ static void memo_slot_release(mtr_memo_slot_t *slot)
   case MTR_MEMO_SX_LOCK:
     rw_lock_sx_unlock(reinterpret_cast<rw_lock_t*>(slot->object));
     break;
-  case MTR_MEMO_SPACE_X_LOCK:
-    {
-      fil_space_t *space= static_cast<fil_space_t*>(slot->object);
-      space->committed_size= space->size;
-      rw_lock_x_unlock(&space->latch);
-    }
-    break;
   case MTR_MEMO_X_LOCK:
     rw_lock_x_unlock(reinterpret_cast<rw_lock_t*>(slot->object));
     break;
@@ -233,7 +226,7 @@ static void memo_slot_release(mtr_memo_slot_t *slot)
   case MTR_MEMO_PAGE_SX_FIX:
   case MTR_MEMO_PAGE_X_FIX:
     buf_block_t *block= reinterpret_cast<buf_block_t*>(slot->object);
-    buf_block_unfix(block);
+    block->unfix();
     buf_page_release_latch(block, slot->type);
     break;
   }
@@ -258,13 +251,6 @@ struct ReleaseLatches {
     case MTR_MEMO_S_LOCK:
       rw_lock_s_unlock(reinterpret_cast<rw_lock_t*>(slot->object));
       break;
-    case MTR_MEMO_SPACE_X_LOCK:
-      {
-        fil_space_t *space= static_cast<fil_space_t*>(slot->object);
-        space->committed_size= space->size;
-        rw_lock_x_unlock(&space->latch);
-      }
-      break;
     case MTR_MEMO_X_LOCK:
       rw_lock_x_unlock(reinterpret_cast<rw_lock_t*>(slot->object));
       break;
@@ -276,7 +262,7 @@ struct ReleaseLatches {
     case MTR_MEMO_PAGE_SX_FIX:
     case MTR_MEMO_PAGE_X_FIX:
       buf_block_t *block= reinterpret_cast<buf_block_t*>(slot->object);
-      buf_block_unfix(block);
+      block->unfix();
       buf_page_release_latch(block, slot->type);
       break;
     }
@@ -382,7 +368,7 @@ mtr_write_log(
 	ut_ad(!recv_no_log_write);
 	DBUG_PRINT("ib_log",
 		   (ULINTPF " extra bytes written at " LSN_PF,
-		    len, log_sys->lsn));
+		    len, log_sys.lsn));
 
 	log_reserve_and_open(len);
 	log->for_each_block(write_log);
@@ -392,7 +378,7 @@ mtr_write_log(
 /** Start a mini-transaction. */
 void mtr_t::start()
 {
-  MEM_UNDEFINED(this, sizeof *this);
+  UNIV_MEM_INVALID(this, sizeof *this);
 
   new(&m_memo) mtr_buf_t();
   new(&m_log) mtr_buf_t();
@@ -404,8 +390,6 @@ void mtr_t::start()
   m_log_mode= MTR_LOG_ALL;
   ut_d(m_user_space_id= TRX_SYS_SPACE);
   m_user_space= NULL;
-  m_undo_space= NULL;
-  m_sys_space= NULL;
   m_state= MTR_STATE_ACTIVE;
   m_flush_observer= NULL;
   m_commit_lsn= 0;
@@ -510,7 +494,7 @@ mtr_t::commit_checkpoint(
 	if (write_mlog_checkpoint) {
 		DBUG_PRINT("ib_log",
 			   ("MLOG_CHECKPOINT(" LSN_PF ") written at " LSN_PF,
-			    checkpoint_lsn, log_sys->lsn));
+			    checkpoint_lsn, log_sys.lsn));
 	}
 }
 
@@ -522,12 +506,7 @@ mtr_t::commit_checkpoint(
 bool
 mtr_t::is_named_space(ulint space) const
 {
-	ut_ad(!m_sys_space || m_sys_space->id == TRX_SYS_SPACE);
-	ut_ad(!m_undo_space || m_undo_space->id != TRX_SYS_SPACE);
 	ut_ad(!m_user_space || m_user_space->id != TRX_SYS_SPACE);
-	ut_ad(!m_sys_space || m_sys_space != m_user_space);
-	ut_ad(!m_sys_space || m_sys_space != m_undo_space);
-	ut_ad(!m_user_space || m_user_space != m_undo_space);
 
 	switch (get_log_mode()) {
 	case MTR_LOG_NONE:
@@ -542,6 +521,26 @@ mtr_t::is_named_space(ulint space) const
 	ut_error;
 	return(false);
 }
+/** Check if a tablespace is associated with the mini-transaction
+(needed for generating a MLOG_FILE_NAME record)
+@param[in]	space	tablespace
+@return whether the mini-transaction is associated with the space */
+bool mtr_t::is_named_space(const fil_space_t* space) const
+{
+  ut_ad(!m_user_space || m_user_space->id != TRX_SYS_SPACE);
+
+  switch (get_log_mode()) {
+  case MTR_LOG_NONE:
+  case MTR_LOG_NO_REDO:
+    return true;
+  case MTR_LOG_ALL:
+  case MTR_LOG_SHORT_INSERTS:
+    return m_user_space == space || is_predefined_tablespace(space->id);
+  }
+
+  ut_error;
+  return false;
+}
 #endif /* UNIV_DEBUG */
 
 /** Acquire a tablespace X-latch.
@@ -558,80 +557,22 @@ mtr_t::x_lock_space(ulint space_id, const char* file, unsigned line)
 	ut_ad(is_active());
 
 	if (space_id == TRX_SYS_SPACE) {
-		space = m_sys_space;
-
-		if (!space) {
-			space = m_sys_space = fil_space_get(space_id);
-		}
+		space = fil_system.sys_space;
 	} else if ((space = m_user_space) && space_id == space->id) {
-	} else if ((space = m_undo_space) && space_id == space->id) {
-	} else if (get_log_mode() == MTR_LOG_NO_REDO) {
-		space = fil_space_get(space_id);
-		ut_ad(space->purpose == FIL_TYPE_TEMPORARY
-		      || space->purpose == FIL_TYPE_IMPORT
-		      || space->redo_skipped_count > 0
-		      || srv_is_tablespace_truncated(space->id));
 	} else {
-		/* called from trx_rseg_create() */
-		space = m_undo_space = fil_space_get(space_id);
+		space = fil_space_get(space_id);
+		ut_ad(get_log_mode() != MTR_LOG_NO_REDO
+		      || space->purpose == FIL_TYPE_TEMPORARY
+		      || space->purpose == FIL_TYPE_IMPORT
+		      || space->redo_skipped_count > 0);
 	}
 
 	ut_ad(space);
 	ut_ad(space->id == space_id);
 	x_lock_space(space, file, line);
-	ut_ad(space->purpose == FIL_TYPE_TEMPORARY
-	      || space->purpose == FIL_TYPE_IMPORT
-	      || space->purpose == FIL_TYPE_TABLESPACE);
 	return(space);
 }
 
-/** Exclusively aqcuire a tablespace latch.
-@param space  tablespace
-@param file   source code file name of the caller
-@param line   source code line number */
-void mtr_t::x_lock_space(fil_space_t *space, const char *file, unsigned line)
-{
-  rw_lock_x_lock_inline(&space->latch, 0, file, line);
-  memo_push(space, MTR_MEMO_SPACE_X_LOCK);
-}
-
-/** Look up the system tablespace. */
-void
-mtr_t::lookup_sys_space()
-{
-	ut_ad(!m_sys_space);
-	m_sys_space = fil_space_get(TRX_SYS_SPACE);
-	ut_ad(m_sys_space);
-}
-
-/** Look up the user tablespace.
-@param[in]	space_id	tablespace ID */
-void
-mtr_t::lookup_user_space(ulint space_id)
-{
-	ut_ad(space_id != TRX_SYS_SPACE);
-	ut_ad(m_user_space_id == space_id);
-	ut_ad(!m_user_space);
-	m_user_space = fil_space_get(space_id);
-	ut_ad(m_user_space);
-}
-
-/** Set the tablespace associated with the mini-transaction
-(needed for generating a MLOG_FILE_NAME record)
-@param[in]	space	user or system tablespace */
-void
-mtr_t::set_named_space(fil_space_t* space)
-{
-	ut_ad(m_user_space_id == TRX_SYS_SPACE);
-	ut_d(m_user_space_id = space->id);
-	if (space->id == TRX_SYS_SPACE) {
-		ut_ad(!m_sys_space || m_sys_space == space);
-		m_sys_space = space;
-	} else {
-		m_user_space = space;
-	}
-}
-
 /** Release an object in the memo stack.
 @return true if released */
 bool
@@ -686,7 +627,7 @@ inline ulint mtr_t::prepare_write()
 		ut_ad(m_log_mode == MTR_LOG_NO_REDO);
 		ut_ad(m_log.size() == 0);
 		log_mutex_enter();
-		m_commit_lsn = log_sys->lsn;
+		m_commit_lsn = log_sys.lsn;
 		return 0;
 	}
 
@@ -695,8 +636,8 @@ inline ulint mtr_t::prepare_write()
 	ut_ad(len > 0);
 	ut_ad(n_recs > 0);
 
-	if (len > log_sys->buf_size / 2) {
-		log_buffer_extend((len + 1) * 2);
+	if (len > srv_log_buffer_size / 2) {
+		log_buffer_extend(ulong((len + 1) * 2));
 	}
 
 	ut_ad(m_n_log_recs == n_recs);
@@ -780,30 +721,6 @@ inline lsn_t mtr_t::finish_write(ulint len)
 	return start_lsn;
 }
 
-/** Release the free extents that was reserved using
-fsp_reserve_free_extents().  This is equivalent to calling
-fil_space_release_free_extents().  This is intended for use
-with index pages.
-@param[in]	n_reserved	number of reserved extents */
-void
-mtr_t::release_free_extents(ulint n_reserved)
-{
-  fil_space_t *space= m_user_space;
-
-  ut_ad(!m_undo_space);
-
-  if (space)
-    ut_ad(m_user_space->id == m_user_space_id);
-  else
-  {
-    ut_ad(m_sys_space->id == TRX_SYS_SPACE);
-    space= m_sys_space;
-  }
-
-  ut_ad(memo_contains(get_memo(), space, MTR_MEMO_SPACE_X_LOCK));
-  space->release_free_extents(n_reserved);
-}
-
 #ifdef UNIV_DEBUG
 /** Check if memo contains the given item.
 @return	true if contains */
@@ -811,7 +728,7 @@ bool
 mtr_t::memo_contains(
 	const mtr_buf_t*	memo,
 	const void*		object,
-	ulint			type)
+	mtr_memo_type_t		type)
 {
 	Iterate<Find> iteration(Find(object, type));
 	if (memo->for_each_block_in_reverse(iteration)) {
@@ -828,6 +745,8 @@ mtr_t::memo_contains(
 	case MTR_MEMO_S_LOCK:
 		ut_ad(rw_lock_own((rw_lock_t*) object, RW_LOCK_S));
 		break;
+	default:
+		break;
 	}
 
 	return(true);
