@@ -25,6 +25,7 @@ Created 7/1/1994 Heikki Tuuri
 
 #include "rem0cmp.h"
 #include "rem0rec.h"
+#include "page0page.h"
 #include "dict0mem.h"
 #include "handler0alter.h"
 
@@ -222,7 +223,6 @@ static
 int
 cmp_geometry_field(
 /*===============*/
-	ulint		mtype,		/*!< in: main type */
 	ulint		prtype,		/*!< in: precise type */
 	const byte*	a,		/*!< in: data field */
 	unsigned int	a_length,	/*!< in: data field length,
@@ -296,12 +296,10 @@ cmp_gis_field(
 					not UNIV_SQL_NULL */
 {
 	if (mode == PAGE_CUR_MBR_EQUAL) {
-		/* TODO: Since the DATA_GEOMETRY is not used in compare
-		function, we could pass it instead of a specific type now */
-		return(cmp_geometry_field(DATA_GEOMETRY, DATA_GIS_MBR,
-					  a, a_length, b, b_length));
+		return cmp_geometry_field(DATA_GIS_MBR,
+					  a, a_length, b, b_length);
 	} else {
-		return(rtree_key_cmp(mode, a, a_length, b, b_length));
+		return rtree_key_cmp(mode, a, int(a_length), b, int(b_length));
 	}
 }
 
@@ -372,8 +370,7 @@ cmp_whole_field(
 		return(innobase_mysql_cmp(prtype,
 					  a, a_length, b, b_length));
 	case DATA_GEOMETRY:
-		return(cmp_geometry_field(mtype, prtype, a, a_length, b,
-				b_length));
+		return cmp_geometry_field(prtype, a, a_length, b, b_length);
 	default:
 		ib::fatal() << "Unknown data type number " << mtype;
 	}
@@ -402,6 +399,9 @@ cmp_data(
 	const byte*	data2,
 	ulint		len2)
 {
+	ut_ad(len1 != UNIV_SQL_DEFAULT);
+	ut_ad(len2 != UNIV_SQL_DEFAULT);
+
 	if (len1 == UNIV_SQL_NULL || len2 == UNIV_SQL_NULL) {
 		if (len1 == len2) {
 			return(0);
@@ -701,6 +701,11 @@ cmp_dtuple_rec_with_match_low(
 		contain externally stored fields, and the first fields
 		(primary key fields) should already differ. */
 		ut_ad(!rec_offs_nth_extern(offsets, cur_field));
+		/* We should never compare against instantly added columns.
+		Columns can only be instantly added to clustered index
+		leaf page records, and the first fields (primary key fields)
+		should already differ. */
+		ut_ad(!rec_offs_nth_default(offsets, cur_field));
 
 		rec_b_ptr = rec_get_nth_field(rec, offsets, cur_field,
 					      &rec_f_len);
@@ -783,20 +788,23 @@ cmp_dtuple_rec_with_match_bytes(
 	ulint*			matched_fields,
 	ulint*			matched_bytes)
 {
-	ulint		n_cmp	= dtuple_get_n_fields_cmp(dtuple);
-	ulint		cur_field;	/* current field number */
-	ulint		cur_bytes;
-	int		ret;		/* return value */
-
 	ut_ad(dtuple_check_typed(dtuple));
 	ut_ad(rec_offs_validate(rec, index, offsets));
 	ut_ad(!(REC_INFO_MIN_REC_FLAG
 		& dtuple_get_info_bits(dtuple)));
-	ut_ad(!(REC_INFO_MIN_REC_FLAG
-		& rec_get_info_bits(rec, rec_offs_comp(offsets))));
 
-	cur_field = *matched_fields;
-	cur_bytes = *matched_bytes;
+	if (UNIV_UNLIKELY(REC_INFO_MIN_REC_FLAG
+			  & rec_get_info_bits(rec, rec_offs_comp(offsets)))) {
+		ut_ad(page_rec_is_first(rec, page_align(rec)));
+		ut_ad(!page_has_prev(page_align(rec)));
+		ut_ad(rec_is_metadata(rec, *index));
+		return 1;
+	}
+
+	ulint cur_field = *matched_fields;
+	ulint cur_bytes = *matched_bytes;
+	ulint n_cmp = dtuple_get_n_fields_cmp(dtuple);
+	int ret;
 
 	ut_ad(n_cmp <= dtuple_get_n_fields(dtuple));
 	ut_ad(cur_field <= n_cmp);
@@ -816,6 +824,8 @@ cmp_dtuple_rec_with_match_bytes(
 
 		dtuple_b_ptr = static_cast<const byte*>(
 			dfield_get_data(dfield));
+
+		ut_ad(!rec_offs_nth_default(offsets, cur_field));
 		rec_b_ptr = rec_get_nth_field(rec, offsets,
 					      cur_field, &rec_f_len);
 		ut_ad(!rec_offs_nth_extern(offsets, cur_field));
@@ -1137,10 +1147,9 @@ cmp_rec_rec_with_match(
 	/* Test if rec is the predefined minimum record */
 	if (UNIV_UNLIKELY(rec_get_info_bits(rec1, comp)
 			  & REC_INFO_MIN_REC_FLAG)) {
-		/* There should only be one such record. */
-		ut_ad(!(rec_get_info_bits(rec2, comp)
-			& REC_INFO_MIN_REC_FLAG));
-		ret = -1;
+		ret = UNIV_UNLIKELY(rec_get_info_bits(rec2, comp)
+				    & REC_INFO_MIN_REC_FLAG)
+			? 0 : -1;
 		goto order_resolved;
 	} else if (UNIV_UNLIKELY
 		   (rec_get_info_bits(rec2, comp)
@@ -1190,6 +1199,8 @@ cmp_rec_rec_with_match(
 		DB_ROLL_PTR, and any externally stored columns. */
 		ut_ad(!rec_offs_nth_extern(offsets1, cur_field));
 		ut_ad(!rec_offs_nth_extern(offsets2, cur_field));
+		ut_ad(!rec_offs_nth_default(offsets1, cur_field));
+		ut_ad(!rec_offs_nth_default(offsets2, cur_field));
 
 		rec1_b_ptr = rec_get_nth_field(rec1, offsets1,
 					       cur_field, &rec1_f_len);
