@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1994, 2016, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2017, 2020, MariaDB Corporation.
+Copyright (c) 2017, 2019, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -260,13 +260,13 @@ void
 rec_init_offsets_comp_ordinary(
 	const rec_t*		rec,
 	const dict_index_t*	index,
-	rec_offs*		offsets,
+	ulint*			offsets,
 	ulint			n_core,
 	const dict_col_t::def_t*def_val,
 	rec_leaf_format		format)
 {
-	rec_offs	offs		= 0;
-	rec_offs	any		= 0;
+	ulint		offs		= 0;
+	ulint		any		= 0;
 	const byte*	nulls		= rec;
 	const byte*	lens		= NULL;
 	ulint		n_fields	= n_core;
@@ -335,16 +335,14 @@ ordinary:
 	}
 
 start:
-#ifdef UNIV_DEBUG
 	/* We cannot invoke rec_offs_make_valid() if format==REC_LEAF_TEMP.
 	Similarly, rec_offs_validate() will fail in that case, because
 	it invokes rec_get_status(). */
-	memcpy(&offsets[RECORD_OFFSET], &rec, sizeof(rec));
-	memcpy(&offsets[INDEX_OFFSET], &index, sizeof(index));
-#endif /* UNIV_DEBUG */
+	ut_d(offsets[2] = ulint(rec));
+	ut_d(offsets[3] = ulint(index));
 
 	/* read the lengths of fields 0..n_fields */
-	rec_offs len;
+	ulint len;
 	ulint i = 0;
 	const dict_field_t* field = index->fields;
 
@@ -352,12 +350,12 @@ start:
 		if (mblob) {
 			if (i == index->first_user_field()) {
 				offs += FIELD_REF_SIZE;
-				len = combine(offs, STORED_OFFPAGE);
+				len = offs | REC_OFFS_EXTERNAL;
 				any |= REC_OFFS_EXTERNAL;
 				field--;
 				continue;
 			} else if (i >= n_fields) {
-				len = combine(offs, DEFAULT);
+				len = offs | REC_OFFS_DEFAULT;
 				any |= REC_OFFS_DEFAULT;
 				continue;
 			}
@@ -366,21 +364,20 @@ start:
 		} else if (!mblob && def_val) {
 			const dict_col_t::def_t& d = def_val[i - n_core];
 			if (!d.data) {
-				len = combine(offs, SQL_NULL);
+				len = offs | REC_OFFS_SQL_NULL;
 				ut_ad(d.len == UNIV_SQL_NULL);
 			} else {
-				len = combine(offs, DEFAULT);
+				len = offs | REC_OFFS_DEFAULT;
 				any |= REC_OFFS_DEFAULT;
 			}
 
 			continue;
 		} else {
-			ulint dlen;
-			if (!index->instant_field_value(i, &dlen)) {
-				len = combine(offs, SQL_NULL);
-				ut_ad(dlen == UNIV_SQL_NULL);
+			if (!index->instant_field_value(i, &len)) {
+				ut_ad(len == UNIV_SQL_NULL);
+				len = offs | REC_OFFS_SQL_NULL;
 			} else {
-				len = combine(offs, DEFAULT);
+				len = offs | REC_OFFS_DEFAULT;
 				any |= REC_OFFS_DEFAULT;
 			}
 
@@ -404,7 +401,7 @@ start:
 				We do not advance offs, and we set
 				the length to zero and enable the
 				SQL NULL flag in offsets[]. */
-				len = combine(offs, SQL_NULL);
+				len = offs | REC_OFFS_SQL_NULL;
 				continue;
 			}
 			null_mask <<= 1;
@@ -427,11 +424,11 @@ start:
 				len <<= 8;
 				len |= *lens--;
 
-				offs += get_value(len);
+				offs += len & 0x3fff;
 				if (UNIV_UNLIKELY(len & 0x4000)) {
 					ut_ad(dict_index_is_clust(index));
 					any |= REC_OFFS_EXTERNAL;
-					len = combine(offs, STORED_OFFPAGE);
+					len = offs | REC_OFFS_EXTERNAL;
 				} else {
 					len = offs;
 				}
@@ -447,7 +444,7 @@ start:
 		 i < rec_offs_n_fields(offsets));
 
 	*rec_offs_base(offsets)
-		= static_cast<rec_offs>(rec - (lens + 1)) | REC_OFFS_COMPACT | any;
+		= ulint(rec - (lens + 1)) | REC_OFFS_COMPACT | any;
 }
 
 #ifdef UNIV_DEBUG
@@ -461,7 +458,7 @@ rec_offs_make_valid(
 	const rec_t*		rec,
 	const dict_index_t*	index,
 	bool			leaf,
-	rec_offs*		offsets)
+	ulint*			offsets)
 {
 	const bool is_alter_metadata = leaf
 		&& rec_is_alter_metadata(rec, *index);
@@ -485,10 +482,10 @@ rec_offs_make_valid(
 	for (; n < rec_offs_n_fields(offsets); n++) {
 		ut_ad(leaf);
 		ut_ad(is_alter_metadata
-		      || get_type(rec_offs_base(offsets)[1 + n]) == DEFAULT);
+		      || rec_offs_base(offsets)[1 + n] & REC_OFFS_DEFAULT);
 	}
-	memcpy(&offsets[RECORD_OFFSET], &rec, sizeof(rec));
-	memcpy(&offsets[INDEX_OFFSET], &index, sizeof(index));
+	offsets[2] = ulint(rec);
+	offsets[3] = ulint(index);
 }
 
 /** Validate offsets returned by rec_get_offsets().
@@ -500,14 +497,14 @@ bool
 rec_offs_validate(
 	const rec_t*		rec,
 	const dict_index_t*	index,
-	const rec_offs*		offsets)
+	const ulint*		offsets)
 {
 	ulint	i	= rec_offs_n_fields(offsets);
 	ulint	last	= ULINT_MAX;
 	ulint	comp	= *rec_offs_base(offsets) & REC_OFFS_COMPACT;
 
 	if (rec) {
-		ut_ad(!memcmp(&rec, &offsets[RECORD_OFFSET], sizeof(rec)));
+		ut_ad(ulint(rec) == offsets[2]);
 		if (!comp) {
 			const bool is_user_rec = rec_get_heap_no_old(rec)
 				>= PAGE_HEAP_NO_USER_LOW;
@@ -519,14 +516,14 @@ rec_offs_validate(
 			      || (n + (index->id == DICT_INDEXES_ID))
 			      >= index->n_core_fields);
 			for (; n < i; n++) {
-				ut_ad(get_type(rec_offs_base(offsets)[1 + n])
-				      == DEFAULT);
+				ut_ad(rec_offs_base(offsets)[1 + n]
+				      & REC_OFFS_DEFAULT);
 			}
 		}
 	}
 	if (index) {
-		ut_ad(!memcmp(&index, &offsets[INDEX_OFFSET], sizeof(index)));
-		ulint max_n_fields = std::max(
+		ut_ad(ulint(index) == offsets[3]);
+		ulint max_n_fields = ut_max(
 			dict_index_get_n_fields(index),
 			dict_index_get_n_unique_in_tree(index) + 1);
 		if (comp && rec) {
@@ -561,7 +558,7 @@ rec_offs_validate(
 		ut_ad(!index->n_def || i <= max_n_fields);
 	}
 	while (i--) {
-		ulint curr = get_value(rec_offs_base(offsets)[1 + i]);
+		ulint	curr = rec_offs_base(offsets)[1 + i] & REC_OFFS_MASK;
 		ut_ad(curr <= last);
 		last = curr;
 	}
@@ -578,9 +575,10 @@ to the extra size (if REC_OFFS_COMPACT is set, the record is in the
 new format; if REC_OFFS_EXTERNAL is set, the record contains externally
 stored columns), and rec_offs_base(offsets)[1..n_fields] will be set to
 offsets past the end of fields 0..n_fields, or to the beginning of
-fields 1..n_fields+1.  When the type of the offset at [i+1]
-is (SQL_NULL), the field i is NULL. When the type of the offset at [i+1]
-is (STORED_OFFPAGE), the field i is stored externally.
+fields 1..n_fields+1.  When the high-order bit of the offset at [i+1]
+is set (REC_OFFS_SQL_NULL), the field i is NULL.  When the second
+high-order bit of the offset at [i+1] is set (REC_OFFS_EXTERNAL), the
+field i is being stored externally.
 @param[in]	rec	record
 @param[in]	index	the index that the record belongs in
 @param[in]	leaf	whether the record resides in a leaf page
@@ -591,10 +589,10 @@ rec_init_offsets(
 	const rec_t*		rec,
 	const dict_index_t*	index,
 	bool			leaf,
-	rec_offs*		offsets)
+	ulint*			offsets)
 {
 	ulint	i	= 0;
-	rec_offs	offs;
+	ulint	offs;
 
 	/* This assertion was relaxed for the btr_cur_open_at_index_side()
 	call in btr_cur_instant_init_low(). We cannot invoke
@@ -603,8 +601,8 @@ rec_init_offsets(
 	dict_table_t::deserialise_columns(). */
 	ut_ad(index->n_core_null_bytes <= UT_BITS_IN_BYTES(index->n_nullable)
 	      || index->in_instant_init);
-	ut_d(memcpy(&offsets[RECORD_OFFSET], &rec, sizeof(rec)));
-	ut_d(memcpy(&offsets[INDEX_OFFSET], &index, sizeof(index)));
+	ut_d(offsets[2] = ulint(rec));
+	ut_d(offsets[3] = ulint(index));
 
 	if (dict_table_is_comp(index->table)) {
 		const byte*	nulls;
@@ -662,7 +660,7 @@ rec_init_offsets(
 
 		/* read the lengths of fields 0..n */
 		do {
-			rec_offs len;
+			ulint	len;
 			if (UNIV_UNLIKELY(i == n_node_ptr_field)) {
 				len = offs += REC_NODE_PTR_SIZE;
 				goto resolved;
@@ -684,7 +682,7 @@ rec_init_offsets(
 					We do not advance offs, and we set
 					the length to zero and enable the
 					SQL NULL flag in offsets[]. */
-					len = combine(offs, SQL_NULL);
+					len = offs | REC_OFFS_SQL_NULL;
 					goto resolved;
 				}
 				null_mask <<= 1;
@@ -716,7 +714,7 @@ rec_init_offsets(
 						stored columns.  Thus
 						the "e" flag must be 0. */
 						ut_a(!(len & 0x4000));
-						offs += get_value(len);
+						offs += len & 0x3fff;
 						len = offs;
 
 						goto resolved;
@@ -732,40 +730,39 @@ resolved:
 		} while (++i < rec_offs_n_fields(offsets));
 
 		*rec_offs_base(offsets)
-			= static_cast<rec_offs>(rec - (lens + 1))
-			  | REC_OFFS_COMPACT;
+			= ulint(rec - (lens + 1)) | REC_OFFS_COMPACT;
 	} else {
 		/* Old-style record: determine extra size and end offsets */
 		offs = REC_N_OLD_EXTRA_BYTES;
 		const ulint n_fields = rec_get_n_fields_old(rec);
 		const ulint n = std::min(n_fields, rec_offs_n_fields(offsets));
-		rec_offs any;
+		ulint any;
 
 		if (rec_get_1byte_offs_flag(rec)) {
-			offs += static_cast<rec_offs>(n_fields);
+			offs += n_fields;
 			any = offs;
 			/* Determine offsets to fields */
 			do {
 				offs = rec_1_get_field_end_info(rec, i);
 				if (offs & REC_1BYTE_SQL_NULL_MASK) {
 					offs &= ~REC_1BYTE_SQL_NULL_MASK;
-					set_type(offs, SQL_NULL);
+					offs |= REC_OFFS_SQL_NULL;
 				}
 				rec_offs_base(offsets)[1 + i] = offs;
 			} while (++i < n);
 		} else {
-			offs += 2 * static_cast<rec_offs>(n_fields);
+			offs += 2 * n_fields;
 			any = offs;
 			/* Determine offsets to fields */
 			do {
 				offs = rec_2_get_field_end_info(rec, i);
 				if (offs & REC_2BYTE_SQL_NULL_MASK) {
 					offs &= ~REC_2BYTE_SQL_NULL_MASK;
-					set_type(offs, SQL_NULL);
+					offs |= REC_OFFS_SQL_NULL;
 				}
 				if (offs & REC_2BYTE_EXTERN_MASK) {
 					offs &= ~REC_2BYTE_EXTERN_MASK;
-					set_type(offs, STORED_OFFPAGE);
+					offs |= REC_OFFS_EXTERNAL;
 					any |= REC_OFFS_EXTERNAL;
 				}
 				rec_offs_base(offsets)[1 + i] = offs;
@@ -777,8 +774,8 @@ resolved:
 			      || i + (index->id == DICT_INDEXES_ID)
 			      == rec_offs_n_fields(offsets));
 
-			ut_ad(i != 0);
-			offs = combine(rec_offs_base(offsets)[i], DEFAULT);
+			offs = (rec_offs_base(offsets)[i] & REC_OFFS_MASK)
+				| REC_OFFS_DEFAULT;
 
 			do {
 				rec_offs_base(offsets)[1 + i] = offs;
@@ -801,11 +798,11 @@ resolved:
 				(ULINT_UNDEFINED to compute all offsets)
 @param[in,out]	heap		memory heap
 @return the new offsets */
-rec_offs*
+ulint*
 rec_get_offsets_func(
 	const rec_t*		rec,
 	const dict_index_t*	index,
-	rec_offs*		offsets,
+	ulint*			offsets,
 	bool			leaf,
 	ulint			n_fields,
 #ifdef UNIV_DEBUG
@@ -891,22 +888,21 @@ rec_get_offsets_func(
 	if (UNIV_UNLIKELY(!offsets)
 	    || UNIV_UNLIKELY(rec_offs_get_n_alloc(offsets) < size)) {
 		if (UNIV_UNLIKELY(!*heap)) {
-			*heap = mem_heap_create_at(size * sizeof(*offsets),
+			*heap = mem_heap_create_at(size * sizeof(ulint),
 						   file, line);
 		}
-		offsets = static_cast<rec_offs*>(
-			mem_heap_alloc(*heap, size * sizeof(*offsets)));
+		offsets = static_cast<ulint*>(
+			mem_heap_alloc(*heap, size * sizeof(ulint)));
 
 		rec_offs_set_n_alloc(offsets, size);
 	}
 
 	rec_offs_set_n_fields(offsets, n);
 
-	if (UNIV_UNLIKELY(alter_metadata) && index->table->not_redundant()) {
-#ifdef UNIV_DEBUG
-		memcpy(&offsets[RECORD_OFFSET], &rec, sizeof rec);
-		memcpy(&offsets[INDEX_OFFSET], &index, sizeof index);
-#endif /* UNIV_DEBUG */
+	if (UNIV_UNLIKELY(alter_metadata)
+	    && dict_table_is_comp(index->table)) {
+		ut_d(offsets[2] = ulint(rec));
+		ut_d(offsets[3] = ulint(index));
 		ut_ad(leaf);
 		ut_ad(index->is_dummy || index->table->instant);
 		ut_ad(index->is_dummy || index->is_instant());
@@ -935,13 +931,13 @@ rec_get_offsets_reverse(
 	const dict_index_t*	index,	/*!< in: record descriptor */
 	ulint			node_ptr,/*!< in: nonzero=node pointer,
 					0=leaf node */
-	rec_offs*		offsets)/*!< in/out: array consisting of
+	ulint*			offsets)/*!< in/out: array consisting of
 					offsets[0] allocated elements */
 {
 	ulint		n;
 	ulint		i;
-	rec_offs	offs;
-	rec_offs	any_ext = 0;
+	ulint		offs;
+	ulint		any_ext;
 	const byte*	nulls;
 	const byte*	lens;
 	dict_field_t*	field;
@@ -967,10 +963,11 @@ rec_get_offsets_reverse(
 	lens = nulls + UT_BITS_IN_BYTES(index->n_nullable);
 	i = offs = 0;
 	null_mask = 1;
+	any_ext = 0;
 
 	/* read the lengths of fields 0..n */
 	do {
-		rec_offs len;
+		ulint	len;
 		if (UNIV_UNLIKELY(i == n_node_ptr_field)) {
 			len = offs += REC_NODE_PTR_SIZE;
 			goto resolved;
@@ -991,7 +988,7 @@ rec_get_offsets_reverse(
 				We do not advance offs, and we set
 				the length to zero and enable the
 				SQL NULL flag in offsets[]. */
-				len = combine(offs, SQL_NULL);
+				len = offs | REC_OFFS_SQL_NULL;
 				goto resolved;
 			}
 			null_mask <<= 1;
@@ -1015,11 +1012,10 @@ rec_get_offsets_reverse(
 					len <<= 8;
 					len |= *lens++;
 
-					offs += get_value(len);
+					offs += len & 0x3fff;
 					if (UNIV_UNLIKELY(len & 0x4000)) {
 						any_ext = REC_OFFS_EXTERNAL;
-						len = combine(offs,
-							      STORED_OFFPAGE);
+						len = offs | REC_OFFS_EXTERNAL;
 					} else {
 						len = offs;
 					}
@@ -1030,16 +1026,15 @@ rec_get_offsets_reverse(
 
 			len = offs += len;
 		} else {
-			len = offs += static_cast<rec_offs>(field->fixed_len);
+			len = offs += field->fixed_len;
 		}
 resolved:
 		rec_offs_base(offsets)[i + 1] = len;
 	} while (++i < rec_offs_n_fields(offsets));
 
 	ut_ad(lens >= extra);
-	*rec_offs_base(offsets)
-		= static_cast<rec_offs>(lens - extra + REC_N_NEW_EXTRA_BYTES)
-		  | REC_OFFS_COMPACT | any_ext;
+	*rec_offs_base(offsets) = (ulint(lens - extra) + REC_N_NEW_EXTRA_BYTES)
+		| REC_OFFS_COMPACT | any_ext;
 }
 
 /************************************************************//**
@@ -1155,10 +1150,7 @@ rec_get_converted_size_comp_prefix_low(
 	for (ulint i = 0; dfield < end; i++, dfield++) {
 		if (mblob && i == index->first_user_field()) {
 			data_size += FIELD_REF_SIZE;
-			if (++dfield == end) {
-				ut_ad(i == index->n_fields);
-				break;
-			}
+			++dfield;
 		}
 
 		ulint len = dfield_get_len(dfield);
@@ -1600,11 +1592,7 @@ start:
 				ut_ad(dfield_is_ext(field));
 				memcpy(end, dfield_get_data(field), len);
 				end += len;
-				if (++field == fend) {
-					ut_ad(i == index->n_fields);
-					break;
-				}
-				len = dfield_get_len(field);
+				len = dfield_get_len(++field);
 			}
 		} else if (UNIV_UNLIKELY(i == n_node_ptr_field)) {
 			ut_ad(field->type.prtype & DATA_NOT_NULL);
@@ -1797,7 +1785,7 @@ void
 rec_init_offsets_temp(
 	const rec_t*		rec,
 	const dict_index_t*	index,
-	rec_offs*		offsets,
+	ulint*			offsets,
 	ulint			n_core,
 	const dict_col_t::def_t*def_val,
 	rec_comp_status_t	status)
@@ -1823,7 +1811,7 @@ void
 rec_init_offsets_temp(
 	const rec_t*		rec,
 	const dict_index_t*	index,
-	rec_offs*		offsets)
+	ulint*			offsets)
 {
 	ut_ad(!index->is_instant());
 	rec_init_offsets_comp_ordinary(rec, index, offsets,
@@ -1866,8 +1854,8 @@ rec_copy_prefix_to_dtuple(
 	ulint			n_fields,
 	mem_heap_t*		heap)
 {
-	rec_offs	offsets_[REC_OFFS_NORMAL_SIZE];
-	rec_offs*	offsets	= offsets_;
+	ulint	offsets_[REC_OFFS_NORMAL_SIZE];
+	ulint*	offsets	= offsets_;
 	rec_offs_init(offsets_);
 
 	ut_ad(is_leaf || n_fields
@@ -1933,7 +1921,7 @@ rec_copy_prefix_to_buf_old(
 		*buf = static_cast<byte*>(ut_malloc_nokey(prefix_len));
 	}
 
-	ut_memcpy(*buf, rec - area_start, prefix_len);
+	memcpy(*buf, rec - area_start, prefix_len);
 
 	copy_rec = *buf + area_start;
 
@@ -2165,7 +2153,7 @@ ibool
 rec_validate(
 /*=========*/
 	const rec_t*	rec,	/*!< in: physical record */
-	const rec_offs*	offsets)/*!< in: array returned by rec_get_offsets() */
+	const ulint*	offsets)/*!< in: array returned by rec_get_offsets() */
 {
 	ulint		len;
 	ulint		n_fields;
@@ -2274,7 +2262,7 @@ rec_print_comp(
 /*===========*/
 	FILE*		file,	/*!< in: file where to print */
 	const rec_t*	rec,	/*!< in: physical record */
-	const rec_offs*	offsets)/*!< in: array returned by rec_get_offsets() */
+	const ulint*	offsets)/*!< in: array returned by rec_get_offsets() */
 {
 	ulint	i;
 
@@ -2400,7 +2388,7 @@ rec_print_mbr_rec(
 /*==============*/
 	FILE*		file,	/*!< in: file where to print */
 	const rec_t*	rec,	/*!< in: physical record */
-	const rec_offs*	offsets)/*!< in: array returned by rec_get_offsets() */
+	const ulint*	offsets)/*!< in: array returned by rec_get_offsets() */
 {
 	ut_ad(rec_offs_validate(rec, NULL, offsets));
 	ut_ad(!rec_offs_any_default(offsets));
@@ -2468,7 +2456,7 @@ rec_print_new(
 /*==========*/
 	FILE*		file,	/*!< in: file where to print */
 	const rec_t*	rec,	/*!< in: physical record */
-	const rec_offs*	offsets)/*!< in: array returned by rec_get_offsets() */
+	const ulint*	offsets)/*!< in: array returned by rec_get_offsets() */
 {
 	ut_ad(rec_offs_validate(rec, NULL, offsets));
 
@@ -2508,7 +2496,7 @@ rec_print(
 		return;
 	} else {
 		mem_heap_t*	heap	= NULL;
-		rec_offs	offsets_[REC_OFFS_NORMAL_SIZE];
+		ulint		offsets_[REC_OFFS_NORMAL_SIZE];
 		rec_offs_init(offsets_);
 
 		rec_print_new(file, rec,
@@ -2531,7 +2519,7 @@ rec_print(
 	std::ostream&	o,
 	const rec_t*	rec,
 	ulint		info,
-	const rec_offs*	offsets)
+	const ulint*	offsets)
 {
 	const ulint	comp	= rec_offs_comp(offsets);
 	const ulint	n	= rec_offs_n_fields(offsets);
@@ -2588,7 +2576,7 @@ std::ostream&
 operator<<(std::ostream& o, const rec_index_print& r)
 {
 	mem_heap_t*	heap	= NULL;
-	rec_offs*	offsets	= rec_get_offsets(
+	ulint*		offsets	= rec_get_offsets(
 		r.m_rec, r.m_index, NULL, page_rec_is_leaf(r.m_rec),
 		ULINT_UNDEFINED, &heap);
 	rec_print(o, r.m_rec,
@@ -2624,9 +2612,9 @@ rec_get_trx_id(
 	const byte*	trx_id;
 	ulint		len;
 	mem_heap_t*	heap		= NULL;
-	rec_offs offsets_[REC_OFFS_HEADER_SIZE + MAX_REF_PARTS + 2];
+	ulint offsets_[REC_OFFS_HEADER_SIZE + MAX_REF_PARTS + 2];
 	rec_offs_init(offsets_);
-	rec_offs* offsets = offsets_;
+	ulint* offsets = offsets_;
 
 	offsets = rec_get_offsets(rec, index, offsets, true,
 				  index->db_trx_id() + 1, &heap);
@@ -2648,11 +2636,11 @@ rec_get_trx_id(
 @param[in]	n		nth field */
 void
 rec_offs_make_nth_extern(
-	rec_offs*	offsets,
+	ulint*		offsets,
 	const ulint	n)
 {
 	ut_ad(!rec_offs_nth_sql_null(offsets, n));
-	set_type(rec_offs_base(offsets)[1 + n], STORED_OFFPAGE);
+	rec_offs_base(offsets)[1 + n] |= REC_OFFS_EXTERNAL;
 }
 #ifdef WITH_WSREP
 # include "ha_prototypes.h"
@@ -2672,8 +2660,8 @@ wsrep_rec_get_foreign_key(
 	ulint		i;
 	uint            key_parts;
 	mem_heap_t*	heap	= NULL;
-	rec_offs	offsets_[REC_OFFS_NORMAL_SIZE];
-	const rec_offs* offsets;
+	ulint		offsets_[REC_OFFS_NORMAL_SIZE];
+        const ulint*    offsets;
 
 	ut_ad(index_for);
 	ut_ad(index_ref);
@@ -2739,7 +2727,7 @@ wsrep_rec_get_foreign_key(
 					}
 					data++;
 				}
-		
+
 				if (!(col_f->prtype & DATA_UNSIGNED)) {
 					buf[len-1] = (byte) (buf[len-1] ^ 128);
 				}
@@ -2751,7 +2739,7 @@ wsrep_rec_get_foreign_key(
 			case DATA_CHAR:
 			case DATA_MYSQL:
 				/* Copy the actual data */
-				ut_memcpy(buf, data, len);
+				memcpy(buf, data, len);
 				len = wsrep_innobase_mysql_sort(
 					(int)
 					(col_f->prtype & DATA_MYSQL_TYPE_MASK),
@@ -2763,7 +2751,7 @@ wsrep_rec_get_foreign_key(
 			case DATA_BINARY:
 				memcpy(buf, data, len);
 				break;
-			default: 
+			default:
 				break;
 			}
 
