@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1996, 2017, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2015, 2020, MariaDB Corporation.
+Copyright (c) 2015, 2019, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -214,7 +214,7 @@ row_upd_check_references_constraints(
 				cursor position is lost in this function! */
 	dict_table_t*	table,	/*!< in: table in question */
 	dict_index_t*	index,	/*!< in: index of the cursor */
-	rec_offs*	offsets,/*!< in/out: rec_get_offsets(pcur.rec, index) */
+	ulint*		offsets,/*!< in/out: rec_get_offsets(pcur.rec, index) */
 	que_thr_t*	thr,	/*!< in: query thread */
 	mtr_t*		mtr)	/*!< in: mtr */
 {
@@ -223,6 +223,7 @@ row_upd_check_references_constraints(
 	dtuple_t*	entry;
 	trx_t*		trx;
 	const rec_t*	rec;
+	ulint		n_ext;
 	dberr_t		err;
 	ibool		got_s_lock	= FALSE;
 
@@ -239,7 +240,7 @@ row_upd_check_references_constraints(
 
 	heap = mem_heap_create(500);
 
-	entry = row_rec_to_index_entry(rec, index, offsets, heap);
+	entry = row_rec_to_index_entry(rec, index, offsets, &n_ext, heap);
 
 	mtr_commit(mtr);
 
@@ -331,7 +332,7 @@ wsrep_row_upd_check_foreign_constraints(
 				cursor position is lost in this function! */
 	dict_table_t*	table,	/*!< in: table in question */
 	dict_index_t*	index,	/*!< in: index of the cursor */
-	rec_offs*	offsets,/*!< in/out: rec_get_offsets(pcur.rec, index) */
+	ulint*		offsets,/*!< in/out: rec_get_offsets(pcur.rec, index) */
 	que_thr_t*	thr,	/*!< in: query thread */
 	mtr_t*		mtr)	/*!< in: mtr */
 {
@@ -340,6 +341,7 @@ wsrep_row_upd_check_foreign_constraints(
 	dtuple_t*	entry;
 	trx_t*		trx;
 	const rec_t*	rec;
+	ulint		n_ext;
 	dberr_t		err;
 	ibool		got_s_lock	= FALSE;
 	ibool		opened     	= FALSE;
@@ -357,7 +359,8 @@ wsrep_row_upd_check_foreign_constraints(
 
 	heap = mem_heap_create(500);
 
-	entry = row_rec_to_index_entry(rec, index, offsets, heap);
+	entry = row_rec_to_index_entry(rec, index, offsets,
+				       &n_ext, heap);
 
 	mtr_commit(mtr);
 
@@ -432,7 +435,7 @@ func_exit:
 
 inline bool wsrep_must_process_fk(const upd_node_t* node, const trx_t* trx)
 {
-	if (!trx->is_wsrep()) {
+	if (!wsrep_on_trx(trx)) {
 		return false;
 	}
 	return que_node_get_type(node->common.parent) != QUE_NODE_UPDATE
@@ -470,7 +473,7 @@ row_upd_rec_sys_fields_in_recovery(
 /*===============================*/
 	rec_t*		rec,	/*!< in/out: record */
 	page_zip_des_t*	page_zip,/*!< in/out: compressed page, or NULL */
-	const rec_offs*	offsets,/*!< in: array returned by rec_get_offsets() */
+	const ulint*	offsets,/*!< in: array returned by rec_get_offsets() */
 	ulint		pos,	/*!< in: TRX_ID position in rec */
 	trx_id_t	trx_id,	/*!< in: transaction id */
 	roll_ptr_t	roll_ptr)/*!< in: roll ptr of the undo log record */
@@ -501,7 +504,7 @@ ibool
 row_upd_changes_field_size_or_external(
 /*===================================*/
 	dict_index_t*	index,	/*!< in: index */
-	const rec_offs*	offsets,/*!< in: rec_get_offsets(rec, index) */
+	const ulint*	offsets,/*!< in: rec_get_offsets(rec, index) */
 	const upd_t*	update)	/*!< in: update vector */
 {
 	const upd_field_t*	upd_field;
@@ -626,7 +629,7 @@ row_upd_rec_in_place(
 /*=================*/
 	rec_t*		rec,	/*!< in/out: record where replaced */
 	dict_index_t*	index,	/*!< in: the index the record belongs to */
-	const rec_offs*	offsets,/*!< in: array returned by rec_get_offsets() */
+	const ulint*	offsets,/*!< in: array returned by rec_get_offsets() */
 	const upd_t*	update,	/*!< in: update vector */
 	page_zip_des_t*	page_zip)/*!< in: compressed page with enough space
 				available, or NULL */
@@ -896,7 +899,7 @@ row_upd_build_sec_rec_difference_binary(
 /*====================================*/
 	const rec_t*	rec,	/*!< in: secondary index record */
 	dict_index_t*	index,	/*!< in: index */
-	const rec_offs*	offsets,/*!< in: rec_get_offsets(rec, index) */
+	const ulint*	offsets,/*!< in: rec_get_offsets(rec, index) */
 	const dtuple_t*	entry,	/*!< in: entry to insert */
 	mem_heap_t*	heap)	/*!< in: memory heap from which allocated */
 {
@@ -977,7 +980,7 @@ row_upd_build_difference_binary(
 	dict_index_t*	index,
 	const dtuple_t*	entry,
 	const rec_t*	rec,
-	const rec_offs*	offsets,
+	const ulint*	offsets,
 	bool		no_sys,
 	trx_t*		trx,
 	mem_heap_t*	heap,
@@ -987,7 +990,7 @@ row_upd_build_difference_binary(
 	ulint		len;
 	upd_t*		update;
 	ulint		n_diff;
-	rec_offs	offsets_[REC_OFFS_NORMAL_SIZE];
+	ulint		offsets_[REC_OFFS_NORMAL_SIZE];
 	const ulint	n_v_fld = dtuple_get_n_v_fields(entry);
 	rec_offs_init(offsets_);
 
@@ -1045,6 +1048,10 @@ row_upd_build_difference_binary(
 	for purge/mvcc purpose) */
 	if (n_v_fld > 0) {
 		row_ext_t*	ext;
+		mem_heap_t*	v_heap = NULL;
+		byte*		record;
+		VCOL_STORAGE*	vcol_storage;
+
 		THD*		thd;
 
 		if (trx == NULL) {
@@ -1055,8 +1062,9 @@ row_upd_build_difference_binary(
 
 		ut_ad(!update->old_vrow);
 
-		ib_vcol_row vc(NULL);
-		uchar *record = vc.record(thd, index, &mysql_table);
+		innobase_allocate_row_for_vcol(thd, index, &v_heap,
+					       &mysql_table,
+					       &record, &vcol_storage);
 
 		for (ulint i = 0; i < n_v_fld; i++) {
 			const dict_v_col_t*     col
@@ -1074,9 +1082,10 @@ row_upd_build_difference_binary(
 
 			dfield_t*	vfield = innobase_get_computed_value(
 				update->old_vrow, col, index,
-				&vc.heap, heap, NULL, thd, mysql_table, record,
+				&v_heap, heap, NULL, thd, mysql_table, record,
 				NULL, NULL, NULL);
 			if (vfield == NULL) {
+				if (v_heap) mem_heap_free(v_heap);
 				*error = DB_COMPUTE_VALUE_FAILED;
 				return(NULL);
 			}
@@ -1097,6 +1106,12 @@ row_upd_build_difference_binary(
 				upd_field_set_v_field_no(uf, i, index);
 			}
 		}
+
+		if (v_heap) {
+			if (vcol_storage)
+				innobase_free_row_for_vcol(vcol_storage);
+			mem_heap_free(v_heap);
+		}
 	}
 
 	update->n_fields = n_diff;
@@ -1115,9 +1130,7 @@ containing also the reference to the external part
 @param[in,out]	len		input - length of prefix to
 fetch; output: fetched length of the prefix
 @param[in,out]	heap		heap where to allocate
-@return BLOB prefix
-@retval NULL if the record is incomplete (should only happen
-in row_vers_vc_matches_cluster() executed concurrently with another purge) */
+@return BLOB prefix */
 static
 byte*
 row_upd_ext_fetch(
@@ -1132,7 +1145,10 @@ row_upd_ext_fetch(
 	*len = btr_copy_externally_stored_field_prefix(
 		buf, *len, zip_size, data, local_len);
 
-	return *len ? buf : NULL;
+	/* We should never update records containing a half-deleted BLOB. */
+	ut_a(*len);
+
+	return(buf);
 }
 
 /** Replaces the new column value stored in the update vector in
@@ -1143,11 +1159,9 @@ the given index entry field.
 @param[in]	uf		update field
 @param[in,out]	heap		memory heap for allocating and copying
 the new value
-@param[in]	zip_size	ROW_FORMAT=COMPRESSED page size, or 0
-@return whether the previous version was built successfully */
-MY_ATTRIBUTE((nonnull, warn_unused_result))
+@param[in]	zip_size	ROW_FORMAT=COMPRESSED page size, or 0 */
 static
-bool
+void
 row_upd_index_replace_new_col_val(
 	dfield_t*		dfield,
 	const dict_field_t*	field,
@@ -1162,13 +1176,13 @@ row_upd_index_replace_new_col_val(
 	dfield_copy_data(dfield, &uf->new_val);
 
 	if (dfield_is_null(dfield)) {
-		return true;
+		return;
 	}
 
 	len = dfield_get_len(dfield);
 	data = static_cast<const byte*>(dfield_get_data(dfield));
 
-	if (field->prefix_len > 0) {
+	if (field && field->prefix_len > 0) {
 		ibool		fetch_ext = dfield_is_ext(dfield)
 			&& len < (ulint) field->prefix_len
 			+ BTR_EXTERN_FIELD_REF_SIZE;
@@ -1180,9 +1194,6 @@ row_upd_index_replace_new_col_val(
 
 			data = row_upd_ext_fetch(data, l, zip_size,
 						 &len, heap);
-			if (UNIV_UNLIKELY(!data)) {
-				return false;
-			}
 		}
 
 		len = dtype_get_at_most_n_mbchars(col->prtype,
@@ -1196,7 +1207,7 @@ row_upd_index_replace_new_col_val(
 			dfield_dup(dfield, heap);
 		}
 
-		return true;
+		return;
 	}
 
 	switch (uf->orig_len) {
@@ -1235,8 +1246,6 @@ row_upd_index_replace_new_col_val(
 		dfield_set_ext(dfield);
 		break;
 	}
-
-	return true;
 }
 
 /** Apply an update vector to an metadata entry.
@@ -1283,11 +1292,8 @@ row_upd_index_replace_metadata(
 
 		f -= f > first;
 		const dict_field_t* field = dict_index_get_nth_field(index, f);
-		if (!row_upd_index_replace_new_col_val(dfield, field,
-						       field->col,
-						       uf, heap, zip_size)) {
-			ut_error;
-		}
+		row_upd_index_replace_new_col_val(dfield, field, field->col,
+						  uf, heap, zip_size);
 	}
 
 	ut_ad(found_mblob);
@@ -1338,57 +1344,68 @@ row_upd_index_replace_new_col_vals_index_pos(
 				update, i, false);
 		}
 
-		if (uf && UNIV_UNLIKELY(!row_upd_index_replace_new_col_val(
-						dtuple_get_nth_field(entry, i),
-						field, col, uf, heap,
-						zip_size))) {
-			ut_error;
+		if (uf) {
+			row_upd_index_replace_new_col_val(
+				dtuple_get_nth_field(entry, i),
+				field, col, uf, heap, zip_size);
 		}
 	}
 }
 
-/** Replace the new column values stored in the update vector,
-during trx_undo_prev_version_build().
-@param entry   clustered index tuple where the values are replaced
-               (the clustered index leaf page latch must be held)
-@param index   clustered index
-@param update  update vector for the clustered index
-@param heap    memory heap for allocating and copying values
-@return whether the previous version was built successfully */
-bool
-row_upd_index_replace_new_col_vals(dtuple_t *entry, const dict_index_t &index,
-                                   const upd_t *update, mem_heap_t *heap)
+/***********************************************************//**
+Replaces the new column values stored in the update vector to the index entry
+given. */
+void
+row_upd_index_replace_new_col_vals(
+/*===============================*/
+	dtuple_t*	entry,	/*!< in/out: index entry where replaced;
+				the clustered index record must be
+				covered by a lock or a page latch to
+				prevent deletion (rollback or purge) */
+	dict_index_t*	index,	/*!< in: index; NOTE that this may also be a
+				non-clustered index */
+	const upd_t*	update,	/*!< in: an update vector built for the
+				CLUSTERED index so that the field number in
+				an upd_field is the clustered index position */
+	mem_heap_t*	heap)	/*!< in: memory heap for allocating and
+				copying the new values */
 {
-  ut_ad(index.is_primary());
-  const ulint zip_size= index.table->space->zip_size();
+	ulint			i;
+	const dict_index_t*	clust_index
+		= dict_table_get_first_index(index->table);
+	const ulint zip_size = index->table->space->zip_size();
 
-  ut_ad(!index.table->skip_alter_undo);
-  dtuple_set_info_bits(entry, update->info_bits);
+	ut_ad(!index->table->skip_alter_undo);
 
-  for (ulint i= 0; i < index.n_fields; i++)
-  {
-   const dict_field_t *field= &index.fields[i];
-   const dict_col_t* col= dict_field_get_col(field);
-   const upd_field_t *uf;
+	dtuple_set_info_bits(entry, update->info_bits);
 
-   if (col->is_virtual())
-   {
-     const dict_v_col_t *vcol= reinterpret_cast<const dict_v_col_t*>(col);
-     uf= upd_get_field_by_field_no(update, vcol->v_pos, true);
-   }
-   else
-     uf= upd_get_field_by_field_no(update, dict_col_get_clust_pos(col, &index),
-                                   false);
+	for (i = 0; i < dict_index_get_n_fields(index); i++) {
+		const dict_field_t*	field;
+		const dict_col_t*	col;
+		const upd_field_t*	uf;
 
-   if (!uf)
-     continue;
+		field = dict_index_get_nth_field(index, i);
+		col = dict_field_get_col(field);
+		if (col->is_virtual()) {
+			const dict_v_col_t*	vcol = reinterpret_cast<
+							const dict_v_col_t*>(
+								col);
 
-   if (!row_upd_index_replace_new_col_val(dtuple_get_nth_field(entry, i),
-                                          field, col, uf, heap, zip_size))
-     return false;
-  }
+			uf = upd_get_field_by_field_no(
+				update, vcol->v_pos, true);
+		} else {
+			uf = upd_get_field_by_field_no(
+				update,
+				dict_col_get_clust_pos(col, clust_index),
+				false);
+		}
 
-  return true;
+		if (uf) {
+			row_upd_index_replace_new_col_val(
+				dtuple_get_nth_field(entry, i),
+				field, col, uf, heap, zip_size);
+		}
+	}
 }
 
 /** Replaces the virtual column values stored in the update vector.
@@ -1835,7 +1852,7 @@ row_upd_changes_ord_field_binary_func(
 			/* Silence a compiler warning without
 			silencing a Valgrind error. */
 			dfield_len = 0;
-			MEM_UNDEFINED(&dfield_len, sizeof dfield_len);
+			UNIV_MEM_INVALID(&dfield_len, sizeof dfield_len);
 			/* See if the column is stored externally. */
 			buf = row_ext_lookup(ext, col_no, &dfield_len);
 
@@ -1850,8 +1867,7 @@ row_upd_changes_ord_field_binary_func(
 					when the server had crashed before
 					storing the field. */
 					ut_ad(thr->graph->trx->is_recovered);
-					ut_ad(thr->graph->trx
-					      == trx_roll_crash_recv_trx);
+					ut_ad(trx_is_recv(thr->graph->trx));
 					return(TRUE);
 				}
 
@@ -2036,7 +2052,7 @@ void
 row_upd_copy_columns(
 /*=================*/
 	rec_t*		rec,	/*!< in: record in a clustered index */
-	const rec_offs*	offsets,/*!< in: array returned by rec_get_offsets() */
+	const ulint*	offsets,/*!< in: array returned by rec_get_offsets() */
 	const dict_index_t*	index, /*!< in: index of rec */
 	sym_node_t*	column)	/*!< in: first column in a column list, or
 				NULL */
@@ -2087,19 +2103,23 @@ row_upd_eval_new_vals(
 @param[in,out]	node		row update node
 @param[in]	update		an update vector if it is update
 @param[in]	thd		mysql thread handle
-@param[in,out]	mysql_table	mysql table object
-@return true if success
-	false if virtual column value computation fails. */
+@param[in,out]	mysql_table	mysql table object */
 static
-bool
+void
 row_upd_store_v_row(
 	upd_node_t*	node,
 	const upd_t*	update,
 	THD*		thd,
 	TABLE*		mysql_table)
 {
+	mem_heap_t*	heap = NULL;
 	dict_index_t*	index = dict_table_get_first_index(node->table);
-	ib_vcol_row	vc(NULL);
+        byte*           record= 0;
+	VCOL_STORAGE	*vcol_storage= 0;
+
+	if (!update)
+	  innobase_allocate_row_for_vcol(thd, index, &heap, &mysql_table,
+					 &record, &vcol_storage);
 
 	for (ulint col_no = 0; col_no < dict_table_get_n_v_cols(node->table);
 	     col_no++) {
@@ -2149,37 +2169,32 @@ row_upd_store_v_row(
 						dfield_dup(dfield, node->heap);
 					}
 				} else {
-					uchar *record = vc.record(thd, index,
-								  &mysql_table);
 					/* Need to compute, this happens when
 					deleting row */
-					dfield_t* vfield =
-						innobase_get_computed_value(
-							node->row, col, index,
-							&vc.heap, node->heap,
-							NULL, thd, mysql_table,
-							record, NULL, NULL,
-							NULL);
-					if (vfield == NULL) {
-						return false;
-					}
+					innobase_get_computed_value(
+						node->row, col, index,
+						&heap, node->heap, NULL,
+						thd, mysql_table, record, NULL,
+						NULL, NULL);
 				}
 			}
 		}
 	}
 
-	return true;
+	if (heap) {
+		if (vcol_storage)
+			innobase_free_row_for_vcol(vcol_storage);
+		mem_heap_free(heap);
+	}
+
 }
 
 /** Stores to the heap the row on which the node->pcur is positioned.
 @param[in]	node		row update node
 @param[in]	thd		mysql thread handle
 @param[in,out]	mysql_table	NULL, or mysql table object when
-				user thread invokes dml
-@return false if virtual column value computation fails
-	true otherwise. */
-static
-bool
+				user thread invokes dml */
+void
 row_upd_store_row(
 	upd_node_t*	node,
 	THD*		thd,
@@ -2189,8 +2204,8 @@ row_upd_store_row(
 	rec_t*		rec;
 	mem_heap_t*	heap		= NULL;
 	row_ext_t**	ext;
-	rec_offs	offsets_[REC_OFFS_NORMAL_SIZE];
-	const rec_offs*	offsets;
+	ulint		offsets_[REC_OFFS_NORMAL_SIZE];
+	const ulint*	offsets;
 	rec_offs_init(offsets_);
 
 	ut_ad(node->pcur->latch_mode != BTR_NO_LATCHES);
@@ -2223,12 +2238,8 @@ row_upd_store_row(
 			      NULL, NULL, NULL, ext, node->heap);
 
 	if (node->table->n_v_cols) {
-		bool ok = row_upd_store_v_row(node,
-				    node->is_delete ? NULL : node->update,
+		row_upd_store_v_row(node, node->is_delete ? NULL : node->update,
 				    thd, mysql_table);
-		if (!ok) {
-			return false;
-		}
 	}
 
 	if (node->is_delete == PLAIN_DELETE) {
@@ -2243,7 +2254,6 @@ row_upd_store_row(
 	if (UNIV_LIKELY_NULL(heap)) {
 		mem_heap_free(heap);
 	}
-	return true;
 }
 
 /***********************************************************//**
@@ -2412,7 +2422,7 @@ row_upd_sec_index_entry(
 #ifdef UNIV_DEBUG
 		mtr_commit(&mtr);
 		mtr_start(&mtr);
-		ut_ad(btr_validate_index(index, 0) == DB_SUCCESS);
+		ut_ad(btr_validate_index(index, 0));
 		ut_ad(0);
 #endif /* UNIV_DEBUG */
 		break;
@@ -2434,7 +2444,7 @@ row_upd_sec_index_entry(
 			    && wsrep_must_process_fk(node, trx)
 			    && !wsrep_thd_is_BF(trx->mysql_thd, FALSE)) {
 
-				rec_offs* offsets = rec_get_offsets(
+				ulint*	offsets = rec_get_offsets(
 					rec, index, NULL, true,
 					ULINT_UNDEFINED, &heap);
 
@@ -2448,7 +2458,7 @@ row_upd_sec_index_entry(
 					err = DB_SUCCESS;
 					break;
 				case DB_LOCK_WAIT:
-					if (UNIV_UNLIKELY(wsrep_debug)) {
+					if (wsrep_debug) {
 						ib::warn() << "WSREP: sec index FK lock wait"
 							   << " index " << index->name
 							   << " table " << index->table->name
@@ -2456,7 +2466,7 @@ row_upd_sec_index_entry(
 					}
 					break;
 				case DB_DEADLOCK:
-					if (UNIV_UNLIKELY(wsrep_debug)) {
+					if (wsrep_get_debug()) {
 						ib::warn() << "WSREP: sec index FK check fail for deadlock"
 							   << " index " << index->name
 							   << " table " << index->table->name
@@ -2464,7 +2474,7 @@ row_upd_sec_index_entry(
 					}
 					break;
 				default:
-					ib::error() << "WSREP: referenced FK check fail: " << err
+					ib::error() << "WSREP: referenced FK check fail: " << ut_strerr(err)
 						    << " index " << index->name
 						    << " table " << index->table->name
 						    << " query " << wsrep_thd_query(trx->mysql_thd);
@@ -2479,7 +2489,7 @@ row_upd_sec_index_entry(
 
 		if (referenced) {
 
-			rec_offs* offsets;
+			ulint*	offsets;
 
 			offsets = rec_get_offsets(
 				rec, index, NULL, true, ULINT_UNDEFINED,
@@ -2562,7 +2572,7 @@ row_upd_clust_rec_by_insert_inherit_func(
 	const rec_t*	rec,	/*!< in: old record, or NULL */
 #ifdef UNIV_DEBUG
 	dict_index_t*	index,	/*!< in: index, or NULL */
-	const rec_offs*	offsets,/*!< in: rec_get_offsets(rec), or NULL */
+	const ulint*	offsets,/*!< in: rec_get_offsets(rec), or NULL */
 #endif /* UNIV_DEBUG */
 	dtuple_t*	entry,	/*!< in/out: updated entry to be
 				inserted into the clustered index */
@@ -2668,13 +2678,10 @@ row_upd_clust_rec_by_insert(
 	dtuple_t*	entry;
 	dberr_t		err;
 	rec_t*		rec;
-	rec_offs	offsets_[REC_OFFS_NORMAL_SIZE];
-	rec_offs*	offsets			= offsets_;
+	ulint*		offsets			= NULL;
 
 	ut_ad(dict_index_is_clust(index));
 
-	rec_offs_init(offsets_);
-
 	trx = thr_get_trx(thr);
 	table = node->table;
 	pcur = node->pcur;
@@ -2707,7 +2714,7 @@ row_upd_clust_rec_by_insert(
 		we update the primary key.  Delete-mark the old record
 		in the clustered index and prepare to insert a new entry. */
 		rec = btr_cur_get_rec(btr_cur);
-		offsets = rec_get_offsets(rec, index, offsets, true,
+		offsets = rec_get_offsets(rec, index, NULL, true,
 					  ULINT_UNDEFINED, &heap);
 		ut_ad(page_rec_is_user_rec(rec));
 
@@ -2775,14 +2782,14 @@ check_fk:
 				err = DB_SUCCESS;
 				break;
 			case DB_DEADLOCK:
-				if (UNIV_UNLIKELY(wsrep_debug)) {
+				if (wsrep_get_debug()) {
 					ib::warn() << "WSREP: sec index FK check fail for deadlock"
 						   << " index " << index->name
 						   << " table " << index->table->name;
 				}
 				goto err_exit;
 			default:
-				ib::error() << "WSREP: referenced FK check fail: " << err
+				ib::error() << "WSREP: referenced FK check fail: " << ut_strerr(err)
 					    << " index " << index->name
 					    << " table " << index->table->name;
 				goto err_exit;
@@ -2793,8 +2800,9 @@ check_fk:
 
 	mtr_commit(mtr);
 
-	err = row_ins_clust_index_entry(index, entry, thr,
-					dtuple_get_n_ext(entry));
+	err = row_ins_clust_index_entry(
+		index, entry, thr,
+		node->upd_ext ? node->upd_ext->n_ext : 0);
 	node->state = UPD_NODE_INSERT_CLUSTERED;
 
 	mem_heap_free(heap);
@@ -2814,7 +2822,7 @@ row_upd_clust_rec(
 	ulint		flags,  /*!< in: undo logging and locking flags */
 	upd_node_t*	node,	/*!< in: row update node */
 	dict_index_t*	index,	/*!< in: clustered index */
-	rec_offs*	offsets,/*!< in: rec_get_offsets() on node->pcur */
+	ulint*		offsets,/*!< in: rec_get_offsets() on node->pcur */
 	mem_heap_t**	offsets_heap,
 				/*!< in/out: memory heap, can be emptied */
 	que_thr_t*	thr,	/*!< in: query thread */
@@ -2945,7 +2953,7 @@ row_upd_del_mark_clust_rec(
 /*=======================*/
 	upd_node_t*	node,	/*!< in: row update node */
 	dict_index_t*	index,	/*!< in: clustered index */
-	rec_offs*	offsets,/*!< in/out: rec_get_offsets() for the
+	ulint*		offsets,/*!< in/out: rec_get_offsets() for the
 				record under the cursor */
 	que_thr_t*	thr,	/*!< in: query thread */
 	ibool		referenced,
@@ -2971,12 +2979,9 @@ row_upd_del_mark_clust_rec(
 	/* Store row because we have to build also the secondary index
 	entries */
 
-	if (!row_upd_store_row(node, trx->mysql_thd,
+	row_upd_store_row(node, trx->mysql_thd,
 			  thr->prebuilt  && thr->prebuilt->table == node->table
-			  ? thr->prebuilt->m_mysql_table : NULL)) {
-		err = DB_COMPUTE_VALUE_FAILED;
-		return err;
-	}
+			  ? thr->prebuilt->m_mysql_table : NULL);
 
 	/* Mark the clustered index record deleted; we do not have to check
 	locks, because we assume that we have an x-lock on the record */
@@ -3004,14 +3009,14 @@ row_upd_del_mark_clust_rec(
 			err = DB_SUCCESS;
 			break;
 		case DB_DEADLOCK:
-			if (UNIV_UNLIKELY(wsrep_debug)) {
+			if (wsrep_get_debug()) {
 				ib::warn() << "WSREP: sec index FK check fail for deadlock"
 					   << " index " << index->name
 					   << " table " << index->table->name;
 			}
 			break;
 		default:
-			ib::error() << "WSREP: referenced FK check fail: " << err
+			ib::error() << "WSREP: referenced FK check fail: " << ut_strerr(err)
 				    << " index " << index->name
 				    << " table " << index->table->name;
 
@@ -3043,8 +3048,8 @@ row_upd_clust_step(
 	mtr_t		mtr;
 	rec_t*		rec;
 	mem_heap_t*	heap	= NULL;
-	rec_offs	offsets_[REC_OFFS_NORMAL_SIZE];
-	rec_offs*	offsets;
+	ulint		offsets_[REC_OFFS_NORMAL_SIZE];
+	ulint*		offsets;
 	ibool		referenced;
 	ulint		flags;
 	trx_t*		trx = thr_get_trx(thr);
@@ -3194,11 +3199,8 @@ row_upd_clust_step(
 		goto exit_func;
 	}
 
-	if(!row_upd_store_row(node, trx->mysql_thd,
-			thr->prebuilt ? thr->prebuilt->m_mysql_table : NULL)) {
-		err = DB_COMPUTE_VALUE_FAILED;
-		goto exit_func;
-	}
+	row_upd_store_row(node, trx->mysql_thd,
+			  thr->prebuilt ? thr->prebuilt->m_mysql_table : NULL);
 
 	if (row_upd_changes_ord_field_binary(index, node->update, thr,
 					     node->row, node->ext)) {
@@ -3474,65 +3476,32 @@ Supposed to be called only by make_versioned_update() and
 make_versioned_delete().
 @param[in]	trx	transaction
 @param[in]	vers_sys_idx	table->row_start or table->row_end */
-void upd_node_t::vers_update_fields(const trx_t *trx, ulint idx)
+void upd_node_t::make_versioned_helper(const trx_t* trx, ulint idx)
 {
 	ut_ad(in_mysql_interface); // otherwise needs to recalculate
 				   // node->cmpl_info
 	ut_ad(idx == table->vers_start || idx == table->vers_end);
 
 	dict_index_t* clust_index = dict_table_get_first_index(table);
-        const dict_col_t *col= dict_table_get_nth_col(table, idx);
-        ulint field_no= dict_col_get_clust_pos(col, clust_index);
-        upd_field_t *ufield;
 
-        for (ulint i= 0; i < update->n_fields; ++i)
-        {
-          if (update->fields[i].field_no == field_no)
-          {
-            ufield= &update->fields[i];
-            goto skip_append;
-          }
-        }
-
-        /* row_create_update_node_for_mysql() pre-allocated this much.
+	/* row_create_update_node_for_mysql() pre-allocated this much.
 	   At least one PK column always remains unchanged. */
 	ut_ad(update->n_fields < ulint(table->n_cols + table->n_v_cols));
 
 	update->n_fields++;
-        ufield= upd_get_nth_field(update, update->n_fields - 1);
-        upd_field_set_field_no(ufield, field_no, clust_index);
+	upd_field_t* ufield = upd_get_nth_field(update, update->n_fields - 1);
+	const dict_col_t* col = dict_table_get_nth_col(table, idx);
 
-skip_append:
-  char *where= reinterpret_cast<char *>(update->vers_sys_value);
-  if (col->vers_native())
-  {
-    mach_write_to_8(where, trx->id);
-  }
-  else
-  {
-    thd_get_query_start_data(trx->mysql_thd, where);
-  }
+	upd_field_set_field_no(ufield, dict_col_get_clust_pos(col, clust_index),
+			       clust_index);
 
-  dfield_set_data(&ufield->new_val, update->vers_sys_value, col->len);
+	char* where = reinterpret_cast<char*>(update->vers_sys_value);
+	if (col->vers_native()) {
+		mach_write_to_8(where, trx->id);
+	} else {
+		thd_get_query_start_data(trx->mysql_thd, where);
+	}
 
-  for (ulint col_no= 0; col_no < dict_table_get_n_v_cols(table); col_no++)
-  {
-
-    const dict_v_col_t *v_col= dict_table_get_nth_v_col(table, col_no);
-    if (!v_col->m_col.ord_part)
-      continue;
-    for (ulint i= 0; i < unsigned(v_col->num_base); i++)
-    {
-      dict_col_t *base_col= v_col->base_col[i];
-      if (base_col->ind == col->ind)
-      {
-        /* Virtual column depends on system field value
-        which we updated above. Remove it from update
-        vector, so it is recalculated in
-        row_upd_store_v_row() (see !update branch). */
-        update->remove(v_col->v_pos);
-        break;
-      }
-    }
-  }
+	dfield_set_data(&ufield->new_val, update->vers_sys_value, col->len);
 }
+
