@@ -417,7 +417,7 @@ static monitor_info_t	innodb_counter_info[] =
 	 MONITOR_DEFAULT_START, MONITOR_FLUSH_AVG_TIME},
 
 	{"buffer_flush_adaptive_avg_pass", "buffer",
-	 "Number of adaptive flushes passed during the recent Avg period.",
+	 "Numner of adaptive flushes passed during the recent Avg period.",
 	 MONITOR_NONE,
 	 MONITOR_DEFAULT_START, MONITOR_FLUSH_ADAPTIVE_AVG_PASS},
 
@@ -1188,11 +1188,6 @@ static monitor_info_t	innodb_counter_info[] =
 	 MONITOR_NONE,
 	 MONITOR_DEFAULT_START, MONITOR_SRV_BACKGROUND_DROP_TABLE_MICROSECOND},
 
-	{"innodb_ibuf_merge_usec", "server",
-	 "Time (in microseconds) spent to process change buffer merge",
-	 MONITOR_NONE,
-	 MONITOR_DEFAULT_START, MONITOR_SRV_IBUF_MERGE_MICROSECOND},
-
 	{"innodb_log_flush_usec", "server",
 	 "Time (in microseconds) spent to flush log records",
 	 MONITOR_NONE,
@@ -1951,35 +1946,35 @@ srv_mon_process_existing_counter(
 		break;
 
 	case MONITOR_OVLD_IBUF_MERGE_INSERT:
-		value = ibuf->n_merged_ops[IBUF_OP_INSERT];
+		value = ibuf.n_merged_ops[IBUF_OP_INSERT];
 		break;
 
 	case MONITOR_OVLD_IBUF_MERGE_DELETE:
-		value = ibuf->n_merged_ops[IBUF_OP_DELETE_MARK];
+		value = ibuf.n_merged_ops[IBUF_OP_DELETE_MARK];
 		break;
 
 	case MONITOR_OVLD_IBUF_MERGE_PURGE:
-		value = ibuf->n_merged_ops[IBUF_OP_DELETE];
+		value = ibuf.n_merged_ops[IBUF_OP_DELETE];
 		break;
 
 	case MONITOR_OVLD_IBUF_MERGE_DISCARD_INSERT:
-		value = ibuf->n_discarded_ops[IBUF_OP_INSERT];
+		value = ibuf.n_discarded_ops[IBUF_OP_INSERT];
 		break;
 
 	case MONITOR_OVLD_IBUF_MERGE_DISCARD_DELETE:
-		value = ibuf->n_discarded_ops[IBUF_OP_DELETE_MARK];
+		value = ibuf.n_discarded_ops[IBUF_OP_DELETE_MARK];
 		break;
 
 	case MONITOR_OVLD_IBUF_MERGE_DISCARD_PURGE:
-		value = ibuf->n_discarded_ops[IBUF_OP_DELETE];
+		value = ibuf.n_discarded_ops[IBUF_OP_DELETE];
 		break;
 
 	case MONITOR_OVLD_IBUF_MERGES:
-		value = ibuf->n_merges;
+		value = ibuf.n_merges;
 		break;
 
 	case MONITOR_OVLD_IBUF_SIZE:
-		value = ibuf->size;
+		value = ibuf.size;
 		break;
 
 	case MONITOR_OVLD_SERVER_ACTIVITY:
