@@ -75,7 +75,7 @@ int maria_create(const char *name, enum data_file_type datafile_type,
   uint max_field_lengths, extra_header_size, column_nr;
   uint internal_table= flags & HA_CREATE_INTERNAL_TABLE;
   ulong reclength, real_reclength,min_pack_length;
-  char kfilename[FN_REFLEN], klinkname[FN_REFLEN], *klinkname_ptr= NullS;
+  char kfilename[FN_REFLEN], klinkname[FN_REFLEN], *klinkname_ptr= 0;
   char dfilename[FN_REFLEN], dlinkname[FN_REFLEN], *dlinkname_ptr= 0;
   ulong pack_reclength;
   ulonglong tot_length,max_rows, tmp;
@@ -95,7 +95,7 @@ int maria_create(const char *name, enum data_file_type datafile_type,
   my_bool forced_packed;
   myf     sync_dir=  0;
   uchar   *log_data= NULL;
-  my_bool encrypted= maria_encrypt_tables && datafile_type == BLOCK_RECORD;
+  my_bool encrypted= ci->encrypted && datafile_type == BLOCK_RECORD;
   my_bool insert_order= MY_TEST(flags & HA_PRESERVE_INSERT_ORDER);
   uint crypt_page_header_space= 0;
   DBUG_ENTER("maria_create");
@@ -318,7 +318,7 @@ int maria_create(const char *name, enum data_file_type datafile_type,
   {
     options|= HA_OPTION_TMP_TABLE;
     tmp_table= TRUE;
-    create_mode|= O_NOFOLLOW | (internal_table ? 0 : O_EXCL);
+    create_mode|= O_NOFOLLOW;
     /* "CREATE TEMPORARY" tables are not crash-safe (dropped at restart) */
     ci->transactional= FALSE;
     flags&= ~HA_CREATE_PAGE_CHECKSUM;
@@ -328,6 +328,8 @@ int maria_create(const char *name, enum data_file_type datafile_type,
   share.base.born_transactional= ci->transactional;
   share.base.max_field_lengths= max_field_lengths;
   share.base.field_offsets= 0;                  /* for future */
+  share.base.compression_algorithm= ci->compression_algorithm;
+  share.base.s3_block_size=         ci->s3_block_size;
 
   if (flags & HA_CREATE_CHECKSUM || (options & HA_OPTION_CHECKSUM))
   {
@@ -827,6 +829,11 @@ int maria_create(const char *name, enum data_file_type datafile_type,
     */
     share.state.skip_redo_lsn= share.state.is_of_horizon=
       share.state.create_rename_lsn= LSN_MAX;
+    /*
+      We have to mark the table as not movable as the table will contain the
+      maria_uuid and create_rename_lsn
+    */
+    share.state.changed|= STATE_NOT_MOVABLE;
   }
 
   if (datafile_type == DYNAMIC_RECORD)
@@ -886,9 +893,10 @@ int maria_create(const char *name, enum data_file_type datafile_type,
   {
     char *iext= strrchr(name, '.');
     int have_iext= iext && !strcmp(iext, MARIA_NAME_IEXT);
-    fn_format(kfilename, name, "", MARIA_NAME_IEXT, MY_UNPACK_FILENAME |
-              (internal_table ? 0 : MY_RETURN_REAL_PATH) |
+    fn_format(kfilename, name, "", MARIA_NAME_IEXT,
+              MY_UNPACK_FILENAME | MY_RETURN_REAL_PATH |
               (have_iext ? MY_REPLACE_EXT : MY_APPEND_EXT));
+    klinkname_ptr= NullS;
     /*
       Replace the current file.
       Don't sync dir now if the data file has the same path.
@@ -1153,7 +1161,7 @@ int maria_create(const char *name, enum data_file_type datafile_type,
     DBUG_EXECUTE_IF("maria_crash_create_table",
                     {
                       DBUG_PRINT("maria_crash_create_table", ("now"));
-                      DBUG_ABORT();
+                      DBUG_SUICIDE();
                     });
     /*
       store LSN into file, needed for Recovery to not be confused if a
@@ -1431,7 +1439,7 @@ int _ma_update_state_lsns(MARIA_SHARE *share, LSN lsn, TrID create_trid,
      @retval 1      error (disk problem)
 */
 
-#if (_MSC_VER == 1310)
+#if defined(_MSC_VER) && (_MSC_VER == 1310)
 /*
  Visual Studio 2003 compiler produces internal compiler error
  in this function. Disable optimizations to workaround.
@@ -1445,6 +1453,7 @@ int _ma_update_state_lsns_sub(MARIA_SHARE *share, LSN lsn, TrID create_trid,
   uchar buf[LSN_STORE_SIZE * 3], *ptr;
   uchar trid_buff[8];
   File file= share->kfile.file;
+  DBUG_ENTER("_ma_update_state_lsns_sub");
   DBUG_ASSERT(file >= 0);
 
   if (lsn == LSN_IMPOSSIBLE)
@@ -1463,7 +1472,7 @@ int _ma_update_state_lsns_sub(MARIA_SHARE *share, LSN lsn, TrID create_trid,
                                               0].length,
                                     sizeof(log_array)/sizeof(log_array[0]),
                                     log_array, NULL, NULL)))
-      return res;
+      DBUG_RETURN(res);
   }
 
   for (ptr= buf; ptr < (buf + sizeof(buf)); ptr+= LSN_STORE_SIZE)
@@ -1496,14 +1505,14 @@ int _ma_update_state_lsns_sub(MARIA_SHARE *share, LSN lsn, TrID create_trid,
   }
   else
     lsn_store(buf, share->state.create_rename_lsn);
-  return (my_pwrite(file, buf, sizeof(buf),
-                    sizeof(share->state.header) +
-                    MARIA_FILE_CREATE_RENAME_LSN_OFFSET, MYF(MY_NABP)) ||
-          my_pwrite(file, trid_buff, sizeof(trid_buff),
-                    sizeof(share->state.header) +
-                    MARIA_FILE_CREATE_TRID_OFFSET, MYF(MY_NABP)) ||
-          (do_sync && mysql_file_sync(file, MYF(0))));
+  DBUG_RETURN(my_pwrite(file, buf, sizeof(buf),
+                        sizeof(share->state.header) +
+                        MARIA_FILE_CREATE_RENAME_LSN_OFFSET, MYF(MY_NABP)) ||
+              my_pwrite(file, trid_buff, sizeof(trid_buff),
+                        sizeof(share->state.header) +
+                        MARIA_FILE_CREATE_TRID_OFFSET, MYF(MY_NABP)) ||
+              (do_sync && mysql_file_sync(file, MYF(0))));
 }
-#if (_MSC_VER == 1310)
+#if defined(_MSC_VER) && (_MSC_VER == 1310)
 #pragma optimize("",on)
 #endif /*VS2003 compiler bug workaround*/
