@@ -19,12 +19,12 @@
 #include <myisamchk.h>
 #include <my_bit.h>
 #include <m_ctype.h>
-#include <stdarg.h>
 #include <my_getopt.h>
 #include <my_check_opt.h>
-#ifdef HAVE_SYS_MMAN_H
-#include <sys/mman.h>
-#endif
+#include <my_handler_errors.h>
+/* Remove next line if you want aria_chk to produce a stack trace */
+#undef HAVE_BACKTRACE
+#include <my_stacktrace.h>
 
 static uint decode_bits;
 static char **default_argv;
@@ -93,25 +93,41 @@ ATTRIBUTE_NORETURN static void my_exit(int exit_code);
 
 HA_CHECK check_param;
 
+/*
+  Register handler error messages for usage with my_error()
+
+  NOTES
+    This is safe to call multiple times as my_error_register()
+    will ignore calls to register already registered error numbers.
+*/
+
+static const char **get_handler_error_messages(int e __attribute__((unused)))
+{
+  return handler_error_messages;
+}
+
+
 /* Free memory and exit */
 
 static void my_exit(int exit_code)
 {
   free_tmpdir(&maria_chk_tmpdir);
   free_defaults(default_argv);
+  my_error_unregister(HA_ERR_FIRST,
+                      HA_ERR_FIRST+ array_elements(handler_error_messages)-1);
   my_end(check_param.testflag & T_INFO ?
          MY_CHECK_ERROR | MY_GIVE_INFO : MY_CHECK_ERROR);
   exit(exit_code);
 }
-  
 
-	/* Main program */
+/* Main program */
 
 int main(int argc, char **argv)
 {
   int error;
   MY_INIT(argv[0]);
 
+  my_init_stacktrace(1);
   default_log_dir= opt_log_dir= maria_data_root= (char *)".";
   maria_chk_init(&check_param);
   check_param.opt_lock_memory= 1;		/* Lock memory if possible */
@@ -120,6 +136,8 @@ int main(int argc, char **argv)
   maria_quick_table_bits=decode_bits;
   error=0;
   maria_init();
+  my_error_register(get_handler_error_messages, HA_ERR_FIRST,
+                    HA_ERR_FIRST+ array_elements(handler_error_messages)-1);
 
   maria_block_size= 0;                 /* Use block size from control file */
   if (!opt_ignore_control_file &&
@@ -160,6 +178,20 @@ int main(int argc, char **argv)
       check_param.testflag&= ~T_REP;
     fflush(stdout);
     fflush(stderr);
+    if (check_param.wrong_trd_printed &&
+        (check_param.testflag & T_FORCE_CREATE) &&
+        !(check_param.error_printed | check_param.warning_printed))
+    {
+      /* Only wrong create_trd. Run zerofill */
+      ulonglong old_testflag= check_param.testflag;
+      check_param.testflag= T_ZEROFILL;
+      error|= maria_chk(&check_param, argv[-1]);
+      check_param.testflag= old_testflag;
+      check_param.error_printed= 0;
+      check_param.warning_printed= 0;
+      fflush(stdout);
+      fflush(stderr);
+    }
     if ((check_param.error_printed | check_param.warning_printed) &&
 	(check_param.testflag & T_FORCE_CREATE) &&
 	(!(check_param.testflag & (T_REP | T_REP_BY_SORT | T_SORT_RECORDS |
@@ -174,7 +206,6 @@ int main(int argc, char **argv)
       fflush(stdout);
       fflush(stderr);
     }
-    else
       error|=new_error;
     if (argc && (!(check_param.testflag & T_SILENT) ||
                  check_param.testflag & T_INFO))
@@ -458,7 +489,7 @@ static void usage(void)
   --ignore-control-file  Don't open the control file. Only use this if you\n\
                          are sure the tables are not in use by another\n\
                          program!\n\
-  --require-control-file  Abort if we can't find/read the maria_log_control\n\
+  --require-control-file  Abort if we can't find/read the aria_log_control\n\
                           file\n\
   -s, --silent	      Only print errors.  One can use two -s to make\n\
 		      maria_chk very silent.\n\
@@ -602,11 +633,10 @@ TYPELIB maria_stats_method_typelib= {
 	 /* Read options */
 
 static my_bool
-get_one_option(int optid,
-	       const struct my_option *opt __attribute__((unused)),
-	       char *argument)
+get_one_option(const struct my_option *opt,
+	       char *argument, const char *filename __attribute__((unused)))
 {
-  switch (optid) {
+  switch (opt->id) {
 #ifdef __NETWARE__
   case OPT_AUTO_CLOSE:
     setscreenmode(SCR_AUTOCLOSE_ON_EXIT);
@@ -693,8 +723,8 @@ get_one_option(int optid,
     }
     else
     {
-     if (check_param.testflag & T_FORCE_CREATE)
-       check_param.testflag= T_FORCE_SORT_MEMORY;
+      if (check_param.testflag & T_FORCE_CREATE)
+        check_param.testflag= T_FORCE_SORT_MEMORY;
       check_param.tmpfile_createflag= O_RDWR | O_TRUNC;
       check_param.testflag|= T_FORCE_CREATE | T_UPDATE_STATE;
     }
@@ -857,7 +887,7 @@ get_one_option(int optid,
     case 2:
       method_conv= MI_STATS_METHOD_IGNORE_NULLS;
       break;
-    default: assert(0);                         /* Impossible */
+    default: abort();                         /* Impossible */
     }
     check_param.stats_method= method_conv;
     break;
@@ -982,8 +1012,8 @@ static int maria_chk(HA_CHECK *param, char *filename)
   MARIA_SHARE *share;
   DBUG_ENTER("maria_chk");
 
-  param->out_flag=error=param->warning_printed=param->error_printed=
-    recreate=0;
+  param->out_flag= error= param->error_printed= recreate= 0;
+  param->warning_printed= param->wrong_trd_printed= 0;
   datafile=0;
   param->isam_file_name=filename;		/* For error messages */
   warning_printed_by_chk_status= 0;
@@ -994,7 +1024,8 @@ static int maria_chk(HA_CHECK *param, char *filename)
                         ((param->testflag & T_WAIT_FOREVER) ?
                          HA_OPEN_WAIT_IF_LOCKED :
                          (param->testflag & T_DESCRIPT) ?
-                         HA_OPEN_IGNORE_IF_LOCKED : HA_OPEN_ABORT_IF_LOCKED))))
+                         HA_OPEN_IGNORE_IF_LOCKED : HA_OPEN_ABORT_IF_LOCKED),
+                        0)))
   {
     /* Avoid twice printing of isam file name */
     param->error_printed=1;
@@ -1062,6 +1093,15 @@ static int maria_chk(HA_CHECK *param, char *filename)
       param->testflag|= T_REP_BY_SORT;
     }
   }
+  if ((share->base.extra_options & MA_EXTRA_OPTIONS_ENCRYPTED) &&
+      !(param->testflag & T_DESCRIPT))
+  {
+    _ma_check_print_warning(param,
+                            "Table %s is encrypted. Only --description (-d) "
+                            "option is supported", filename);
+    param->warning_printed= 0;
+    goto end2;
+  }
 
   /*
     Skip the checking of the file if:
@@ -1514,6 +1554,8 @@ static void descript(HA_CHECK *param, register MARIA_HA *info, char *name)
 
   if (param->testflag & T_VERBOSE)
   {
+    if (share->base.extra_options & MA_EXTRA_OPTIONS_ENCRYPTED)
+      printf("Encrypted:           yes\n");
     printf("File-version:        %d\n",
 	   (int) share->state.header.file_version[3]);
     if (share->state.create_time)
@@ -1619,7 +1661,7 @@ static void descript(HA_CHECK *param, register MARIA_HA *info, char *name)
 	   buff, share->base.keys);
   }
   puts("\nTable description:");
-  printf("Key Start Len Index   Type");
+  printf("Key Start Len Index    Type");
   if (param->testflag & T_VERBOSE)
     printf("                     Rec/key         Root  Blocksize");
   putchar('\n');
@@ -1651,7 +1693,7 @@ static void descript(HA_CHECK *param, register MARIA_HA *info, char *name)
       pos=strmov(pos,null_txt);
     *pos=0;
 
-    printf("%-4d%-6ld%-3d %-8s%-23s",
+    printf("%-4d%-6ld%-3d %-9s%-23s",
 	   key+1,(long) keyseg->start+1,keyseg->length,text,buff);
     if (share->state.key_root[key] != HA_OFFSET_ERROR)
       llstr(share->state.key_root[key],buff);
@@ -1676,7 +1718,7 @@ static void descript(HA_CHECK *param, register MARIA_HA *info, char *name)
       if (keyseg->flag & HA_NULL_PART)
 	pos=strmov(pos,null_txt);
       *pos=0;
-      printf("    %-6ld%-3d         %-21s",
+      printf("    %-6ld%-3d          %-21s",
 	     (long) keyseg->start+1,keyseg->length,buff);
       if (param->testflag & T_VERBOSE)
 	printf("%11.0f", share->state.rec_per_key_part[keyseg_nr++]);
@@ -1976,9 +2018,10 @@ static int sort_record_index(MARIA_SORT_PARAM *sort_param,
   MARIA_HA *info= ma_page->info;
   MARIA_SHARE *share= info->s;
   uint	page_flag, nod_flag,used_length;
+  my_bool buff_alloced;
   uchar *temp_buff,*keypos,*endpos;
   my_off_t next_page,rec_pos;
-  uchar lastkey[MARIA_MAX_KEY_BUFF];
+  uchar *lastkey;
   char llbuff[22];
   MARIA_SORT_INFO *sort_info= sort_param->sort_info;
   HA_CHECK *param=sort_info->param;
@@ -1987,20 +2030,24 @@ static int sort_record_index(MARIA_SORT_PARAM *sort_param,
   const MARIA_KEYDEF *keyinfo= ma_page->keyinfo;
   DBUG_ENTER("sort_record_index");
 
+  temp_buff=0;
   page_flag= ma_page->flag;
   nod_flag=  ma_page->node;
-  temp_buff=0;
   tmp_key.keyinfo= (MARIA_KEYDEF*) keyinfo;
+
+  alloc_on_stack(*info->stack_end_ptr, lastkey, buff_alloced,
+                 (nod_flag ? keyinfo->block_length  : 0) +
+                 ALIGN_SIZE(keyinfo->max_store_length));
+  if (!lastkey)
+  {
+    _ma_check_print_error(param,"Not Enough memory");
+    DBUG_RETURN(-1);
+  }
+  if (nod_flag)
+    temp_buff= lastkey + ALIGN_SIZE(keyinfo->max_store_length);
+
   tmp_key.data=    lastkey;
 
-  if (nod_flag)
-  {
-    if (!(temp_buff= (uchar*) my_alloca(tmp_key.keyinfo->block_length)))
-    {
-      _ma_check_print_error(param,"Not Enough memory");
-      DBUG_RETURN(-1);
-    }
-  }
   used_length= ma_page->size;
   keypos= ma_page->buff + share->keypage_header + nod_flag;
   endpos= ma_page->buff + used_length;
@@ -2055,12 +2102,11 @@ static int sort_record_index(MARIA_SORT_PARAM *sort_param,
     _ma_check_print_error(param,"%d when updating keyblock",my_errno);
     goto err;
   }
-  if (temp_buff)
-    my_afree(temp_buff);
+  stack_alloc_free(lastkey, buff_alloced);
   DBUG_RETURN(0);
+
 err:
-  if (temp_buff)
-    my_afree(temp_buff);
+  stack_alloc_free(lastkey, buff_alloced);
   DBUG_RETURN(1);
 } /* sort_record_index */
 
@@ -2071,7 +2117,7 @@ static my_bool write_log_record(HA_CHECK *param)
     Now that all operations including O_NEW_DATA|INDEX are successfully
     done, we can write a log record.
   */
-  MARIA_HA *info= maria_open(param->isam_file_name, O_RDWR, 0);
+  MARIA_HA *info= maria_open(param->isam_file_name, O_RDWR, 0, 0);
   if (info == NULL)
     _ma_check_print_error(param, default_open_errmsg, my_errno,
                           param->isam_file_name);
