@@ -1,20 +1,13 @@
 /* Copyright (c) 2008, 2015, Oracle and/or its affiliates. All rights reserved.
 
   This program is free software; you can redistribute it and/or modify
-  it under the terms of the GNU General Public License, version 2.0,
-  as published by the Free Software Foundation.
-
-  This program is also distributed with certain software (including
-  but not limited to OpenSSL) that is licensed under separate terms,
-  as designated in a particular file or component or in included license
-  documentation.  The authors of MySQL hereby grant you an additional
-  permission to link the program and your derivative works with the
-  separately licensed software that they have included with MySQL.
+  it under the terms of the GNU General Public License as published by
+  the Free Software Foundation; version 2 of the License.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-  GNU General Public License, version 2.0, for more details.
+  GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
@@ -375,7 +368,7 @@ void PFS_engine_table::set_field_char_utf8(Field *f, const char* str,
 {
   DBUG_ASSERT(f->real_type() == MYSQL_TYPE_STRING);
   Field_string *f2= (Field_string*) f;
-  f2->store(str, len, &my_charset_utf8_bin);
+  f2->store(str, len, &my_charset_utf8mb3_bin);
 }
 
 void PFS_engine_table::set_field_varchar_utf8(Field *f, const char* str,
@@ -383,7 +376,7 @@ void PFS_engine_table::set_field_varchar_utf8(Field *f, const char* str,
 {
   DBUG_ASSERT(f->real_type() == MYSQL_TYPE_VARCHAR);
   Field_varstring *f2= (Field_varstring*) f;
-  f2->store(str, len, &my_charset_utf8_bin);
+  f2->store(str, len, &my_charset_utf8mb3_bin);
 }
 
 void PFS_engine_table::set_field_longtext_utf8(Field *f, const char* str,
@@ -391,7 +384,7 @@ void PFS_engine_table::set_field_longtext_utf8(Field *f, const char* str,
 {
   DBUG_ASSERT(f->real_type() == MYSQL_TYPE_BLOB);
   Field_blob *f2= (Field_blob*) f;
-  f2->store(str, len, &my_charset_utf8_bin);
+  f2->store(str, len, &my_charset_utf8mb3_bin);
 }
 
 void PFS_engine_table::set_field_enum(Field *f, ulonglong value)
