@@ -742,7 +742,6 @@ static int sphinx_init_func ( void * p )
 
 		#if MYSQL_VERSION_ID > 50100
 		handlerton * hton = (handlerton*) p;
-		hton->state = SHOW_OPTION_YES;
 		hton->db_type = DB_TYPE_AUTOASSIGN;
 		hton->create = sphinx_create_handler;
 		hton->close_connection = sphinx_close_connection;
@@ -769,10 +768,8 @@ static int sphinx_close_connection ( handlerton * hton, THD * thd )
 {
 	// deallocate common handler data
 	SPH_ENTER_FUNC();
-	void ** tmp = thd_ha_data ( thd, hton );
-	CSphTLS * pTls = (CSphTLS *) (*tmp);
+	CSphTLS * pTls = (CSphTLS *) thd_get_ha_data ( thd, hton );
 	SafeDelete ( pTls );
-	*tmp = NULL;
 	SPH_RET(0);
 }
 
@@ -844,7 +841,7 @@ bool sphinx_show_status ( THD * thd )
 
 #if MYSQL_VERSION_ID>50100
 	// 5.1.x style stats
-	CSphTLS * pTls = (CSphTLS*) ( *thd_ha_data ( thd, hton ) );
+	CSphTLS * pTls = (CSphTLS*) ( thd_get_ha_data ( thd, hton ) );
 
 #define LOC_STATS(_key,_keylen,_val,_vallen) \
 	stat_print ( thd, sphinx_hton_name, strlen(sphinx_hton_name), _key, _keylen, _val, _vallen );
@@ -2118,11 +2115,7 @@ int ha_sphinx::open ( const char * name, int, uint )
 
 	thr_lock_data_init ( &m_pShare->m_tLock, &m_tLock, NULL );
 
-	#if MYSQL_VERSION_ID>50100
-	*thd_ha_data ( table->in_use, ht ) = NULL;
-	#else
-	table->in_use->ha_data [ sphinx_hton.slot ] = NULL;
-	#endif
+	thd_set_ha_data ( table->in_use, ht, 0 );
 
 	SPH_RET(0);
 }
@@ -2163,7 +2156,7 @@ int ha_sphinx::Connect ( const char * sHost, ushort uPort )
 #if MYSQL_VERSION_ID>=50515
 			struct addrinfo *hp = NULL;
 			tmp_errno = getaddrinfo ( sHost, NULL, NULL, &hp );
-			if ( !tmp_errno || !hp || !hp->ai_addr )
+			if ( tmp_errno || !hp || !hp->ai_addr )
 			{
 				bError = true;
 				if ( hp )
@@ -2190,8 +2183,9 @@ int ha_sphinx::Connect ( const char * sHost, ushort uPort )
 			}
 
 #if MYSQL_VERSION_ID>=50515
-			memcpy ( &sin.sin_addr, hp->ai_addr, Min ( sizeof(sin.sin_addr), (size_t)hp->ai_addrlen ) );
-			freeaddrinfo ( hp );
+			struct sockaddr_in *in = (sockaddr_in *)hp->ai_addr;
+			memcpy ( &sin.sin_addr, &in->sin_addr, Min ( sizeof(sin.sin_addr), sizeof(in->sin_addr) ) );
+ 			freeaddrinfo ( hp );
 #else
 			memcpy ( &sin.sin_addr, hp->h_addr, Min ( sizeof(sin.sin_addr), (size_t)hp->h_length ) );
 			my_gethostbyname_r_free();
@@ -2804,23 +2798,16 @@ CSphSEThreadTable * ha_sphinx::GetTls()
 {
 	SPH_ENTER_METHOD()
 	// where do we store that pointer in today's version?
-	CSphTLS ** ppTls;
-#if MYSQL_VERSION_ID>50100
-	ppTls = (CSphTLS**) thd_ha_data ( table->in_use, ht );
-#else
-	ppTls = (CSphTLS**) &current_thd->ha_data[sphinx_hton.slot];
-#endif // >50100
+	CSphTLS * pTls = (CSphTLS*) thd_get_ha_data ( table->in_use, ht );
 
 	CSphSEThreadTable * pTable = NULL;
 	// allocate if needed
-	if ( !*ppTls )
+	if ( !pTls )
 	{
-		*ppTls = new CSphTLS ( this );
-		pTable = (*ppTls)->m_pHeadTable;
-	} else
-	{
-		pTable = (*ppTls)->m_pHeadTable;
+		pTls = new CSphTLS ( this );
+		thd_set_ha_data(table->in_use, ht, pTls);
 	}
+	pTable = pTls->m_pHeadTable;
 
 	while ( pTable && pTable->m_pHandler!=this )
 		pTable = pTable->m_pTableNext;
@@ -2828,8 +2815,8 @@ CSphSEThreadTable * ha_sphinx::GetTls()
 	if ( !pTable )
 	{
 		pTable = new CSphSEThreadTable ( this );
-		pTable->m_pTableNext = (*ppTls)->m_pHeadTable;
-		(*ppTls)->m_pHeadTable = pTable;
+		pTable->m_pTableNext = pTls->m_pHeadTable;
+		pTls->m_pHeadTable = pTable;
 	}
 
 	// errors will be handled by caller
@@ -3531,7 +3518,7 @@ CSphSEStats * sphinx_get_stats ( THD * thd, SHOW_VAR * out )
 #if MYSQL_VERSION_ID>50100
 	if ( sphinx_hton_ptr )
 	{
-		CSphTLS * pTls = (CSphTLS *) *thd_ha_data ( thd, sphinx_hton_ptr );
+		CSphTLS * pTls = (CSphTLS *) thd_get_ha_data ( thd, sphinx_hton_ptr );
 
 		if ( pTls && pTls->m_pHeadTable && pTls->m_pHeadTable->m_bStats )
 			return &pTls->m_pHeadTable->m_tStats;
@@ -3596,7 +3583,7 @@ int sphinx_showfunc_words ( THD * thd, SHOW_VAR * out, char * sBuffer )
 #if MYSQL_VERSION_ID>50100
 	if ( sphinx_hton_ptr )
 	{
-		CSphTLS * pTls = (CSphTLS *) *thd_ha_data ( thd, sphinx_hton_ptr );
+		CSphTLS * pTls = (CSphTLS *) thd_get_ha_data ( thd, sphinx_hton_ptr );
 #else
 	{
 		CSphTLS * pTls = (CSphTLS *) thd->ha_data[sphinx_hton.slot];
